/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France) 

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell-aes.h
  * \brief Declaration of aes functions.
  */

#ifndef AES_H
#define AES_H

#include <string.h>

#include "mphell-errors.h"

/**
 * @fn    void aes_expand_key(const uint8_t *key, const uint8_t Nr, const uint8_t Nk, uint32_t expanded_key[Nr+1][4])
 * @brief Create Nr + 1 round keys (128 bits) from the cipher key
 * @param key           Cipher key
 * @param Nr            Number of round
 * @param Nk            Size (in 32-bits word) of the cipher key
 * @param expanded_key  Array containing the Nr + 1 round key.
 */
void 
aes_expand_key(const uint8_t *key, const uint8_t Nr,
        const uint8_t Nk, uint32_t expanded_key[Nr+1][4]);

/**
 * @fn    void aes128_cipher(uint8_t *encrypted_data, const uint8_t* key, const uint8_t* data);
 * @brief Set encrypted_data to the result of AES using a 128 bits key on data
 * @param encrypted_data Cipher text (destination)
 * @param key            128 bits key
 * @param data           Plaintext (source)
 */
void
aes128_cipher(uint8_t *encrypted_data, const uint8_t* key, 
        const uint8_t* data);

/**
 * @fn    void aes192_cipher(uint8_t *encrypted_data, const uint8_t* key, const uint8_t* data);
 * @brief Set encrypted_data to the result of AES using a 192 bits key on data
 * @param encrypted_data Cipher text (destination)
 * @param key            192 bits key
 * @param data           Plaintext (source)
 */
void 
aes192_cipher(uint8_t *encrypted_data, const uint8_t* key, 
        const uint8_t* data);

/**
 * @fn    void aes256_cipher(uint8_t *encrypted_data, const uint8_t* key, const uint8_t* data_in);
 * @brief Set encrypted_data to the result of AES using a 256 bits key on data
 * @param encrypted_data Cipher text (destination)
 * @param key            256 bits key
 * @param data           Plaintext (source)
 */
void 
aes256_cipher(uint8_t *encrypted_data, const uint8_t* key,
        const uint8_t* data_in);


/**
 * @fn    void aes128_decipher(uint8_t *deciphered_data, const uint8_t* key, const uint8_t* encrypted_data);
 * @brief Set deciphered_data to the result of inverting AES using a 128 bits key on encrypted_data
 * @param deciphered_data   Plain text (destination)
 * @param key               128 bits key
 * @param encrypted_data    Cipher text (source)
 */
void
aes128_decipher(uint8_t *deciphered_data, const uint8_t* key, 
        const uint8_t* encrypted_data);

/**
 * @fn    void aes192_decipher(uint8_t *deciphered_data, const uint8_t* key, const uint8_t* encrypted_data);
 * @brief Set deciphered_data to the result of inverting AES using a 192 bits key on encrypted_data
 * @param deciphered_data   Plain text (destination)
 * @param key               192 bits key
 * @param encrypted_data    Cipher text (source)
 */
void 
aes192_decipher(uint8_t *deciphered_data, const uint8_t* key, 
        const uint8_t* encrypted_data);

/**
 * @fn    void aes256_decipher(uint8_t *deciphered_data, const uint8_t* key, const uint8_t* encrypted_data);
 * @brief Set deciphered_data to the result of inverting AES using a 256 bits key on encrypted_data
 * @param deciphered_data   Plain text (destination)
 * @param key               256 bits key
 * @param encrypted_data    Cipher text (source)
 */
void 
aes256_decipher(uint8_t *deciphered_data, const uint8_t* key,
        const uint8_t* encrypted_data);
#endif
