/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France) 

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell-curve.h
  * \brief Declaration of ECC functions.
  */

#ifndef MPHELL_CURVE_H
#define MPHELL_CURVE_H

#include "mphell-field.h"


/**************************************DATA STRUCTURE**********************************/


/**
 * @enum  ec_type_enum
 * @brief Define the type of curve.
 */
enum ec_type_enum 
{
    WEIERSTRASS,        /*!< Weierstrass curve, E: y^2 = x^3 + ax + b, 4a^3 +27b^2 != 0         */
    EDWARDS,            /*!< Edwards curve, E: x^2 + y^2 = 1 + d x^2 y^2; d(1-d)!=0             */
    JACOBI_QUARTIC      /*!< Jacobi quartic curve, E: y^2 = z^2 + 2ax^2 + t^2; x^2 = zt; a != 1 */
};

/**
 * @typedef  ec_type
 * @brief    Define the type of curve.
 */
typedef enum ec_type_enum ec_type;

/**
 * @enum  ec_algo_enum
 * @brief Define the type of coordinate.
 */
enum ec_algo_enum
{
    PROJECTIVE,                         /*!< (X,Y,Z) matching the affine point (x,y) where x=X/Z, y=Y/Z, the point at infinity is (0,1,0)                   */
    JACOBIAN,                           /*!< (X,Y,Z) matching the affine point (x,y) where x=X/Z^2, y=Y/Z^3, the point at infinity is (a^2,a^3,0) with a!=0 */
    EXTENDED_HOMOGENEOUS_PROJECTIVE,    /*!< (X,Y,T,Z) satisfying Y^2 = Z^2 + 2aX^2 + T^2; X^2 = ZT; a != 1 Here (X:Y:T:Z) = (sX:sY:T:sZ) for all nonzero s */
    EXTENDED_EDWARDS                    /*!< (X,Y,T,Z) matching the affine point (x,y) where x=X/Z, y=Y/Z, and T=XY/Z the point at infinity is (0,1,0,1)    */
};

/**
 * @typedef ec_formula
 * @brief   Define the type of coordinate.
 */
typedef enum ec_algo_enum ec_formula;

/**
 * @enum  ec_known_curve_enum
 * @brief Define the hardcoded curves
 */
enum ec_known_curve_enum
{
    NIST_192,                           /*!<    NIST 192 bits           */
    NIST_224,                           /*!<    NIST 224 bits           */
    NIST_256,                           /*!<    NIST 256 bits           */
    NIST_384,                           /*!<    NIST 384 bits           */
    NIST_521,                           /*!<    NIST 521 bits           */
    BRAINPOOL_160,                      /*!<    BRAINPOOL 160 bits      */
    BRAINPOOL_192,                      /*!<    BRAINPOOL 192 bits      */
    BRAINPOOL_224,                      /*!<    BRAINPOOL 224 bits      */
    BRAINPOOL_256,                      /*!<    BRAINPOOL 256 bits      */
    BRAINPOOL_320,                      /*!<    BRAINPOOL 320 bits      */
    BRAINPOOL_384,                      /*!<    BRAINPOOL 384 bits      */
    BRAINPOOL_512,                      /*!<    BRAINPOOL 521 bits      */
    FR_256,                             /*!<    ANSI 256 bits           */
    JQ_256_3,                           /*!<    JACOBI QUARTIC 256 bits */
    ED_25519                            /*!<    EDWARDS 25519           */
};

/**
 * @typedef ec_known_curve
 * @brief   Define the hardcoded curves
 */
typedef enum ec_known_curve_enum ec_known_curve;

/**
 * @struct ec_point
 * @brief  Define an elliptic curve point 
 */
struct ec_point 
{
    field_elt x; /*!< x-coordinate of the point */
    field_elt y; /*!< y-coordinate of the point */
    field_elt z; /*!< z-coordinate of the point */
    field_elt t; /*!< t-coordinate of the point */
};

/**
 * @typedef ec_point_t
 * @brief   Define an elliptic curve point 
 */
typedef struct ec_point ec_point_t;

/**
 * @typedef ec_point
 * @brief   Address of an ec_point structure 
 */
typedef ec_point_t ec_point[1];

/**
 * @typedef ec_point_ptr
 * @brief   Define ec_point_ptr, use in all functions
 */
typedef ec_point_t * ec_point_ptr;

/**
 * @typedef ec_point_srcptr
 * @brief   Define ec_point_srcptr, use in all functions where the parameter is const
 */
typedef const ec_point_t * ec_point_srcptr;

/**
 * @struct ec_curve
 * @brief  Define an elliptic curve
 */
struct ec_curve 
{
    char *id_curve;     /*!< Name of the curve                      */
    field_ptr k;        /*!< Base field                             */
    number q;           /*!< Size of the base field                 */
    field_elt a, b;     /*!< Coefficients of the curve              */
    ec_type type;       /*!< Type of the curve                      */
    ec_formula f;       /*!< Type of coordinate  system             */
    ec_point G;         /*!< Base point                             */
    number n;           /*!< Cardinality of <G>                     */
    number h;           /*!< Co-factor of Cardinality of <G>        */
    field_elt disc;     /*!< Discriminant of the curve              */
    field_elt D;        /*!< Discriminant without its coefficient   */
    bool ec_spec1;      /*!< Set to true if E->a = -3 mod p         */
};

/**
 * @typedef ec_curve_t
 * @brief  Define an elliptic curve
 */
typedef struct ec_curve ec_curve_t;

/**
 * @typedef ec_curve
 * @brief   Address of an ec_curve structure 
 */
typedef ec_curve_t ec_curve[1];

/**
 * @typedef ec_curve_ptr
 * @brief   Define ec_curve_ptr, use in all functions
 */
typedef ec_curve_t * ec_curve_ptr;

/**
 * @typedef ec_curve_srcptr
 * @brief   Define ec_curve_srcptr, use in all functions where the parameter is const
 */
typedef const ec_curve_t * ec_curve_srcptr;


/**********************************WEIESTRASS*****************************/

#include "mphell-weierstrass.h"

/********************************JACOBI QUARTIC***************************/

#include "mphell-jacobi.h"

/***********************************EDWARDS*******************************/

#include "mphell-edwards.h"

/*************************************TMP*********************************/

/**
 * @fn    void ec_point_get_pool_elt (ec_point_ptr P, field_ptr k, uint8_t stack);
 * @brief Get an initialised point from the pool
 * @param P     Point
 * @param k     Base field
 * @param stack Temporary memory stack to use
 */
static inline void
ec_point_get_pool_elt (ec_point_ptr P, field_ptr k, uint8_t stack)
{
    field_elt_get_pool_elt(&(P->x), k, stack);
    field_elt_get_pool_elt(&(P->y), k, stack);
    field_elt_get_pool_elt(&(P->z), k, stack);
    field_elt_get_pool_elt(&(P->t), k, stack);
}

/**
 * @fn    void ec_point_relax_pool_elt (ec_point_ptr P, field_ptr k, uint8_t stack)
 * @brief Relax an initialised point from the pool
 * @param P     Point
 * @param k     Base field
 * @param stack Temporary memory stack to use
 */
static inline void
ec_point_relax_pool_elt (ec_point_ptr P, field_ptr k, uint8_t stack)
{
    field_elt_relax_pool_elt(&(P->x), k, stack);
    field_elt_relax_pool_elt(&(P->y), k, stack);
    field_elt_relax_pool_elt(&(P->z), k, stack);
    field_elt_relax_pool_elt(&(P->t), k, stack);
}

/************************************SETTERS**********************************/


/**
 * @fn    void ec_alloc (ec_curve_ptr E, field_srcptr k);
 * @brief Allocate a curve
 * @param E    Elliptic curve to allocate
 * @param k    Base field
 */
void 
ec_alloc (ec_curve_ptr E, field_srcptr k);

/**
 * @fn    void ec_init (ec_curve_ptr E, field_srcptr k);
 * @brief Initialise a curve
 * @param E    Elliptic curve to initialise
 * @param k    Base field
 */
void 
ec_init (ec_curve_ptr E, field_srcptr k);

/**
 * @fn    void ec_create (ec_curve_ptr E, const char *id_curve, field_srcptr k, fe_srcptr a, fe_srcptr b, ec_point_srcptr G, number_srcptr h, number_srcptr n, const ec_type type,  const ec_formula f, uint8_t stack);
 * @brief Create an elliptic curve E, the curve must be allocated and initialised (ec_alloc & ec_init)
 * @param E         Elliptic curve, previously allocated
 * @param id_curve  Name of the curve
 * @param k         Base field
 * @param a         a coefficient for short Weierstrass and Jacobi Quartic elliptic curve
 * @param b         b coefficient for short Weierstrass elliptic curve
 * @param G         Base point
 * @param h         Cofactor of the order of G 
 * @param n         Order of G
 * @param type      Type of elliptic curve
 * @param f         Type of coordinate system
 * @param stack     Temporary memory stack to use
 */
void 
ec_create (ec_curve_ptr E, const char *id_curve, field_srcptr k, fe_srcptr a,
        fe_srcptr b, ec_point_srcptr G, number_srcptr h, number_srcptr n, 
        const ec_type type, const ec_formula f, uint8_t stack);

/**
 * @fn    void ec_use_curve (ec_curve_ptr E, field_srcptr k, const ec_known_curve id_curve, const ec_formula f, uint8_t stack)
 * @brief Create the elliptic curve (and the associated base field) id_curve, the curve and the field must me allocated (field_alloc & ec_alloc).
 * @param E         Elliptic curve, previously allocated
 * @param k         Base field, previously allocated
 * @param id_curve  Name of the curve, see ec_known_curve_enum
 * @param f         Type of coordinate system
 * @param stack     Temporary memory stack to use
 */
void 
ec_use_curve (ec_curve_ptr E, field_ptr k, const ec_known_curve id_curve, const ec_formula f, uint8_t stack);

/**
 * @fn    void ec_random(ec_curve_ptr E, const char *id_curve, char * seed_res, field_srcptr k, const ec_type type, const ec_formula f, uint8_t stack);
 * @brief Create a random elliptic curve E, not cryptographically secure (not tested to be). Algorithm A.3.3.2 from ANS X9.62-1998.
 * @param E         Elliptic curve, previously allocated
 * @param id_curve  Name of the curve
 * @param seed_res  Random seed use to generate the curve, 160 bits under hexadecimal form. Use a string of length 41 to store it.
 * @param type      Type of elliptic curve
 * @param f         Type of coordinate system
 * @param stack     Temporary memory stack to use
 */
void 
ec_random(ec_curve_ptr E, const char *id_curve, char * seed_res, field_srcptr k, const ec_type type, const ec_formula f, uint8_t stack);

/**
 * @fn     void ec_verify_random_generation(ec_curve_ptr E, const char * seed, uint8_t stack)
 * @brief  Test if E if generated from the seed "seed", using algorithm A.3.4.2 from ANS X9.62-1998. Only curve generated with the algorithm A.3.3.2 can be tested.
 * @param  E     Elliptic curve to test
 * @param  seed  Seed (of 160 bits) used to generate an elliptic curve, under hexadecimal form.
 * @param  stack Temporary memory stack to use
 * @return true  if E is generated from seed
 * @return false otherwise
 */
bool 
ec_verify_random_generation(ec_curve_ptr E, const char * seed, uint8_t stack);

/**
 * @fn     void ec_set_fast_unified_coordinates(ec_curve_ptr E);
 * @brief  Set the fastest unified coordinates system
 * @param  E    Elliptic curve
 */
void 
ec_set_fast_unified_coordinates(ec_curve_ptr E);

/**
 * @fn     void ec_set_fast_dedicated_coordinates(ec_curve_ptr E);
 * @brief  Set the fastest dedicated coordinates system
 * @param  E    Elliptic curve
 */
void 
ec_set_fast_dedicated_coordinates(ec_curve_ptr E);

/**
 * @fn    void ec_compute_disc (ec_curve_ptr E, uint8_t stack);
 * @brief Set the discriminant of E
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void
ec_compute_disc (ec_curve_ptr E, uint8_t stack);

/**
 * @fn    void ec_test_spec (ec_curve_ptr E, uint8_t stack);
 * @brief Set the E->ec_spec1 to true if E->a = -3 mod p, false otherwise.
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void
ec_test_spec (ec_curve_ptr E, uint8_t stack);

/**
 * @fn    void ec_copy (ec_curve_ptr E_res, ec_curve_srcptr E);
 * @brief Copy E into E_res which has been previously allocated, beware: the same field is used, do not free it 2 times
 * @param E_res Resulting elliptic curve
 * @param E     Elliptic curve
 */
void
ec_copy (ec_curve_ptr E_res, ec_curve_srcptr E);

/**
 * @fn    void ec_clear (ec_curve_ptr E);
 * @brief Clear the elliptic curve E (remove the action of ec_init, but let the one of ec_alloc)
 * @param E Elliptic curve
 */
void 
ec_clear (ec_curve_ptr E);

/**
 * @fn    void ec_free (ec_curve_ptr E);
 * @brief Free the elliptic curve E
 * @param E Elliptic curve
 */
void 
ec_free (ec_curve_ptr E);

/**
 * @fn    void ec_point_alloc (ec_point_ptr P, field_srcptr k);
 * @brief Allocate an elliptic curve point
 * @param P     Point to allocate
 * @param k     Base field
 */
void 
ec_point_alloc (ec_point_ptr P, field_srcptr k);

/**
 * @fn    void ec_point_init (ec_point_ptr P, field_srcptr k);
 * @brief Initialise an elliptic curve point
 * @param P     Point to allocate
 * @param k     Base field
 */
void 
ec_point_init (ec_point_ptr P, field_srcptr k);

/**
 * @fn    void ec_point_clear (ec_point_ptr P, field_srcptr k);
 * @brief Clear the point P (remove the action of ec_point_init, but let the one of ec_point_alloc)
 * @param P     Point to clear
 * @param k     Base field
 */
void 
ec_point_clear (ec_point_ptr P, field_srcptr k);

/**
 * @fn    void ec_point_free (ec_point_ptr P, field_srcptr k);
 * @brief Free the point P
 * @param P     Point to free
 * @param k     Base field
 */
void 
ec_point_free (ec_point_ptr P, field_srcptr k);

/**
 * @fn    void ec_point_copy (ec_point_ptr P3, ec_point_srcptr P, field_srcptr k);
 * @brief Copy P into P3
 * @param P3    Destination point
 * @param P     Point
 * @param k     Base field
 */
void
ec_point_copy (ec_point_ptr P3, ec_point_srcptr P, field_srcptr k);

/**
 * @fn    void ec_point_set (ec_point_ptr P, fe_srcptr x, fe_srcptr y, fe_srcptr z, fe_srcptr t, field_srcptr k);
 * @brief Set a point from its coordinates
 * @param P Destination point
 * @param x x-coordinate of P
 * @param y y-coordinate of P
 * @param z z-coordinate of P
 * @param t t-coordinate of P
 * @param k Base field
 */
void 
ec_point_set (ec_point_ptr P, fe_srcptr x, fe_srcptr y, 
        fe_srcptr z, fe_srcptr t, field_srcptr k);

/**
 * @fn    void ec_point_set_str (ec_point_ptr P, const char *str_x, const char *str_y, const char *str_z, const char *str_t, const bool is_reduced, const uint8_t base, field_srcptr k, uint8_t stack);
 * @brief Set a point from its coordinates under string format
 * @param P           Destination point
 * @param x           x-coordinate of P under string form
 * @param y           y-coordinate of P under string form
 * @param z           z-coordinate of P under string form
 * @param t           t-coordinate of P under string form
 * @param is_reduced  True if the coordinates are already in Montgomery form, false otherwise
 * @param base        Base used to read the coordinate (either 10 or 16)
 * @param k           Base field
 * @param stack       Temporary memory stack to use
 */
void 
ec_point_set_str (ec_point_ptr P, const char *str_x, const char *str_y, 
        const char *str_z, const char *str_t, const bool is_reduced, 
        const uint8_t base, field_srcptr k, uint8_t stack);

/**
 * @fn    void ec_point_set_aff (ec_point_ptr P, fe_srcptr x, fe_srcptr y, const ec_type type, field_srcptr k, uint8_t stack);
 * @brief Set a point from its affine coordinates
 * @param P     Destination point
 * @param x     x-coordinate of P
 * @param y     y-coordinate of P
 * @param type  Type of elliptic curve
 * @param stack Temporary memory stack to use
 */
void 
ec_point_set_aff (ec_point_ptr P, fe_srcptr x, fe_srcptr y,
        const ec_type type, field_srcptr k, uint8_t stack);

/**
 * @fn    void ec_point_set_aff_str (ec_point_ptr P, const char *str_x, const char *str_y, const bool is_reduced, const uint8_t base, const ec_type type, field_srcptr k, uint8_t stack);
 * @brief Set a point from its affine coordinates under string format
 * @param P           Destination point
 * @param x           x-coordinate of P under string form
 * @param y           y-coordinate of P under string form
 * @param is_reduced  True if the coordinates are already reduced in k, false otherwise
 * @param base        Base used to read the coordinate (either 10 or 16)
 * @param type        Type of elliptic curve
 * @param k           Base field
 * @param stack       Temporary memory stack to use
 */
void 
ec_point_set_aff_str (ec_point_ptr P, const char *str_x, const char *str_y,
        const bool is_reduced, const uint8_t base, const ec_type type,
        field_srcptr k, uint8_t stack);

/**
 * @fn    void ec_point_set_neutral (ec_point_ptr dst, ec_curve_srcptr E, uint8_t stack);
 * @brief Set dst to the neutral element
 * @param dst   Destination point
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void 
ec_point_set_neutral (ec_point_ptr dst, ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void ec_point_norm (ec_point_ptr P, ec_curve_srcptr E, uint8_t stack)
 * @brief Convert a point in projective or jacobian coordinate to an affine point (x,y)
 * @param P     Point
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void
ec_point_norm (ec_point_ptr P, ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void ec_point_get_x_affine (field_elt x, ec_point_ptr P, ec_curve_srcptr E, uint8_t stack);
 * @brief Convert P->x to its affine representation
 * @param x     P->x in affine coordinates
 * @param P     Point
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void
ec_point_get_x_affine (field_elt x, ec_point_ptr P, ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void ec_point_get_y_affine (field_elt y, ec_point_ptr P, ec_curve_srcptr E, uint8_t stack);
 * @brief Convert P->y to its affine representation
 * @param y     P->y in affine coordinates
 * @param P     Point
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void
ec_point_get_y_affine (field_elt y, ec_point_ptr P, ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void ec_point_lift (ec_point_ptr P, field_srcptr k, uint8_t stack)
 * @brief Lift the coordinates from Montgomery basis to classical arithmetic
 * @param P     Point
 * @param k     Base field
 * @param stack Temporary memory stack to use
 */
void
ec_point_lift (ec_point_ptr P, field_srcptr k, uint8_t stack);

/**
 * @fn     bool ec_belongs (ec_point_srcptr P, ec_curve_srcptr E, uint8_t stack)
 * @brief  Test if P belongs to E
 * @param  P     Point
 * @param  E     Elliptic curve
 * @param  stack Temporary memory stack to use
 * @return True if P belongs to E, false otherwise
 */
bool
ec_belongs (ec_point_srcptr P, ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void ec_point_random (ec_point_ptr P, ec_curve_srcptr E, uint8_t stack)
 * @brief Create a random point P on the elliptic curve E
 * @param P     Destination point
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void
ec_point_random (ec_point_ptr P, ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void ec_curve_str (char **str, ec_curve_srcptr E, const uint8_t base, uint8_t stack);
 * @brief Allocate *str and write in it the description of E
 * @param str   Pointer to a string, allocated by the function.
 * @param E     Elliptic curve
 * @param base  Base used to write the parameters (either 10 or 16)
 * @param stack Temporary memory stack to use
 */
void
ec_curve_str (char **str, ec_curve_srcptr E, const uint8_t base, uint8_t stack);

/**
 * @fn    void ec_curve_print (ec_curve_srcptr E, const uint8_t base, uint8_t stack);
 * @brief Print a description of E
 * @param E     Elliptic curve
 * @param base  Base used to print the parameters (either 10 or 16)
 * @param stack Temporary memory stack to use
 */
void
ec_curve_print (ec_curve_srcptr E, const uint8_t base, uint8_t stack);

/**
 * @fn    void ec_point_str (char **str, ec_point_srcptr P, const uint8_t base, const bool lift, field_srcptr k, uint8_t stack);
 * @brief Allocate *str and write in it the description of P
 * @param str   Pointer to a string, allocated by the function.
 * @param P     Point
 * @param base  Base used to write the coordinates (either 10 or 16)
 * @param lift  If Montgomery is used, lift P coordinates to classical fp.
 * @param k     Base field
 * @param stack Temporary memory stack to use
 */
void
ec_point_str (char **str, ec_point_srcptr P, const uint8_t base, 
        const bool lift, field_srcptr k, uint8_t stack);

/**
 * @fn    void ec_point_print (ec_point_srcptr P, const uint8_t base, const bool lift, field_srcptr k, uint8_t stack);
 * @brief Print a description of P
 * @param P     Point
 * @param base  Base used to print the coordinates (either 10 or 16)
 * @param lift  If Montgomery is used, lift P coordinates to classical fp.
 * @param k     Base field
 * @param stack Temporary memory stack to use
 */
void
ec_point_print (ec_point_srcptr P, const uint8_t base, const bool lift, field_srcptr k, uint8_t stack);


/*******************************COMPARISON************************************/


/**
 * @fn     bool ec_point_is_neutral (ec_point_srcptr P, ec_curve_srcptr E, uint8_t stack);
 * @brief  Test if P is the neutral element
 * @param  P     Point to test
 * @param  E     Elliptic curve
 * @param  stack Temporary memory stack to use
 * @return True if P is neutral, False otherwise
 */
bool 
ec_point_is_neutral (ec_point_srcptr P, ec_curve_srcptr E, uint8_t stack);

/**
 * @fn     bool ec_point_are_equal (ec_point_srcptr P1, ec_point_srcptr P2, ec_curve_srcptr E, uint8_t stack);
 * @brief  Test if P1 and P2 are equal on E
 * @param  P1    Point
 * @param  P2    Point
 * @param  E     Elliptic curve
 * @param  stack Temporary memory stack to use
 * @return True if P1 == P2, False otherwise
 */
bool
ec_point_are_equal (ec_point_srcptr P1, ec_point_srcptr P2, const ec_curve_srcptr E, uint8_t stack);


/*******************************OPERATIONS************************************/

/**
 * @fn    ec_point_neg (ec_point_ptr P3, ec_point_srcptr P1, ec_curve_srcptr E)
 * @brief Set P3 to -P1
 * @param P3 Resulting point
 * @param P1 Source point 
 * @param E  Elliptic curve
 */
void 
ec_point_neg (ec_point_ptr P3, ec_point_srcptr P1, ec_curve_srcptr E);

/* UNIFIED => Resistant to SPA */

/**
 * @fn    void ec_point_add_unified (ec_point_ptr P3, ec_point_srcptr P1, ec_point_srcptr P2, ec_curve_srcptr E, uint8_t stack)
 * @brief Set P3 to P1 + P2, using unified formulae (protection against SPA)
 * @param P3    Resulting point 
 * @param P1    Source point
 * @param P2    Source point 
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void 
ec_point_add_unified (ec_point_ptr P3, ec_point_srcptr P1, ec_point_srcptr P2, 
        ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    ec_point_dbl_unified (ec_point_ptr P3, ec_point_srcptr P1, ec_curve_srcptr E, uint8_t stack)
 * @brief Set P3 to 2*P1, using unified formulae (protection against SPA)
 * @param P3    Resulting point
 * @param P1    Source point 
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void
ec_point_dbl_unified (ec_point_ptr P3, ec_point_srcptr P1, ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void ec_point_sub_unified (ec_point_ptr P3, ec_point_srcptr P1, ec_point_srcptr P2, ec_curve_srcptr E, uint8_t stack)
 * @brief Set P3 to P1 - P2, using unified formulae (protection against SPA)
 * @param P3    Resulting point 
 * @param P1    Source point
 * @param P2    Source point 
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void 
ec_point_sub_unified (ec_point_ptr P3, ec_point_srcptr P1, ec_point_srcptr P2, 
        ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void ec_point_mul_unified(ec_point_ptr P3, number_srcptr n, ec_point_srcptr P1, ec_curve_srcptr E, uint8_t stack);
 * @brief Set P3 to n * P1 using Montgomery for Weierstrass elliptic curve, and naive method for other elliptic curve type, using unified formulae (protection against SPA)
 * @param P3    Destination point
 * @param n     Scalar to multiply P1 with
 * @param P1    Point
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void
ec_point_mul_unified (ec_point_ptr P3, number_srcptr n, ec_point_srcptr P1, 
        ec_curve_srcptr E, uint8_t stack);


/* DEDICATED => Not resistant to SPA, but faster, execpt Montgomery and joye which are anyhow faster and unified */

/**
 * @fn    void ec_point_add (ec_point_ptr P3, ec_point_srcptr P1, ec_point_srcptr P2, ec_curve_srcptr E, uint8_t stack)
 * @brief Set P3 to P1 + P2, using dedicated formulae (not protected against SPA, but faster)
 * @param P3    Resulting point 
 * @param P1    Source point
 * @param P2    Source point 
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void 
ec_point_add (ec_point_ptr P3, ec_point_srcptr P1, ec_point_srcptr P2, 
        ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    ec_point_dbl (ec_point_ptr P3, ec_point_srcptr P1, ec_curve_srcptr E, uint8_t stack)
 * @brief Set P3 to 2*P1, using dedicated formulae (not protected against SPA, but faster)
 * @param P3    Resulting point
 * @param P1    Source point 
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void
ec_point_dbl (ec_point_ptr P3, ec_point_srcptr P1, ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void ec_point_sub (ec_point_ptr P3, ec_point_srcptr P1, ec_point_srcptr P2, ec_curve_srcptr E, uint8_t stack)
 * @brief Set P3 to P1 - P2, using dedicated formulae (not protected against SPA, but faster)
 * @param P3    Resulting point 
 * @param P1    Source point
 * @param P2    Source point 
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void 
ec_point_sub (ec_point_ptr P3, ec_point_srcptr P1, ec_point_srcptr P2, 
        ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void ec_point_mul(ec_point_ptr P3, number_srcptr n, ec_point_srcptr P1, ec_curve_srcptr E, uint8_t stack);
 * @brief Set P3 to n * P1 using Montgomery for Weierstrass elliptic curve, and naive method for other elliptic curve type, using dedicated formulae (not protected against SPA execpt in WEIESTRASS / JACOBIAN case, but faster)
 * @param P3    Destination point
 * @param n     Scalar to multiply P1 with
 * @param P1    Point
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void
ec_point_mul (ec_point_ptr P3, number_srcptr n, ec_point_srcptr P1, 
        ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void ec_point_mul_with_precomp(ec_point_ptr P3, number_srcptr n, ec_point_srcptr * tab_P1, int16_t window_size,  ec_curve_srcptr E, uint8_t stack);
 * @brief Set P3 to n * P1 using naive double and add method, Montgomery and Joye multiplication are also available for Weierstrass elliptic curve, using dedicated formulae (not protected against SPA, but faster)
 * @param P3        Destination point
 * @param n         Scalar to multiply P1 with
 * @param tab_P1    Array of precomputed [P1, 3P1, 5P1, ..., (2^e - 1)P1]
 * @param win_size  Size of the windows
 * @param E         Elliptic curve
 * @param stack     Temporary memory stack to use
 */
void 
ec_point_mul_with_precomp (ec_point_ptr P3, number_srcptr n, ec_point * tab_P1, 
        int16_t window_size, ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void ec_point_2mul_with_precomp(ec_point_ptr P3, number_srcptr n1, ec_point_srcptr * tab_P1, number_srcptr n2, ec_point_srcptr * tab_P2, int16_t win_size, ec_curve_srcptr E, uint8_t stack);
 * @brief Set P3 to n1 * P1 + n2 * P2 using 2 precomputated array
 * @param P3        Destination point
 * @param n1        Scalar to multiply P1 with
 * @param tab_P1    Array of precomputed [P1, 2P1, 3P1, ..., (2^e - 1)P1]
 * @param n2        Scalar to multiply P2 with
 * @param tab_P2    Array of precomputed [P2, 2P2, 3P2, ..., (2^e - 1)P2]
 * @param win_size  Size of the windows
 * @param E         Elliptic curve
 * @param stack     Temporary memory stack to use
 */
void 
ec_point_2mul_with_precomp (ec_point_ptr P3, number_srcptr n1, ec_point * tab_P1,
        number_srcptr n2, ec_point * tab_P2, 
        int16_t win_size, ec_curve_srcptr E, uint8_t stack);

/***********************************CONVERSION********************************/

/**
 * @fn    void jacobi_quartic_to_weierstrass (ec_curve_ptr E_res, ec_curve_srcptr E, uint8_t stack)
 * @brief Convert a jacobi quartic elliptic curve into a Weierstrass elliptic curve
 * @param E_res Resulting elliptic curve (Weierstrass)
 * @param E     Source elliptic curve (Jacobi quartic)
 * @param stack Temporary memory stack to use
 */
void
jacobi_quartic_to_weierstrass (ec_curve_ptr E_res, ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void jacobi_quartic_point_to_weierstrass_point (ec_point_ptr dst, ec_point_srcptr P, ec_curve_srcptr E, uint8_t stack)
 * @brief Convert a point on a jacobi quartic elliptic curve to a point on a weierstrass elliptic curve
 * @param dst   Resulting point (weierstrass point)
 * @param P     Source point (jacobi quartic point)
 * @param E     Elliptic curve (Jacobi quartic)
 * @param stack Temporary memory stack to use
 */
void
jacobi_quartic_point_to_weierstrass_point (ec_point_ptr dst, ec_point_srcptr P, 
        ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void weierstrass_to_jacobi_quartic (ec_curve_ptr E_res, ec_curve_srcptr E, const bool determined, fe_ptr teta, uint8_t stack);
 * @brief Convert a Weierstrass elliptic curve into a jacobi quartic elliptic curve
 * @param E_res      Resulting elliptic curve (Jacobi quartic)
 * @param E          Source elliptic curve (Weierstrass)
 * @param determined Boolean equals to false if theta is not specified then we make the assumption that b_j=1 otherwise theta is specified
 * @param teta       Field element such as (teta, 0) is a 2-torsion point on E
 * @param stack      Temporary memory stack to use
 */
void
weierstrass_to_jacobi_quartic (ec_curve_ptr E_res, ec_curve_srcptr E, const bool determined, fe_ptr teta, uint8_t stack);

/**
 * @fn    void weierstrass_point_to_jacobi_quartic_point (ec_point_ptr dst, ec_point_srcptr P, ec_curve_srcptr E, uint8_t stack)
 * @brief Convert a point on a weierstrass elliptic curve to a point on a jacobi quartic elliptic curve
 * @param dst   Resulting point (jacobi quartic point)
 * @param P     Source point (Weierstrass point)
 * @param E     Elliptic curve (Jacobi quartic)
 * @param stack Temporary memory stack to use
 */
void
weierstrass_point_to_jacobi_quartic_point (ec_point_ptr dst, ec_point_srcptr P, 
        ec_curve_srcptr E, uint8_t stack);
        
/**
 * @fn     void edwards_to_weierstrass (ec_curve_ptr E_res, ec_curve_srcptr E, uint8_t stack);
 * @brief  Convert the Edwards elliptic curve E to the corresponding Weierstrass elliptic curve E_res 
 * @param  E_res  Weierstrass Elliptic curve
 * @param  E      Edwards Elliptic curve
 * @param  stack  Temporary memory stack to use
 */
void
edwards_to_weierstrass (ec_curve_ptr E_res, ec_curve_srcptr E, uint8_t stack);

/**
 * @fn     void edwards_point_to_weierstrass_point (ec_point_ptr dst, ec_point_srcptr P, ec_curve_srcptr E, uint8_t stack);
 * @brief  Convert the Edwards point P of the elliptic curve E to the corresponding Weierstrass elliptic curve point dst
 * @param  dst   Elliptic curve destination point 
 * @param  P     Elliptic curve point
 * @param  E     Edwards Elliptic curve
 * @param  stack Temporary memory stack to use
 */
void
edwards_point_to_weierstrass_point (ec_point_ptr dst, ec_point_srcptr P, 
        ec_curve_srcptr E, uint8_t stack);

/**
 * @fn     void weierstrass_to_edwards (ec_curve_ptr E_res, ec_curve_srcptr E, const uint8_t n, fe_ptr alpha, fe_ptr beta, uint8_t stack);
 * @brief  Convert the Weierstrass elliptic curve E to the one of the corresponding Edwards elliptic curve according to the choice (or not) of alpha and/or beta 
 * @param  E_res  Edwards Elliptic curve
 * @param  E      Weierstrass Elliptic curve
 * @param  n      Integer indicating the number of variables between alpha and beta fixed by the user 2 if the both are fixed 1 if only alpha is fixed and 0 if the choice is let to the program
 * @param  alpha  Field element
 * @param  beta   Field element 
 * @param  stack  Temporary memory stack to use
 */
void
weierstrass_to_edwards (ec_curve_ptr E_res, ec_curve_srcptr E, const uint8_t n, 
        fe_ptr alpha, fe_ptr beta, uint8_t stack);

/**
 * @fn     void weierstrass_point_to_edwards_point (ec_point_ptr dst, ec_point_srcptr P, ec_curve_srcptr E, const uint8_t n, fe_ptr alpha, fe_ptr beta, uint8_t stack);
 * @brief  Convert the point of the Weierstrass elliptic curve E to the elliptic curve of one of the corresponding Edwards elliptic curve according to the choice (or not) of alpha and/or beta 
 * @param  dst    Edwards Elliptic curve destination point 
 * @param  P      Weierstrass Elliptic curve point
 * @param  E      Edwards Elliptic curve
 * @param  n      Integer indicating the number of variables between alpha and beta fixed by the user 2 if the both are fixed 1 if only alpha is fixed and 0 if the choice is let to the program
 * @param  alpha  Field element
 * @param  beta   Field element 
 * @param  stack  Temporary memory stack to use
 */
void
weierstrass_point_to_edwards_point (ec_point_ptr dst, ec_point_srcptr P, 
        ec_curve_srcptr E, const uint8_t n, fe_ptr alpha, fe_ptr beta, uint8_t stack);


#endif

