/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France) 

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell-drbg.h
  * \brief Declaration of the Deterministic Random Bit Generator functions. Details can be found in "NIST Special Publication 800-90A, Recommendation for Random Number Generation Using Deterministic Random Bit Generators"
  */

#ifndef MPHELL_DRBG_H
#define MPHELL_DRBG_H

#include "mphell-drbg_internal.h"

/**
 * @fn    void drbg_instantiate (mphell_status ret, int8_t *state_handle, const uint16_t requested_instantiation_security_strength, const drbg_type type, const entropy_type entropy_src, uint64_t reseed_interval, const uint8_t * entropy, const uint16_t entropy_length, const uint8_t * nonce, const uint16_t nonce_length, const uint8_t* personalization_string, const uint64_t pers_length);
 * @brief Instantiate one DRBG state, "acquires entropy input and may combine it with a nonce and a personalization string to create a seed from which the initial internal state is created".
 * @param ret                                       Set to MPHELL_ERROR_RANDOM if an error occur, MPHELL_SUCCESS otherwise
 * @param state_handle                              Destination integer, store the number of the DRBG state used.
 * @param requested_instantiation_security_strength Security (in bits) required
 * @param type                                      DRBG algorithm to use (DRBG_SHA1, DRBG_SHA256, DRBG_AES128, DRBG_AES192, DRBG_AES256)
 * @param entropy_src                               Entropy to use (to seed the DRBG): INTEL, DEVURANDOM, DEVRANDOM
 * @param reseed_interval                           Reseed interval, must be between 1 (prediction resistance == true) and 2^48
 * @param entropy                                   Entropy (byte string)
 * @param entropy_length                            Entropy length (in bytes)
 * @param nonce                                     Nonce (byte string)
 * @param nonce_length                              Nonce length (in bytes)
 * @param personalization_string                    A string given by the user can be added to the material seed to instantiate the DRBG, see section 8.7.1 of NIST Special Publication 800-90A for details
 * @param pers_length                               Length of personalization_string
 */
void 
drbg_instantiate (mphell_status ret, int8_t *state_handle, 
        const uint16_t requested_instantiation_security_strength, 
        const drbg_type type, const entropy_type entropy_src, uint64_t reseed_interval,
        const uint8_t * entropy, const uint16_t entropy_length,
        const uint8_t * nonce, const uint16_t nonce_length,
        const uint8_t* personalization_string, const uint64_t pers_length);

/**
 * @fn    void drbg_reseed(mphell_status ret, const int8_t state_handle, const uint8_t * entropy, const uint16_t entropy_length, const uint8_t* additional_input, const uint64_t add_length);
 * @brief Reseed the DRBG, "acquires new entropy input and combines it with the current internal state and any additional input that is provided to create a new seed and a new internal state"
 * @param ret              Set to MPHELL_ERROR_RANDOM if an error occur, MPHELL_SUCCESS otherwise
 * @param state_handle     Define which DRBG state to use
 * @param entropy          Entropy (byte string)
 * @param entropy_length   Entropy length (in bytes)
 * @param additional_input A string given by the user can be added to the material seed to reseed the DRBG, "can provide more assurance when recovering from the compromise of the entropy input, the seed or one or more DRBG internal states"
 * @param add_length       Length of additional_input
 */
void 
drbg_reseed(mphell_status ret, const int8_t state_handle,
        const uint8_t * entropy, const uint16_t entropy_length,
        const uint8_t* additional_input, const uint64_t add_length);

/**
 * @fn    void drbg_generate(mphell_status ret, uint8_t* returned_bytes, const int8_t state_handle, const uint32_t requested_number_of_bytes, const uint16_t requested_security_strength, const uint8_t* additional_input, const uint64_t add_length);
 * @brief "Generates pseudorandom bits upon request, using the current internal state, and generates a new internal state for the next request".
 * @param ret                           Set to MPHELL_ERROR_RANDOM if an error occur, MPHELL_SUCCESS otherwise
 * @param returned_bytes                Destination array (of requested_number_of_bytes random bits)
 * @param state_handle                  Define which DRBG state to use
 * @param requested_number_of_bytes     Number of random bytes required
 * @param requested_security_strength   Security (in bits) required
 * @param additionnal_input             A string given by the user can be added to the material seed to reseed the DRBG, "can provide more assurance when recovering from the compromise of the entropy input, the seed or one or more DRBG internal states"
 * @param add_length                    Length of additional_input
 */
void 
drbg_generate(mphell_status ret, uint8_t* returned_bytes, 
        const int8_t state_handle, const uint32_t requested_number_of_bytes, 
        const uint16_t requested_security_strength,
        const uint8_t* additional_input, const uint64_t add_length);

/**
 * @fn    void drbg_uninstantiate(mphell_status ret, const int8_t state_handle);
 * @brief Uninstantiate a DRBG state
 * @param ret           Set to MPHELL_ERROR_RANDOM if an error occur, MPHELL_SUCCESS otherwise
 * @param state_handle  DRBG state to uninstantiate
 */
void 
drbg_uninstantiate(mphell_status ret, const int8_t state_handle);

#endif
