/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France) 

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell-drbg_internal.h
  * \brief Declaration of the Deterministic Random Bit Generator internal functions.
  */

#ifndef DRBG_INTERNAL_H
#define DRBG_INTERNAL_H

#include "mphell-drbg_types.h"
#include "mphell-sha1.h"
#include "mphell-sha256.h"
#include "mphell-sha512.h"
#include "mphell-aes.h"

/* These constants comes from Table 3: Definitions for the CTR_DRBG, in section 10.2.1 of "NIST Special Publication 800-90A" */

#define DRBG_MAX_ENTROPY_LENGTH 4294967296                 /* (2^35 / 8)    */
#define DRBG_MAX_PERSONALIZATION_STRING_LENGTH 4294967296  /* (2^35 / 8)    */
#define DRBG_MAX_ADDITIONAL_INPUT_LENGTH 4294967296        /* (2^35 / 8)    */
#define DRBG_MAX_NUMBER_OF_BYTES_PER_REQUEST 65536         /* (2^19 / 8)    */
#define DRBG_MAX_RESEED_INTERVAL 281474976710656           /* 2^48          */

/* Utility functions */

/**
 * @fn     uint32_t drbg_ceil(uint32_t n1, uint32_t n2);
 * @brief  Compute ceil(n1/n2), ie smallest integer >= (n1/n2)
 * @param  n1 Integer
 * @param  n2 Integer
 * @return ceil(n1/n2)
 */
uint32_t drbg_ceil(uint32_t n1, uint32_t n2);

/**
 * @fn    void drbg_incr_data(uint8_t* data, const uint16_t data_len);
 * @brief Increment data (Set data to data + 1 where data is an array of data_len bytes)
 * @param data     Number (under bytes array form) to increment
 * @param data_len Length (in bytes) of data
 */
void drbg_incr_data(uint8_t* data, const uint16_t data_len);

/**
 * @fn    void drbg_add_data(uint8_t* data, const uint16_t data_len, const uint8_t* data1, const uint16_t data1_len);
 * @brief Set data to data + data1
 * @param data      Number (under bytes array form), source and destination
 * @param data_len  Length (in bytes) of data
 * @param data1     Number (under bytes array form), to add to data
 * @param data1_len Length (in bytes) of data1, must be <= data_len
 */
void drbg_add_data(uint8_t* data, const uint16_t data_len, const uint8_t* data1, 
        const uint16_t data1_len);

/**
 * @fn    void drbg_add_int(uint8_t* data, const uint16_t data_len, const uint64_t n);
 * @brief Set data to data + n
 * @param data     Number (under bytes array form), source and destination
 * @param data_len Length (in bytes) of data, must be >= 8
 * @param n        Integer of size 8 bytes
 */
void drbg_add_int(uint8_t* data, const uint16_t data_len, const uint64_t n);

/**
 * @fn    void drbg_xor_data(uint8_t* res, const uint8_t* data1, const uint8_t* data2, const uint16_t data_length);
 * @brief Set res to data1 XOR data2 where XOR is the bitwise operator
 * @param res         Resulting data
 * @param data1       Source data, must be of size >= data_length bytes
 * @param data2       Source data, must be of size >= data_length bytes
 * @param data_length Length of data1 and data2
 */
void drbg_xor_data(uint8_t* res, const uint8_t* data1, const uint8_t* data2,
        const uint16_t data_length);

/* HASH DRBG functions */
/* Details can be found in "NIST Special Publication 800-90A, section 10.1.1 */

/**
 * @fn    void hash_drbg_instantiate(mphell_status ret, drbg_internal_state initial_state, const uint8_t security_strength, const uint8_t* entropy_input, const uint16_t entropy_length, const uint8_t* personalization_string, const uint64_t pers_length, const drbg_type type);
 * @brief Instantiate the DRBG state (V, C_Key) using entropy_input and personalization_string
 * @param ret                       Set to MPHELL_ERROR_RANDOM if an error occur, MPHELL_SUCCESS otherwise
 * @param initial_state             DRBG state to instantiate
 * @param security_strength         Security (in bytes) required
 * @param entropy_input             Entropy (byte string)
 * @param entropy_length            Length (in bytes) of entropy_input
 * @param personalization_string    A string given by the user can be added to the material seed to instantiate the DRBG, see section 8.7.1 of NIST Special Publication 800-90A for details
 * @param pers_length               Length of personalization_string
 * @param type                      DRBG algorithm to use (DRBG_SHA1, DRBG_SHA256)
 */
void hash_drbg_instantiate(mphell_status ret, drbg_internal_state initial_state,
        const uint8_t security_strength, const uint8_t* entropy_input, const uint16_t entropy_length,
        const uint8_t* personalization_string, 
        const uint64_t pers_length, const drbg_type type);

/**
 * @fn    void hash_drbg_reseed(mphell_status ret, drbg_internal_state working_state, const uint8_t* entropy_input, const uint16_t entropy_length, const uint8_t* additional_input, const uint64_t add_length);
 * @brief Reseed the DRBG state (V, C_Key) using entropy_input and additional_input
 * @param ret               Set to MPHELL_ERROR_RANDOM if an error occur, MPHELL_SUCCESS otherwise
 * @param working_state     Current state of the DRBG (current value of V, C_key and reseed_counter)
 * @param entropy_input     Entropy (byte string)
 * @param entropy_length    Length (in bytes) of entropy_input
 * @param additionnal_input A string given by the user can be added to the material seed to reseed the DRBG, "can provide more assurance when recovering from the compromise of the entropy input, the seed or one or more DRBG internal states"
 * @param add_length        Length of additional_input
 */
void hash_drbg_reseed(mphell_status ret, drbg_internal_state working_state,
        const uint8_t* entropy_input, const uint16_t entropy_length, 
        const uint8_t* additional_input, const uint64_t add_length);

/**
 * @fn    void hash_drbg_generate(mphell_status ret, uint8_t* returned_bytes, drbg_internal_state working_state, const uint32_t requested_no_of_bytes, const uint8_t* additional_input, const uint64_t add_length);
 * @brief "Generates pseudorandom bits upon request, using the current internal state, and generates a new internal state for the next request". 
 * @param ret                   Set to MPHELL_ERROR_RANDOM if an error occur, MPHELL_RESEED_REQUIRED if the working state need to be reseeded, MPHELL_SUCCESS otherwise
 * @param returned_bytes        Destination random bytes string
 * @param working_state         Current state of the DRBG (current value of V, C_key and reseed_counter)
 * @param requested_no_of_bytes Number of requested random bytes
 * @param additional_input      A string given by the user can be added to the material seed to reseed the DRBG, "can provide more assurance when recovering from the compromise of the entropy input, the seed or one or more DRBG internal states"
 * @param add_length            Length of additional_input
 */
void hash_drbg_generate(mphell_status ret, uint8_t* returned_bytes, 
        drbg_internal_state working_state, const uint32_t requested_no_of_bytes,
        const uint8_t* additional_input, const uint64_t add_length);

/* CTR DRBG functions */
/* Details can be found in "NIST Special Publication 800-90A, section 10.2.1 */

/**
 * @fn    void ctr_drbg_instantiate(mphell_status ret, drbg_internal_state initial_state, const uint8_t security_strength, const uint8_t* entropy_input, const uint16_t entropy_length, const uint8_t* personalization_string, const uint64_t pers_length, const drbg_type type);
 * @brief Instantiate the DRBG state (V, C_Key) using entropy_input and personalization_string
 * @param ret                       Set to MPHELL_ERROR_RANDOM if an error occur, MPHELL_SUCCESS otherwise
 * @param initial_state             DRBG state to instantiate
 * @param security_strength         Security (in bytes) required
 * @param entropy_input             Entropy (byte string)
 * @param entropy_length            Length (in bytes) of entropy_input
 * @param personalization_string    A string given by the user can be added to the material seed to instantiate the DRBG, see section 8.7.1 of NIST Special Publication 800-90A for details
 * @param pers_length               Length of personalization_string
 * @param type                      DRBG algorithm to use (DRBG_AES128, DRBG_AES192, DRBG_AES256)
 */
void ctr_drbg_instantiate(mphell_status ret, drbg_internal_state initial_state,
        const uint8_t security_strength, const uint8_t* entropy_input, const uint16_t entropy_length,
        const uint8_t* personalization_string, const uint64_t pers_length, const drbg_type type);

/**
 * @fn    void ctr_drbg_reseed(mphell_status ret, drbg_internal_state working_state, const uint8_t* entropy_input, const uint16_t entropy_length, const uint8_t* additional_input, const uint64_t add_length);
 * @brief Reseed the DRBG state (V, C_Key) using entropy_input and additional_input
 * @param ret               Set to MPHELL_ERROR_RANDOM if an error occur, MPHELL_SUCCESS otherwise
 * @param working_state     Current state of the DRBG (current value of V, C_key and reseed_counter)
 * @param entropy_input     Entropy (byte string)
 * @param entropy_length    Length (in bytes) of entropy_input
 * @param additionnal_input A string given by the user can be added to the material seed to reseed the DRBG, "can provide more assurance when recovering from the compromise of the entropy input, the seed or one or more DRBG internal states"
 * @param add_length        Length of additional_input
 */
void ctr_drbg_reseed(mphell_status ret, drbg_internal_state working_state,
        const uint8_t* entropy_input, const uint16_t entropy_length, 
        const uint8_t* additional_input, const uint64_t add_length);

/**
 * @fn    void ctr_drbg_generate(mphell_status ret, uint8_t* returned_bytes, drbg_internal_state working_state, const uint32_t requested_no_of_bytes, const uint8_t* additional_input, const uint64_t add_length);
 * @brief "Generates pseudorandom bits upon request, using the current internal state, and generates a new internal state for the next request". 
 * @param ret                   Set to MPHELL_ERROR_RANDOM if an error occur, MPHELL_RESEED_REQUIRED if the working state need to be reseeded, MPHELL_SUCCESS otherwise
 * @param returned_bytes        Destination random bytes string
 * @param working_state         Current state of the DRBG (current value of V, C_key and reseed_counter)
 * @param requested_no_of_bytes Number of requested random bytes
 * @param additional_input      A string given by the user can be added to the material seed to reseed the DRBG, "can provide more assurance when recovering from the compromise of the entropy input, the seed or one or more DRBG internal states"
 * @param add_length            Length of additional_input
 */
void ctr_drbg_generate(mphell_status ret, uint8_t* returned_bytes, 
        drbg_internal_state working_state, const uint32_t requested_no_of_bytes,
        const uint8_t* additional_input, const uint64_t add_length);
#endif
