/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France) 

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell-drbg_types.h
  * \brief Declaration of the Deterministic Random Bit Generator types.
  */

#include "mphell-entropy.h"

#ifndef DRBG_TYPES_H
#define DRBG_TYPES_H

/**
 * @struct drbg_working_state_t
 * @brief  One or more values that are derived from the seed and become part of the internal state, and a reseed counter
 */
typedef struct {
    uint8_t *V;                 /*!< Secret value (CTR-DRBG -> outlen bits, HASH-DRBG -> seedlen bits) updated during each call to the DRBG */
    uint8_t *C_Key;             /*!< Secret value (CTR-DRBG -> keylen bits, HASH-DRBG -> seedlen bits) updated when the seed is updated     */
    uint32_t reseed_counter;    /*!< Number of requests since the instantiation was seeded or reseeded                                      */
} drbg_working_state_t;

/**
 * @typedef drbg_working_state
 * @brief   Address of a drbg_working_state_t structure
 */
typedef drbg_working_state_t drbg_working_state[1];

/**
 * @enum  drbg_algorithm
 * @brief Define the DRBG algorithm
 */
enum drbg_algorithm {
    DRBG_UNSET,     /*!< No algorithm specified                     */
    DRBG_SHA1,      /*!< Use SHA1 hash algorithm                    */
    DRBG_SHA256,    /*!< Use SHA256 hash algorithm                  */
    DRBG_AES128,    /*!< Use AES with 128 bits keys in Counter mode */
    DRBG_AES192,    /*!< Use AES with 192 bits keys in Counter mode */
    DRBG_AES256     /*!< Use AES with 256 bits keys in Counter mode */
};

/**
 * @typedef  drbg_type
 * @brief Define the DRBG algorithm
 */
typedef enum drbg_algorithm drbg_type;

/**
 * @struct drbg_info_data_t
 * @brief  Administrative informations on the DRBG state
 */
typedef struct {
    uint8_t  seedlen;            /*!< Length (in bits) of the seed (outlen + keylen)                         */
    uint8_t  outlen;             /*!< Length (in bits) of the output block of the cryptographic primitive    */
    uint8_t  security_strength;  /*!< Security (in bytes) required                                           */
    uint64_t reseed_interval;    /*!< Reseed interval (in number of drbg_generate request)                   */
} drbg_info_data_t;

/**
 * @typedef drbg_info_data
 * @brief   Address of a drbg_info_data_t structure
 */
typedef drbg_info_data_t drbg_info_data[1];

#define DRBG_MAX_INTERNAL_STATES 10

/**
 * @struct drbg_internal_state_t
 * @brief The working state and administrative informations
 */
typedef struct {
    drbg_working_state state;   /*!< Values that are derived from the seed and become part of the internal state, and a reseed counter  */
    drbg_info_data info;        /*!< Informations on the DRBG state                                                                     */
    drbg_type type;             /*!< Used DRBG algorithm                                                                                */
    entropy_type entropy_src;   /*!< Used entropy source                                                                                */
} drbg_internal_state_t;

/**
 * @typedef drbg_internal_state
 * @brief   Address of a drbg_internal_state_t structure
 */
typedef drbg_internal_state_t drbg_internal_state[1];

#endif
