/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France) 

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell-edwards.h
  * \brief Declaration of Edward ECC functions.
  */

#ifndef MPHELL_EDWARDS_H
#define MPHELL_EDWARDS_H

/* E: x^2 + y^2 = 1 + d x^2 y^2 */
/* Edwards form  */
/* Projective coordinates represent an affine point (x,y) on a Edwards-form elliptic curve x^2 + y^2 = 1 + d x^2 y^2 as (X:Y:Z) satisfying (X^2 + Y^2) Z^2 = Z^4 + d X^2 Y^2. Here (X:Y:Z) = (sX:sY:sZ) for all nonzero s. */
/* The triple (X, Y, Z) represents the affine point (X / Z, Y / Z) */

/* Jacobian coordinates represent an affine point (x,y) on a Edwards-form elliptic curve  x^2 + y^2 = 1 + d x^2 y^2 as (X:Y:Z) satisfying (X^2 Z^6 + Y^2 Z^4)  = Z^10 + d X^2 Y^2. Here (X:Y:Z) = (s^2 X:s^3 Y:sZ) for all nonzero s.*/
/* The triple (X, Y, Z) represents the affine point (X / Z^2, Y / Z^3). */

/* Extended twisted Edwards coordinates represent an affine point (x,y) on a Edwards-form elliptic curve x^2 + y^2 = 1 + d x^2 y^2 as (X:Y:Z) satisfying ( X^2 + Y^2) Z^2 = Z^4 + d X^2 Y^2. Here (X:Y:T:Z) = (sX:sY:sT:sZ) for all nonzero s. The auxiliary coordinate T has the property T=XY/Z. */

/* E: a x^2 + y^2 = 1 + d x^2 y^2 */
/* Twisted Edwards form  */
/* Projective coordinates represent an affine point (x,y) on a Edwards-form elliptic curve a x^2 + y^2 = 1 + d x^2 y^2 as (X:Y:Z) satisfying (X^2 + Y^2) Z^2 = Z^4 + d X^2 Y^2. Here (X:Y:Z) = (sX:sY:sZ) for all nonzero s. */
/* The triple (X, Y, Z) represents the affine point (X / Z, Y / Z) */

/* Jacobian coordinates represent an affine point (x,y) on a Weierstrass-form elliptic curve a x^2 + y^2 = 1 + d x^2 y^2 as (X:Y:Z) satisfying (a X^2 Z^6 + Y^2 Z^4)  = Z^10 + d X^2 Y^2. Here (X:Y:Z) = (s^2 X:s^3 Y:sZ) for all nonzero s.*/
/* The triple (X, Y, Z) represents the affine point (X / Z^2, Y / Z^3). */

/* Extended twisted Edwards coordinates represent an affine point (x,y) on a Edwards-form elliptic curve a x^2 + y^2 = 1 + d x^2 y^2 as (X:Y:Z) satisfying (a X^2 + Y^2) Z^2 = Z^4 + d X^2 Y^2. Here (X:Y:T:Z) = (sX:sY:sT:sZ) for all nonzero s. The auxiliary coordinate T has the property T=XY/Z. */

/* In this software the coefficients a and d that represent a(n) (Twisted) Edwards Curve are stored in E->a for a and E->b for d.*/

/* More details can be found for example in Twisted Edwards curves by Daniel J. Bernstein, Peter Birkner, Marc Joye, Tanja Lange, and Christiane Peters
  published in International Conference on Cryptology in Africa. Springer, Berlin, Heidelberg, 2008. p. 389-405. */

/************************************SETTERS**********************************/

/**
 * @fn    void edwards_compute_disc(ec_curve E, uint8_t stack);
 * @brief Set the discriminant of E: disc = a.d.(a-d)^4
 * @param E    Elliptic curve
 * @param stack Temporary memory stack to use
 */
void
edwards_compute_disc(ec_curve E, uint8_t stack);

/**
 * @fn     bool edwards_verify_random_generation(ec_curve E, const char * seed, uint8_t stack)
 * @brief  Test if E if generated from the seed "seed"
 * @param  E     Elliptic curve to test
 * @param  seed  Seed (of 160 bits) used to generate an elliptic curve
 * @param  stack Temporary memory stack to use
 * @return true  if E is generated from seed
 * @return false otherwise
 */
bool 
edwards_verify_random_generation(ec_curve E, const char * seed, uint8_t stack);

/**
 * @fn    void edwards_curve_random_generation(fe_ptr d, char * seed_res, field_srcptr k, uint8_t stack)
 * @brief Generate a 160 bits seed and coefficients a and b defining a Weiestrass elliptic curve. The curve is not tested to be secure !!!
 * @param d        Destination coefficient, generated from a random seed 
 * @param seed_res Destination seed (of 160 bits) used to generate an elliptic curve
 * @param k        Base field, must be created by the user before use.
 * @param stack    Temporary memory stack to use
 */
void 
edwards_curve_random_generation(fe_ptr d, char * seed_res, field_srcptr k, uint8_t stack);

/**
 * @fn    void edwards_point_set_neutral (ec_point_ptr dst, ec_curve_srcptr E, uint8_t stack);
 * @brief Set dst to the neutral element: (0,1,0,1) for extended twisted Edwards coordinates.
 * @param dst   Destination point
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void 
edwards_point_set_neutral (ec_point_ptr dst, ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void edwards_point_set_aff (ec_point_ptr P, fe_srcptr x, fe_srcptr y, field_srcptr k, uint8_t stack);
 * @brief Set dest to the extended twisted Edwards coordinates point (x,y,1,xy)
 * @param P     Destination point
 * @param x     x-coordinate of the extended twisted Edwards coordinates point (x,y,1,xy) 
 * @param y     y-coordinate of the extended twisted Edwards coordinates point (x,y,1,xy)
 * @param k     Base field
 * @param stack Temporary memory stack to use
 */
void 
edwards_point_set_aff (ec_point_ptr P, fe_srcptr x, fe_srcptr y, field_srcptr k, uint8_t stack);

/**
 * @fn    void edwards_point_set_aff_str (ec_point_ptr P, const char *str_x, const char *str_y, const bool is_reduced, const uint8_t base, field_srcptr k, uint8_t stack);
 * @brief Set dest to the affine point twisted Edwards coordinates point (str_x,str_y,1,str_x str_y)
 * @param P             Destination point
 * @param str_x         x-coordinate of the extended twisted Edwards coordinates point (x,y,1,xy) under string form
 * @param str_y         y-coordinate of the extended twisted Edwards coordinates point (x,y,1,xy) under string form
 * @param is_reduced    set true if str_x and str_y are already reduced in k, false otherwise
 * @param base          Base (either 10 or 16) used for str_x and str_y
 * @param k             Base field
 * @param stack         Temporary memory stack to use
 */
void 
edwards_point_set_aff_str (ec_point_ptr P, const char *str_x, const char *str_y,
        const bool is_reduced, const uint8_t base, field_srcptr k, uint8_t stack);

/**
 * @fn    bool edwards_belongs (ec_point_srcptr P, ec_curve_srcptr E, uint8_t stack);
 * @brief Test if P belongs to E
 * @param  P     Point 
 * @param  E     Edwards Elliptic curve
 * @param  stack Temporary memory stack to use
 * @return true  if P belongs to E, false otherwise
 */
bool
edwards_belongs (ec_point_srcptr P, ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void edwards_point_random (ec_point_ptr P, ec_curve_srcptr E, uint8_t stack);
 * @brief Set P to a random point on E
 * @param P     Destination point
 * @param E     Edwards Elliptic curve
 * @param stack Temporary memory stack to use
 */
void 
edwards_point_random (ec_point_ptr P, ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void edwards_point_norm (ec_point_ptr P, ec_curve_srcptr E, uint8_t stack);
 * @brief Set P in affine coordinates
 * @param P     Point
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void
edwards_point_norm (ec_point_ptr P, ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void edwards_point_get_x_affine(field_elt x, ec_point_ptr P, ec_curve_srcptr E, uint8_t stack);
 * @brief Convert P->x to its affine representation
 * @param x     P->x in affine coordinates
 * @param P     Point
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void
edwards_point_get_x_affine(field_elt x, ec_point_ptr P, ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void edwards_point_get_y_affine(field_elt y, ec_point_ptr P, ec_curve_srcptr E, uint8_t stack);
 * @brief Convert P->y to its affine representation
 * @param x     P->y in affine coordinates
 * @param P     Point
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void
edwards_point_get_y_affine(field_elt y, ec_point_ptr P, ec_curve_srcptr E, uint8_t stack);

/*******************************COMPARISON************************************/

/**
 * @fn     bool edwards_point_is_neutral (ec_point_srcptr P, ec_curve_srcptr E);
 * @brief  Test if P is the neutral element
 * @param  P Point to test
 * @param  E Elliptic curve
 * @return True if P is neutral, False otherwise
 */
bool 
edwards_point_is_neutral (ec_point_srcptr P, ec_curve_srcptr E);

/**
 * @fn     bool edwards_point_are_equal (ec_point_srcptr P1, ec_point_srcptr P2, ec_curve_srcptr E, uint8_t stack);
 * @brief  Test if P1 and P2 are equal on E (BUT do not test if the points belongs to the curve )
 * @param  P1    Point
 * @param  P2    Point
 * @param  E     Elliptic curve
 * @param  stack Temporary memory stack to use
 * @return True if P1 == P2, False otherwise
 */
bool
edwards_point_are_equal (ec_point_srcptr P1, ec_point_srcptr P2, 
        ec_curve_srcptr E, uint8_t stack);

/*******************************OPERATIONS************************************/

/**
 * @fn     edwards_point_neg (ec_point_ptr P3, ec_point_srcptr P1, ec_curve_srcptr E);
 * @brief  Compute the opposite of a point of the the Edwards elliptic curve E
 * @param  P3 Elliptic curve destination point
 * @param  P1 Elliptic curve point
 * @param  E  Edwards Elliptic curve
 */
void 
edwards_point_neg (ec_point_ptr P3, ec_point_srcptr P1, ec_curve_srcptr E);

/* UNIFIED => Resistant to SPA */

/**
 * @fn     edwards_point_add_unified (ec_point_ptr P3, ec_point_srcptr P1, ec_point_srcptr P2, ec_curve_srcptr E, uint8_t stack);
 * @brief  Compute the unified addition of 2 points P1 and P2 of the Edwards elliptic curve E, using unified formulae (protection against SPA). Only available with Extended Edwards coordinates.
 * @param  P3    Elliptic curve destination point
 * @param  P2    Elliptic curve point
 * @param  P1    Elliptic curve point
 * @param  E     Edwards Elliptic curve
 * @param  stack Temporary memory stack to use
 */
void 
edwards_point_add_unified (ec_point_ptr P3, ec_point_srcptr P1, ec_point_srcptr P2,
        ec_curve_srcptr E, uint8_t stack);

/* DEDICATED => Not resistant to SPA, but faster  */

/**
 * @fn     edwards_point_add_dedicated (ec_point_ptr P3, ec_point_srcptr P1, ec_point_srcptr P2, ec_curve_srcptr E, uint8_t stack);
 * @brief  Compute the addition of 2 points P1 and P2 of the Edwards elliptic curve E (not protected against SPA, but faster)
 * @param  P3    Elliptic curve destination point
 * @param  P2    Elliptic curve point
 * @param  P1    Elliptic curve point
 * @param  E     Edwards Elliptic curve
 * @param  stack Temporary memory stack to use
 */
void 
edwards_point_add_dedicated (ec_point_ptr P3, ec_point_srcptr P1, ec_point_srcptr P2,
        ec_curve_srcptr E, uint8_t stack);

/**
 * @fn     edwards_point_dbl_dedicated (ec_point_ptr P3, ec_point_srcptr P1, ec_curve_srcptr E, uint8_t stack);
 * @brief  Compute the double of a point of the Edwards elliptic curve E (not protected against SPA, but faster)
 * @param  P3    Elliptic curve destination point
 * @param  P1    Elliptic curve point
 * @param  E     Edwards Elliptic curve
 * @param  stack Temporary memory stack to use
 */
void 
edwards_point_dbl_dedicated (ec_point_ptr P3, ec_point_srcptr P1, ec_curve_srcptr E, uint8_t stack);

#endif

