/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France) 

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell-errors.h
  * \brief Declaration of error functions.
  */
 
#ifndef ERRORS_H
#define ERRORS_H

#include <stdio.h>
#include <stdlib.h>

#include "mphell-define.h"

#ifdef __LINE__
#define MPHELL_ASSERT_LINE  __LINE__
#else
#define MPHELL_ASSERT_LINE  -1
#endif

#ifdef __FILE__
#define MPHELL_ASSERT_FILE  __FILE__
#else
#define MPHELL_ASSERT_FILE  ""
#endif

#define MPHELL_ASSERT_FAIL(info)  mphell_assert_error (MPHELL_ASSERT_FILE, \
        MPHELL_ASSERT_LINE, info)

#define MPHELL_ASSERT_ALWAYS(expr, info)                        \
  do {                                  \
    if (!(expr))                        \
      MPHELL_ASSERT_FAIL (info);                        \
  } while (0)


#if MPHELL_USE_ASSERT == 1
#define MPHELL_ASSERT(expr, info)   MPHELL_ASSERT_ALWAYS (expr, info)
#else
#define MPHELL_ASSERT(expr, info)   do {} while (0)
#endif

/**
 * @enum mphell_status_flag_enum
 * @brief Enumerate status flag
 */
enum mphell_status_flag_enum
{
    MPHELL_SUCCESS,                     /*!< Success                                            */
    MPHELL_ERROR_RANDOM,                /*!< Error with the Random Number Generator             */
    MPHELL_CATASTROPHIC_ERROR_RANDOM,   /*!< Error with the Random Number Generator, must stop  */
    MPHELL_RESEED_REQUIRED              /*!< The RNG need to be resseeded                       */
};

/**
 * @typedef mphell_status_flag
 * @brief   Enumerate status flag
 */
typedef enum mphell_status_flag_enum mphell_status_flag;

/**
 * @struct mphell_status_t
 * @brief The status is a couple (flag, information)
 */
typedef struct
{
    mphell_status_flag flag;            /*!< Flags              */
    char *info;                         /*!< Information string */
} mphell_status_t;

/**
 * @typedef mphell_status
 * @brief The status is a couple (flag, information)
 */
typedef mphell_status_t mphell_status[1];

/**
 * @fn    void mphell_assert_error (const char *filename, int line, const char *expr);
 * @brief Write in stderr, filename, line and expr, free mphell.
 * @param filename File where the error occured
 * @param line     Line where the rrors occured
 * @param expr     Error description
 */
void
mphell_assert_error (const char *filename, int line, const char *expr);

/**
 * @fn    void mphell_error (char *expr);
 * @brief Write in stderr, filename, line and expr, free mphell.
 * @param expr Error description
 */
void
mphell_error (char *expr);

/**
 * @fn    void mphell_error_free (char *expr);
 * @brief Write in stderr, filename, line and expr, free expr and mphell.
 * @param expr Error description, allocated dynamically
 */
void
mphell_error_free (char *expr);

#endif
