/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France) 

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell-field.h
  * \brief Declaration of field functions.
  */

#ifndef MPHELL_FIELD_H
#define MPHELL_FIELD_H

#include "mphell-fp.h"
#include "mphell-fp2.h"
#include "mphell-fp3.h"

#if MPHELL_USE_FP_ONLY == 0

/**
 * @typedef field_elt
 * @brief   Generic field element
 */
typedef fp_elt * field_elt;

/**
 * @typedef fe_ptr
 * @brief   Pointer on a field element
 */
typedef fp_elt * fe_ptr;

/**
 * @typedef fe_srcptr
 * @brief Pointer on a field element, the field element cannot be modified through this pointer
 */
typedef const fp_elt * fe_srcptr;

#elif MPHELL_USE_FP_ONLY == 1

/**
 * @typedef field_elt
 * @brief   Generic field element
 */
typedef fp_elt field_elt;

/**
 * @typedef fe_ptr
 * @brief   Pointer on a field element
 */
typedef fp_elt_ptr fe_ptr;

/**
 * @typedef fe_srcptr
 * @brief Pointer on a field element, the field element cannot be modified through this pointer
 */
typedef const fp_elt_srcptr fe_srcptr;

#endif

/**
 * @enum field_type
 * @brief Define the field type
 */
typedef enum
{
    FP,     /*!< Prime field                    */
    FP2,    /*!< Finite field of p^2 element    */
    FP3     /*!< Finite field of p^3 element    */
} field_type;

/**
 * @struct field_t
 * @brief  Define a field
 */
typedef struct
{
    void *param;                        /*!< Pointer to a field information structure: fp_param_t for FP, fp2_param_t for FP2 and fp3_param_t for FP3       */
    field_type type;                    /*!< Type of field, FP, FP2 or FP3                                                                                  */
    uint8_t size;                       /*!< Size in blocks of the base prime field (ex: 4 for a 4 * 64 = 256 bits prime field on a 64 bits architecture)   */
    uint16_t bit_size;                  /*!< Size in bit                                                                                                    */
    field_elt pool_1[POOL_SIZE_FIELD];  /*!< Pool of already initialised field element to accelerate field operations, use when STACK_1 is specified        */
    uint8_t i_1;                        /*!< Index for pool 1                                                                                               */
#if MPHELL_USE_MULTITHREADING == 1
    field_elt pool_2[POOL_SIZE_FIELD];  /*!< Pool of already initialised field element to accelerate field operations, use when STACK_2 is specified        */
    uint8_t i_2;                        /*!< Index for pool 2                                                                                               */
#endif
    char id[20];                        /*!< Field name                                                                                                     */
}
field_t;

/**
 * @typedef field
 * @brief Address of a field structure
 */
typedef field_t field[1];

/**
 * @typedef field_ptr
 * @brief Address of a field structure, use in all functions
 */
typedef field_t * field_ptr;

/**
 * @typedef field_srcptr
 * @brief Address of a field structure, use in all functions where the parameter is const
 */
typedef const field_t * field_srcptr;

/**************************************TMP**********************************/

/**
 * @fn    void field_elt_get_pool_elt (field_elt * dst, field_ptr k, uint8_t stack);
 * @brief Get an initialised field element from the pool
 * @param dst   Address of a field element
 * @param k     Information on the field, k must have been created with field_create.
 * @param stack Temporary memory stack to use
 */
static inline void
field_elt_get_pool_elt (field_elt * dst, field_ptr k, uint8_t stack)
{
#if MPHELL_USE_MULTITHREADING == 0
    MPHELL_ASSERT(stack == STACK_1, "field_elt_get_pool_elt, unknow stack \n");
    MPHELL_ASSERT(k->i_1 < POOL_SIZE_FIELD, "field_elt_get_pool_elt, stack is too small \n");
    *dst = (k->pool_1)[(k->i_1)++];
#elif MPHELL_USE_MULTITHREADING == 1
    if(stack == STACK_1)
    {
        MPHELL_ASSERT(k->i_1 < POOL_SIZE_FIELD, "field_elt_get_pool_elt, stack is too small \n");
        *dst = (k->pool_1)[(k->i_1)++];
    }
    else if (stack == STACK_2)
    {
        MPHELL_ASSERT(k->i_2 < POOL_SIZE_FIELD, "field_elt_get_pool_elt, stack is too small \n");
        *dst = (k->pool_2)[(k->i_2)++];
    }
    else
    {
        *dst = 0;
        mphell_error("field_elt_get_pool_elt, unknow stack \n");
    }
#endif
}

/**
 * @fn    void field_elt_relax_pool_elt (field_elt * dst, field_ptr k, uint8_t stack);
 * @brief Relax an initialised field element from the pool
 * @param dst   Address of a field element
 * @param k     Information on the field, k must have been created with field_create.
 * @param stack Temporary memory stack to use
 */
static inline void
field_elt_relax_pool_elt (field_elt * dst, field_ptr k, uint8_t stack)
{
#if MPHELL_USE_MULTITHREADING == 0
    MPHELL_ASSERT(stack == STACK_1, "field_elt_relax_pool_elt, unknow stack \n");
    (k->i_1)--;
    MPHELL_ASSERT(k->i_1 >= 0, "k->i_1 is < 0 in pool 1\n");
#elif MPHELL_USE_MULTITHREADING == 1
    if(stack == STACK_1)
    {
        (k->i_1)--;
        MPHELL_ASSERT(k->i_1 >= 0, "k->i_1 is < 0 in pool 1\n");
    }
    else if (stack == STACK_2)
    {
        (k->i_2)--;
        MPHELL_ASSERT(k->i_2 >= 0, "k->i_2 is < 0 in pool 2\n");
    }
    else
    {
        mphell_error("field_elt_relax_pool_elt, unknow stack \n");
    }
#endif
}

/************************************SETTERS**********************************/

 /**
 * @fn    void field_alloc (field_ptr k, const field_type type, const uint8_t size, field_ptr base);
 * @brief Allocates space for the different fields of the structure pointed by k
 * @param k     Structure for the field informations to allocate
 * @param type  Type of field, available fields are FP, FP2 and FP3
 * @param size  Size in blocks of the field (ex: 4 for a 4 * 64 = 256 bits prime field on a 64 bits architecture)
 * @param base  Base prime field or NULL if k is already a prime field
 */
void
field_alloc (field_ptr k, const field_type type, const uint8_t size, field_ptr base);

 /**
 * @brief Initialize the different fields of the structure pointed by k.
 * @param k     Informations structure on the field to fill (must be allocated)
 * @param id    Name of the field
 * @param stack Temporary memory stack to use
 * @param n     Number of arguments that follow (use 1 for FP and 2 for FP2 and FP3)
 * @param ...   Characteristic (number) for FP and base field (prime field) and non_residu (field_elt) for FP2 and FP3
 */
void
field_create (field_ptr k, const char* id, uint8_t stack, const uint32_t n, ...); 

/**
 * @fn    void field_copy (field_ptr k_res, field_srcptr k);
 * @brief Copy the field structure param into param_res
 * @param k_res Destination field 
 * @param k     Source field 
 */
void
field_copy (field_ptr k_res, field_srcptr k);

/**
 * @fn    void field_free (field_ptr k);
 * @brief Free the space of the field informations structure
 * @param k Field informations structure
 */
void
field_free (field_ptr k);

/**
 * @fn    void field_get_size (number_ptr c, field_srcptr k);
 * @brief Get the size of the field "k"
 * @param c     Destination number (to store the characteristic)
 * @param k     Field 
 */
void
field_get_size (number_ptr c, field_srcptr k);

/**
 * @fn    void field_get_characteristic (number_ptr c, field_srcptr k);
 * @brief Get the charactristic of the field "k"
 * @param c     Destination number (to store the characteristic)
 * @param k     Field 
 */
void
field_get_characteristic (number_ptr c, field_srcptr k);

 /**
 * @fn    void field_elt_alloc (fe_ptr * dst, field_srcptr k);
 * @brief Allocate space for a field element
 * @param dst  Field element to allocate
 * @param k    Field
 */
void
field_elt_alloc (fe_ptr * dst, field_srcptr k);

 /**
 * @fn    void field_elt_init (fe_ptr dst, field_srcptr k);
 * @brief Initialise the field element
 * @param dst  Field element to initialise
 * @param k    Field
 */
void
field_elt_init (fe_ptr dst, field_srcptr k);

/**
 * @fn    void field_elt_copy (fe_ptr dst, fe_srcptr src, field_srcptr k);
 * @brief Copy src into dst, src and dst must belong to the same field
 * @param dst  Destination field element 
 * @param src  Source field element
 * @param k    Field 
 */
void
field_elt_copy (fe_ptr dst, fe_srcptr src, field_srcptr k);

/**
 * @fn    void field_elt_clear (fe_ptr * src, field_srcptr k);
 * @brief Clear space used by src (remove the action of field_elt_init but let the one of field_elt_alloc)
 * @param src  Field element to clear
 * @param k    Field 
 */
void
field_elt_clear (fe_ptr * src, field_srcptr k);

/**
 * @fn    void field_elt_free (fe_ptr * src, field_srcptr k);
 * @brief Free space used by src
 * @param src  Field element to free
 * @param k    Field 
 */
void
field_elt_free (fe_ptr * src, field_srcptr k);

/**
 * @fn    void field_elt_set_one (fe_ptr dst, field_srcptr k);
 * @brief Set dst to one (or its Montgomery form if Montgomery arithmetic is used)
 * @param dst Destination field element (must be allocated)
 * @param k   Field 
 */
void
field_elt_set_one (fe_ptr dst, field_srcptr k);

/**
 * @fn    void field_elt_set_ui (fe_ptr dst, const block src, const bool isreduced, field_srcptr k, uint8_t stack);
 * @brief Set dst to src, if Montgomery arithmetic is used, is_reduced == false -> transform dst into its Montgomery form
 * @param dst       Destination field element
 * @param src       Source block
 * @param isreduced If Montgomery arithmetic is used, is_reduced == false -> transform dst into its Montgomery form, is_reduced == true -> simple copy
 * @param k         Field 
 * @param stack     Temporary memory stack to use
 */
void
field_elt_set_ui (fe_ptr dst, const block src, const bool isreduced,
        field_srcptr k, uint8_t stack);

/**
 * @brief Set dst to src, if Montgomery arithmetic is used, is_reduced == false -> transform dst into its Montgomery form
 * @param dst       Destination field element
 * @param isreduced If Montgomery arithmetic is used, is_reduced == false -> transform dst into its Montgomery form, is_reduced == true -> simple copy
 * @param k         Field
 * @param stack     Temporary memory stack to use
 * @param n         Number of arguments that follow (use 1 for FP, 1 or 2 for FP2 and 1 or 3 FP3)
 * @param ...       Number for FP, coefficients of the polynomial for FP2 and FP3
 */
void
field_elt_set_number (fe_ptr dst, const bool isreduced, field_srcptr k,
        uint8_t stack, const uint32_t n, ...);

/**
 * @fn    void field_elt_set_fp_elts (fe_ptr dst, field_srcptr k, const uint32_t n, ...);
 * @brief Set dst to src(s) 
 * @param dst       Destination field element
 * @param k         Field
 * @param n         Number of arguments that follow (use 1 for FP, 1 or 2 for FP2 and 1 or 3 FP3)
 * @param ...       fp_elt for FP, coefficients of the polynomial for FP2 and FP3
 */
void
field_elt_set_fp_elts (fe_ptr dst, field_srcptr k, const uint32_t n, ...);

/**
 * @fn    void field_elt_set_str (fe_ptr dst, const char *str, const uint8_t base, const bool isreduced, field_ptr k, uint8_t stack); 
 * @brief Set dst to str, if Montgomery arithmetic is used, is_reduced == false -> transform dst into its Montgomery form
 * @param dst       Destination field element
 * @param str       Source number under string format in base "base". If FP2 or FP3 is use the polynomials coefficients are given separated by a "," without space.
 * @param base      Base used to write str, can be 2, 10 or 16
 * @param isreduced If Montgomery arithmetic is used, is_reduced == false -> transform dst into its Montgomery form, is_reduced == true -> simple copy
 * @param k         Field
 * @param stack     Temporary memory stack to use
 */
void
field_elt_set_str (fe_ptr dst, const char *str, const uint8_t base, 
        const bool isreduced, field_srcptr k, uint8_t stack); 

/**
 * @fn    void field_elt_random (fe_ptr dst, field_ptr k, uint8_t stack);
 * @brief Set dst to a random element of k, the random process is chosen at the MHELL initialisation
 * @param dst   Destination field element
 * @param k     Field 
 * @param stack Temporary memory stack to use
 */
void
field_elt_random (fe_ptr dst, field_srcptr k, uint8_t stack);
 
/**
 * @fn    void field_elt_lift (fe_ptr dst, fe_srcptr src, field_srcptr k, uint8_t stack);
 * @brief If Montgomery arithmetic is used, lift src (which is into Montgomery form) to classical fp (or its coefficients). 
 * @param dst   Destination field element (in FP or with coefficients in Fp for FP2 and FP3)
 * @param src   Source field element (eventually under Montgomery form)
 * @param k     Field 
 * @param stack Temporary memory stack to use
 */
void
field_elt_lift (fe_ptr dst, fe_srcptr src, field_srcptr k, uint8_t stack);

/**
 * @fn    void field_elt_get_number (number_ptr dst, fe_srcptr src, uint8_t pos, field_srcptr k, uint8_t stack);
 * @brief If Montgomery arithmetic is used, lift src (which is into Montgomery form) to classical fp (or its coefficients). 
 * @param dst   Destination number
 * @param src   Source field element (eventually under Montgomery form)
 * @param pos   1 for FP, 1 or 2 for FP2, 1 or 2 or 3 for FP3 to get the wanted part of the field element.
 * @param k     Field 
 * @param stack Temporary memory stack to use
 */
void
field_elt_get_number (number_ptr dst, fe_srcptr src, uint8_t pos, field_srcptr k, uint8_t stack);

/**
 * @fn    void field_elt_get_fp_elt (fe_ptr dst, fe_srcptr src, uint8_t pos, field_srcptr k)
 * @brief Get the field_elt in position 1, 2 or 3 of the src element. Rmq: if src element is in FP, it is just a copy.
 * @param dst   Destination field element in FP
 * @param src   Source field element
 * @param pos   1 for FP, 1 or 2 for FP2, 1 or 2 or 3 for FP3 to get the wanted part of the field element.
 * @param k     Field from src
 */
void
field_elt_get_fp_elt (fe_ptr dst, fe_srcptr src, uint8_t pos, field_srcptr k);
 
/**
 * @fn    void field_str (char **str, field_srcptr k, const uint8_t base, uint8_t stack);
 * @brief Converts k to string format in base specified by base.
 * @param str   Destination string, allocated by the function, must be free by the user.
 * @param k     Field 
 * @param base  Base to use to write str, can be 2, 10 or 16
 * @param stack Temporary memory stack to use
 */
void
field_str (char **str, field_srcptr k, const uint8_t base, uint8_t stack);
 
/**
 * @fn    void field_print (field_srcptr k, const uint8_t base, uint8_t stack);
 * @brief Print the different fields of the structure pointed by k.
 * @param k     Field 
 * @param base  Base to use to print k, can be 2, 10 or 16
 * @param stack Temporary memory stack to use
 */
void
field_print (field_srcptr k, const uint8_t base, uint8_t stack);

/**
 * @fn    void field_elt_str (char **str, fe_srcptr src, const uint8_t base, const bool lift, field_srcptr k, uint8_t stack);
 * @brief Converts src to string format in base specified by base.
 * @param str   Destination string, allocated by the function, must be free by the user.
 * @param src   Source field element (under Montgomery form if Montgomery arithmetic is used)
 * @param base  Base to use to write str, can be 2, 10 or 16
 * @param lift  If Montgomery arithmetic is used, lift == true -> lift src into Fp (or its coefficients to FP) before conversion
 * @param k     Field 
 * @param stack Temporary memory stack to use
 */
void
field_elt_str (char **str, fe_srcptr src, const uint8_t base, const bool lift, field_srcptr k, uint8_t stack);

/**
 * @fn    void field_elt_print (fe_srcptr src, const uint8_t base, const bool lift, field_srcptr k, uint8_t stack);
 * @brief Print src in base specified by base.
 * @param src   Source field element (under Montgomery form if Montgomery arithmetic is used)
 * @param base  Base to use to write str, can be 2, 10 or 16
 * @param lift  If Montgomery arithmetic is used, lift == true -> lift src into Fp (or its coefficients to FP) before conversion
 * @param k     Field 
 * @param stack Temporary memory stack to use
 */
void
field_elt_print (fe_srcptr src, const uint8_t base, const bool lift, field_srcptr k, uint8_t stack);

/**
 * @fn     bool field_isequal (field_srcptr k1, field_srcptr k2);
 * @brief  Test if k1 and k2 are equals
 * @param  k1 Field
 * @param  k2 Field
 * @return true  if k1 == k2
 * @return false if k1 != k2
 */
bool 
field_isequal (field_srcptr k1, field_srcptr k2);


/*************************COMPARISON AND LOGICAL******************************/

/**
 * @fn     int8_t field_elt_cmp (fe_srcptr src1, fe_srcptr src2, field_srcptr k);
 * @brief  Compare src1 and src2
 * @param  src1 Source field element
 * @param  src2 Source field element
 * @param  k    Field 
 * @return -1 if src1 < src2
 * @return  0 if src1 == src2
 * @return  1 if src1 > src2
 */
int8_t
field_elt_cmp (fe_srcptr src1, fe_srcptr src2, field_srcptr k);

/**
 * @fn     int8_t field_elt_cmp_fp_elt (fe_srcptr src1, fe_srcptr src2, field_srcptr k);
 * @brief  Compare src1 and src2
 * @param  src1 Source field element
 * @param  src2 Source field element in FP
 * @param  k    Field 
 * @return -1 if src1 < src2
 * @return  0 if src1 == src2
 * @return  1 if src1 > src2
 */
int8_t
field_elt_cmp_fp_elt (fe_srcptr src1, fe_srcptr src2, field_srcptr k);

/**
 * @fn     bool field_elt_isone (fe_srcptr src, field_srcptr k);
 * @brief  Test if src is one
 * @param  src Source field element
 * @param  k   Field 
 * @return true if scr is one, false otherwise
 */
bool
field_elt_isone (fe_srcptr src, field_srcptr k);

/**
 * @fn     bool field_elt_iszero (fe_srcptr src, field_srcptr k);
 * @brief  Test if src is zero
 * @param  src Source field element
 * @param  k   Field 
 * @return true if scr is zero, false otherwise
 */
static inline bool
field_elt_iszero (fe_srcptr src, field_srcptr k)
{
    bool res = false;
    switch(k->type)
    {
        case FP :
            res = fp_elt_iszero((fp_elt_srcptr)(*src), k->param);
            break;

        case FP2 :
            res = fp2_elt_iszero((fp2_elt_srcptr)src, k->param);
            break;

        case FP3 :
            res = fp3_elt_iszero((fp3_elt_srcptr)src, k->param);
            break;
    }
    return res;
}


/***************************ADDITION SUBTRACTION******************************/

/**
 * @fn    void field_elt_inc (fe_ptr dst, fe_srcptr src, field_srcptr k);
 * @brief Set dst <- src + 1
 * @param dst   To store the result
 * @param src   Source field element
 * @param k     Field 
 */
static inline void
field_elt_inc (fe_ptr dst, fe_srcptr src, field_srcptr k)
{
    switch(k->type)
    {
        case FP :
            fp_elt_inc((fp_elt_ptr)(*dst), (fp_elt_srcptr)(*src), k->param);
            break;
        
        case FP2 :
            fp2_elt_inc((fp2_elt_ptr)dst, (fp2_elt_srcptr)src, k->param);
            break;

        case FP3 :
            fp3_elt_inc((fp3_elt_ptr)dst, (fp3_elt_srcptr)src, k->param);
            break;
    }
}

/**
 * @fn    void field_elt_add (fe_ptr dst, fe_srcptr src1, fe_srcptr src2, field_srcptr k);
 * @brief Set dst <- src1 + src2
 * @param dst   To store the result
 * @param src1  Source field element
 * @param src2  Source field element
 * @param k     Field 
 */
static inline void
field_elt_add (fe_ptr dst, fe_srcptr src1, fe_srcptr src2, field_srcptr k)
{
    switch(k->type)
    {
        case FP :
            fp_elt_add((fp_elt_ptr)(*dst), (fp_elt_srcptr)(*src1), (fp_elt_srcptr)(*src2), k->param);
            break;
 
        case FP2 :
            fp2_elt_add((fp2_elt_ptr)dst, (fp2_elt_srcptr)src1, (fp2_elt_srcptr)src2, k->param);
            break;
 
        case FP3 :
            fp3_elt_add((fp3_elt_ptr)dst, (fp3_elt_srcptr)src1, (fp3_elt_srcptr)src2, k->param);
            break;
    }
}

/**
 * @fn    void field_elt_add_fp_elt (fe_ptr dst, fe_srcptr src1, fe_srcptr src2, field_srcptr k);
 * @brief Set dst <- src1 + src2
 * @param dst   To store the result
 * @param src1  Source field element
 * @param src2  Source field element in FP
 * @param k     Field 
 */
static inline void
field_elt_add_fp_elt (fe_ptr dst, fe_srcptr src1, fe_srcptr src2, field_srcptr k)
{
    switch(k->type)
    {
        case FP :
            fp_elt_add((fp_elt_ptr)(*dst), (fp_elt_srcptr)(*src1), (fp_elt_srcptr)(*src2), k->param);
            break;
 
        case FP2 :
            fp2_elt_add_fp_elt((fp2_elt_ptr)dst, (fp2_elt_srcptr)src1, (fp_elt_srcptr)(*src2), k->param);
            break;
 
        case FP3 :
            fp3_elt_add_fp_elt((fp3_elt_ptr)dst, (fp3_elt_srcptr)src1, (fp_elt_srcptr)(*src2), k->param);
            break;
    }
}

/**
 * @fn    void field_elt_dec (fe_ptr dst, fe_srcptr src, field_srcptr k);
 * @brief Set dst <- src - 1
 * @param dst   To store the result
 * @param src   Source field element
 * @param k     Field 
 */
static inline void
field_elt_dec (fe_ptr dst, fe_srcptr src, field_srcptr k)
{
    switch(k->type)
    {
        case FP :
            fp_elt_dec((fp_elt_ptr)(*dst), (fp_elt_srcptr)(*src), k->param);
            break;
        
        case FP2 :
            fp2_elt_dec((fp2_elt_ptr)dst, (fp2_elt_srcptr)src, k->param);
            break;

        case FP3 :
            fp3_elt_dec((fp3_elt_ptr)dst, (fp3_elt_srcptr)src, k->param);
            break;
    }
}

/**
 * @fn    void field_elt_sub (fe_ptr dst, fe_srcptr src1, fe_srcptr src2, field_srcptr k);
 * @brief Set dst <- src1 - src2
 * @param dst   To store the result
 * @param src1  Source field element
 * @param src2  Source field element
 * @param k     Field 
 */
static inline void
field_elt_sub (fe_ptr dst, fe_srcptr src1, fe_srcptr src2, field_srcptr k)
{
    switch(k->type)
    {
        case FP :
            fp_elt_sub((fp_elt_ptr)(*dst), (fp_elt_srcptr)(*src1), (fp_elt_srcptr)(*src2), k->param);
            break;

        case FP2 :
            fp2_elt_sub((fp2_elt_ptr)dst, (fp2_elt_srcptr)src1, (fp2_elt_srcptr)src2, k->param);
            break;

        case FP3 :
            fp3_elt_sub((fp3_elt_ptr)dst, (fp3_elt_srcptr)src1, (fp3_elt_srcptr)src2, k->param);
            break;
    }
}

/**
 * @fn    void field_elt_sub_fp_elt (fe_ptr dst, fe_srcptr src1, fe_srcptr src2, field_srcptr k);
 * @brief Set dst <- src1 - src2
 * @param dst   To store the result
 * @param src1  Source field element
 * @param src2  Source field element in FP
 * @param k     Field 
 */
static inline void
field_elt_sub_fp_elt (fe_ptr dst, fe_srcptr src1, fe_srcptr src2, field_srcptr k)
{
    switch(k->type)
    {
        case FP :
            fp_elt_sub((fp_elt_ptr)(*dst), (fp_elt_srcptr)(*src1), (fp_elt_srcptr)(*src2), k->param);
            break;

        case FP2 :
            fp2_elt_sub_fp_elt((fp2_elt_ptr)dst, (fp2_elt_srcptr)src1, (fp_elt_srcptr)(*src2), k->param);
            break;

        case FP3 :
            fp3_elt_sub_fp_elt((fp3_elt_ptr)dst, (fp3_elt_srcptr)src1, (fp_elt_srcptr)(*src2), k->param);
            break;
    }
}

/**
 * @fn    void field_elt_neg (fe_ptr dst, fe_srcptr src, field_srcptr k);
 * @brief Set dst <- (-src)
 * @param dst   To store the result
 * @param src   Source field element
 * @param k     Field 
 */
static inline void
field_elt_neg (fe_ptr dst, fe_srcptr src, field_srcptr k)
{
    switch(k->type)
    {
        case FP :
            fp_elt_neg((fp_elt_ptr)(*dst), (fp_elt_srcptr)(*src), k->param);
            break;

        case FP2 :
            fp2_elt_neg((fp2_elt_ptr)dst, (fp2_elt_srcptr)src, k->param);
            break;

        case FP3 :
            fp3_elt_neg((fp3_elt_ptr)dst, (fp3_elt_srcptr)src, k->param);
            break;
    }
}

/**
 * @fn    void field_elt_neg_fp_elt (fe_ptr dst, fe_srcptr src, field_srcptr k);
 * @brief Set dst <- (-src)
 * @param dst   To store the result
 * @param src   Source field element in FP
 * @param k     Field 
 */
static inline void
field_elt_neg_fp_elt (fe_ptr dst, fe_srcptr src, field_srcptr k)
{
    switch(k->type)
    {
        case FP :
            fp_elt_neg((fp_elt_ptr)(*dst), (fp_elt_srcptr)(*src), k->param);
            break;

        case FP2 :
            fp2_elt_neg_fp_elt((fp2_elt_ptr)dst, (fp_elt_srcptr)(*src), k->param);
            break;

        case FP3 :
            fp3_elt_neg_fp_elt((fp3_elt_ptr)dst, (fp_elt_srcptr)(*src), k->param);
            break;
    }
}



/*******************************MULTIPLICATION********************************/

/**
 * @fn    void field_elt_mul (fe_ptr dst, fe_srcptr src1, fe_srcptr src2, field_srcptr k, uint8_t stack)
 * @brief Set dst <- src1 * src2, if Montgomery arithmetic is used, the Montgomery multiplication will be used !
 * @param dst   To store the result
 * @param src1  Source field element, must be of the same type than dst (FP, FP2 or FP3)
 * @param src2  Source field element, can be of the same type than dst (FP, FP2, FP3) or FP.
 * @param k     Field 
 * @param stack Temporary memory stack to use
 */
static inline void
field_elt_mul (fe_ptr dst, fe_srcptr src1, fe_srcptr src2, field_srcptr k, uint8_t stack)
{
    switch(k->type)
    {
        case FP :
            fp_elt_mul((fp_elt_ptr)(*dst), (fp_elt_srcptr)(*src1), (fp_elt_srcptr)(*src2), k->param, stack);
            break;
 
        case FP2 :
            fp2_elt_mul((fp2_elt_ptr)dst, (fp2_elt_srcptr)src1, (fp2_elt_srcptr)src2, k->param, stack);
            break;
 
        case FP3 :
            fp3_elt_mul((fp3_elt_ptr)dst, (fp3_elt_srcptr)src1, (fp3_elt_srcptr)src2, k->param, stack);
            break;
    }
}

/**
 * @fn    void field_elt_mul_fp_elt (fe_ptr dst, fe_srcptr src1, fe_srcptr src2, field_srcptr k, uint8_t stack);
 * @brief Set dst <- src1 * src2, if Montgomery arithmetic is used, the Montgomery multiplication will be used !
 * @param dst   To store the result
 * @param src1  Source field element, must be of the same type than dst (FP, FP2 or FP3)
 * @param src2  Source field element in FP.
 * @param k     Field 
 * @param stack Temporary memory stack to use
 */
static inline void
field_elt_mul_fp_elt (fe_ptr dst, fe_srcptr src1, fe_srcptr src2, field_srcptr k, uint8_t stack)
{
    switch(k->type)
    {
        case FP :
            fp_elt_mul((fp_elt_ptr)(*dst), (fp_elt_srcptr)(*src1), (fp_elt_srcptr)(*src2), k->param, stack);
            break;
 
        case FP2 :
            fp2_elt_mul_fp_elt((fp2_elt_ptr)dst, (fp2_elt_srcptr)src1, (fp_elt_srcptr)(*src2), k->param, stack);
            break;
 
        case FP3 :
            fp3_elt_mul_fp_elt((fp3_elt_ptr)dst, (fp3_elt_srcptr)src1, (fp_elt_srcptr)(*src2), k->param, stack);
            break;
    }
}

/**
 * @fn    void field_elt_mul2 (fe_ptr dst, fe_srcptr src, field_srcptr k);
 * @brief Set dst <- 2 * src.
 * @param dst   To store the result
 * @param src   Source field element, must be of the same type than dst (FP, FP2 or FP3)
 * @param k     Field 
 */
static inline void
field_elt_mul2 (fe_ptr dst, fe_srcptr src, field_srcptr k)
{
    switch(k->type)
    {
        case FP :
            fp_elt_mul2((fp_elt_ptr)(*dst), (fp_elt_srcptr)(*src), k->param);
            break;
 
        case FP2 :
            fp2_elt_mul2((fp2_elt_ptr)dst, (fp2_elt_srcptr)src, k->param);
            break;
 
        case FP3 :
            fp3_elt_mul2((fp3_elt_ptr)dst, (fp3_elt_srcptr)src, k->param);
            break;
    }
}

/**
 * @fn    void field_elt_mul4 (fe_ptr dst, fe_srcptr src, field_srcptr k);
 * @brief Set dst <- 4 * src.
 * @param dst   To store the result
 * @param src   Source field element, must be of the same type than dst (FP, FP2 or FP3)
 * @param k     Field 
 */
static inline void
field_elt_mul4 (fe_ptr dst, fe_srcptr src, field_srcptr k)
{
    switch(k->type)
    {
        case FP :
            fp_elt_mul4((fp_elt_ptr)(*dst), (fp_elt_srcptr)(*src), k->param);
            break;
 
        case FP2 :
            fp2_elt_mul4((fp2_elt_ptr)dst, (fp2_elt_srcptr)src, k->param);
            break;
 
        case FP3 :
            fp3_elt_mul4((fp3_elt_ptr)dst, (fp3_elt_srcptr)src, k->param);
            break;
    }
}

/**
 * @fn    void field_elt_mul8 (fe_ptr dst, fe_srcptr src, field_srcptr k);
 * @brief Set dst <- 8 * src.
 * @param dst   To store the result
 * @param src   Source field element, must be of the same type than dst (FP, FP2 or FP3)
 * @param k     Field 
 */
static inline void
field_elt_mul8 (fe_ptr dst, fe_srcptr src, field_srcptr k)
{
    switch(k->type)
    {
        case FP :
            fp_elt_mul8((fp_elt_ptr)(*dst), (fp_elt_srcptr)(*src), k->param);
            break;
 
        case FP2 :
            fp2_elt_mul8((fp2_elt_ptr)dst, (fp2_elt_srcptr)src, k->param);
            break;
 
        case FP3 :
            fp3_elt_mul8((fp3_elt_ptr)dst, (fp3_elt_srcptr)src, k->param);
            break;
    }
}

/**
 * @fn    void field_elt_mul3 (fe_ptr dst, fe_srcptr src, field_srcptr k, uint8_t stack);
 * @brief Set dst <- 3 * src.
 * @param dst   To store the result
 * @param src   Source field element, must be of the same type than dst (FP, FP2 or FP3)
 * @param k     Field 
 * @param stack Temporary memory stack to use
 */
static inline void
field_elt_mul3 (fe_ptr dst, fe_srcptr src, field_srcptr k, uint8_t stack)
{
    switch(k->type)
    {
        case FP :
            fp_elt_mul3((fp_elt_ptr)(*dst), (fp_elt_srcptr)(*src), k->param, stack);
            break;
 
        case FP2 :
            fp2_elt_mul3((fp2_elt_ptr)dst, (fp2_elt_srcptr)src, k->param, stack);
            break;
 
        case FP3 :
            fp3_elt_mul3((fp3_elt_ptr)dst, (fp3_elt_srcptr)src, k->param, stack);
            break;
    }
}

/**
 * @fn    void field_elt_sqr (fe_ptr dst, fe_srcptr src, field_srcptr k, uint8_t stack);
 * @brief Set dst <- src^2
 * @param dst   To store the result
 * @param src   Source field element
 * @param k     Field 
 * @param stack Temporary memory stack to use
 */
static inline void
field_elt_sqr (fe_ptr dst, fe_srcptr src, field_srcptr k, uint8_t stack)
{
    switch(k->type)
    {
        case FP :
            fp_elt_sqr((fp_elt_ptr)(*dst), (fp_elt_srcptr)(*src), k->param, stack);
            break;

        case FP2 :
            fp2_elt_sqr((fp2_elt_ptr)dst, (fp2_elt_srcptr)src, k->param, stack);
            break;

        case FP3 :
            fp3_elt_sqr((fp3_elt_ptr)dst, (fp3_elt_srcptr)src, k->param, stack);
            break;
    }
}

/**
 * @fn    void field_elt_sqr_fp_elt (fe_ptr dst, fe_srcptr src, field_srcptr k, uint8_t stack);
 * @brief Set dst <- src^2
 * @param dst   To store the result
 * @param src   Source field element in FP
 * @param k     Field 
 * @param stack Temporary memory stack to use
 */
static inline void
field_elt_sqr_fp_elt (fe_ptr dst, fe_srcptr src, field_srcptr k, uint8_t stack)
{
    switch(k->type)
    {
        case FP :
            fp_elt_sqr((fp_elt_ptr)(*dst), (fp_elt_srcptr)(*src), k->param, stack);
            break;

        case FP2 :
            fp2_elt_sqr_fp_elt((fp2_elt_ptr)dst, (fp_elt_srcptr)(*src), k->param, stack);
            break;

        case FP3 :
            fp3_elt_sqr_fp_elt((fp3_elt_ptr)dst, (fp_elt_srcptr)(*src), k->param, stack);
            break;
    }
}

/**
 * @fn    void field_elt_pow_ui (fe_ptr dst, fe_srcptr src, const block n, field_srcptr k, uint8_t stack);
 * @brief Set dst <- src^n
 * @param dst   To store the result
 * @param src   Source field element
 * @param n     Source block
 * @param k     Field 
 * @param stack Temporary memory stack to use
 */
void
field_elt_pow_ui (fe_ptr dst, fe_srcptr src, const block n, field_srcptr k, uint8_t stack);

/**
 * @fn    void field_elt_pow_number (fe_ptr dst, fe_srcptr src, number_srcptr n, field_srcptr k, uint8_t stack);
 * @brief Set dst <- src^n
 * @param dst   To store the result
 * @param src   Source field element
 * @param n     Source number
 * @param k     Field 
 * @param stack Temporary memory stack to use
 */
void
field_elt_pow_number (fe_ptr dst, fe_srcptr src, number_srcptr n, field_srcptr k, uint8_t stack);

/**
 * @fn     bool field_elt_issquare (fe_srcptr src, field_srcptr k, uint8_t stack);
 * @brief  Test if src is a square using the Lengendre symbol 
 * @param  src   Source field element
 * @param  k     Field 
 * @param  stack Temporary memory stack to use
 * @return true  if src is a square in src->k, false otherwise
 */
bool
field_elt_issquare (fe_srcptr src, field_srcptr k, uint8_t stack);

/**
 * @fn     bool field_elt_ispower_ui (fe_srcptr src, const block n, field_srcptr k, uint8_t stack);
 * @brief  Test if src is a n-power in src->k
 * @param  src   Source field element
 * @param  n     Source block
 * @param  k     Field 
 * @param  stack Temporary memory stack to use
 * @return 2 if all element in src->k are some n-power
 * @return 1 if src is a n-power in src->k
 * @return false otherwise
 */
int8_t
field_elt_ispower_ui (fe_srcptr src, const block n, field_srcptr k, uint8_t stack);

/**
 * @fn     bool field_elt_ispower_number (fe_srcptr src, number_srcptr n, field_srcptr k, uint8_t stack);
 * @brief  Test if src is a n-power in src->k
 * @param  src   Source field element
 * @param  n     Source number
 * @param  k     Field 
 * @param  stack Temporary memory stack to use
 * @return 2 if all element in src->k are some n-power
 * @return 1 if src is a n-power in src->k
 * @return false otherwise
 */
int8_t
field_elt_ispower_number (fe_srcptr src, number_srcptr n, field_srcptr k, uint8_t stack);

/**
 * @fn     void field_find_nonsquare (fe_ptr dst, field_ptr k, uint8_t stack);
 * @brief  Look for a random non square element in k
 * @param  dst   Destination field element
 * @param  k     Field
 * @param  stack Temporary memory stack to use
 */
void 
field_find_nonsquare (fe_ptr dst, field_ptr k, uint8_t stack);

/**
 * @fn     void field_find_nonpower_ui (fe_ptr dst, const block n, field_ptr k, uint8_t stack);
 * @brief  Look for a random non n-power element in k
 * @param  dst   Destination field element
 * @param  n     Source block
 * @param  k     Field
 * @param  stack Temporary memory stack to use
 * @return true if a n-power is found in k
 * @return false if k does not have any n-power
 */
bool
field_find_nonpower_ui (fe_ptr dst, const block n, field_ptr k, uint8_t stack);

/**
 * @fn     void field_find_nonpower_number (fe_ptr dst, number_srcptr n, field_ptr k, uint8_t stack);
 * @brief  Look for a random non n-power element in k
 * @param  dst   Destination field element
 * @param  n     Source number
 * @param  k     Field
 * @param  stack Temporary memory stack to use
 * @return true if a n-power is found in k
 * @return false if k does not have any n-power
 */
bool
field_find_nonpower_number (fe_ptr dst, number_srcptr n, field_ptr k, uint8_t stack);

/**
 * @fn    void field_elt_sqrt (fe_ptr dst, fe_srcptr src, field_srcptr k, uint8_t stack);
 * @brief Set dst <- src^(1/2)
 * @param dst   To store the result
 * @param src   Source field element
 * @param k     Field 
 * @param stack Temporary memory stack to use
 */
void
field_elt_sqrt (fe_ptr dst, fe_srcptr src, field_srcptr k, uint8_t stack);

/**
 * @fn    void field_elt_cube_root (fe_ptr dst, fe_srcptr src, field_srcptr k, uint8_t stack);
 * @brief Set dst <- src^(1/3)
 * @param dst   To store the result
 * @param src   Source field element
 * @param k     Field 
 * @param stack Temporary memory stack to use
 */
void
field_elt_cube_root (fe_ptr dst, fe_srcptr src, field_srcptr k, uint8_t stack);

/**
 * @fn    void field_elt_unity_nth_root (fe_ptr dst, const block n, field_ptr k, uint8_t stack);
 * @brief Set dst to a non trivial n-th root of unity if it exists (ie n divides order(k)-1), 1 otherwise.
 * @param dst   To store the result
 * @param n     Source block
 * @param param Field
 * @param stack Temporary memory stack to use
 */
void
field_elt_unity_nth_root (fe_ptr dst, const block n, field_ptr k, uint8_t stack);

/**********************************DIVISION***********************************/

/**
 * @fn    void field_elt_inv (fe_ptr dst, fe_srcptr src, field_srcptr k, uint8_t stack);
 * @brief Set dst <- src^(-1)
 * @param dst   To store the result
 * @param src   Source field element
 * @param k     Field 
 * @param stack Temporary memory stack to use
 */
void
field_elt_inv (fe_ptr dst, fe_srcptr src, field_srcptr k, uint8_t stack);

/**
 * @fn    void field_elt_inv_fp_elt (fe_ptr dst, fe_srcptr src, field_srcptr k, uint8_t stack);
 * @brief Set dst <- src^(-1)
 * @param dst   To store the result
 * @param src   Source field element in FP
 * @param k     Field 
 * @param stack Temporary memory stack to use
 */
void
field_elt_inv_fp_elt (fe_ptr dst, fe_srcptr src, field_srcptr k, uint8_t stack);

/**
 * @fn    void field_elt_div (fe_ptr dst, fe_srcptr src1, fe_srcptr src2, field_srcptr k, uint8_t stack);
 * @brief Set dst <- src1 / src2
 * @param dst   To store the result
 * @param src1  Source field element
 * @param src2  Source field element
 * @param k     Field 
 * @param stack Temporary memory stack to use
 */
void
field_elt_div (fe_ptr dst, fe_srcptr src1, fe_srcptr src2, field_srcptr k, uint8_t stack);

/**
 * @fn    void field_elt_div_fp_elt (fe_ptr dst, fe_srcptr src1, fe_srcptr src2, field_srcptr k, uint8_t stack);
 * @brief Set dst <- src1 / src2
 * @param dst   To store the result
 * @param src1  Source field element
 * @param src2  Source field element in FP
 * @param k     Field 
 * @param stack Temporary memory stack to use
 */
void
field_elt_div_fp_elt (fe_ptr dst, fe_srcptr src1, fe_srcptr src2, field_srcptr k, uint8_t stack);

#endif
