/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France) 

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell-fp.h
  * \brief Declaration of primary field functions, if Montgomery arithmetic is used, the Montgomery arithmetic will be used !
  */

#ifndef MPHELL_FP_H
#define MPHELL_FP_H

#include "mphell-number.h"

#if (MPHELL_USE_GMP == 1 || MPHELL_USE_MBEDTLS == 1)
    /**
     * @typedef fp_elt
     * @brief Primary element
     */
    typedef number * fp_elt;
    typedef number fp_elt_t;
    /**
     * @typedef fp_elt_ptr
     * @brief   Define fp_elt_ptr, use in all functions
     */
    typedef number * fp_elt_ptr;

    /**
     * @typedef fp_elt_srcptr
     * @brief   Define fp_elt_srcptr, use in all functions where the parameter is const
     */
    typedef const number * fp_elt_srcptr;
#elif  MPHELL_USE_IPP == 1
    /**
     * @typedef fp_elt
     * @brief Primary element
     */
    typedef IppsGFpElement * fp_elt;
    typedef IppsGFpElement fp_elt_t;
    /**
     * @typedef fp_elt_ptr
     * @brief   Define fp_elt_ptr, use in all functions
     */
    typedef IppsGFpElement * fp_elt_ptr;

    /**
     * @typedef fp_elt_srcptr
     * @brief   Define fp_elt_srcptr, use in all functions where the parameter is const
     */
    typedef const IppsGFpElement * fp_elt_srcptr;
#endif

/**
 * @struct fp_param_t
 * @brief Primary field parameters
 */
#if (MPHELL_USE_GMP == 1 || MPHELL_USE_MBEDTLS == 1)
typedef struct
{
    number p;                       /*!< Modulus (prime number)                                                                                         */
    number pm[7];                   /*!< Modulus multiples                                                                                              */
    /*char * pm2;                     !< Binary representation of p-2 (Fermat little theorem)                                                           */
#if MPHELL_USE_MONTGOMERY == 1
    number R2;                      /*!< Square of the Montgomery basis: R^2 mod p                                                                      */
#if BLOCK_SIZE == 64
    uint64_t invp;                  /*!< p' such that (-p)*p' = 1 mod B, B=2^BLOCK_SIZE                                                                 */
#else
    uint32_t invp;                  /*!< p' such that (-p)*p' = 1 mod B, B=2^BLOCK_SIZE                                                                 */
#endif
#endif
    fp_elt one;                     /*!< Neutral element for the multiplication                                                                         */
    fp_elt non_residue;             /*!< Non square residue                                                                                             */
    number p_odd;                   /*!< Write p-1 as p_odd.(2^p_even)                                                                                  */
    fp_elt gen_2sylow;              /*!< Non quadratic residue power p_odd                                                                              */
    uint32_t p_even;                /*!< Write p-1 as p_odd.(2^p_even)                                                                                  */
    uint8_t size;                   /*!< Number of block required by p                                                                                  */
    fp_elt pool_1[POOL_SIZE_FP];    /*!< Pool of already initialised field element to accelerate field operations, use when STACK_1 is specified        */
    uint8_t i_1;                    /*!< Index for pool 1                                                                                               */
#if MPHELL_USE_MULTITHREADING == 1
    fp_elt pool_2[POOL_SIZE_FP];    /*!< Pool of already initialised field element to accelerate field operations, use when STACK_2 is specified        */
    uint8_t i_2;                    /*!< Index for pool 1                                                                                               */
#endif
}
fp_param_t;
#elif MPHELL_USE_IPP == 1
typedef struct
{
    IppsGFpState * gf;              /*!< First member = field structure of IPPCP                                                                        */
    uint8_t size;                   /*!< Number of block required by p                                                                                  */
    number p;                       /*!< Modulus (prime number)                                                                                         */
    number pm[7];                   /*!< Modulus multiples                                                                                              */
    /*char * pm2;                     !< Binary representation of p-2 (Fermat little theorem)                                                           */
    fp_elt one;                     /*!< Neutral element for the multiplication                                                                         */
    fp_elt pool_1[POOL_SIZE_FP];    /*!< Pool of already initialised field element to accelerate field operations, use when STACK_1 is specified        */
    uint8_t i_1;                    /*!< Index for pool 1                                                                                               */
#if MPHELL_USE_MULTITHREADING == 1
    fp_elt pool_2[POOL_SIZE_FP];    /*!< Pool of already initialised field element to accelerate field operations, use when STACK_2 is specified        */
    uint8_t i_2;                    /*!< Index for pool 1                                                                                               */
#endif
}
fp_param_t;
#endif

/**
 * @typedef fp_param
 * @brief Pointer on a primary field parameters structure
 */
typedef fp_param_t * fp_param;

/**
 * @enum fp_id_e
 * @brief Identifier for known field, use by IPPCP to accelerate the field arithmetic.
 */
enum fp_id_e {
    ARBITRARY,      /*!< Arbitrary field */
    P192R1,         /*!< Nist prime field use for the curve p192r1 */
    P224R1,         /*!< Nist prime field use for the curve p224r1 */
    P256R1,         /*!< Nist prime field use for the curve p256r1 */
    P384R1,         /*!< Nist prime field use for the curve p384r1 */
    P521R1,         /*!< Nist prime field use for the curve p521r1 */
};

/**
 * @typedef fp_id
 * @brief   Identifier for known field, use by IPPCP to accelerate the field arithmetic.
 */
typedef enum fp_id_e fp_id;

void
fp_elt_print (fp_elt_srcptr src, const uint8_t base, const bool lift, const fp_param param, uint8_t stack);

/**************************************TMP************************************/

/**
 * @fn    void fp_elt_get_pool_elt (fp_elt * dst, const fp_param param, uint8_t stack);
 * @brief Get an initialised field element from the pool
 * @param dst    Adress of the temporary element to allocate
 * @param param  Information on prime field 
 * @param stack  Temporary memory stack to use
 */
static inline void
fp_elt_get_pool_elt (fp_elt * dst, const fp_param param, uint8_t stack)
{
#if MPHELL_USE_MULTITHREADING == 0
    MPHELL_ASSERT(stack == STACK_1, "fp_elt_get_pool_elt, unknow stack \n");
    MPHELL_ASSERT(param->i_1 < POOL_SIZE_FP, "fp_elt_get_pool_elt, stack is too small \n");
    *dst = (param->pool_1)[(param->i_1)++];
#elif MPHELL_USE_MULTITHREADING == 1
    if(stack == STACK_1)
    {
        MPHELL_ASSERT(param->i_1 < POOL_SIZE_FP, "fp_elt_get_pool_elt, stack is too small \n");
        *dst = (param->pool_1)[(param->i_1)++];
#if MPHELL_USE_MBEDTLS == 1
#endif
    }
    else if (stack == STACK_2)
    {
        MPHELL_ASSERT(param->i_2 < POOL_SIZE_FP, "fp_elt_get_pool_elt, stack is too small \n");
        *dst = (param->pool_2)[(param->i_2)++];
#if MPHELL_USE_MBEDTLS == 1
#endif
    }
    else
    {
        *dst = 0;
        mphell_error("fp_elt_get_pool_elt, unknow stack \n");
    }
#endif
}

/**
 * @fn    void fp_elt_relax_pool_elt (fp_elt * dst, const fp_param param, uint8_t stack);
 * @brief Relax an initialised field element from the pool
 * @param dst    Adress of the temporary element to free
 * @param param  Information on prime field 
 * @param stack  Temporary memory stack to use
 */
static inline void
fp_elt_relax_pool_elt (fp_elt * dst, const fp_param param, uint8_t stack)
{
#if MPHELL_USE_MULTITHREADING == 0
    MPHELL_ASSERT(stack == STACK_1, "fp_elt_relax_pool_elt, unknow stack \n");
    (param->i_1)--;
    MPHELL_ASSERT(param->i_1 >= 0, "param->i_1 is < 0 in pool 1\n");
#elif MPHELL_USE_MULTITHREADING == 1
    if(stack == STACK_1)
    {
        (param->i_1)--;
        MPHELL_ASSERT(param->i_1 >= 0, "param->i_1 is < 0 in pool 1\n");
    }
    else if (stack == STACK_2)
    {
        (param->i_2)--;
        MPHELL_ASSERT(param->i_2 >= 0, "param->i_2 is < 0 in pool 2\n");
    }
    else
    {
        mphell_error("fp_elt_relax_pool_elt, unknow stack \n");
    }
#endif
}

/************************************SETTERS**********************************/

/**
 * @fn    void fp_alloc (fp_param param, const uint8_t size);
 * @brief Allocate space for the prime field informations structure
 * @param param Structure for the prime field informations to allocate
 * @param size  Size in blocks of the prime field (ex: 4 for a 4 * 64 = 256 bits prime field on a 64 bits architecture)
 */
void
fp_alloc (fp_param param, const uint8_t size);

/**
 * @fn    void fp_create (fp_param param, number_srcptr p, fp_id id, uint8_t stack);
 * @brief Create a prime field of characteristic p
 * @param param Informations structure on prime field to fill (must be allocated)
 * @param p     Characteristic of the prime field
 * @param id    Name of the field, use P192R1, P224R1, P256R1, P384R1, P521R1 for the NIST curve, ARBITRARY otherwise.
 * @param stack Temporary memory stack to use
 */
void
fp_create (fp_param param, number_srcptr p, fp_id id, uint8_t stack);

/**
 * @fn    void fp_copy (fp_param param_res, const fp_param param);
 * @brief Copy the prime field structure param into param_res
 * @param param_res Destination prime field 
 * @param param     Source prime field 
 */
void
fp_copy (fp_param param_res, const fp_param param);
 
/**
 * @fn    void fp_free (fp_param param);
 * @brief Free the space of the prime field informations structure
 * @param param Prime field informations structure
 */
void
fp_free (fp_param param);

/**
 * @fn    void fp_get_characteristic (number_ptr c, const fp_param param);
 * @brief Get the characteristic of the prime field "param"
 * @param c     Destination number (to store the characteristic)
 * @param param Information on prime field 
 */
void
fp_get_characteristic (number_ptr c, const fp_param param);

/*void*/
/*fp_get_pm2 (char * str, const fp_param param);*/
 
/**
 * @fn    void fp_elt_alloc (fp_elt * dst, const fp_param param);
 * @brief Allocate space for a primary field element
 * @param dst   Primary field element to allocate
 * @param param Information on prime field
 */
void
fp_elt_alloc (fp_elt * dst, const fp_param param);

/**
 * @fn    void fp_elt_init (fp_elt_ptr dst, const fp_param param);
 * @brief Initialise a primary field element
 * @param dst   Primary field element to initialise
 * @param param Information on prime field
 */
void
fp_elt_init (fp_elt_ptr dst, const fp_param param);
 
/**
 * @fn    void fp_elt_copy (fp_elt_ptr dst, fp_elt_srcptr src, const fp_param param);
 * @brief Copy src into dst, src and dst must belong to the same Fp
 * @param dst   Destination primary field element 
 * @param src   Source primary field element
 * @param param Information on prime field
 */
void
fp_elt_copy (fp_elt_ptr dst, fp_elt_srcptr src, const fp_param param);

/**
 * @fn    void fp_elt_clear (fp_elt * src);
 * @brief Clear space used by src (remove the action of fp_elt_init but let the one of fp_elt_alloc)
 * @param src Primary field element to free
 */
void
fp_elt_clear (fp_elt * src);

/**
 * @fn    void fp_elt_free (fp_elt * src);
 * @brief Free space used by src
 * @param src Primary field element to free
 */
void
fp_elt_free (fp_elt * src);

/**
 * @fn    void fp_elt_set_one (fp_elt_ptr dst, const fp_param param);
 * @brief Set dst to one (or its Montgomery form if Montgomery arithmetic is used)
 * @param dst   Destination primary field element (must be allocated)
 * @param param Information on prime field 
 */
void
fp_elt_set_one (fp_elt_ptr dst, const fp_param param);

/**
 * @fn    void fp_elt_set_zero (fp_elt_ptr dst, const fp_param param);
 * @brief Set dst to zero
 * @param dst   Destination primary field element (must be allocated)
 * @param param Information on prime field 
 */
void
fp_elt_set_zero (fp_elt_ptr dst, const fp_param param);

/**
 * @fn    void fp_elt_set_ui (fp_elt_ptr dst, const block src, const bool isreduced, const fp_param param, uint8_t stack);
 * @brief Set dst to src, if Montgomery arithmetic is used, is_reduced == false -> transform dst into its Montgomery form
 * @param dst       Destination primary field element
 * @param src       Source block
 * @param isreduced If Montgomery arithmetic is used, is_reduced == false -> transform dst into its Montgomery form, is_reduced == true -> simple copy
 * @param param     Information on prime field 
 * @param stack     Temporary memory stack to use
 */
void
fp_elt_set_ui (fp_elt_ptr dst, const block src, const bool isreduced, 
        const fp_param param, uint8_t stack);

/**
 * @fn    void fp_elt_set_number (fp_elt_ptr dst, number_srcptr src, const bool isreduced, const fp_param param, uint8_t stack);
 * @brief Set dst to src, if Montgomery arithmetic is used, is_reduced == false -> transform dst into its Montgomery form
 * @param dst       Destination primary field element
 * @param src       Source number
 * @param isreduced If Montgomery arithmetic is used, is_reduced == false -> transform dst into its Montgomery form, is_reduced == true -> simple copy
 * @param param     Information on prime field
 * @param stack     Temporary memory stack to use
 */
void
fp_elt_set_number (fp_elt_ptr dst, number_srcptr src, const bool isreduced,
        const fp_param param, uint8_t stack);

/**
 * @fn    void fp_elt_set_str (fp_elt_ptr dst, const char *str, const uint8_t base, const bool isreduced, const fp_param param, uint8_t stack); 
 * @brief Set dst to str, if Montgomery arithmetic is used, is_reduced == false -> transform dst into its Montgomery form
 * @param dst       Destination primary field element
 * @param str       Source number under string format in base "base"
 * @param base      Base used to write str, can be 2, 10 or 16
 * @param isreduced If Montgomery arithmetic is used, is_reduced == false -> transform dst into its Montgomery form, is_reduced == true -> simple copy
 * @param param     Information on prime field
 * @param stack     Temporary memory stack to use
 */
void
fp_elt_set_str (fp_elt_ptr dst, const char *str, const uint8_t base, 
        const bool isreduced, const fp_param param, uint8_t stack); 

/**
 * @fn    void fp_elt_random (fp_elt_ptr dst, const fp_param param, uint8_t stack);
 * @brief Set dst to a random element of Fp, the random process is chosen at the MHELL initialisation
 * @param dst   Destination primary field element
 * @param param Information on prime field 
 * @param stack Temporary memory stack to use
 */
void
fp_elt_random (fp_elt_ptr dst, const fp_param param, uint8_t stack);

/**
 * @fn    void fp_elt_lift (fp_elt_ptr dst, fp_elt_srcptr src, const fp_param param, uint8_t stack);
 * @brief If Montgomery arithmetic is used, lift src (which is into Montgomery form) to classical fp. 
 * @param dst   Destination primary field element (in Fp)
 * @param src   Source primary field element (under Montgomery form)
 * @param param Information on prime field 
 * @param stack Temporary memory stack to use
 */
void
fp_elt_lift (fp_elt_ptr dst, fp_elt_srcptr src, const fp_param param, uint8_t stack);

/**
 * @fn    void fp_elt_get_number (number_ptr dst, fp_elt_srcptr src, const fp_param param, uint8_t stack);
 * @brief If Montgomery arithmetic is used, lift src (which is into Montgomery form) to classical number (in FP). 
 * @param dst   Destination number
 * @param src   Source field element (eventually under Montgomery form)
 * @param param Information on prime field 
 * @param stack Temporary memory stack to use
 */
void
fp_elt_get_number (number_ptr dst, fp_elt_srcptr src, const fp_param param, uint8_t stack);

/**
 * @fn    void fp_str (char **str, const fp_param param, const uint8_t base, uint8_t stack);
 * @brief Converts fp_param param to string format in base specified by base.
 * @param str   Destination string, allocated by the function, must be free by the user.
 * @param param Information on prime field 
 * @param base  Base to use to write str, can be 2, 10 or 16
 * @param stack Temporary memory stack to use
 */
void
fp_str (char **str, const fp_param param, const uint8_t base, uint8_t stack);

/**
 * @fn    void fp_elt_str (char **str, fp_elt_srcptr src, const uint8_t base, const bool lift, const fp_param param, uint8_t stack);
 * @brief Converts src to string format in base specified by base.
 * @param str   Destination string, allocated by the function, must be free by the user.
 * @param src   Source primary field element (under Montgomery form if Montgomery arithmetic is used)
 * @param base  Base to use to write str, can be 2, 10 or 16
 * @param lift  If Montgomery arithmetic is used, lift == true -> lift src into Fp before conversion
 * @param param Information on prime field 
 * @param stack Temporary memory stack to use
 */
void
fp_elt_str (char **str, fp_elt_srcptr src, const uint8_t base, 
        const bool lift, const fp_param param, uint8_t stack);

/*************************COMPARISON AND LOGICAL******************************/

/**
 * @fn     int8_t fp_elt_cmp (fp_elt_srcptr src1, fp_elt_srcptr src2, const fp_param param);
 * @brief  Compare src1 and src2 in Fp
 * @param  src1  Source field element
 * @param  src2  Source field element
 * @param  param Information on prime field 
 * @return -1 if src1 < src2
 * @return  0 if src1 == src2
 * @return  1 if src1 > src2
 */
int8_t
fp_elt_cmp (fp_elt_srcptr src1, fp_elt_srcptr src2, const fp_param param);

/**
 * @fn     bool fp_elt_isone (fp_elt_srcptr src, const fp_param param);
 * @brief  Test if src is one
 * @param  src   Source field element
 * @param  param Information on prime field 
 * @return true if scr is one, false otherwise
 */
bool
fp_elt_isone (fp_elt_srcptr src, const fp_param param);

/**
 * @fn     bool fp_elt_iszero (fp_elt_srcptr src, const fp_param param);
 * @brief  Test if src is zero
 * @param  src   Source field element
 * @param  param Information on prime field 
 * @return true if scr is zero, false otherwise
 */
static inline bool
fp_elt_iszero (fp_elt_srcptr src, const fp_param param)
{
#if (MPHELL_USE_GMP == 1 || MPHELL_USE_MBEDTLS == 1)
    return number_iszero(*src);
#elif  MPHELL_USE_IPP == 1
    int res;
    ippsGFpIsZeroElement(src, &res, param->gf);
    return (res == IPP_IS_EQ);
#endif
}

/***************************ADDITION SUBTRACTION******************************/

#if MPHELL_USE_MBEDTLS == 1
static inline void mpi_sub_hlp_mphell( size_t n, mbedtls_mpi_uint *s, mbedtls_mpi_uint *d )
{
    size_t i;
    mbedtls_mpi_uint c, z;

    for( i = c = 0; i < n; i++, s++, d++ )
    {
        z = ( *d <  c );     *d -=  c;
        c = ( *d < *s ) + z; *d -= *s;
    }

    while( c != 0 )
    {
        z = ( *d < c ); *d -= c;
        c = z; d++;
    }
}

static inline bool mpi_mod_add_mphell(number_ptr dst, number_srcptr mod)
{
    if( mbedtls_mpi_cmp_abs(dst, mod) >= 0 )
    {
        mpi_sub_hlp_mphell(mod->n, mod->p, dst->p );
        return true;
    }
    return false;
}
#endif

/**
 * @fn    void fp_elt_add (fp_elt_ptr dst, fp_elt_srcptr src1, fp_elt_srcptr src2, const fp_param param);
 * @brief Set dst <- src1 + src2
 * @param dst   To store the result
 * @param src1  Source field element
 * @param src2  Source field element
 * @param param Information on prime field 
 */
static inline void
fp_elt_add (fp_elt_ptr dst, fp_elt_srcptr src1, fp_elt_srcptr src2, 
        const fp_param param)
{
#if MPHELL_USE_GMP == 1
    number_add(*dst, *src1, *src2);
    if(number_isgreatereq(*dst, param->p))
    {
        number_sub(*dst, *dst, param->p);
    }
#elif MPHELL_USE_MBEDTLS == 1
    mbedtls_mpi_add_abs(*dst, *src1, *src2);
    mpi_mod_add_mphell(*dst, param->p);
#elif  MPHELL_USE_IPP == 1
    ippsGFpAdd(src1, src2, dst, param->gf);
#endif
}

/**
 * @fn    void fp_elt_inc (fp_elt_ptr dst, fp_elt_srcptr src, const fp_param param);
 * @brief Set dst <- src + 1
 * @param dst   To store the result
 * @param src   Source field element
 * @param param Information on prime field 
 */
static inline void
fp_elt_inc (fp_elt_ptr dst, fp_elt_srcptr src, const fp_param param)
{
#if (MPHELL_USE_GMP == 1 || MPHELL_USE_MBEDTLS == 1)
    fp_elt_add(dst, src, param->one, param);
#elif  MPHELL_USE_IPP == 1
    ippsGFpAdd(src, param->one, dst, param->gf);
#endif
}

/**
 * @fn    void fp_elt_sub (fp_elt_ptr dst, fp_elt_srcptr src1, fp_elt_srcptr src2, const fp_param param);
 * @brief Set dst <- src1 - src2
 * @param dst   To store the result
 * @param src1  Source field element
 * @param src2  Source field element
 * @param param Information on prime field 
 */
static inline void
fp_elt_sub (fp_elt_ptr dst, fp_elt_srcptr src1, fp_elt_srcptr src2,
        const fp_param param)
{
#if MPHELL_USE_GMP == 1
    number_sub(*dst, *src1, *src2);
    if(number_islower_ui(*dst, 0))
    {
        number_add(*dst, *dst, param->p);
    }
#elif MPHELL_USE_MBEDTLS == 1
    if( mbedtls_mpi_cmp_abs( *src1, *src2 ) >= 0 )
    {
        mbedtls_mpi_sub_abs( *dst, *src1, *src2 );
    }
    else
    {
        mbedtls_mpi B;
        mbedtls_mpi_uint p[(*src2)->n];
        B.s = 1;
        B.n = (*src2)->n;
        B.p = p;
        mbedtls_mpi_copy(&B, *src2);
        mpi_sub_hlp_mphell((*src1)->n, (*src1)->p, B.p);
        mbedtls_mpi_copy(*dst, param->p);
        mpi_sub_hlp_mphell(B.n, B.p, (*dst)->p);
    }
#elif  MPHELL_USE_IPP == 1
    ippsGFpSub(src1, src2, dst, param->gf);
#endif
}

/**
 * @fn    void fp_elt_dec (fp_elt_ptr dst, fp_elt_srcptr src, const fp_param param);
 * @brief Set dst <- src - 1
 * @param dst   To store the result
 * @param src   Source field element
 * @param param Information on prime field 
 */
static inline void
fp_elt_dec (fp_elt_ptr dst, fp_elt_srcptr src, const fp_param param)
{
#if (MPHELL_USE_GMP == 1 || MPHELL_USE_MBEDTLS == 1)
    fp_elt_sub(dst, src, param->one, param);
#elif  MPHELL_USE_IPP == 1
    ippsGFpSub(src, param->one, dst, param->gf);
#endif
}

/**
 * @fn    void fp_elt_neg (fp_elt_ptr dst, fp_elt_srcptr src, const fp_param param);
 * @brief Set dst <- (-src) mod p
 * @param dst   To store the result
 * @param src   Source field element
 * @param param Information on prime field 
 */
static inline void
fp_elt_neg (fp_elt_ptr dst, fp_elt_srcptr src, const fp_param param)
{
#if MPHELL_USE_GMP == 1
    mpz_neg(*dst, *src);
    if(number_islower_ui(*dst, 0))
    {
        number_add(*dst, *dst, param->p);
    }
#elif MPHELL_USE_IPP == 1
    ippsGFpNeg(src, dst, param->gf);
#elif MPHELL_USE_MBEDTLS == 1
    number_copy(*dst, *src);
    (*dst)->s = -1;
    number_add(*dst, *dst, param->p);
#endif
}


/*******************************MULTIPLICATION********************************/

/**
 * @fn    void fp_elt_mul (fp_elt_ptr dst, fp_elt_srcptr src1, fp_elt_srcptr src2, const fp_param param, uint8_t stack);
 * @brief Set dst <- src1 * src2, if Montgomery arithmetic is used, the Montgomery multiplication will be used !
 * @param dst  To store the result
 * @param src1  Source field element
 * @param src2  Source field element
 * @param param Information on prime field 
 * @param stack Temporary memory stack to use
 */
static inline void
fp_elt_mul (fp_elt_ptr dst, fp_elt_srcptr src1, fp_elt_srcptr src2, 
        const fp_param param, uint8_t stack)
{ 
#if (MPHELL_USE_GMP == 1 || MPHELL_USE_MBEDTLS == 1)
#if MPHELL_USE_MONTGOMERY == 1
    number_mul_montgomery(*dst, *src1, *src2, param->p, param->invp, stack);
#else
    number_mul_mod(*dst, *src1, *src2, param->p, stack);
#endif
#elif  MPHELL_USE_IPP == 1
    ippsGFpMul(src1, src2, dst, param->gf);
#endif
}

/**
 * @fn    void fp_elt_mul2 (fp_elt_ptr dst, fp_elt_srcptr src, const fp_param param);
 * @brief Set dst <- 2 * src
 * @param dst   To store the result
 * @param src   Source field element
 * @param param Information on prime field 
 */
static inline void
fp_elt_mul2 (fp_elt_ptr dst, fp_elt_srcptr src,
        const fp_param param)
{
#if MPHELL_USE_GMP == 1
    number_lshift(*dst, *src, 1);
    if(number_isgreatereq(*dst, param->p))
    {
        number_sub(*dst, *dst, param->p);
    }
#elif MPHELL_USE_MBEDTLS == 1
    number_lshift(*dst, *src, 1);
    mpi_mod_add_mphell(*dst, param->p);
#elif  MPHELL_USE_IPP == 1
    fp_elt_add(dst, src, src, param);
#endif
}

/**
 * @fn    void fp_elt_mul4 (fp_elt_ptr dst, fp_elt_srcptr src, const fp_param param);
 * @brief Set dst <- 4 * src
 * @param dst   To store the result
 * @param src   Source field element
 * @param param Information on prime field 
 */
static inline void
fp_elt_mul4 (fp_elt_ptr dst, fp_elt_srcptr src,
        const fp_param param)
{
#if MPHELL_USE_GMP == 1
    number_lshift(*dst, *src, 2);
    int i;
    for (i=2; i>=0; i--)
    {
        if(number_isgreatereq(*dst, param->pm[i]))
        {
            number_sub(*dst, *dst, param->pm[i]);
            break;
        }
    }
#elif MPHELL_USE_MBEDTLS == 1
    number_lshift(*dst, *src, 2);
    int i;
    for (i=2; i>=0; i--)
    {
        if(mpi_mod_add_mphell(*dst, param->pm[i]))
        {
            break;
        }
    }
#elif  MPHELL_USE_IPP == 1
    fp_elt_add(dst, src, src, param);
    fp_elt_add(dst, dst, dst, param);
#endif
}

/**
 * @fn    void fp_elt_mul8 (fp_elt_ptr dst, fp_elt_srcptr src, const fp_param param);
 * @brief Set dst <- 8 * src
 * @param dst   To store the result
 * @param src   Source field element
 * @param param Information on prime field 
 */
static inline void
fp_elt_mul8 (fp_elt_ptr dst, fp_elt_srcptr src,
        const fp_param param)
{
#if MPHELL_USE_GMP == 1
    number_lshift(*dst, *src, 3);
    int i;
    for (i=6; i>=0; i--)
    {
        if(number_isgreatereq(*dst, param->pm[i]))
        {
            number_sub(*dst, *dst, param->pm[i]);
            break;
        }
    }
#elif MPHELL_USE_MBEDTLS == 1
    number_lshift(*dst, *src, 3);
    int i;
    for (i=6; i>=0; i--)
    {
        if(mpi_mod_add_mphell(*dst, param->pm[i]))
        {
            break;
        }
    }
#elif  MPHELL_USE_IPP == 1
    fp_elt_add(dst, src, src, param);
    fp_elt_add(dst, dst, dst, param);
    fp_elt_add(dst, dst, dst, param);
#endif
}

/**
 * @fn    void fp_elt_mul3 (fp_elt_ptr dst, fp_elt_srcptr src, const fp_param param, uint8_t stack);
 * @brief Set dst <- 3 * src
 * @param dst   To store the result
 * @param src   Source field element
 * @param param Information on prime field 
 * @param stack Temporary memory stack to use
 */
static inline void
fp_elt_mul3 (fp_elt_ptr dst, fp_elt_srcptr src,
        const fp_param param, uint8_t stack)
{
#if MPHELL_USE_GMP == 1
    fp_elt tmp;
    fp_elt_get_pool_elt(&tmp, param, stack);
    number_lshift(*tmp, *src, 1);
    if(number_isgreatereq(*tmp, param->p))
    {
        number_sub(*tmp, *tmp, param->p);
    }
    fp_elt_add(dst, tmp, src, param); 
    fp_elt_relax_pool_elt(&tmp, param, stack);
#elif MPHELL_USE_MBEDTLS == 1
    fp_elt tmp;
    fp_elt_get_pool_elt(&tmp, param, stack);
    number_lshift(*tmp, *src, 1);
    mpi_mod_add_mphell(*tmp, param->p);
    fp_elt_add(dst, tmp, src, param); 
    fp_elt_relax_pool_elt(&tmp, param, stack);
#elif  MPHELL_USE_IPP == 1
    fp_elt tmp;
    fp_elt_get_pool_elt(&tmp, param, stack);
    fp_elt_add(tmp, src, src, param);
    fp_elt_add(dst, tmp, src, param);
    fp_elt_relax_pool_elt(&tmp, param, stack);
#endif
}

/**
 * @fn    void fp_elt_sqr (fp_elt_ptr dst, fp_elt_srcptr src, const fp_param param, uint8_t stack);
 * @brief Set dst <- src^2
 * @param dst   To store the result
 * @param src   Source field element
 * @param param Information on prime field 
 * @param stack Temporary memory stack to use
 */
static inline void
fp_elt_sqr (fp_elt_ptr dst, fp_elt_srcptr src, const fp_param param, uint8_t stack)
{
#if (MPHELL_USE_GMP == 1 || MPHELL_USE_MBEDTLS == 1)
    fp_elt_mul(dst, src, src, param, stack);
#elif  MPHELL_USE_IPP == 1
    ippsGFpSqr(src, dst, param->gf);
#endif
}

/**
 * @fn    void fp_elt_inv_flt (fp_elt_ptr dst, fp_elt_srcptr src, const fp_param param, uint8_t stack);
 * @brief Set dst <- src^(-1) using Fermat Little Theorem
 * @param dst   To store the result
 * @param src   Source field element
 * @param param Information on prime field 
 * @param stack Temporary memory stack to use
 */
void
fp_elt_inv_flt (fp_elt_ptr dst, fp_elt_srcptr src, const fp_param param, uint8_t stack);


/**
 * @fn    void fp_elt_inv (fp_elt_ptr dst, fp_elt_srcptr src, const fp_param param, uint8_t stack);
 * @brief Set dst <- src^(-1)
 * @param dst   To store the result
 * @param src   Source field element
 * @param param Information on prime field 
 * @param stack Temporary memory stack to use
 */
void
fp_elt_inv (fp_elt_ptr dst, fp_elt_srcptr src, const fp_param param, uint8_t stack);

/**
 * @fn    void fp_elt_div (fp_elt_ptr dst, fp_elt_srcptr src1, fp_elt_srcptr src2, const fp_param param, uint8_t stack);
 * @brief Set dst <- src1 / src2
 * @param dst   To store the result
 * @param src1  Source field element
 * @param src2  Source field element
 * @param param Information on prime field 
 * @param stack Temporary memory stack to use
 */
void
fp_elt_div (fp_elt_ptr dst, fp_elt_srcptr src1, fp_elt_srcptr src2, 
        const fp_param param, uint8_t stack);

/**
 * @fn    void fp_elt_pow_ui (fp_elt_ptr dst, fp_elt_srcptr src, const block n, const fp_param param, uint8_t stack);
 * @brief Set dst <- src^n
 * @param dst   To store the result
 * @param src   Source field element
 * @param n     Source block
 * @param param Information on prime field 
 * @param stack Temporary memory stack to use
 */
void
fp_elt_pow_ui (fp_elt_ptr dst, fp_elt_srcptr src, const block n,
        const fp_param param, uint8_t stack);

/**
 * @fn    void fp_elt_pow_number (fp_elt_ptr dst, fp_elt_srcptr src, number_srcptr n, const fp_param param, uint8_t stack);
 * @brief Set dst <- src^n
 * @param dst   To store the result
 * @param src   Source field element
 * @param n     Source number
 * @param param Information on prime field 
 * @param stack Temporary memory stack to use
 */
void
fp_elt_pow_number (fp_elt_ptr dst, fp_elt_srcptr src, number_srcptr n,
        const fp_param param, uint8_t stack);

/**
 * @fn     bool fp_elt_issquare (fp_elt_srcptr src, const fp_param param, uint8_t stack);
 * @brief  Test if src is a square using the Lengendre symbol 
 * @param  src   Source field element
 * @param  param Information on prime field 
 * @param  stack Temporary memory stack to use
 * @return true if src is a square in Fp, false otherwise
 */
bool
fp_elt_issquare (fp_elt_srcptr src, const fp_param param, uint8_t stack);

/**
 * @fn     bool fp_elt_ispower_ui (fp_elt_srcptr src, const block n, const fp_param param, uint8_t stack);
 * @brief  Test if src is a n-power in Fp
 * @param  src   Source field element
 * @param  n     Source block
 * @param  param Information on prime field 
 * @param  stack Temporary memory stack to use
 * @return 2 if all element in Fp are some n-power
 * @return 1 if src is a n-power in Fp
 * @return false otherwise
 */
int8_t
fp_elt_ispower_ui (fp_elt_srcptr src, const block n, const fp_param param, uint8_t stack);

/**
 * @fn     bool fp_elt_ispower_number (fp_elt_srcptr src, number_srcptr n, const fp_param param, uint8_t stack);
 * @brief  Test if src is a n-power in Fp
 * @param  src   Source field element
 * @param  n     Source number
 * @param  param Information on prime field 
 * @param  stack Temporary memory stack to use
 * @return 2 if all element in Fp are some n-power
 * @return 1 if src is a n-power in Fp
 * @return false otherwise
 */
int8_t
fp_elt_ispower_number (fp_elt_srcptr src, number_srcptr n, const fp_param param, uint8_t stack);


/*******************************ROOTS********************************/

/**
 * @fn    void fp_elt_sqrt (fp_elt_ptr dst, fp_elt_srcptr src, const fp_param param, uint8_t stack);
 * @brief Set dst <- src^(1/2) mod p, using Tonelli–Shanks algorithm
 * @param dst   To store the result
 * @param src   Source field element
 * @param param Information on prime field 
 * @param stack Temporary memory stack to use
 */
void
fp_elt_sqrt (fp_elt_ptr dst, fp_elt_srcptr src, const fp_param param, uint8_t stack);

/**
 * @fn    void fp_elt_cube_root (fp_elt_ptr dst, fp_elt_srcptr src, const fp_param param, uint8_t stack);
 * @brief Set dst <- src^(1/3) mod p
 * @param dst    To store the result
 * @param src    Source field element
 * @param param  Information on prime field 
 * @param stack  Temporary memory stack to use
 */
void
fp_elt_cube_root (fp_elt_ptr dst, fp_elt_srcptr src, const fp_param param, uint8_t stack);

/**
 * @fn    void fp_elt_unity_nth_root (fp_elt_ptr dst, block n, const fp_param param, uint8_t stack);
 * @brief Set dst to a non trivial n-th root of unity if it exists (ie n divides p-1), 1 otherwise.
 * @param dst    To store the result
 * @param n      Source block
 * @param param  Information on prime field 
 * @param stack  Temporary memory stack to use
 */
void
fp_elt_unity_nth_root (fp_elt_ptr dst, const block n, const fp_param param, uint8_t stack);

#endif

