/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France)

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell-fp2.h
  * \brief Declaration of binary field (finite field constructed with an irreducible polynomial of degree 2) functions.
  */

#ifndef MPHELL_FP2_H
#define MPHELL_FP2_H

#include "mphell-fp.h"

/**
 * @struct fp2_elt_t
 * @brief  Quadratic extension field element structure
 */
typedef struct
{
    fp_elt v0;  /*!< Coefficient of degree 0    */
    fp_elt v1;  /*!< Coefficient of degree 1    */
} fp2_elt_t;

/**
 * @typedef fp2_elt
 * @brief   Quadratic extension field element structure
 */
typedef fp2_elt_t fp2_elt;

/**
 * @typedef fp2_elt_ptr
 * @brief   Define fp2_elt_ptr, use in all functions
 */
typedef fp2_elt_t * fp2_elt_ptr;

/**
 * @typedef fp2_elt_srcptr
 * @brief   Define fp2_elt_srcptr, use in all functions where the parameter is const
 */
typedef const fp2_elt_t * fp2_elt_srcptr;

/**
 * @struct fp2_param_t
 * @brief  Quadratic extension field structure
 */
typedef struct
{
    fp_param base;          /*!< Base prime field                           */
    fp_elt non_residue;     /*!< Non quadratic residue                      */
    number q;               /*!< Order of FP2 = p^2                         */
    /*char * pm2;             !< Binary representation of (p-1)p-1 (FLT)    */
    fp2_elt non_res;        /*!< Non square residue                         */
    fp2_elt gen_sylow;      /*!< Non quadratic residue power p_odd          */
    number p_odd;           /*!< Write q-1 as p_odd.(2^p_even)              */
    uint32_t p_even;        /*!< Write q-1 as p_odd.(2^p_even)              */
    uint8_t size;           /*!< Number of block required by p              */
}
fp2_param_t;

/**
 * @typedef fp2_param
 * @brief   Pointer on a quadratic extension field structure
 */
typedef fp2_param_t *fp2_param;

/**************************************TMP************************************/

/**
 * @fn    void fp2_elt_get_pool_elt (fp2_elt * dst, const fp2_param param, uint8_t stack);
 * @brief Allocate and initialise space for a temporary quadratic extension field element
 * @param dst    Temporary element to allocate
 * @param param  Information on the quadratic extension field 
 * @param stack  Temporary memory stack to use
 */
void
fp2_elt_get_pool_elt (fp2_elt * dst, const fp2_param param, uint8_t stack);

/**
 * @fn    void fp2_elt_relax_pool_elt (fp2_elt * dst, const fp2_param param, uint8_t stack);
 * @brief Free space of a temporary quadratic extension field element
 * @param dst    Temporary element to free
 * @param param  Information on the quadratic extension field 
 * @param stack  Temporary memory stack to use
 */
void
fp2_elt_relax_pool_elt (fp2_elt * dst, const fp2_param param, uint8_t stack);

/************************************SETTERS**********************************/

/**
 * @fn    void fp2_alloc (fp2_param param, const fp_param base);
 * @brief Allocate space for the quadratic extension field informations structure
 * @param param Structure for the quadratic extension field informations to allocate
 * @param base  Base prime field
 */
void
fp2_alloc (fp2_param param, const fp_param base);

/**
 * @fn    void fp2_create (fp2_param param, const fp_param base, fp_elt_srcptr non_residue, uint8_t stack);
 * @brief Create a quadratic extension of the field base
 * @param param       Informations structure on quadratic extension field to fill (must be allocated)
 * @param base        Base prime field
 * @param non_residue Non quadratic residue
 * @param stack       Temporary memory stack to use
 */
void
fp2_create (fp2_param param, const fp_param base, fp_elt_srcptr non_residue, uint8_t stack);

/**
 * @fn    void fp2_copy (fp2_param param_res, const fp2_param param);
 * @brief Copy the quadratic extension field structure param into param_res
 * @param param_res Destination quadratic extension field 
 * @param param     Source quadratic extension field 
 */
void
fp2_copy (fp2_param param_res, const fp2_param param);

/**
 * @fn    void fp2_free (fp2_param param);
 * @brief Free the space of the quadratic extension field informations structure
 * @param param Quadratic extension field informations structure
 */
void
fp2_free (fp2_param param);

/**
 * @fn    void fp2_get_characteristic (number_ptr c, const fp2_param param);
 * @brief Get the characteristic of the quadratic extension field "param"
 * @param c     Destination number_ptr (to store the characteristic)
 * @param param Information on quadratic extension field 
 */
void
fp2_get_characteristic (number_ptr c, const fp2_param param);

/**
 * @fn    void fp2_get_size (number_ptr c, const fp2_param param);
 * @brief Get the size of the quadratic extension field "param"
 * @param c     Destination number_ptr (to store the characteristic)
 * @param param Information on quadratic extension field 
 */
void
fp2_get_size (number_ptr c, const fp2_param param);

/*void*/
/*fp2_get_pm2 (char * str, const fp2_param param);*/

/**
 * @fn    void fp2_elt_alloc (fp2_elt * dst, const fp2_param param);
 * @brief Allocate space for a quadratic extension field element
 * @param dst   Quadratic extension field element to allocate
 * @param param Information on quadratic extension field 
 */
void
fp2_elt_alloc (fp2_elt * dst, const fp2_param param);

/**
 * @fn    void fp2_elt_init (fp2_elt_ptr dst, const fp2_param param);
 * @brief Initialise a quadratic extension field element
 * @param dst   Quadratic extension field element to initialise
 * @param param Information on quadratic extension field 
 */
void
fp2_elt_init (fp2_elt_ptr dst, const fp2_param param);

/**
 * @fn    void fp2_elt_copy (fp2_elt_ptr dst, fp2_elt_srcptr src, const fp2_param param);
 * @brief Copy src into dst, src and dst must belong to the same FP2
 * @param dst   Destination quadratic extension field element 
 * @param src   Source quadratic extension field element
 * @param param Information on quadratic extension field 
 */
void
fp2_elt_copy (fp2_elt_ptr dst, fp2_elt_srcptr src, const fp2_param param);

/**
 * @fn    void fp2_elt_clear (fp2_elt * src);
 * @brief Clear space used by src (remove the action of fp2_elt_init but less the one of fp2_elt_alloc)
 * @param src Quadratic extension field element to clear
 */
void
fp2_elt_clear (fp2_elt * src);

/**
 * @fn    void fp2_elt_free (fp2_elt * src);
 * @brief Free space used by src
 * @param src Quadratic extension field element to free
 */
void
fp2_elt_free (fp2_elt * src);

/**
 * @fn    void fp2_elt_set_one (fp2_elt_ptr dst, const fp2_param param);
 * @brief Set dst to one (or its Montgomery form if Montgomery arithmetic is used)
 * @param dst   Destination quadratic extension field element (must be allocated)
 * @param param Information on the quadratic extension field 
 */
void
fp2_elt_set_one (fp2_elt_ptr dst, const fp2_param param);

/**
 * @fn    void fp2_elt_set_ui (fp2_elt_ptr dst, const uint64_t src, const bool isreduced, const fp2_param param, uint8_t stack);
 * @brief Set dst to src, if Montgomery arithmetic is used, is_reduced == false -> transform dst into its Montgomery form
 * @param dst       Destination quadratic extension field element
 * @param src       Source block
 * @param isreduced If Montgomery arithmetic is used, is_reduced == false -> transform dst into its Montgomery form, is_reduced == true -> simple copy
 * @param param     Information on the quadratic extension field 
 * @param stack     Temporary memory stack to use
 */
void
fp2_elt_set_ui (fp2_elt_ptr dst, const uint64_t src, const bool isreduced, 
        const fp2_param param, uint8_t stack);

/**
 * @fn    void fp2_elt_set_number (fp2_elt_ptr dst, number_srcptr src, const bool isreduced, const fp2_param param, uint8_t stack);
 * @brief Set dst to src, if Montgomery arithmetic is used, is_reduced == false -> transform dst into its Montgomery form
 * @param dst       Destination quadratic extension field element
 * @param src       Source number
 * @param isreduced If Montgomery arithmetic is used, is_reduced == false -> transform dst into its Montgomery form, is_reduced == true -> simple copy
 * @param param     Information on the quadratic extension field
 * @param stack     Temporary memory stack to use
 */
void
fp2_elt_set_number (fp2_elt_ptr dst, number_srcptr src, const bool isreduced, 
        const fp2_param param, uint8_t stack);

/**
 * @fn    void fp2_elt_set_fp_elt (fp2_elt_ptr dst, fp_elt_srcptr src, const fp2_param param);
 * @brief Set dst to src
 * @param dst       Destination quadratic extension field element
 * @param src       Source fp_elt
 * @param param     Information on the quadratic extension field
 */
void
fp2_elt_set_fp_elt (fp2_elt_ptr dst, fp_elt_srcptr src, const fp2_param param);

/**
 * @fn    void fp2_elt_set (fp2_elt_ptr dst, number_srcptr src1, number_srcptr src2, const bool isreduced, const fp2_param param, uint8_t stack);
 * @brief Set dst to src1 + src2*x, if Montgomery arithmetic is used, is_reduced == false -> transform dst into its Montgomery form
 * @param dst       Destination quadratic extension field element
 * @param src1      Source number
 * @param src2      Source number
 * @param isreduced If Montgomery arithmetic is used, is_reduced == false -> transform dst into its Montgomery form, is_reduced == true -> simple copy
 * @param param     Information on the quadratic extension field
 * @param stack     Temporary memory stack to use
 */
void
fp2_elt_set (fp2_elt_ptr dst, number_srcptr src1, number_srcptr src2, 
        const bool isreduced, const fp2_param param, uint8_t stack);

/**
 * @fn    void fp2_elt_set_fp_elts (fp2_elt_ptr dst, fp_elt_srcptr src1, fp_elt_srcptr src2, const fp2_param param);
 * @brief Set dst to src1 + src2*x
 * @param dst       Destination quadratic extension field element
 * @param src1      Source fp_elt
 * @param src2      Source fp_elt
 * @param param     Information on the quadratic extension field
 */
void
fp2_elt_set_fp_elts (fp2_elt_ptr dst, fp_elt_srcptr src1, fp_elt_srcptr src2, const fp2_param param);

/**
 * @fn    void fp2_elt_set_str (fp2_elt_ptr dst, const char *str, const uint8_t base, const bool isreduced, const fp2_param param, uint8_t stack);
 * @brief Set dst to str, if Montgomery arithmetic is used, is_reduced == false -> transform dst into its Montgomery form
 * @param dst       Destination quadratic extension field element
 * @param str       Source polynomial coefficient under string format in base "base", separated by a "," without space. ex: "123,456" will create the polynomial 123 + 456.X
 * @param base      Base used to write str, can be 2, 10 or 16
 * @param isreduced If Montgomery arithmetic is used, is_reduced == false -> transform dst into its Montgomery form, is_reduced == true -> simple copy
 * @param param     Information on the quadratic extension field
 * @param stack     Temporary memory stack to use
 */
void
fp2_elt_set_str (fp2_elt_ptr dst, const char *str, const uint8_t base, 
        const bool isreduced, const fp2_param param, uint8_t stack); 

/**
 * @fn    void fp2_elt_random (fp2_elt_ptr dst, const fp2_param param, uint8_t stack);
 * @brief Set dst to a random element of FP2, the random process is chosen at the MHELL initialisation
 * @param dst   Destination quadratic extension field element
 * @param param Information on the quadratic extension field 
 * @param stack Temporary memory stack to use
 */
void
fp2_elt_random (fp2_elt_ptr dst, const fp2_param param, uint8_t stack);

/**
 * @fn    void fp2_elt_lift (fp2_elt_ptr dst, fp2_elt_srcptr src, const fp2_param param, uint8_t stack);
 * @brief If Montgomery arithmetic is used, lift src (which is into Montgomery form) to classical FP2. 
 * @param dst   Destination primary field element (in FP)
 * @param src   Source quadratic extension field element (under Montgomery form)
 * @param param Information on the quadratic extension field 
 * @param stack Temporary memory stack to use
 */
void
fp2_elt_lift (fp2_elt_ptr dst, fp2_elt_srcptr src, const fp2_param param, uint8_t stack);

/**
 * @fn    void fp2_str (char **str, const fp2_param param, const uint8_t base, uint8_t stack);
 * @brief Converts fp2_param param to string format in base specified by base.
 * @param str   Destination string, allocated by the function, must be free by the user.
 * @param param Information on the quadratic extension field 
 * @param base  Base to use to write str, can be 2, 10 or 16
 * @param stack Temporary memory stack to use
 */
void
fp2_str (char **str, const fp2_param param, const uint8_t base, uint8_t stack);

/**
 * @fn    void fp2_elt_str (char **str, fp2_elt_srcptr src, const uint8_t base, const bool lift, const fp2_param param, uint8_t stack);
 * @brief Converts src to string format in base specified by base.
 * @param str   Destination string, allocated by the function, must be free by the user. Polynomial coefficient are written separated by ",". ex: 123 + 456.X will be written as "123,456"
 * @param src   Source quadratic extension field element (under Montgomery form if Montgomery arithmetic is used)
 * @param base  Base to use to write str, can be 2, 10 or 16
 * @param lift  If Montgomery arithmetic is used, lift == true -> lift src into FP2 before conversion
 * @param param Information on the quadratic extension field
 * @param stack Temporary memory stack to use
 */
void
fp2_elt_str (char **str, fp2_elt_srcptr src, const uint8_t base, 
        const bool lift, const fp2_param param, uint8_t stack);

/*************************COMPARISON AND LOGICAL******************************/

/**
 * @fn     int8_t fp2_elt_cmp_fp_elt (fp2_elt_srcptr src1, fp_elt_srcptr src2, const fp2_param param);
 * @brief  Compare src1 and src2 in FP2
 * @param  src1  Source field element in FP2
 * @param  src2  Source number_ptr in FP
 * @param  param Information on the quadratic extension field 
 * @return -1 if src1 < src2
 * @return  0 if src1 == src2
 * @return  1 if src1 > src2
 */
int8_t
fp2_elt_cmp_fp_elt (fp2_elt_srcptr src1, fp_elt_srcptr src2, const fp2_param param);

/**
 * @fn     int8_t fp2_elt_cmp (fp2_elt_srcptr src1, fp2_elt_srcptr src2, const fp2_param param);
 * @brief  Compare src1 and src2 in FP2
 * @param  src1  Source field element in FP2
 * @param  src2  Source field element in FP2
 * @param  param Information on the quadratic extension field 
 * @return -1 if src1 < src2
 * @return  0 if src1 == src2
 * @return  1 if src1 > src2
 */
int8_t
fp2_elt_cmp (fp2_elt_srcptr src1, fp2_elt_srcptr src2, const fp2_param param);

/**
 * @fn     bool fp2_elt_isone (fp2_elt_srcptr src, const fp2_param param);
 * @brief  Test if src is one
 * @param  src   Source field element
 * @param  param Information on the quadratic extension field 
 * @return true  if scr is one, false otherwise
 */
bool
fp2_elt_isone (fp2_elt_srcptr src, const fp2_param param);

/**
 * @fn     bool fp2_elt_iszero (fp2_elt_srcptr src, const fp2_param param);
 * @brief  Test if src is zero
 * @param  src   Source field element
 * @param  param Information on the quadratic extension field 
 * @return true if scr is zero, false otherwise
 */
bool
fp2_elt_iszero (fp2_elt_srcptr src, const fp2_param param);

/***************************ADDITION SUBTRACTION******************************/

/**
 * @fn    void fp2_elt_inc (fp2_elt_ptr dst, fp2_elt_srcptr src, const fp2_param param);
 * @brief Set dst <- src + 1
 * @param dst   To store the result
 * @param src   Source field element
 * @param param Information on the quadratic extension field 
 */
void
fp2_elt_inc (fp2_elt_ptr dst, fp2_elt_srcptr src, const fp2_param param);

/**
 * @fn    void fp2_elt_add_fp_elt (fp2_elt_ptr dst, fp2_elt_srcptr src1, fp_elt_srcptr src2, const fp2_param param);
 * @brief Set dst <- src1 + src2
 * @param dst   To store the result
 * @param src1  Source field element
 * @param src2  Source field element in the prime field
 * @param param Information on the quadratic extension field 
 */
void
fp2_elt_add_fp_elt (fp2_elt_ptr dst, fp2_elt_srcptr src1, fp_elt_srcptr src2, 
        const fp2_param param);

/**
 * @fn    void fp2_elt_add (fp2_elt_ptr dst, fp2_elt_srcptr src1, fp2_elt_srcptr src2, const fp2_param param);
 * @brief Set dst <- src1 + src2
 * @param dst   To store the result
 * @param src1  Source field element
 * @param src2  Source field element
 * @param param Information on the quadratic extension field 
 */
void
fp2_elt_add (fp2_elt_ptr dst, fp2_elt_srcptr src1, fp2_elt_srcptr src2, 
        const fp2_param param);

/**
 * @fn    void fp2_elt_dec (fp2_elt_ptr dst, fp2_elt_srcptr src, const fp2_param param);
 * @brief Set dst <- src - 1
 * @param dst   To store the result
 * @param src   Source field element
 * @param param Information on the quadratic extension field 
 */
void
fp2_elt_dec (fp2_elt_ptr dst, fp2_elt_srcptr src, const fp2_param param);

/**
 * @fn    void fp2_elt_sub_fp_elt (fp2_elt_ptr dst, fp2_elt_srcptr src1, fp_elt_srcptr src2, const fp2_param param);
 * @brief Set dst <- src1 - src2
 * @param dst   To store the result
 * @param src1  Source field element
 * @param src2  Source field element in the prime field
 * @param param Information on the quadratic extension field 
 */
void
fp2_elt_sub_fp_elt (fp2_elt_ptr dst, fp2_elt_srcptr src1, fp_elt_srcptr src2, 
        const fp2_param param);

/**
 * @fn    void fp2_elt_sub (fp2_elt_ptr dst, fp2_elt_srcptr src1, fp2_elt_srcptr src2, const fp2_param param);
 * @brief Set dst <- src1 - src2
 * @param dst   To store the result
 * @param src1  Source field element
 * @param src2  Source field element
 * @param param Information on the quadratic extension field 
 */
void
fp2_elt_sub (fp2_elt_ptr dst, fp2_elt_srcptr src1, fp2_elt_srcptr src2,
        const fp2_param param);

/**
 * @fn    void fp2_elt_neg_fp_elt (fp2_elt_ptr dst, fp_elt_srcptr src, const fp2_param param);
 * @brief Set dst <- (-src) 
 * @param dst   To store the result
 * @param src   Source field element in the prime field
 * @param param Information on the quadratic extension field 
 */
void
fp2_elt_neg_fp_elt (fp2_elt_ptr dst, fp_elt_srcptr src, const fp2_param param);

/**
 * @fn    void fp2_elt_neg (fp2_elt_ptr dst, fp2_elt_srcptr src, const fp2_param param);
 * @brief Set dst <- (-src)
 * @param dst   To store the result
 * @param src   Source field element
 * @param param Information on the quadratic extension field 
 */
void
fp2_elt_neg (fp2_elt_ptr dst, fp2_elt_srcptr src, const fp2_param param);


/*******************************MULTIPLICATION********************************/

/**
 * @fn    void fp2_elt_mul_fp_elt (fp2_elt_ptr dst, fp2_elt_srcptr src1, fp_elt_srcptr src2, const fp2_param param, uint8_t stack);
 * @brief Set dst <- src1 * src2, if Montgomery arithmetic is used, the Montgomery multiplication will be used !
 * @param dst   To store the result
 * @param src1  Source field element
 * @param src2  Source field element in the prime field
 * @param param Information the quadratic extension prime field 
 * @param stack Temporary memory stack to use
 */
void
fp2_elt_mul_fp_elt (fp2_elt_ptr dst, fp2_elt_srcptr src1, fp_elt_srcptr src2, 
        const fp2_param param, uint8_t stack);

/**
 * @fn    void fp2_elt_mul2 (fp2_elt_ptr dst, fp2_elt_srcptr src, const fp2_param param);
 * @brief Set dst <- 2 * src
 * @param dst   To store the result
 * @param src   Source field element
 * @param param Information the quadratic extension prime field 
 */
void
fp2_elt_mul2 (fp2_elt_ptr dst, fp2_elt_srcptr src,
        const fp2_param param);

/**
 * @fn    void fp2_elt_mul4 (fp2_elt_ptr dst, fp2_elt_srcptr src, const fp2_param param);
 * @brief Set dst <- 4 * src
 * @param dst   To store the result
 * @param src   Source field element
 * @param param Information the quadratic extension prime field 
 */
void
fp2_elt_mul4 (fp2_elt_ptr dst, fp2_elt_srcptr src,
        const fp2_param param);

/**
 * @fn    void fp2_elt_mul8 (fp2_elt_ptr dst, fp2_elt_srcptr src, const fp2_param param);
 * @brief Set dst <- 8 * src
 * @param dst   To store the result
 * @param src   Source field element
 * @param param Information the quadratic extension prime field
 */
void
fp2_elt_mul8 (fp2_elt_ptr dst, fp2_elt_srcptr src,
        const fp2_param param);

/**
 * @fn    void fp2_elt_mul3 (fp2_elt_ptr dst, fp2_elt_srcptr src, const fp2_param param, uint8_t stack);
 * @brief Set dst <- 3 * src
 * @param dst   To store the result
 * @param src   Source field element
 * @param param Information the quadratic extension prime field 
 * @param stack Temporary memory stack to use
 */
void
fp2_elt_mul3 (fp2_elt_ptr dst, fp2_elt_srcptr src,
        const fp2_param param, uint8_t stack);

/**
 * @fn    void fp2_elt_mul (fp2_elt_ptr dst, fp2_elt_srcptr src1, fp2_elt_srcptr src2, const fp2_param param, uint8_t stack);
 * @brief Set dst <- src1 * src2, if Montgomery arithmetic is used, the Montgomery multiplication will be used !
 * @param dst   To store the result
 * @param src1  Source field element
 * @param src2  Source field element
 * @param param Information the quadratic extension prime field 
 * @param stack Temporary memory stack to use
 */
void
fp2_elt_mul (fp2_elt_ptr dst, fp2_elt_srcptr src1, fp2_elt_srcptr src2, 
        const fp2_param param, uint8_t stack);

/**
 * @fn    void fp2_elt_sqr_fp_elt (fp2_elt_ptr dst, fp_elt_srcptr src, const fp2_param param, uint8_t stack);
 * @brief Set dst <- src^2
 * @param dst   To store the result
 * @param src   Source field element in the prime field
 * @param param Information on the quadratic extension field 
 * @param stack Temporary memory stack to use
 */
void
fp2_elt_sqr_fp_elt (fp2_elt_ptr dst, fp_elt_srcptr src, const fp2_param param, uint8_t stack);

/**
 * @fn    void fp2_elt_sqr (fp2_elt_ptr dst, fp2_elt_srcptr src, const fp2_param param, uint8_t stack);
 * @brief Set dst <- src^2
 * @param dst   To store the result
 * @param src   Source field element
 * @param param Information on the quadratic extension field 
 * @param stack Temporary memory stack to use
 */
void
fp2_elt_sqr (fp2_elt_ptr dst, fp2_elt_srcptr src, const fp2_param param, uint8_t stack);

/**
 * @fn    void fp2_elt_inv_fp_elt (fp2_elt_ptr dst, fp_elt_srcptr src, const fp2_param param, uint8_t stack);
 * @brief Set dst <- src^(-1)
 * @param dst   To store the result
 * @param src   Source field element in the prime field
 * @param param Information on the quadratic extension field 
 * @param stack Temporary memory stack to use
 */
void
fp2_elt_inv_fp_elt (fp2_elt_ptr dst, fp_elt_srcptr src, const fp2_param param, uint8_t stack);

/**
 * @fn    void fp2_elt_inv (fp2_elt_ptr dst, fp2_elt_srcptr src, const fp2_param param, uint8_t stack);
 * @brief Set dst <- src^(-1)
 * @param dst   To store the result
 * @param src   Source field element
 * @param param Information on the quadratic extension field 
 * @param stack Temporary memory stack to use
 */
void
fp2_elt_inv (fp2_elt_ptr dst, fp2_elt_srcptr src, const fp2_param param, uint8_t stack);

/**
 * @fn    void fp2_elt_div_fp_elt (fp2_elt_ptr dst, fp2_elt_srcptr src1, fp_elt_srcptr src2, const fp2_param param, uint8_t stack);
 * @brief Set dst <- src1 / src2
 * @param dst   To store the result
 * @param src1  Source field element
 * @param src2  Source field element in the prime field
 * @param param Information on the quadratic extension field 
 * @param stack Temporary memory stack to use
 */
void
fp2_elt_div_fp_elt (fp2_elt_ptr dst, fp2_elt_srcptr src1, fp_elt_srcptr src2, 
        const fp2_param param, uint8_t stack);

/**
 * @fn    void fp2_elt_div (fp2_elt_ptr dst, fp2_elt_srcptr src1, fp2_elt_srcptr src2, const fp2_param param, uint8_t stack);
 * @brief Set dst <- src1 / src2
 * @param dst   To store the result
 * @param src1  Source field element
 * @param src2  Source field element
 * @param param Information on the quadratic extension field 
 * @param stack Temporary memory stack to use
 */
void
fp2_elt_div (fp2_elt_ptr dst, fp2_elt_srcptr src1, fp2_elt_srcptr src2, 
        const fp2_param param, uint8_t stack);
/**
 * @fn    void fp2_elt_pow_ui (fp2_elt_ptr dst, fp2_elt_srcptr src, const block n, const fp2_param param, uint8_t stack);
 * @brief Set dst <- src^n
 * @param dst   To store the result
 * @param src   Source field element
 * @param n     Source block
 * @param param Information on the quadratic extension field 
 * @param stack Temporary memory stack to use
 */
void
fp2_elt_pow_ui (fp2_elt_ptr dst, fp2_elt_srcptr src, const block n,
        const fp2_param param, uint8_t stack);

/**
 * @fn    void fp2_elt_pow_number (fp2_elt_ptr dst, fp2_elt_srcptr src, number_srcptr n, const fp2_param param, uint8_t stack);
 * @brief Set dst <- src^n
 * @param dst   To store the result
 * @param src   Source field element
 * @param n     Source number
 * @param param Information on the quadratic extension field 
 * @param stack Temporary memory stack to use
 */
void
fp2_elt_pow_number (fp2_elt_ptr dst, fp2_elt_srcptr src, number_srcptr n,
        const fp2_param param, uint8_t stack);

/**
 * @fn     bool fp2_elt_issquare (fp2_elt_srcptr src, const fp2_param param, uint8_t stack);
 * @brief  Test if src is a square, using A NOTE ON POWERS IN FINITE FIELDS from ANDREAS AABRANDT AND VAGN LUNDSGAARD HANSEN
 * @param  src   Source field element
 * @param  param Information on the quadratic extension field 
 * @param  stack Temporary memory stack to use
 * @return true if src is a square in Fp2, false otherwise
 */
bool
fp2_elt_issquare (fp2_elt_srcptr src, const fp2_param param, uint8_t stack);

/**
 * @fn     int8_t fp2_elt_ispower_ui (fp2_elt_srcptr src, const block n, const fp2_param param, uint8_t stack);
 * @brief  Test if src is a n-power in FP2, using A NOTE ON POWERS IN FINITE FIELDS from ANDREAS AABRANDT AND VAGN LUNDSGAARD HANSEN
 * @param  src   Source field element
 * @param  n     Source block
 * @param  stack Temporary memory stack to use
 * @param  param Information on the quadratic extension field 
 * @return 2 if all element in FP2 are some n-power
 * @return 1 if src is a n-power in FP2
 * @return false otherwise
 */
int8_t
fp2_elt_ispower_ui (fp2_elt_srcptr src, const block n, const fp2_param param, uint8_t stack);

/**
 * @fn     int8_t fp2_elt_ispower_number (fp2_elt_srcptr src, number_srcptr n, const fp2_param param, uint8_t stack);
 * @brief  Test if src is a n-power in FP2, using A NOTE ON POWERS IN FINITE FIELDS from ANDREAS AABRANDT AND VAGN LUNDSGAARD HANSEN
 * @param  src   Source field element
 * @param  n     Source number
 * @param  param Information on the quadratic extension field 
 * @param  stack Temporary memory stack to use
 * @return 2 if all element in FP2 are some n-power
 * @return 1 if src is a n-power in FP2
 * @return false otherwise
 */
int8_t
fp2_elt_ispower_number (fp2_elt_srcptr src, number_srcptr n,
        const fp2_param param, uint8_t stack);

/**
 * @fn    void fp2_elt_sqrt (fp2_elt_ptr dst, fp2_elt_srcptr src, const fp2_param param, uint8_t stack);
 * @brief Set dst <- src^(1/2), using Tonelli–Shanks algorithm
 * @param dst   To store the result
 * @param src   Source field element
 * @param param Information on the quadratic extension field 
 * @param stack Temporary memory stack to use
 */
void
fp2_elt_sqrt (fp2_elt_ptr dst, fp2_elt_srcptr src, const fp2_param param, uint8_t stack);

/**
 * @fn    void fp2_elt_cube_root (fp2_elt_ptr dst, fp2_elt_srcptr src, const fp2_param param, uint8_t stack);
 * @brief Set dst <- src^(1/3)
 * @param dst   To store the result
 * @param src   Source field element
 * @param param Information on the quadratic extension field 
 * @param stack Temporary memory stack to use
 */
void
fp2_elt_cube_root (fp2_elt_ptr dst, fp2_elt_srcptr src, const fp2_param param, uint8_t stack);

/**
 * @fn    void fp2_elt_unity_nth_root (fp2_elt_ptr dst, const block n, const fp2_param param, uint8_t stack);
 * @brief Set dst to a non trivial n-th root of unity if it exists (ie n divides q-1), 1 otherwise.
 * @param dst   To store the result
 * @param n     Source block
 * @param param Information on prime field 
 * @param stack Temporary memory stack to use
 */
void
fp2_elt_unity_nth_root (fp2_elt_ptr dst, const block n, const fp2_param param, uint8_t stack);

#endif


