/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France)

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell-fp3.h
  * \brief Declaration of cubic field (finite field constructed with an irreducible polynomial of degree 3) functions.
  */

#ifndef MPHELL_fp3_H
#define MPHELL_fp3_H

#include "mphell-fp.h"

/**
 * @struct fp3_elt_t
 * @brief  Cubic extension field element structure
 */
typedef struct
{
    fp_elt v0;  /*!< Coefficient of degree 0    */
    fp_elt v1;  /*!< Coefficient of degree 1    */
    fp_elt v2;  /*!< Coefficient of degree 2    */
} fp3_elt_t;

/**
 * @typedef fp3_elt
 * @brief   Cubic extension field element structure
 */
typedef fp3_elt_t fp3_elt;

/**
 * @typedef fp3_elt_ptr
 * @brief   Define fp3_elt_ptr, use in all functions
 */
typedef fp3_elt_t * fp3_elt_ptr;

/**
 * @typedef fp3_elt_srcptr
 * @brief   Define fp3_elt_srcptr, use in all functions where the parameter is const
 */
typedef const fp3_elt_t * fp3_elt_srcptr;

/**
 * @struct fp3_param_t
 * @brief  Cubic extension field structure
 */
typedef struct
{
    fp_param_t *base;       /*!< Base prime field                           */
    fp_elt non_residue;     /*!< Non cubic residue                          */
    number q;               /*!< Order of FP3 = p^3                         */
    /*char * pm2;             !< Binary representation of (p-1)p^2-1 (FLT)  */
    fp3_elt non_res;        /*!< Non square residue                         */
    fp3_elt gen_sylow;      /*!< Non quadratic residue power p_odd          */
    number p_odd;           /*!< Write q-1 as p_odd.(2^p_even)              */
    uint32_t p_even;        /*!< Write q-1 as p_odd.(2^p_even)              */
    uint8_t size;           /*!< Number of block required by p              */
}
fp3_param_t;

/**
 * @typedef fp3_param
 * @brief   Address of a cubic extension field structure
 */
typedef fp3_param_t *fp3_param;

/**************************************TMP************************************/

/**
 * @fn    void fp3_elt_get_pool_elt (fp3_elt * dst, const fp3_param param, uint8_t stack);
 * @brief Allocate and initialise space for a temporary cubic extension field element
 * @param dst    Temporary element to allocate
 * @param param  Information on the cubic extension field 
 * @param stack  Temporary memory stack to use
 */
void
fp3_elt_get_pool_elt (fp3_elt * dst, const fp3_param param, uint8_t stack);

/**
 * @fn    void fp3_elt_relax_pool_elt (fp3_elt * dst, const fp3_param param, uint8_t stack);
 * @brief Free space of a temporary cubic extension field element
 * @param dst    Temporary element to free
 * @param param  Information on the cubic extension field 
 * @param stack  Temporary memory stack to use
 */
void
fp3_elt_relax_pool_elt (fp3_elt * dst, const fp3_param param, uint8_t stack);

/************************************SETTERS**********************************/

/**
 * @fn    void fp3_alloc (fp3_param param, const fp_param base);
 * @brief Allocate space for the cubic extension field informations structure
 * @param param Structure for the cubic extension field informations to allocate
 * @param base  Base prime field
 */
void
fp3_alloc (fp3_param param, const fp_param base);

/**
 * @fn    void fp3_create (fp3_param param, const fp_param base, fp_elt_srcptr non_residue, uint8_t stack);
 * @brief Create a cubic extension of the field base
 * @param param       Informations structure on cubic extension field to fill (must be allocated)
 * @param base        Base prime field
 * @param non_residue Non cubic residue
 * @param stack       Temporary memory stack to use
 */
void
fp3_create (fp3_param param, const fp_param base, fp_elt_srcptr non_residue, uint8_t stack);

/**
 * @fn    void fp3_copy (fp3_param param_res, const fp3_param param);
 * @brief Copy the cubic extension field structure param into param_res
 * @param param_res Destination cubic extension field 
 * @param param     Source cubic extension field 
 */
void
fp3_copy (fp3_param param_res, const fp3_param param);

/**
 * @fn    void fp3_free (fp3_param param);
 * @brief Free the space of the cubic extension field informations structure
 * @param param cubic extension field informations structure
 */
void
fp3_free (fp3_param param);

/**
 * @fn    void fp3_get_characteristic (number_ptr c, const fp3_param param);
 * @brief Get the characteristic of the cubic extension field "param"
 * @param c     Destination number (to store the characteristic)
 * @param param Information on cubic extension field 
 */
void
fp3_get_characteristic (number_ptr c, const fp3_param param);

/**
 * @fn    void fp3_get_size (number_ptr c, const fp3_param param);
 * @brief Get the size of the cubic extension field "param"
 * @param c     Destination number (to store the characteristic)
 * @param param Information on cubic extension field 
 */
void
fp3_get_size (number_ptr c, const fp3_param param);

/*void*/
/*fp3_get_pm2 (char * str, const fp3_param param);*/

/**
 * @fn    void fp3_elt_alloc (fp3_elt * dst, const fp3_param param);
 * @brief Allocate space for a cubic extension field element
 * @param dst   Cubic extension field element to allocate
 * @param param Information on cubic extension field 
 */
void
fp3_elt_alloc (fp3_elt * dst, const fp3_param param);

/**
 * @fn    void fp3_elt_init (fp3_elt_ptr dst, const fp3_param param);
 * @brief Initialise a cubic extension field element
 * @param dst   Cubic extension field element to initialise
 * @param param Information on cubic extension field 
 */
void
fp3_elt_init (fp3_elt_ptr dst, const fp3_param param);

/**
 * @fn    void fp3_elt_copy (fp3_elt_ptr dst, fp3_elt_srcptr src, const fp3_param param);
 * @brief Copy src into dst, src and dst must belong to the same FP3
 * @param dst   Destination cubic extension field element 
 * @param src   Source cubic extension field element
 * @param param Information on cubic extension field 
 */
void
fp3_elt_copy (fp3_elt_ptr dst, fp3_elt_srcptr src, const fp3_param param);

/**
 * @fn    void fp3_elt_clear (fp3_elt * src);
 * @brief Clear space used by src (remove the action of fp3_elt_init but let the one of fp3_elt_alloc)
 * @param src cubic extension field element to clear
 */
void
fp3_elt_clear (fp3_elt * src);

/**
 * @fn    void fp3_elt_free (fp3_elt * src);
 * @brief Free space used by src
 * @param src cubic extension field element to free
 */
void
fp3_elt_free (fp3_elt * src);

/**
 * @fn    void fp3_elt_set_one (fp3_elt_ptr dst, const fp3_param param);
 * @brief Set dst to one (or its Montgomery form if Montgomery arithmetic is used)
 * @param dst   Destination cubic extension field element (must be allocated)
 * @param param Information on the cubic extension field 
 */
void
fp3_elt_set_one (fp3_elt_ptr dst, const fp3_param param);

/**
 * @fn    void fp3_elt_set_ui (fp3_elt_ptr dst, const uint64_t src, const bool isreduced, const fp3_param param, uint8_t stack);
 * @brief Set dst to src, if Montgomery arithmetic is used, is_reduced == false -> transform dst into its Montgomery form
 * @param dst       Destination cubic extension field element
 * @param src       Source block
 * @param isreduced If Montgomery arithmetic is used, is_reduced == false -> transform dst into its Montgomery form, is_reduced == true -> simple copy
 * @param param     Information on the cubic extension field 
 * @param stack     Temporary memory stack to use
 */
void
fp3_elt_set_ui (fp3_elt_ptr dst, const uint64_t src, const bool isreduced, 
        const fp3_param param, uint8_t stack);

/**
 * @fn    void fp3_elt_set_number (fp3_elt_ptr dst, number_srcptr src, const bool isreduced, const fp3_param param, uint8_t stack);
 * @brief Set dst to src, if Montgomery arithmetic is used, is_reduced == false -> transform dst into its Montgomery form
 * @param dst       Destination cubic extension field element
 * @param src       Source number
 * @param isreduced If Montgomery arithmetic is used, is_reduced == false -> transform dst into its Montgomery form, is_reduced == true -> simple copy
 * @param param     Information on the cubic extension field
 * @param stack     Temporary memory stack to use
 */
void
fp3_elt_set_number (fp3_elt_ptr dst, number_srcptr src, const bool isreduced, 
        const fp3_param param, uint8_t stack);

/**
 * @fn    void fp3_elt_set_fp_elt (fp3_elt_ptr dst, fp_elt_srcptr src, const fp3_param param);
 * @brief Set dst to src
 * @param dst       Destination cubic extension field element
 * @param src       Source fp_elt
 * @param param     Information on the cubic extension field
 */
void
fp3_elt_set_fp_elt (fp3_elt_ptr dst, fp_elt_srcptr src, const fp3_param param);

/**
 * @fn    void fp3_elt_set (fp3_elt_ptr dst, number_srcptr src1, number_srcptr src2, number_srcptr src3, const bool isreduced, const fp3_param param, uint8_t stack);
 * @brief Set dst to src1 + src2*x + src3*x^2, if Montgomery arithmetic is used, is_reduced == false -> transform dst into its Montgomery form
 * @param dst       Destination cubic extension field element
 * @param src1      Source number
 * @param src2      Source number
 * @param src3      Source number
 * @param isreduced If Montgomery arithmetic is used, is_reduced == false -> transform dst into its Montgomery form, is_reduced == true -> simple copy
 * @param param     Information on the cubic extension field
 * @param stack     Temporary memory stack to use
 */
void
fp3_elt_set (fp3_elt_ptr dst, number_srcptr src1, number_srcptr src2, 
        number_srcptr src3, const bool isreduced, const fp3_param param, uint8_t stack);

/**
 * @fn    void fp3_elt_set_fp_elts (fp3_elt_ptr dst, fp_elt_srcptr src1, fp_elt_srcptr src2, fp_elt_srcptr src3, const fp3_param param);
 * @brief Set dst to src1 + src2*x + src3*x^2
 * @param dst       Destination cubic extension field element
 * @param src1      Source fp_elt
 * @param src2      Source fp_elt
 * @param src3      Source fp_elt
 * @param param     Information on the cubic extension field
 */
void
fp3_elt_set_fp_elts (fp3_elt_ptr dst, fp_elt_srcptr src1, fp_elt_srcptr src2, fp_elt_srcptr src3, const fp3_param param);

/**
 * @fn    void fp3_elt_set_str (fp3_elt_ptr dst, const char *str, const uint8_t base, const bool isreduced, const fp3_param param, uint8_t stack); 
 * @brief Set dst to str, if Montgomery arithmetic is used, is_reduced == false -> transform dst into its Montgomery form
 * @param dst       Destination cubic extension field element
 * @param str       Source polynomial coefficient under string format in base "base", separated by a "," without space. ex: "123,456,789" will create the polynomial 123 + 456.X + 789.X^2
 * @param base      Base used to write str, can be 2, 10 or 16
 * @param isreduced If Montgomery arithmetic is used, is_reduced == false -> transform dst into its Montgomery form, is_reduced == true -> simple copy
 * @param param     Information on the cubic extension field
 * @param stack     Temporary memory stack to use
 */
void
fp3_elt_set_str (fp3_elt_ptr dst, const char *str, const uint8_t base, 
        const bool isreduced, const fp3_param param, uint8_t stack);

/**
 * @fn    void fp3_elt_random (fp3_elt_ptr dst, const fp3_param param, uint8_t stack);
 * @brief Set dst to a random element of FP3, the random process is chosen at the MHELL initialisation
 * @param dst   Destination cubic extension field element
 * @param param Information on the cubic extension field 
 * @param stack Temporary memory stack to use
 */
void
fp3_elt_random (fp3_elt_ptr dst, const fp3_param param, uint8_t stack);

/**
 * @fn    void fp3_elt_lift (fp3_elt_ptr dst, fp3_elt_srcptr src, const fp3_param param, uint8_t stack);
 * @brief If Montgomery arithmetic is used, lift src (which is into Montgomery form) to classical FP3. 
 * @param dst   Destination primary field element (in FP)
 * @param src   Source cubic extension field element (under Montgomery form)
 * @param param Information on the cubic extension field 
 * @param stack Temporary memory stack to use
 */
void
fp3_elt_lift (fp3_elt_ptr dst, fp3_elt_srcptr src, const fp3_param param, uint8_t stack);

/**
 * @fn    void fp3_str (char **str, const fp3_param param, const uint8_t base, uint8_t stack);
 * @brief Converts fp2_param param to string format in base specified by base.
 * @param str   Destination string, allocated by the function, must be free by the user.
 * @param param Information on the cubic extension field 
 * @param base  Base to use to write str, can be 2, 10 or 16
 * @param stack Temporary memory stack to use
 */
void
fp3_str (char **str, const fp3_param param, const uint8_t base, uint8_t stack);

/**
 * @fn    void fp3_elt_str (char **str, fp3_elt_srcptr src, const uint8_t base, const bool lift, const fp3_param param, uint8_t stack);
 * @brief Converts src to string format in base specified by base.
 * @param str   Destination string, allocated by the function, must be free by the user. Polynomial coefficient are written separated by ",". ex: 123 + 456.X + 789.X^2 will be written as "123,456,789"
 * @param src   Source cubic extension field element (under Montgomery form if Montgomery arithmetic is used)
 * @param base  Base to use to write str, can be 2, 10 or 16
 * @param lift  If Montgomery arithmetic is used, lift == true -> lift src into Fp2 before conversion
 * @param param Information on the cubic extension field
 * @param stack Temporary memory stack to use
 */
void
fp3_elt_str (char **str, fp3_elt_srcptr src, const uint8_t base, 
        const bool lift, const fp3_param param, uint8_t stack);

/*************************COMPARISON AND LOGICAL******************************/

/**
 * @fn     int8_t fp3_elt_cmp_fp_elt (fp3_elt_srcptr src1, fp_elt_srcptr src2, const fp3_param param);
 * @brief  Compare src1 and src2 in FP3
 * @param  src1  Source field element in FP3
 * @param  src2  Source number in FP
 * @param  param Information on the cubic extension field 
 * @return -1 if src1 < src2
 * @return  0 if src1 == src2
 * @return  1 if src1 > src2
 */
int8_t
fp3_elt_cmp_fp_elt (fp3_elt_srcptr src1, fp_elt_srcptr src2, const fp3_param param);

/**
 * @fn     int8_t fp3_elt_cmp (fp3_elt_srcptr src1, fp3_elt_srcptr src2, const fp3_param param);
 * @brief  Compare src1 and src2 in FP3
 * @param  src1  Source field element in FP3
 * @param  src2  Source field element in FP3
 * @param  param Information on the cubic extension field 
 * @return -1 if src1 < src2
 * @return  0 if src1 == src2
 * @return  1 if src1 > src2
 */
int8_t
fp3_elt_cmp (fp3_elt_srcptr src1, fp3_elt_srcptr src2, const fp3_param param);

/**
 * @fn     bool fp3_elt_isone (fp3_elt_srcptr src, const fp3_param param);
 * @brief  Test if src is one
 * @param  src   Source field element
 * @param  param Information on the cubic extension field 
 * @return true if scr is one, false otherwise
 */
bool
fp3_elt_isone (fp3_elt_srcptr src, const fp3_param param);

/**
 * @fn     bool fp3_elt_iszero (fp3_elt_srcptr src, const fp3_param param);
 * @brief  Test if src is zero
 * @param  src   Source field element
 * @param  param Information on the cubic extension field 
 * @return true if scr is zero, false otherwise
 */
bool
fp3_elt_iszero (fp3_elt_srcptr src, const fp3_param param);

/***************************ADDITION SUBTRACTION******************************/

/**
 * @fn    void fp3_elt_inc (fp3_elt_ptr dst, fp3_elt_srcptr src, const fp3_param param);
 * @brief Set dst <- src + 1
 * @param dst   To store the result
 * @param src   Source field element
 * @param param Information on the cubic extension field 
 */
void
fp3_elt_inc (fp3_elt_ptr dst, fp3_elt_srcptr src, const fp3_param param);

/**
 * @fn    void fp3_elt_add_fp_elt (fp3_elt_ptr dst, fp3_elt_srcptr src1, fp_elt_srcptr src2, const fp3_param param);
 * @brief Set dst <- src1 + src2
 * @param dst   To store the result
 * @param src1  Source field element
 * @param src2  Source field element in the prime field
 * @param param Information on the cubic extension field 
 */
void
fp3_elt_add_fp_elt (fp3_elt_ptr dst, fp3_elt_srcptr src1, fp_elt_srcptr src2, 
        const fp3_param param);

/**
 * @fn    void fp3_elt_add (fp3_elt_ptr dst, fp3_elt_srcptr src1, fp3_elt_srcptr src2, const fp3_param param);
 * @brief Set dst <- src1 + src2
 * @param dst   To store the result
 * @param src1  Source field element
 * @param src2  Source field element
 * @param param Information on the cubic extension field 
 */
void
fp3_elt_add (fp3_elt_ptr dst, fp3_elt_srcptr src1, fp3_elt_srcptr src2, 
        const fp3_param param);

/**
 * @fn    void fp3_elt_dec (fp3_elt_ptr dst, fp3_elt_srcptr src, const fp3_param param);
 * @brief Set dst <- src - 1
 * @param dst   To store the result
 * @param src   Source field element
 * @param param Information on the cubic extension field 
 */
void
fp3_elt_dec (fp3_elt_ptr dst, fp3_elt_srcptr src, const fp3_param param);

/**
 * @fn    void fp3_elt_sub_fp_elt (fp3_elt_ptr dst, fp3_elt_srcptr src1, fp_elt_srcptr src2, const fp3_param param);
 * @brief Set dst <- src1 - src2
 * @param dst   To store the result
 * @param src1  Source field element
 * @param src2  Source field element in the prime field
 * @param param Information on the cubic extension field 
 */
void
fp3_elt_sub_fp_elt (fp3_elt_ptr dst, fp3_elt_srcptr src1, fp_elt_srcptr src2, 
        const fp3_param param);

/**
 * @fn    void fp3_elt_sub (fp3_elt_ptr dst, fp3_elt_srcptr src1, fp3_elt_srcptr src2, const fp3_param param);
 * @brief Set dst <- src1 - src2
 * @param dst   To store the result
 * @param src1  Source field element
 * @param src2  Source field element
 * @param param Information on the cubic extension field 
 */
void
fp3_elt_sub (fp3_elt_ptr dst, fp3_elt_srcptr src1, fp3_elt_srcptr src2,
        const fp3_param param);

/**
 * @fn    void fp3_elt_neg_fp_elt (fp3_elt_ptr dst, fp_elt_srcptr src, const fp3_param param);
 * @brief Set dst <- (-src) 
 * @param dst   To store the result
 * @param src   Source field element in the prime fields
 * @param param Information on the cubic extension field 
 */
void
fp3_elt_neg_fp_elt (fp3_elt_ptr dst, fp_elt_srcptr src, const fp3_param param);

/**
 * @fn    void fp3_elt_neg (fp3_elt_ptr dst, fp3_elt_srcptr src, const fp3_param param);
 * @brief Set dst <- (-src)
 * @param dst   To store the result
 * @param src   Source field element
 * @param param Information on the cubic extension field 
 */
void
fp3_elt_neg (fp3_elt_ptr dst, fp3_elt_srcptr src, const fp3_param param);


/*******************************MULTIPLICATION********************************/

/**
 * @fn    void fp3_elt_mul_fp_elt (fp3_elt_ptr dst, fp3_elt_srcptr src1, fp_elt_srcptr src2, const fp3_param param, uint8_t stack);
 * @brief Set dst <- src1 * src2, if Montgomery arithmetic is used, the Montgomery multiplication will be used !
 * @param dst  To store the result
 * @param src1  Source field element
 * @param src2  Source field element in the prime field
 * @param param Information the cubic extension prime field 
 * @param stack Temporary memory stack to use
 */
void
fp3_elt_mul_fp_elt (fp3_elt_ptr dst, fp3_elt_srcptr src1, fp_elt_srcptr src2, 
        const fp3_param param, uint8_t stack);

/**
 * @fn    void fp3_elt_mul2 (fp3_elt_ptr dst, fp3_elt_srcptr src, const fp3_param param);
 * @brief Set dst <- 2 * src
 * @param dst   To store the result
 * @param src   Source field element
 * @param param Information the cubic extension prime field 
 */
void
fp3_elt_mul2 (fp3_elt_ptr dst, fp3_elt_srcptr src,
        const fp3_param param);

/**
 * @fn    void fp3_elt_mul4 (fp3_elt_ptr dst, fp3_elt_srcptr src, const fp3_param param);
 * @brief Set dst <- 4 * src
 * @param dst   To store the result
 * @param src   Source field element
 * @param param Information the cubic extension prime field 
 */
void
fp3_elt_mul4 (fp3_elt_ptr dst, fp3_elt_srcptr src,
        const fp3_param param);

/**
 * @fn    void fp3_elt_mul8 (fp3_elt_ptr dst, fp3_elt_srcptr src, const fp3_param param);
 * @brief Set dst <- 8 * src
 * @param dst   To store the result
 * @param src   Source field element
 * @param param Information the cubic extension prime field 
 */
void
fp3_elt_mul8 (fp3_elt_ptr dst, fp3_elt_srcptr src,
        const fp3_param param);

/**
 * @fn    void fp3_elt_mul3 (fp3_elt_ptr dst, fp3_elt_srcptr src, const fp3_param param, uint8_t stack);
 * @brief Set dst <- 3 * src
 * @param dst   To store the result
 * @param src   Source field element
 * @param param Information the cubic extension prime field 
 * @param stack Temporary memory stack to use
 */
void
fp3_elt_mul3 (fp3_elt_ptr dst, fp3_elt_srcptr src,
        const fp3_param param, uint8_t stack);

/**
 * @fn    void fp3_elt_mul (fp3_elt_ptr dst, fp3_elt_srcptr src1, fp3_elt_srcptr src2, const fp3_param param, uint8_t stack);
 * @brief Set dst <- src1 * src2, if Montgomery arithmetic is used, the Montgomery multiplication will be used !
 * @param dst  To store the result
 * @param src1  Source field element
 * @param src2  Source field element
 * @param param Information the cubic extension prime field 
 * @param stack Temporary memory stack to use
 */
void
fp3_elt_mul (fp3_elt_ptr dst, fp3_elt_srcptr src1, fp3_elt_srcptr src2, 
        const fp3_param param, uint8_t stack);

/**
 * @fn    void fp3_elt_sqr_fp_elt (fp3_elt_ptr dst, fp_elt_srcptr src, const fp3_param param, uint8_t stack);
 * @brief Set dst <- src^2
 * @param dst   To store the result
 * @param src   Source field element in the prime field
 * @param param Information on the cubic extension field 
 * @param stack Temporary memory stack to use
 */
void
fp3_elt_sqr_fp_elt (fp3_elt_ptr dst, fp_elt_srcptr src, const fp3_param param, uint8_t stack);

/**
 * @fn    void fp3_elt_sqr (fp3_elt_ptr dst, fp3_elt_srcptr src, const fp3_param param, uint8_t stack);
 * @brief Set dst <- src^2
 * @param dst   To store the result
 * @param src   Source field element
 * @param param Information on the cubic extension field 
 */
void
fp3_elt_sqr (fp3_elt_ptr dst, fp3_elt_srcptr src, const fp3_param param, uint8_t stack);

/**
 * @fn    void fp3_elt_inv_fp_elt (fp3_elt_ptr dst, fp_elt_srcptr src, const fp3_param param, uint8_t stack);
 * @brief Set dst <- src^(-1)
 * @param dst   To store the result
 * @param src   Source field element in the prime field
 * @param param Information on the cubic extension field 
 */
void
fp3_elt_inv_fp_elt (fp3_elt_ptr dst, fp_elt_srcptr src, const fp3_param param, uint8_t stack);

/**
 * @fn    void fp3_elt_inv (fp3_elt_ptr dst, fp3_elt_srcptr src, const fp3_param param, uint8_t stack);
 * @brief Set dst <- src^(-1)
 * @param dst   To store the result
 * @param src   Source field element
 * @param param Information on the cubic extension field 
 * @param stack Temporary memory stack to use
 */
void
fp3_elt_inv (fp3_elt_ptr dst, fp3_elt_srcptr src, const fp3_param param, uint8_t stack);

/**
 * @fn    void fp3_elt_div_fp_elt (fp3_elt_ptr dst, fp3_elt_srcptr src1, fp_elt_srcptr src2, const fp3_param param, uint8_t stack);
 * @brief Set dst <- src1 / src2
 * @param dst   To store the result
 * @param src1  Source field element
 * @param src2  Source field element in the prime field
 * @param param Information on the cubic extension field 
 * @param stack Temporary memory stack to use
 */
void
fp3_elt_div_fp_elt (fp3_elt_ptr dst, fp3_elt_srcptr src1, fp_elt_srcptr src2, 
        const fp3_param param, uint8_t stack);

/**
 * @fn    void fp3_elt_div (fp3_elt_ptr dst, fp3_elt_srcptr src1, fp3_elt_srcptr src2, const fp3_param param, uint8_t stack);
 * @brief Set dst <- src1 / src2
 * @param dst   To store the result
 * @param src1  Source field element
 * @param src2  Source field element
 * @param param Information on the cubic extension field 
 * @param stack Temporary memory stack to use
 */
void
fp3_elt_div (fp3_elt_ptr dst, fp3_elt_srcptr src1, fp3_elt_srcptr src2, 
        const fp3_param param, uint8_t stack);

/**
 * @fn    void fp3_elt_pow_ui (fp3_elt_ptr dst, fp3_elt_srcptr src, const block n, const fp3_param param, uint8_t stack);
 * @brief Set dst <- src^n
 * @param dst   To store the result
 * @param src   Source field element
 * @param n     Source block
 * @param param Information on the cubic extension field 
 * @param stack Temporary memory stack to use
 */
void
fp3_elt_pow_ui (fp3_elt_ptr dst, fp3_elt_srcptr src, const block n,
        const fp3_param param, uint8_t stack);

/**
 * @fn    void fp3_elt_pow_number (fp3_elt_ptr dst, fp3_elt_srcptr src, number_srcptr n, const fp3_param param, uint8_t stack);
 * @brief Set dst <- src^n
 * @param dst   To store the result
 * @param src   Source field element
 * @param n     Source number
 * @param param Information on the cubic extension field 
 * @param stack Temporary memory stack to use
 */
void
fp3_elt_pow_number (fp3_elt_ptr dst, fp3_elt_srcptr src, number_srcptr n,
        const fp3_param param, uint8_t stack);

/**
 * @fn     bool fp3_elt_issquare (fp3_elt_srcptr src, const fp3_param param, uint8_t stack);
 * @brief  Test if src is a square, using A NOTE ON POWERS IN FINITE FIELDS from ANDREAS AABRANDT AND VAGN LUNDSGAARD HANSEN
 * @param  src   Source field element
 * @param  param Information on the cubic extension field 
 * @param  stack Temporary memory stack to use
 * @return true if src is a square in FP3, false otherwise
 */
bool
fp3_elt_issquare (fp3_elt_srcptr src, const fp3_param param, uint8_t stack);

/**
 * @fn     int8_t fp3_elt_ispower_ui (fp3_elt_srcptr src, const block n, const fp3_param param, uint8_t stack);
 * @brief  Test if src is a n-power in FP3, using A NOTE ON POWERS IN FINITE FIELDS from ANDREAS AABRANDT AND VAGN LUNDSGAARD HANSEN
 * @param  src   Source field element
 * @param  n     Source block
 * @param  param Information on the cubic extension field 
 * @param  stack Temporary memory stack to use
 * @return 2 if all element in FP3 are some n-power
 * @return 1 if src is a n-power in FP3
 * @return false otherwise
 */
int8_t
fp3_elt_ispower_ui (fp3_elt_srcptr src, const block n, const fp3_param param, uint8_t stack);

/**
 * @fn     int8_t fp3_elt_ispower_number (fp3_elt_srcptr src, number_srcptr n, const fp3_param param, uint8_t stack);
 * @brief  Test if src is a n-power in FP3, using A NOTE ON POWERS IN FINITE FIELDS from ANDREAS AABRANDT AND VAGN LUNDSGAARD HANSEN
 * @param  src   Source field element
 * @param  n     Source number
 * @param  param Information on the cubic extension field 
 * @param  stack Temporary memory stack to use
 * @return 2 if all element in FP3 are some n-power
 * @return 1 if src is a n-power in FP3
 * @return false otherwise
 */
int8_t
fp3_elt_ispower_number (fp3_elt_srcptr src, number_srcptr n, 
        const fp3_param param, uint8_t stack);

/**
 * @fn    void fp3_elt_sqrt (fp3_elt_ptr dst, fp3_elt_srcptr src, const fp3_param param, uint8_t stack);
 * @brief Set dst <- src^(1/2), using Tonelli–Shanks algorithm
 * @param dst   To store the result
 * @param src   Source field element
 * @param param Information on the cubic extension field 
 * @param stack Temporary memory stack to use
 */
void
fp3_elt_sqrt (fp3_elt_ptr dst, fp3_elt_srcptr src, const fp3_param param, uint8_t stack);

/**
 * @fn    void fp3_elt_cube_root (fp3_elt_ptr dst, fp3_elt_srcptr src, const fp3_param param, uint8_t stack);
 * @brief Set dst <- src^(1/3)
 * @param dst   To store the result
 * @param src   Source field element
 * @param param Information on the cubic extension field 
 * @param stack Temporary memory stack to use
 */
void
fp3_elt_cube_root (fp3_elt_ptr dst, fp3_elt_srcptr src, const fp3_param param, uint8_t stack);

/**
 * @fn    void fp3_elt_unity_nth_root (fp3_elt_ptr dst, const block n, const fp3_param param, uint8_t stack);
 * @brief Set dst to a non trivial n-th root of unity if it exists (ie n divides q-1), 1 otherwise.
 * @param dst   To store the result
 * @param n     Source block
 * @param param Information on prime field 
 * @param stack Temporary memory stack to use
 */
void
fp3_elt_unity_nth_root (fp3_elt_ptr dst, const block n, const fp3_param param, uint8_t stack);

#endif



