/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France) 

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell-jacobi.h
  * \brief Declaration of Jacobi ECC functions.
  */

#ifndef MPHELL_JACOBI_H
#define MPHELL_JACOBI_H
/* E: y^2 = x^4 + 2ax^2 + 1 */
/* extended Jacobi Quartic with d=1 in affine coordinates */
/* which can be extented to the extended Jacobi Quartic (with Projective equation) E_ext: y^2 = bx^4 + 2ax^2z^2 + z^4 */
/* Homogeneous projective coordinates represent an affine point (x,y) on an extended Jacobi Quartic  E: y^2 = x^4 + 2ax^2 + 1  as (X:Y:T:Z) satisfying Y^2 = Z^2 + 2aX^2 + T^2; X^2 = ZT; a != 1 Here (X:Y:T:Z) = (sX:sY:T:sZ) for all nonzero s. */
/* The Quadruplet (X, Y, T, Z) with X^2=ZT represents the affine point (X / Z, Y / Z) */
/* We work here only with homogeneous projective coordinates. */


/************************************SETTERS**********************************/

/**
 * @fn    void jacobi_quartic_compute_disc(ec_curve E, uint8_t stack);
 * @brief Set the discriminant of E: disc = (2^8).(a^2 - 1)^2
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void
jacobi_quartic_compute_disc(ec_curve E, uint8_t stack);

/**
 * @fn     bool jacobi_quartic_verify_random_generation(ec_curve E, const char * seed, uint8_t stack)
 * @brief  Test if E if generated from the seed "seed"
 * @param  E     Elliptic curve to test
 * @param  seed  Seed (of 160 bits) used to generate an elliptic curve
 * @return true  if E is generated from seed
 * @return false otherwise
 */
bool 
jacobi_quartic_verify_random_generation(ec_curve E, const char * seed, uint8_t stack);

/**
 * @fn    void jacobi_quartic_curve_random_generation(fe_ptr a, char * seed_res, field_srcptr k, uint8_t stack)
 * @brief Generate a 160 bits seed and coefficients a and b defining a Weiestrass elliptic curve. The curve is not tested to be secure !!!
 * @param a        Destination coefficient, generated from a random seed 
 * @param seed_res Destination seed (of 160 bits) used to generate an elliptic curve
 * @param k        Base field, must be created by the user before use.
 * @param stack    Temporary memory stack to use
 */
void 
jacobi_quartic_curve_random_generation(fe_ptr a, char * seed_res, field_srcptr k, uint8_t stack);

/**
 * @fn    void jacobi_quartic_point_set_neutral (ec_point_ptr dst, ec_curve_srcptr E, uint8_t stack);
 * @brief Set dst to the neutral element: (0,1,0,1).
 * @param dst   Destination point
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void 
jacobi_quartic_point_set_neutral (ec_point_ptr dst, ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void jacobi_quartic_point_set_aff (ec_point_ptr P, fe_srcptr x, fe_srcptr y, field_srcptr k, uint8_t stack);
 * @brief Set dest to the affine point (x,y)
 * @param P     Destination point
 * @param x     x-coordinate of the affine point (x,y) 
 * @param y     y-coordinate of the affine point (x,y) 
 * @param k     Base field
 * @param stack Temporary memory stack to use
 */
void 
jacobi_quartic_point_set_aff (ec_point_ptr P, fe_srcptr x, fe_srcptr y, field_srcptr k, uint8_t stack);

/**
 * @fn    void jacobi_quartic_point_set_aff_str (ec_point_ptr P, const char *str_x, const char *str_y, const bool is_reduced, const uint8_t base, field_srcptr k, uint8_t stack);
 * @brief Set dest to the affine point (str_x,str_y)
 * @param P Destination point
 * @param str_x      x-coordinate of the affine point (x,y) under string form
 * @param str_y      y-coordinate of the affine point (x,y) under string form
 * @param is_reduced Set true if str_x and str_y are already reduced in k, false otherwise
 * @param base       Base (either 10 or 16) used for str_x and str_y
 * @param k          Base field
 * @param stack      Temporary memory stack to use
 */
void 
jacobi_quartic_point_set_aff_str (ec_point_ptr P, const char *str_x, 
        const char *str_y, const bool is_reduced, const uint8_t base, 
        field_srcptr k, uint8_t stack);

/**
 * @fn     bool jacobi_quartic_belongs (ec_point_srcptr P, ec_curve_srcptr E, uint8_t stack)
 * @brief  Test if P belongs to E
 * @param  P     Jacobi quartic point
 * @param  E     Elliptic curve
 * @param  stack Temporary memory stack to use
 * @return True  if P belongs to E, false otherwise
 */
bool
jacobi_quartic_belongs (ec_point_srcptr P, ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void jacobi_quartic_point_random (ec_point_ptr P, ec_curve_srcptr E, uint8_t stack)
 * @brief Create a random point P on the elliptic curve E
 * @param P     Destination point in extented projective coordinate (X,Y,T,Z) with T=(X^2)/Z
 * @param E     Jacobi quartic elliptic curve
 * @param stack Temporary memory stack to use
 */
void 
jacobi_quartic_point_random (ec_point_ptr P, ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void jacobi_quartic_point_norm (ec_point_ptr P, ec_curve_srcptr E, uint8_t stack)
 * @brief Convert a point in extented projective coordinate (X,Y,T,Z) to an affine point (x,y) with x=X/Z, y=Y/Z 
 * @param P     Point in extented projective coordinate (X,Y,T,Z) with T=(X^2)/Z
 * @param E     Jacobi quartic elliptic curve
 * @param stack Temporary memory stack to use
 */
void
jacobi_quartic_point_norm (ec_point_ptr P, ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void jacobi_quartic_point_get_x_affine (field_elt x, ec_point_ptr P, ec_curve_srcptr E, uint8_t stack);
 * @brief Convert P->x to its affine representation
 * @param x     P->x in affine coordinates
 * @param P     Point
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void
jacobi_quartic_point_get_x_affine (field_elt x, ec_point_ptr P, ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void jacobi_quartic_point_get_y_affine (field_elt y, ec_point_ptr P, ec_curve_srcptr E, uint8_t stack);
 * @brief Convert P->y to its affine representation
 * @param x     P->y in affine coordinates
 * @param P     Point
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void
jacobi_quartic_point_get_y_affine (field_elt y, ec_point_ptr P, ec_curve_srcptr E, uint8_t stack);



/*******************************COMPARISON************************************/

/**
 * @fn     bool jacobi_quartic_point_is_neutral (ec_point_srcptr P, ec_curve_srcptr E);
 * @brief  Test if P is the neutral element
 * @param  P Point to test
 * @param  E Elliptic curve
 * @return True if P is neutral, False otherwise
 */
bool 
jacobi_quartic_point_is_neutral (ec_point_srcptr P, ec_curve_srcptr E);

/**
 * @fn     bool jacobi_quartic_point_are_equal (ec_point_srcptr P1, ec_point_srcptr P2, ec_curve_srcptr E, uint8_t stack);
 * @brief  Test if P1 and P2 are equal on E
 * @param  P1    Point
 * @param  P2    Point
 * @param  E     Elliptic curve
 * @param  stack Temporary memory stack to use
 * @return True if P1 == P2, False otherwise
 */
bool
jacobi_quartic_point_are_equal (ec_point_srcptr P1, ec_point_srcptr P2, 
        ec_curve_srcptr E, uint8_t stack);


/*******************************OPERATIONS************************************/

/**
 * @fn    jacobi_quartic_point_neg (ec_point_ptr P3, ec_point_srcptr P1, ec_curve_srcptr E)
 * @brief Set P3 to -P1
 * @param P3 Resulting point
 * @param P1 Source point 
 * @param E  Elliptic curve (Jacobi quartic)
 */
void 
jacobi_quartic_point_neg (ec_point_ptr P3, ec_point_srcptr P1, ec_curve_srcptr E);

/* UNIFIED => Resistant to SPA */

/**
 * @fn    void jacobi_quartic_point_add_unified (ec_point_ptr P3, ec_point_srcptr P1, ec_point_srcptr P2, ec_curve_srcptr E, uint8_t stack)
 * @brief Set P3 to P1 + P2 using extended projective coordinate (T=(X^2)/Z), and unified formulae (protection against SPA)
 * @param P3    Resulting point 
 * @param P1    Source point
 * @param P2    Source point 
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void 
jacobi_quartic_point_add_unified (ec_point_ptr P3, ec_point_srcptr P1, ec_point_srcptr P2,
        ec_curve_srcptr E, uint8_t stack);

/* DEDICATED => Not resistant to SPA, but faster 
 * Only the doubling is implemented as dedicated because the dedicated addition is not faster than the unified addition 
 */

/**
 * @fn    jacobi_quartic_point_dbl_dedicated (ec_point_ptr P3, ec_point_srcptr P1, ec_curve_srcptr E, uint8_t stack)
 * @brief Set P3 to 2*P1, using dedicated formulae (not protected against SPA, but faster)
 * @param P3    Resulting point
 * @param P1    Source point 
 * @param E     Elliptic curve (Jacobi quartic)
 * @param stack Temporary memory stack to use
 */
void 
jacobi_quartic_point_dbl_dedicated (ec_point_ptr P3, ec_point_srcptr P1, ec_curve_srcptr E, uint8_t stack);


#endif
