/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France) 

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell-number.h
  * \brief Declaration of arithmetic functions, interface to chose either GMP mpz or number as base type for arithmetic.
  */

#ifndef MPHELL_NUMBER_H
#define MPHELL_NUMBER_H

#include <stdarg.h>

#include "mphell-init.h"
#include "mphell-util.h"

#if MPHELL_USE_GMP == 1
/**
 * @typedef number
 * @brief   Define number using GMP mpz_t type
 */
typedef mpz_t number;

/**
 * @typedef number_t
 * @brief   Define number_t using GMP __mpz_struct type
 */
typedef __mpz_struct number_t;

/**
 * @typedef number_ptr
 * @brief   Define number_ptr, use in all functions
 */
typedef __mpz_struct * number_ptr;

/**
 * @typedef number_srcptr
 * @brief   Define number_srcptr, use in all functions where the parameter is const
 */
typedef const __mpz_struct * number_srcptr;

#define LIMB(x) x->_mp_d
#define SIZ(x) x->_mp_size
#define ALLOC_SIZ(x) x->_mp_alloc
#define SIGN(x) x->_mp_size >= 0 ?  1 : -1

#elif MPHELL_USE_IPP == 1

/**
 * @typedef number
 * @brief   Define number using IPP IppsBigNumState * type
 */
typedef IppsBigNumState * number;

/**
 * @typedef number_t
 * @brief   Define number_t using IPP IppsBigNumState type
 */
typedef IppsBigNumState number_t;

/**
 * @typedef number_ptr
 * @brief   Define number_ptr, use in all functions
 */
typedef IppsBigNumState * number_ptr;

/**
 * @typedef number_srcptr
 * @brief   Define number_srcptr, use in all functions where the parameter is const
 */
typedef IppsBigNumState * number_srcptr;

#elif MPHELL_USE_MBEDTLS == 1

/**
 * @typedef number
 * @brief   Define number using IPP IppsBigNumState * type
 */
typedef mbedtls_mpi number[1];

/**
 * @typedef number_t
 * @brief   Define number_t using IPP IppsBigNumState type
 */
typedef mbedtls_mpi number_t;

/**
 * @typedef number_ptr
 * @brief   Define number_ptr, use in all functions
 */
typedef mbedtls_mpi * number_ptr;

/**
 * @typedef number_srcptr
 * @brief   Define number_srcptr, use in all functions where the parameter is const
 */
typedef const mbedtls_mpi * number_srcptr;

#define LIMB(x) x->p
#define SIZ(x) x->n
#define ALLOC_SIZ(x) x->n
#define SIGN(x) x->s
    
#endif

/*********************************** TMP ALLOC *******************************/

/**
 * @fn    void number_tmp_alloc (number * t, const uint8_t size, uint8_t stack);
 * @brief Allocate a temporary number
 * @param t     Adress of the destination number
 * @param size  Number of block to allocate for t (ex: 4 for a 4 * 64 = 256 bits number_ptr on a 64 bits architecture)
 * @param stack Temporary memory stack to use
 */
void
number_tmp_alloc (number * t, const uint8_t size, uint8_t stack);

/**
 * @fn    void number_tmp_free (number * t, const uint8_t size, uint8_t stack);
 * @brief Free a temporary number
 * @param t     Adress of the destination number_ptr to free
 * @param size  Number of block to allocate for t (ex: 4 for a 4 * 64 = 256 bits number_ptr on a 64 bits architecture)
 * @param stack Temporary memory stack to use
 */
void
number_tmp_free (number * t, const uint8_t size, uint8_t stack);

/************************************ SETTERS ********************************/

/**
 * @fn    void number_init (number * dst, const uint8_t n);
 * @brief Allocate a number_ptr on the RAM memory (malloc)
 * @param dst Adress of the destination number
 * @param n   Number of block to allocate for t (ex: 4 for a 4 * 64 = 256 bits number_ptr on a 64 bits architecture)
 */
void
number_init (number * dst, const uint8_t n);

/**
 * @fn    void number_free (number * dst);
 * @brief Free a number_ptr allocated on the RAM memory (malloc)
 * @param dst Adress of the destination number_ptr to free
 */
void
number_free (number * dst);

/**
 * @fn    void number_copy (number_ptr dst, number_srcptr src);
 * @brief Copy src into dst
 * @param src Source number
 * @param dst Destination number
 */
void
number_copy (number_ptr dst, number_srcptr src);

/**
 * @fn    void number_set_ui (number_ptr dst, const block src);
 * @brief Set dst to src
 * @param dst Destination number
 * @param src Source block (uint32_t or uint64_t according to your architecture)
 */
void
number_set_ui (number_ptr dst, const block src);

/**
 * @brief Set dst to [block_0, block_1, ..., block_(size-1)]. SIZ(dst) must be set before use
 * @param dst  Destination number: dst <- [block_0, block_1, ..., block_(size-1)] where block_0 is the smallest and block_(size-1) the biggest
 * @param sign 1 for positive number, -1 for negative number
 * @param size Number of block that follow
 * @param ...  List of blocks
 */
void 
number_set (number_ptr dst, const int8_t sign, const uint32_t size, ...);

/**
 * @fn    void number_set_bit (number_ptr dst, const uint32_t bit);
 * @brief Set the bit-th of dst to 1, all the others to 0.
 * @param bit Bit to set to 1
 * @param dst Destination number
 */
void
number_set_bit (number_ptr dst, const uint32_t bit);

/**
 * @fn    void number_random (number_ptr dst, const uint32_t size);
 * @brief Set dst to a random number, the random process is chosen at the MPHELL initialisation. SIZ(dst) must be set before use.
 * @param size Number of block of dst
 * @param dst Destination number
 */
void 
number_random (number_ptr dst, const uint32_t size);

/**
 * @fn    void number_random1 (number_ptr dst, number_srcptr bound, uint8_t stack);
 * @brief Set dst to a random number_ptr between 0 and bound, the random process is chosen at the MPHELL initialisation
 * @param dst   Destination number
 * @param bound Bound for the random number
 * @param stack Temporary memory stack to use
 */
void 
number_random1 (number_ptr dst, number_srcptr bound, uint8_t stack);

/*********************************** STRING **********************************/

/**
 * @fn     uint16_t number_log2 (number_srcptr src);
 * @brief  Calculate log2(src), which is the binary size of src
 * @param  src Source number
 * @return log2(src)
 */
uint16_t 
number_log2 (number_srcptr src);

/**
 * @fn     uint8_t number_block_size (number_srcptr src);
 * @brief  Calculate the number of block used by src
 * @param  src Source number
 * @return Size in block used by src
 */
uint8_t
number_block_size (number_srcptr src);

/**
 * @fn     uint8_t number_buffer_size (number_srcptr src);
 * @brief  Give the number of block that src can use
 * @param  src Source number
 * @return Size in block dedicated for src
 */
uint8_t 
number_buffer_size (number_srcptr src);

/**
 * @fn     void number_set_str (number_ptr dst, const char *str, const uint8_t base);
 * @brief  Set dst to str
 * @param  dst  Destination number
 * @param  str  Source number_ptr under string format in base "base"
 * @param  base Base used to write str, can be 2, 10 or 16
 */
void 
number_set_str (number_ptr dst, const char *str, 
        const uint8_t base);

/**
 * @fn     void number_str (char **str, number_srcptr src, const uint8_t base);
 * @brief  Converts src to string format in base specified by base.
 * @param  str  Destination string, allocated by the function, must be free by the user.
 * @param  src  Source number
 * @param  base Base to use to write str, can be 2, 10 or 16
 */
void 
number_str (char **str, number_srcptr src, const uint8_t base);

/**
 * @fn     void number_get_ui (block * dst, number_srcptr src);
 * @brief  Get the less significant block of src.
 * @param  src Source number
 * @param  dst Unsigned integer 
 */
void 
number_get_ui (block * dst, number_srcptr src);

/**
 * @fn     void number_to_bit_string (char *str, number_srcptr src);
 * @brief  Converts src (must be positive) to a bit string
 * @param  str  Destination string. Must be large enough.
 * @param  src  Source number
 */
void 
number_to_bit_string (char *str, number_srcptr src);

/**
 * @fn     void number_print (number_srcptr src, const uint8_t base);
 * @brief  Print src in base "base"
 * @param  src  Source number
 * @param  base Base to use to write str, can be 2, 10 or 16
 */
void 
number_print (number_srcptr src, const uint8_t base);

/*********************************** COMPARISON ******************************/

/**
 * @fn     bool number_iszero(number_srcptr src);
 * @brief  Test if src is zero
 * @param  src Source number
 * @return true if src == 0
 * @return false otherwise
 */
bool
number_iszero(number_srcptr src);

/**
 * @fn     bool number_isequal_ui (number_srcptr src1, const block src2);
 * @brief  Test if src1 == src2
 * @param  src1  Source number
 * @param  src2  Source block 
 * @return true  if src1 == src2
 * @return false if src1 != src2
 */
bool 
number_isequal_ui (number_srcptr src1, const block src2);

/**
 * @fn     bool number_isequal (number_srcptr src1, number_srcptr src2);
 * @brief  Test if src1 == src2
 * @param  src1  Source number
 * @param  src2  Source number
 * @return true  if src1 == src2
 * @return false if src1 != src2
 */
bool 
number_isequal (number_srcptr src1, number_srcptr src2);

/**
 * @fn     bool number_isdiff_ui (number_srcptr src1, const block src2);
 * @brief  Test if src1 != src2
 * @param  src1  Source number
 * @param  src2  Source block
 * @return true  if src1 != src2
 * @return false if src1 == src2
 */
bool 
number_isdiff_ui (number_srcptr src1, const block src2);

/**
 * @fn     bool number_isdiff (number_srcptr src1, number_srcptr src2);
 * @brief  Test if src1 != src2
 * @param  src1  Source number
 * @param  src2  Source number
 * @return true  if src1 != src2
 * @return false if src1 == src2
 */
bool 
number_isdiff (number_srcptr src1, number_srcptr src2);

/**
 * @fn     bool number_islower_ui (number_srcptr src1, const block src2);
 * @brief  Test if src1 < src2
 * @param  src1  Source number
 * @param  src2  Source block
 * @return true  if src1 < src2
 * @return false otherwise
 */
bool 
number_islower_ui (number_srcptr src1, const block src2);

/**
 * @fn     bool number_islower (number_srcptr src1, number_srcptr src2);
 * @brief  Test if src1 < src2
 * @param  src1  Source number
 * @param  src2  Source number
 * @return true  if src1 < src2
 * @return false otherwise
 */
bool 
number_islower (number_srcptr src1, number_srcptr src2);

/**
 * @fn     bool number_isgreater_ui (number_srcptr src1, const block src2);
 * @brief  Test if src1 > src2
 * @param  src1  Source number
 * @param  src2  Source block
 * @return true  if src1 > src2
 * @return false otherwise
 */
bool
number_isgreater_ui (number_srcptr src1, const block src2);

/**
 * @fn     bool number_isgreater (number_srcptr src1, number_srcptr src2);
 * @brief  Test if src1 > src2
 * @param  src1  Source number
 * @param  src2  Source number
 * @return true  if src1 > src2
 * @return false otherwise
 */
bool 
number_isgreater (number_srcptr src1, number_srcptr src2);

/**
 * @fn     bool number_islowereq_ui (number_srcptr src1, const block src2);
 * @brief  Test if src1 <= src2
 * @param  src1  Source number
 * @param  src2  Source block
 * @return true  if src1 <= src2
 * @return false otherwise
 */
bool 
number_islowereq_ui (number_srcptr src1, const block src2);

/**
 * @fn     bool number_islowereq (number_srcptr src1, number_srcptr src2);
 * @brief  Test if src1 <= src2
 * @param  src1  Source number
 * @param  src2  Source number
 * @return true  if src1 <= src2
 * @return false otherwise
 */
bool
number_islowereq (number_srcptr src1, number_srcptr src2);

/**
 * @fn     bool number_isgreatereq_ui (number_srcptr src1, const block src2);
 * @brief  Test if src1 >= src2
 * @param  src1  Source number
 * @param  src2  Source block
 * @return true  if src1 >= src2
 * @return false otherwise
 */
bool 
number_isgreatereq_ui (number_srcptr src1, const block src2);

/**
 * @fn     bool number_isgreatereq (number_srcptr src1, number_srcptr src2);
 * @brief  Test if src1 >= src2
 * @param  src1  Source number
 * @param  src2  Source number
 * @return true  if src1 >= src2
 * @return false otherwise
 */
bool 
number_isgreatereq (number_srcptr src1, number_srcptr src2);

/**
 * @fn     int8_t number_cmp_ui (number_srcptr src1, const block src2);
 * @brief  Compare src1 and src2
 * @param  src1  Source number
 * @param  src2  Source block
 * @return  0 if src1 == src2
 * @return  1 if src1 > src2
 * @return -1 if src1 < src2
 */
int8_t
number_cmp_ui (number_srcptr src1, const block src2);

/**
 * @fn     int8_t number_cmp (number_srcptr src1, number_srcptr src2);
 * @brief  Compare src1 and src2
 * @param  src1 Source number
 * @param  src2 Source number
 * @return  0 if src1 == src2
 * @return  1 if src1 > src2
 * @return -1 if src1 < src2 
 */
int8_t
number_cmp (number_srcptr src1, number_srcptr src2);

/**
 * @fn     bool number_iseven(number_srcptr src);
 * @brief  Test if src is even
 * @param  src Source number
 * @return true if src == 0 % 2
 * @return false otherwise
 */
bool
number_iseven(number_srcptr src);

/**
 * @fn     bool number_isodd(number_srcptr src);
 * @brief  Test if src is odd
 * @param  src Source number
 * @return true if src == 1 % 2
 * @return false otherwise
 */
bool
number_isodd(number_srcptr src);

/************************************ LOGICAL ********************************/

/**
 * @fn     block number_and_ui (number_srcptr src1, const block src2, uint8_t stack);
 * @brief  Apply logical bitwise AND operator between src1 and src2
 * @param  src1  Source number
 * @param  src2  Source block
 * @param  stack Temporary memory stack to use
 * @return src1 AND src2
 */
block 
number_and_ui (number_srcptr src1, const block src2, uint8_t stack);

/**
 * @fn    void number_and (number_ptr dst, number_srcptr src1, number_srcptr src2);
 * @brief Apply logical bitwise AND operator between src1 and src2
 * @param dst  Destination number
 * @param src1 Source number
 * @param src2 Source number
 */
void 
number_and (number_ptr dst, number_srcptr src1, number_srcptr src2);

/**
 * @fn    void number_or_ui (number_ptr dst, number_srcptr src1, const block src2, uint8_t stack);
 * @brief Apply logical bitwise OR operator between src1 and src2
 * @param dst   Destination number
 * @param src1  Source number
 * @param src2  Source block
 * @param stack Temporary memory stack to use
 */
void 
number_or_ui (number_ptr dst, number_srcptr src1, const block src2, uint8_t stack);

/**
 * @fn    void number_or (number_ptr dst, number_srcptr src1, number_srcptr src2);
 * @brief Apply logical bitwise OR operator between src1 and src2
 * @param dst  Destination number
 * @param src1 Source number
 * @param src2 Source number
 */
void 
number_or (number_ptr dst, number_srcptr src1, number_srcptr src2);

/**
 * @fn    void number_xor_ui (number_ptr dst, number_srcptr src1, const block src2, uint8_t stack);
 * @brief Apply logical bitwise XOR operator between src1 and src2
 * @param dst   Destination number
 * @param src1  Source number
 * @param src2  Source block
 * @param stack Temporary memory stack to use
 */
void 
number_xor_ui (number_ptr dst, number_srcptr src1, const block src2, uint8_t stack);

/**
 * @fn    void number_xor (number_ptr dst, number_srcptr src1, number_srcptr src2);
 * @brief Apply logical bitwise XOR operator between src1 and src2
 * @param dst  Destination number
 * @param src1 Source number
 * @param src2 Source number
 */
void 
number_xor (number_ptr dst, number_srcptr src1, number_srcptr src2);

/**
 * @fn     void number_lshift (number_ptr dst, number_srcptr src, const uint16_t shift);
 * @brief  Set dst to src << shift
 * @param  dst   Destination number
 * @param  src   Source number
 * @param  shift Number of bits to shift
 */
void 
number_lshift (number_ptr dst, number_srcptr src, const uint16_t shift);

/**
 * @fn     void number_rshift (number_ptr dst, number_srcptr src, const uint16_t shift);
 * @brief  Set dst to src >> shift
 * @param  dst   Destination number
 * @param  src   Source number
 * @param  shift Number of bits to shift
 */
void 
number_rshift (number_ptr dst, number_srcptr src, const uint16_t shift);

/********************************* ARITHMETIC ********************************/

/**
 * @fn    void number_add_ui (number_ptr dst, number_srcptr src1, const block src2);
 * @brief Set dst to src1 + src2 if src1 + src2 fit in dst
 * @param dst   Destination number
 * @param src1  Source number_ptr (array of blocks)
 * @param src2  Source number_ptr (block)
 */
void 
number_add_ui (number_ptr dst, number_srcptr src1,
        const block src2);

/**
 * @fn    void number_add (number_ptr dst, number_srcptr src1, number_srcptr src2);
 * @brief Set dst to src1 + src2 if src1 + src2 fit in dst
 * @param dst   Destination number
 * @param src1  Source number_ptr (array of blocks)
 * @param src2  Source number_ptr (array of blocks)
 */
void 
number_add (number_ptr dst, number_srcptr src1, 
        number_srcptr src2);

/**
 * @fn    void number_inc (number_ptr dst, number_srcptr src);
 * @brief Set dst to src + 1 if src + 1 fit in dst
 * @param dst   Destination number
 * @param src   Source number
 */
void
number_inc (number_ptr dst, number_srcptr src);

/**
 * @fn    void number_sub_ui (number_ptr dst, number_srcptr src1, const block src2);
 * @brief Set dst to src1 - src2 if src1 - src2 fit in dst
 * @param dst   Destination number
 * @param src1  Source number_ptr (array of blocks)
 * @param src2  Source number_ptr (block)
 */
void 
number_sub_ui (number_ptr dst, number_srcptr src1, 
        const block src2);

/**
 * @fn    void number_sub (number_ptr dst, number_srcptr src1, number_srcptr src2);
 * @brief Set dst to src1 - src2 if src1 - src2 fit in dst
 * @param dst   Destination number
 * @param src1  Source number_ptr (array of blocks)
 * @param src2  Source number_ptr (array of blocks)
 */
void 
number_sub (number_ptr dst, number_srcptr src1, 
        number_srcptr src2);

/**
 * @fn    void number_dec (number_ptr dst, number_srcptr src);
 * @brief Set dst to src - 1 if src - 1 fit in dst
 * @param dst   Destination number
 * @param src   Source number
 */
void
number_dec (number_ptr dst, number_srcptr src);

/**
 * @fn    void number_mul_ui (number_ptr dst, number_srcptr src1, const block src2);
 * @brief Set dst to src1 * src2
 * @param dst   Destination number, must be of size >= size(src1) + 1
 * @param src1  Source number_ptr (array of blocks)
 * @param src2  Source number_ptr (block)
 */
void 
number_mul_ui (number_ptr dst, number_srcptr src1, const block src2);

/**
 * @fn    void number_addmul_ui (number_ptr dst, number_srcptr src1, const block src2);
 * @brief Set dst to dst + src1 * src2
 * @param dst   Destination number, must be of size >= size(src1) + 1
 * @param src1  Source number_ptr (array of blocks)
 * @param src2  Source number_ptr (block)
 */
void 
number_addmul_ui (number_ptr dst, number_srcptr src1, const block src2);

/**
 * @fn    void number_mul (number_ptr dst, number_srcptr src1, number_srcptr src2);
 * @brief Set dst to src1 * src2
 * @param dst   Destination number, must be of size == size(src1) * 2
 * @param src1  Source number_ptr (array of blocks)
 * @param src2  Source number_ptr (array of blocks) must be of size <= size(src1)
 */
void 
number_mul (number_ptr dst, number_srcptr src1, number_srcptr src2);

/**
 * @fn    void static inline  number_mul_mod (number_ptr dst, number_srcptr src1, number_srcptr src2, number_srcptr mod, uint8_t stack)
 * @brief Set dst to (src1 * src2) % mod
 * @param dst   Destination number, must be of size == size(mod)
 * @param src1  Source number_ptr (array of blocks)
 * @param src2  Source number_ptr (array of blocks) must be of size <= size(src1)
 * @param mod   Modulus number
 * @param stack Temporary memory stack to use
 */
void number_mul_mod (number_ptr dst, number_srcptr src1, number_srcptr src2, number_srcptr mod, uint8_t stack);

/**
 * @fn    void number_addmul (number_ptr dst, number_srcptr src1, number_srcptr src2);
 * @brief Set dst to dst + src1 * src2
 * @param dst   Destination number, must be of size == size(src1) * 2
 * @param src1  Source number_ptr (array of blocks)
 * @param src2  Source number_ptr (array of blocks) must be of size <= size(src1)
 */
void 
number_addmul (number_ptr dst, number_srcptr src1, number_srcptr src2);

/**
 * @fn    void number_pow (number_ptr dst, number_srcptr src1, number_srcptr src2, uint8_t stack);
 * @brief Set dst to src1^src2
 * @param dst   Destination number, must be of size == size(src1) * size(src2)
 * @param src1  Source number_ptr (array of blocks)
 * @param src2  Source number_ptr (array of blocks)
 * @param stack Temporary memory stack to use
 */
void 
number_pow (number_ptr dst, number_srcptr src1, number_srcptr src2, uint8_t stack);

/**
 * @fn    void number_powm (number_ptr dst, number_srcptr src1, number_srcptr src2, number_srcptr mod, uint8_t stack);
 * @brief Set dst to src1^src2 modulo mod
 * @param dst   Destination number, must be of size of mod
 * @param src1  Source number_ptr (array of blocks)
 * @param src2  Source number_ptr (array of blocks)
 * @param mod   Modulus number_ptr (array of blocks)
 * @param stack Temporary memory stack to use
 */
void 
number_powm (number_ptr dst, number_srcptr src1, number_srcptr src2, number_srcptr mod, uint8_t stack);

/**
 * @fn    void number_powm_ui (number_ptr dst, number_srcptr src1, const block src2, number_srcptr mod, uint8_t stack);
 * @brief Set dst to src1^src2 modulo mod
 * @param dst   Destination number, must be of size of mod
 * @param src1  Source number_ptr (array of blocks)
 * @param src2  Source number_ptr (array of blocks)
 * @param mod   Modulus number_ptr (array of blocks)
 * @param stack Temporary memory stack to use
 */
void 
number_powm_ui (number_ptr dst, number_srcptr src1, const block src2, number_srcptr mod, uint8_t stack);

/**
 * @fn    void number_sqr (number_ptr dst, number_srcptr src);
 * @brief Set dst to src1^2
 * @param dst   Destination number, must be of size == size(src1) * 2
 * @param src1  Source number_ptr (array of blocks)
 */
void 
number_sqr (number_ptr dst, number_srcptr src);

/**
 * @fn    void number_divmod_ui (number_ptr q, block *r, number_srcptr src1, const block src2);
 * @brief Compute (q, r) such that src1 = q * src2 + r ; r < src2
 * @param q    Destination quotient (array of blocks)
 * @param r    Destination rest (block)
 * @param src1 Source dividende (array of blocks)
 * @param src2 Source divisor (block)
 */
void
number_divmod_ui (number_ptr q, block *r, number_srcptr src1, 
        const block src2);

/**
 * @fn    void number_divmod (number_ptr q, number_ptr r, number_srcptr src1, number_srcptr src2);
 * @brief Compute (q, r) such that src1 = q * src2 + r ; r < src2
 * @param q    Destination quotient (array of blocks)
 * @param r    Destination rest (array block)
 * @param src1 Source dividende (array of blocks)
 * @param src2 Source divisor (array of blocks)
 */
void 
number_divmod (number_ptr q, number_ptr r, number_srcptr src1, 
        number_srcptr src2);

/**
 * @fn    void number_div_ui (number_ptr dst, number_srcptr src1, const block src2);
 * @brief Compute dst such that src1 = dst * src2 + r ; r < src2
 * @param dst  Destination quotient (array of blocks)
 * @param src1 Source dividende (array of blocks)
 * @param src2 Source divisor (block)
 */
void
number_div_ui (number_ptr dst, number_srcptr src1, const block src2);

/**
 * @fn    void number_div (number_ptr dst, number_srcptr src1, number_srcptr src2);
 * @brief Compute dst such that src1 = dst * src2 + r ; r < src2
 * @param dst  Destination quotient (array of blocks)
 * @param src1 Source dividende (array of blocks)
 * @param src2 Source divisor (array of blocks)
 */
void 
number_div (number_ptr dst, number_srcptr src1, number_srcptr src2);

/**
 * @fn    void number_mod_ui (block *dst, number_srcptr src1, const block src2);
 * @brief Compute dst such that src1 = q * src2 + dst ; dst < src2
 * @param dst  Destination rest (block)
 * @param src1 Source dividende (array of blocks)
 * @param src2 Source divisor (block)
 */
void 
number_mod_ui (block *dst, number_srcptr src1, const block src2);

/**
 * @fn    void number_mod (number_ptr dst, number_srcptr src1, number_srcptr src2);
 * @brief Compute dst such that src1 = q * src2 + dst ; dst < src2
 * @param dst  Destination rest (array block)
 * @param src1 Source dividende (array of blocks)
 * @param src2 Source divisor (array of blocks)
 */
void 
number_mod (number_ptr dst, number_srcptr src1, number_srcptr src2);

/**
 * @fn    void number_invmod (number_ptr dst, number_srcptr src1, number_srcptr src2);
 * @brief Compute dst such that dst = src1^(-1) mod src2
 * @param dst  Destination rest (array block)
 * @param src1 Source (array of blocks)
 * @param src2 Source modulus (array of blocks)
 */
void 
number_invmod (number_ptr dst, number_srcptr src1, number_srcptr src2);

/**
 * @fn    void number_inv_mod_basis(block * dst, const block src);
 * @brief Compute dst such that dst * (-src) = -1 mod (2^(BLOCK_SIZE))
 * @param dst Destination (block)
 * @param src Source (block)
 */
void 
number_inv_mod_basis(block * dst, const block src);

/**
 * @fn    void redcify(number_ptr dst, number_srcptr src, number_srcptr p, uint8_t stack);
 * @brief Set dst to redc form dst=B^n*src mod p
 * @param dst Output number_ptr and MontgomeryForm of src
 * @param src Input number_ptr in "classical form"
 * @param p   Modulus
*/
void
redcify(number_ptr dst, number_srcptr src, number_srcptr p, uint8_t stack);

/**
 * @fn    void number_mul_montgomery(number_ptr dst, number_srcptr src1, number_srcptr src2, number_srcptr p, const block invp, uint8_t stack);
 * @brief Compute dst such that dst = (src1 * src2) mod(p) into the Montgomery form
 * @param dst   Destination (array block)
 * @param src1  Source (array of blocks)
 * @param src2  Source (array of blocks)
 * @param p     Prime number_ptr (array of blocks)
 * @param invp  Computed with: number_inv_mod_basis(invp, p->limb[0]);
 * @param stack Temporary memory stack to use
 */
void
number_mul_montgomery(number_ptr dst, number_srcptr src1, number_srcptr src2, number_srcptr p, const block invp, uint8_t stack);

/**
 * @fn    void number_mul_montgomery_block(number_ptr dst, number_srcptr src1, number_srcptr src2, number_srcptr p, const block invp, uint8_t stack);
 * @brief Compute dst such that dst = (src1 * src2) mod(p) into the Montgomery form, use when SIZ(src1) or SIZ(src2) != SIZ(p)
 * @param dst   Destination (array block)
 * @param src1  Source (array of blocks)
 * @param src2  Source (array of blocks)
 * @param p     Prime number_ptr (array of blocks)
 * @param invp  Computed with: number_inv_mod_basis(invp, p->limb[0]);
 * @param stack Temporary memory stack to use
 */
void
number_mul_montgomery_block(number_ptr dst, number_srcptr src1, number_srcptr src2, number_srcptr p, const block invp, uint8_t stack);

/**
 * @fn    void number_inv_montgomery(number_ptr dst, number_srcptr src, number_srcptr p, const block invp, number_srcptr R2, number_srcptr one, uint8_t stack);
 * @brief Compute dst such that dst = src^(-1) mod(p) into the Montgomery form
 * @param dst   Destination (array block)
 * @param src   Source (array of blocks)
 * @param p     Prime number_ptr (array of blocks)
 * @param invp  Computed with: number_inv_mod_basis(invp, p->limb[0]);
 * @param R2    Square of the Montgomery basis
 * @param one   Neutral element for the multiplication
 * @param stack Temporary memory stack to use
 */
void 
number_inv_montgomery(number_ptr dst, number_srcptr src, number_srcptr p, const block invp, number_srcptr R2, number_srcptr one, uint8_t stack);

/**
 * @fn     int8_t number_legendre (number_srcptr src1, number_srcptr src2, uint8_t stack);
 * @brief  Compute the legendre symbole of src1 and src2
 * @param  dst  Destination (array block)
 * @param  src1 Source (array of blocks)
 * @param  src2 Source (array of blocks), must be a prime number
 * @param  stack Temporary memory stack to use
 * @return  O if p divide src2
 * @return  1 if src1 is a square mod src2
 * @return -1 if src1 is not a square mod src2
 */
int8_t 
number_legendre (number_srcptr src1, number_srcptr src2, uint8_t stack);

/**
 * @fn    void number_gcd_ui (block *dst, number_srcptr src1, const block src2);
 * @brief Set dst to GCD(src1, src2)
 * @param dst  Destination block
 * @param src1 Source (array of blocks)
 * @param src2 Source block
 */
void 
number_gcd_ui (block *dst, number_srcptr src1, const block src2);

/**
 * @fn    void number_gcd (number_ptr dst, number_srcptr src1, number_srcptr src2);
 * @brief Set dst to GCD(src1, src2)
 * @param dst  Destination (array block)
 * @param src1 Source (array of blocks)
 * @param src2 Source (array of blocks)
 */
void 
number_gcd (number_ptr dst, number_srcptr src1, number_srcptr src2);

/**
 * @fn     uint8_t number_Rabin_Miller (number_srcptr src, const uint16_t bound);
 * @brief  Apply the Rabin-Miller test to src with the bound 
 * @param  src   Source (array of blocks)
 * @param  bound Positive integer, 50 is a good choice
 * @return 0 if src is not prime
 * @return 2 if src is prime
 * @return 1 if src is probably prime, a composite number_ptr will be identified as a prime with a probability less than 4^(-bound)
 */
uint8_t
number_Rabin_Miller (number_srcptr src, const uint16_t bound);

#endif

