/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France) 

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell-random.h
  * \brief Declaration of random functions.
  */

#ifndef RANDOM_H
#define RANDOM_H

#include "mphell-drbg.h"

/**
 * @enum  random_algo
 * @brief Define the random algorithm (DRBG algorithm)
 */
enum random_algo {
    RANDOM_INTEL,       /*!< Use the RDRAND instruction, available since Ivy Bridge ("third generation" of the Intel Core processors) */
    RANDOM_AES128,      /*!< Use AES with 128 bits key                  */
    RANDOM_AES192,      /*!< Use AES with 192 bits key                  */
    RANDOM_AES256,      /*!< Use AES with 256 bits key                  */
    RANDOM_SHA1,        /*!< Use SHA1                                   */
    RANDOM_SHA256,      /*!< Use SHA256                                 */
    STM32F4_RNG,        /*!< Use the STM32F4 True Random Generator      */
    RANDOM_UNSET        /*!< Used when no random algorithm is specified */
};

/**
 * @typedef random_type
 * @brief   Define the random algorithm (DRBG algorithm)
 */
typedef enum random_algo random_type;

/**
 * @struct random_state_t
 * @brief  Define a random state
 */
typedef struct {
    int8_t state_handle;    /*!< Store the used DRBG state number (we can use until DRBG_MAX_INTERNAL_STATES states)    */
    random_type type;       /*!< Random algorithm                                                                       */
}random_state_t;

/**
 * @typedef random_state
 * @brief   Address of a random_state_t structure
 */
typedef random_state_t random_state[1];

/**
 * @fn    void random_instantiate(mphell_status ret, random_state initial_state, const uint16_t requested_instantiation_security_strength, const uint8_t* personalization_string, const uint64_t pers_length, const random_type type, const entropy_type entropy, uint64_t reseed_interval);
 * @brief Instantiate one random state
 * @param ret                                       Set to MPHELL_ERROR_RANDOM if an error occur, MPHELL_SUCCESS otherwise
 * @param initial_state                             Destination structure, store the random algorithm and the number of the DRBG state used.
 * @param requested_instantiation_security_strength Security (in bits) required
 * @param personalization_string                    A string given by the user can be added to the material seed to instantiate the DRBG
 * @param pers_length                               Length of personalization_string
 * @param type                                      Random algorithm to use (RANDOM_INTEL, RANDOM_AES128, RANDOM_AES192, RANDOM_AES256, RANDOM_SHA1, RANDOM_SHA256)
 * @param entropy                                   Entropy to use (to seed the DRBG): INTEL, DEVURANDOM, DEVRANDOM
 * @param reseed_interval                           Reseed interval, must be between 1 (prediction resistance == true) and 2^48
 */
void
random_instantiate(mphell_status ret, random_state initial_state, 
        const uint16_t requested_instantiation_security_strength, 
        const uint8_t* personalization_string,
        const uint64_t pers_length, const random_type type,
        const entropy_type entropy, uint64_t reseed_interval);

/**
 * @fn    void random_reseed(mphell_status ret, random_state working_state, const uint16_t entropy_length, const entropy_type entropy_src, const uint8_t* additional_input, const uint64_t add_length);
 * @brief Reseed the DRBG (if you want to reseed it before it does by itself)
 * @param ret              Set to MPHELL_ERROR_RANDOM if an error occur, MPHELL_SUCCESS otherwise
 * @param working_state    Define which DRBG state to use
 * @param entropy_length   Number of bits to extract from the entropy source
 * @param entropy_src      Entropy to use (to seed the DRBG): INTEL, DEVURANDOM, DEVRANDOM
 * @param additional_input A string given by the user can be added to the material seed to reseed the DRBG
 * @param add_length       Length of additional_input
 */
void
random_reseed(mphell_status ret, random_state working_state, 
        const uint16_t entropy_length, const entropy_type entropy_src,
        const uint8_t* additional_input,
        const uint64_t add_length);

/**
 * @fn    void random_generate(mphell_status ret, uint8_t* returned_bytes, random_state working_state, const uint32_t requested_number_of_bytes, const uint16_t requested_security_strength, const uint8_t* additionnal_input, const uint64_t add_length);
 * @brief Gererate an array of random bits
 * @param ret                           Set to MPHELL_ERROR_RANDOM if an error occur, MPHELL_SUCCESS otherwise
 * @param returned_bytes                Destination array (of requested_number_of_bytes random bits)
 * @param working_state                 Define which DRBG state to use
 * @param requested_number_of_bytes     Number of random bits required
 * @param requested_security_strength   Security (in bits) required
 * @param additionnal_input             A string given by the user can be added to the material seed to reseed the DRBG
 * @param add_length                    Length of additional_input
 */
void
random_generate(mphell_status ret, uint8_t* returned_bytes, 
        random_state working_state,
        const uint32_t requested_number_of_bytes, 
        const uint16_t requested_security_strength,
        const uint8_t* additionnal_input, const uint64_t add_length);

/**
 * @fn    void random_uninstantiate(mphell_status ret, random_state working_state);
 * @brief Uninstantiate a random state
 * @param ret           Set to MPHELL_ERROR_RANDOM if an error occur, MPHELL_SUCCESS otherwise
 * @param working_state Random state to uninstantiate
 */
void
random_uninstantiate(mphell_status ret, random_state working_state);

/**
 * @var   stm32f4_rng
 * @brief This function pointer must be set in the stm32f4 application before calling init_mphell
 */
extern void (*stm32f4_rng)(mphell_status ret, uint8_t *data, const uint16_t length);

#endif
