/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France) 

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
 * \file  mphell-sha256.h
 * \brief Declaration of sha256 functions.
 */

#ifndef SHA256_H 
#define SHA256_H

#include <stdint.h>
#include <string.h>

/* DBL_INT_ADD treats two unsigned ints a and b as one 64-bit integer and adds c to it */
#define DBL_INT_ADD(a,b,c) if (a > 0xffffffff - (c)) ++b; a += c;
#define ROTLEFT(a,b) (((a) << (b)) | ((a) >> (32-(b))))
#define ROTRIGHT(a,b) (((a) >> (b)) | ((a) << (32-(b))))

#define CH(x,y,z) (((x) & (y)) ^ (~(x) & (z)))
#define MAJ(x,y,z) (((x) & (y)) ^ ((x) & (z)) ^ ((y) & (z)))
#define EP0(x) (ROTRIGHT(x,2) ^ ROTRIGHT(x,13) ^ ROTRIGHT(x,22))
#define EP1(x) (ROTRIGHT(x,6) ^ ROTRIGHT(x,11) ^ ROTRIGHT(x,25))
#define SIG0(x) (ROTRIGHT(x,7) ^ ROTRIGHT(x,18) ^ ((x) >> 3))
#define SIG1(x) (ROTRIGHT(x,17) ^ ROTRIGHT(x,19) ^ ((x) >> 10))

/**
 * @struct sha256_state_t
 * @brief  All the sha256 computations are done on this structure
 */
struct sha256_state_t
{
    unsigned char data[64]; /*!< block to store user data, and work on it                                   */
    uint8_t datalen;        /*!< counter, each time datalen==64, states are modified, and datalen is reset. */
    uint32_t bitlen[2];     /*!< contain the user message bit length.                                       */
    uint32_t state[8];      /*!< stores the value H0, H1, H2, H3, H4, H5, H6, H7.                           */
};

/**
 * @typedef sha256_state
 * @brief   All the sha256 computations are done on this structure
 */
typedef struct sha256_state_t sha256_state;

/**
 * @fn    void sha256(uint8_t *hashvalue, const uint8_t* data, const uint64_t data_len);
 * @brief Compute the Sha256 hash of "data"
 * @param hashvalue Resulting hash
 * @param data      User message
 * @param data_len  Data length (in bits)
 */
void
sha256(uint8_t *hashvalue, const uint8_t* data, const uint64_t data_len);
#endif
