/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France) 

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
 * \file  mphell-sha512.h
 * \brief Declaration of sha512 functions.
 */

#ifndef SHA512_H 
#define SHA512_H

#include <stdint.h>
#include <string.h>

/* DBL_INT_ADD treats two unsigned ints a and b as one 64-bit integer and adds c to it */
#define DBL_INT_ADD(a,b,c) if (a > 0xffffffff - (c)) ++b; a += c;
#define ROTLEFT_512(a,b) (((a) << (b)) | ((a) >> (64-(b))))
#define ROTRIGHT_512(a,b) (((a) >> (b)) | ((a) << (64-(b))))

#define CH_512(x,y,z) (((x) & (y)) ^ (~(x) & (z)))
#define MAJ_512(x,y,z) (((x) & (y)) ^ ((x) & (z)) ^ ((y) & (z)))
#define EP0_512(x) (ROTRIGHT_512(x,28) ^ ROTRIGHT_512(x,34) ^ ROTRIGHT_512(x,39))
#define EP1_512(x) (ROTRIGHT_512(x,14) ^ ROTRIGHT_512(x,18) ^ ROTRIGHT_512(x,41))
#define SIG0_512(x) (ROTRIGHT_512(x,1) ^ ROTRIGHT_512(x,8) ^ ((x) >> 7))
#define SIG1_512(x) (ROTRIGHT_512(x,19) ^ ROTRIGHT_512(x,61) ^ ((x) >> 6))

/**
 * @struct sha512_state_t
 * @brief  All the sha512 computations are done on this structure
 */
struct sha512_state_t
{
    unsigned char data[128]; /*!< block to store user data, and work on it                                      */
    uint8_t datalen;         /*!< counter, each time datalen==64, states are modified, and datalen is reset.    */
    uint64_t bitlen[2];      /*!< contain the user message bit length.                                          */
    uint64_t state[8];       /*!< stores the value H0, H1, H2, H3, H4, H5, H6, H7.                              */
};

/**
 * @typedef sha512_state
 * @brief   All the sha512 computations are done on this structure
 */
typedef struct sha512_state_t sha512_state;

/**
 * @fn    void sha512(uint8_t *hashvalue, const uint8_t* data, const uint64_t data_len);
 * @brief Compute the Sha512 hash of "data"
 * @param hashvalue Resulting hash
 * @param data      User message
 * @param data_len  Data length (in bits)
 */
void
sha512(uint8_t *hashvalue, const uint8_t* data, const uint64_t data_len);
#endif
