/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France) 

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell-tmp.h
  * \brief Declaration of internal MPHELL temporary memory functions.
  */

#ifndef MPHELL_TMP_H
#define MPHELL_TMP_H

#include <stdbool.h>

#include "mphell-errors.h"

/*********************************** TMP ALLOC *******************************/

/**
 * @fn    void alloc_tmp_memory();
 * @brief Allocate the stack for temporary elements, its size is defined in mphell-define.h
 */
void
alloc_tmp_memory();

/**
 * @fn    void free_tmp_memory();
 * @brief Free the stack for temporary elements
 */
void
free_tmp_memory();

#if MPHELL_USE_GMP == 1

/**
 * @fn    void gmp_tmp_alloc(mpz_t t, const uint8_t size, uint8_t stack);
 * @brief Allocate a temporary mpz_t
 * @param t     Destination numer
 * @param size  Number of block to allocate for t (ex: 4 for a 4 * 64 = 256 bits mpz_t on a 64 bits architecture)
 * @param stack Temporary memory stack to use
 */
void
gmp_tmp_alloc(__mpz_struct * t, const uint8_t size, uint8_t stack);

/**
 * @fn    void gmp_tmp_free(__mpz_struct * t, const uint8_t size, uint8_t stack);
 * @brief Free a temporary number
 * @param t     Destination number to free
 * @param size  Number of block to allocate for t (ex: 4 for a 4 * 64 = 256 bits mpz_t on a 64 bits architecture)
 * @param stack Temporary memory stack to use
 */
void
gmp_tmp_free(__mpz_struct * t, const uint8_t size, uint8_t stack);

#elif MPHELL_USE_IPP == 1

/**
 * @fn    void ippcp_bn_tmp_alloc(number * t, const uint8_t size, uint8_t stack);
 * @brief Allocate a temporary IppsBigNumState
 * @param t     Destination IppsBigNumState
 * @param size  Number of block to allocate for t (ex: 8 for a 8 * 32 = 256 bits number)
 * @param stack Temporary memory stack to use
 */
void
ippcp_bn_tmp_alloc(IppsBigNumState ** t, const uint8_t size, uint8_t stack);

/**
 * @fn    void ippcp_bn_tmp_free(IppsBigNumState ** t, const uint8_t size, uint8_t stack);
 * @brief Free a temporary IppsBigNumState
 * @param t     Destination IppsBigNumState to free
 * @param size  Number of block to allocate for t (ex: 8 for a 8 * 32 = 256 bits number)
 * @param stack Temporary memory stack to use
 */
void
ippcp_bn_tmp_free(IppsBigNumState ** t, const uint8_t size, uint8_t stack);

#elif MPHELL_USE_MBEDTLS == 1

/**
 * @fn    void mbedtls_tmp_alloc(mbedtls_mpi * t, const uint8_t size, uint8_t stack);
 * @brief Allocate a temporary mbedtls_mpi
 * @param t     Destination mbedtls_mpi
 * @param size  Number of block to allocate for t (ex: 8 for a 8 * 32 = 256 bits number)
 * @param stack Temporary memory stack to use
 */
void
mbedtls_tmp_alloc(mbedtls_mpi * t, const uint8_t size, uint8_t stack);

/**
 * @fn    void mbedtls_tmp_free(mbedtls_mpi * t, const uint8_t size, uint8_t stack);
 * @brief Free a temporary mbedtls_mpi
 * @param t     Destination mbedtls_mpi to free
 * @param size  Number of block to allocate for t (ex: 8 for a 8 * 32 = 256 bits number)
 * @param stack Temporary memory stack to use
 */
void
mbedtls_tmp_free(mbedtls_mpi * t, const uint8_t size, uint8_t stack);
    
#endif

/**
 * @fn     uint32_t get_stack_size(uint8_t stack);
 * @brief  Get the current size of stack nb stack
 * @param  stack Temporary memory stack to use
 * @return Current size of stack nb stack
 */
uint32_t get_stack_size(uint8_t stack);

/**
 * @fn     bool verify_stack_size(uint32_t prev_size, uint8_t stack);
 * @brief  Compare the current size and the previous size of stack nb stack
 * @param  prev_size Previous size of the stack to compare with
 * @param  stack     Temporary memory stack to use
 * @return True if current size == previous size
 * @return False otherwise
 */
bool verify_stack_size(uint32_t prev_size, uint8_t stack);

#endif
