/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France) 

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell-util.h
  * \brief Declaration of miscellaneous function
  */

#ifndef MPHELL_UTIL_H
#define MPHELL_UTIL_H

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>

#include "mphell-define.h"

/* For all tests */

/**
 * @fn      uint8_t bits_to_nblock (const uint16_t nbits);
 * @brief   Return the number of blocks required to store a nbits number
 * @param   nbits Number of bits
 * @return  Number of block required to store a nbits number
 */
uint8_t 
bits_to_nblock (const uint16_t nbits);

/**
 * @fn      int max_size(uint8_t a, uint8_t b);
 * @brief   Return the max of a and b
 * @param   a size of an element in number blocks (< 256)
 * @param   b size of an element in number blocks (< 256)
 * @return  Max between a and b
 */
uint8_t max_size(uint8_t a, uint8_t b);

/**
 * @fn      time_t get_s(struct timespec * start,  struct timespec * end);
 * @brief   Return the elapsed time in second between "start" and "end".
 * @param   start  Begin of chrono  
 * @param   end    End of chrono 
 * @return  Elapse time in second
 */
time_t get_s(struct timespec * start,  struct timespec * end);

/**
 * @fn      long get_ns(struct timespec * start,  struct timespec * end);
 * @brief   Return the elapsed time in nano second to add to the result of get_s().
 * @param   start  Begin of chrono  
 * @param   end    End of chrono 
 * @return  Elapse time in nano second to add to the result of get_s().
 */
unsigned int get_ns(struct timespec * start,  struct timespec * end);

/* For DRBG tests */

/**
 * @fn    void print_bytes_string_hex(const uint8_t * bytes_string, uint16_t length); 
 * @brief Print the byte string "bytes_string" under hexadecimal form
 * @param bytes_string Byte string to print
 * @param length       Length of bytes_string (in byte)
 */
void 
print_bytes_string_hex(const uint8_t * bytes_string, uint16_t length);

/**
 * @fn    void bytes_string2hex(char * hex_string, const uint8_t * bytes_string, uint16_t length);
 * @brief Convert the byte string "bytes_string" under hexadecimal form
 * @param hex_string   Resulting hexadecimal string, must be allocated of size length*2
 * @param bytes_string Byte string to convert
 * @param length       Length of bytes_string (in byte)
 */
void 
bytes_string2hex(char * hex_string, const uint8_t * bytes_string, uint16_t length);

/**
 * @fn    void hex_string2bytes(uint8_t * bytes_string, const char * hex_string, uint16_t length)
 * @brief Convert the hexadecimal string "hex_string" under bytes string form
 * @param bytes_string Resulting byte string, must be allocated of size length/2
 * @param hex_string   Hexadecimal string to convert
 * @param length       Length of resulting byte string
 */
void 
hex_string2bytes(uint8_t * bytes_string, const char * hex_string, uint16_t length);

/**
 * @fn     void bin2hex(const unsigned char *old, const uint32_t oldlen, unsigned char* result);
 * @brief  Converts a string of ASCII characters to hexadecimal values
 * @param  old    ASCII string to convert
 * @param  oldlen Length of old
 * @param  result Resulting hexadecimal string, must be of size 2*oldlen+1.
 */
void 
bin2hex(const unsigned char *old, const uint32_t oldlen, unsigned char* result);

/**
 * @fn     uint32_t hex2bin(const char *in, unsigned char *out);
 * @brief  Convert an hexadecimal string into a binary string
 * @param  in  Hexadecimal string
 * @param  out Binary string
 * @return Length (int bytes) of "out"
 */
uint32_t 
hex2bin(const char *in, unsigned char *out);

/**
 * @fn     int FindMarker (FILE *infile, const char *marker);
 * @brief  Find a marker in a file
 * @param  infile Filename
 * @param  marker Marker to find
 * @return O if marker is not find
 * @return 1 if marker is not find
 */
int 
FindMarker (FILE *infile, const char *marker);

/**
 * @fn     int ReadHex (FILE *infile, char *A, const int Length, char *str);
 * @brief  Read an hexadecimal string in file "infile"
 * @param  infile Filename
 * @param  A      Destination string
 * @param  Length Size in bytes of the string to read
 * @param  str    Marker used to detect the string
 * @return O if marker is not find
 * @return 1 if marker is not find
 */
int 
ReadHex (FILE *infile, char *A, const int Length, char *str);

/* For ecdsa benchmark */

/**
 * @fn    void char2bin(char c, uint8_t * b);
 * @brief Convert a character into a byte (according to its ascii value)
 * @param c character to convert
 * @param b destination byte
 */
void 
char2bin(char c, uint8_t * b);

/**
 * @fn    void printchar2bin(uint8_t b);
 * @brief print a byte
 * @param b byte to print
 */
void 
printchar2bin(uint8_t b);

/**
 * @fn    void  string2bin(char* s, uint8_t * output);
 * @brief Convert a string into a byte array
 * @param s string to convert
 * @param output destination byte array
 */
void 
string2bin(char* s, uint8_t * output);

/**
 * @fn    void printstring2bin(uint8_t * s, uint32_t len);
 * @brief Print a byte array
 * @param s byte array
 * @param len length of "s"
 */
void 
printstring2bin(uint8_t * s, uint32_t len);

#endif
