/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France) 

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell-weierstrass.h
  * \brief Declaration of Weierstrass ECC function.
  */

#ifndef MPHELL_WEIERSTRASS_H
#define MPHELL_WEIERSTRASS_H

/* E: y^2 = x^3 + ax + b */
/* Projective coordinates represent an affine point (x,y) on a Weierstrass-form elliptic curve y^2 = x^3 + ax + b as (X:Y:Z) satisfying Y^2 Z = X^3 + aXZ^2 + bZ^3. Here (X:Y:Z) = (sX:sY:sZ) for all nonzero s. */
/* The triple (X, Y, Z) represents the affine point (X / Z, Y / Z) */

/* Jacobian coordinates represent an affine point (x,y) on a Weierstrass-form elliptic curve y^2 = x^3 + ax + b as (X:Y:Z) satisfying Y^2 = X^3 + aXZ^4 + bZ^6. Here (X:Y:Z) = (s^2 X:s^3 Y:sZ) for all nonzero s.*/
/* The triple (X, Y, Z) represents the affine point (X / Z^2, Y / Z^3). */


/************************************SETTERS**********************************/

/**
 * @fn    void weierstrass_compute_disc(ec_curve E, uint8_t stack);
 * @brief Set the discriminant of E: disc = -16(4a^3 + 27b^2)
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void
weierstrass_compute_disc(ec_curve E, uint8_t stack);

/**
 * @fn     bool weierstrass_verify_random_generation(ec_curve E, const char * seed, uint8_t stack)
 * @brief  Test if E if generated from the seed "seed"
 * @param  E     Elliptic curve to test
 * @param  seed  Seed (of 160 bits) used to generate an elliptic curve
 * @param  stack Temporary memory stack to use
 * @return true  if E is generated from seed
 * @return false otherwise
 */
bool 
weierstrass_verify_random_generation(ec_curve E, const char * seed, uint8_t stack);

/**
 * @fn    void weierstrass_curve_random_generation(fe_ptr a, fe_ptr b, char * seed_res, field_srcptr k, uint8_t stack)
 * @brief Generate a 160 bits seed and coefficients a and b defining a Weiestrass elliptic curve. The curve is not tested to be secure !!!
 * @param a        Destination coefficient, generated from a random seed 
 * @param b        Destination coefficient, generated from a random seed 
 * @param seed_res Destination seed (of 160 bits) used to generate an elliptic curve
 * @param k        Base field, must be created by the user before use.
 * @param stack    Temporary memory stack to use
 */
void 
weierstrass_curve_random_generation(fe_ptr a, fe_ptr b, char * seed_res, field_srcptr k, uint8_t stack);

/**
 * @fn    void weierstrass_point_set_neutral (ec_point_ptr dst, ec_curve_srcptr E, uint8_t stack);
 * @brief Set dst to the neutral element: (0,1,0) for projective coordinates and (1,1,0) for jacobian coordinates.
 * @param dst   Destination point
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void 
weierstrass_point_set_neutral (ec_point_ptr dst, ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void weierstrass_point_set_aff (ec_point_ptr P, fe_srcptr x, fe_srcptr y, field_srcptr k);
 * @brief Set dest to the affine point (x, y)
 * @param P Destination point
 * @param x x-coordinate of the affine point (x, y) 
 * @param y y-coordinate of the affine point (x, y) 
 * @param k Base field
 */
void 
weierstrass_point_set_aff (ec_point_ptr P, fe_srcptr x, fe_srcptr y, field_srcptr k);

/**
 * @fn    void weierstrass_point_set_aff_str (ec_point_ptr P, const char *str_x, const char *str_y, const bool is_reduced, const uint8_t base, field_srcptr k, uint8_t stack);
 * @brief Set dest to the affine point (str_x,str_y)
 * @param P Destination point
 * @param str_x      x-coordinate of the affine point (x, y) under string form
 * @param str_y      y-coordinate of the affine point (x, y) under string form
 * @param is_reduced Set true if str_x and str_y are already reduced in k, false otherwise
 * @param base       Base (either 10 or 16) used for str_x and str_y
 * @param k          Base field
 * @param stack      Temporary memory stack to use
 */
void 
weierstrass_point_set_aff_str (ec_point_ptr P, const char *str_x, const char *str_y,
        const bool is_reduced, const uint8_t base, field_srcptr k, uint8_t stack);

/**
 * @fn    bool weierstrass_belongs (ec_point_srcptr P, ec_curve_srcptr E, uint8_t stack);
 * @brief Test if P belongs to E
 * @param  P     Point 
 * @param  E     Elliptic curve
 * @param  stack Temporary memory stack to use
 * @return true  if P belongs to E
 * @return false otherwise
 */
bool
weierstrass_belongs (ec_point_srcptr P, ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void weierstrass_point_random (ec_point_ptr P, ec_curve_srcptr E, uint8_t stack);
 * @brief Set P to a random point on E
 * @param P     Destination point
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void 
weierstrass_point_random (ec_point_ptr P, ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void weierstrass_point_norm (ec_point_ptr P, ec_curve_srcptr E, uint8_t stack);
 * @brief Set P in affine coordinates
 * @param P     Point
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void
weierstrass_point_norm (ec_point_ptr P, ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void weierstrass_point_get_x_affine (field_elt x, ec_point_ptr P, ec_curve_srcptr E, uint8_t stack)
 * @brief Convert P->x to its affine representation
 * @param x     P->x in affine coordinates
 * @param P     Point
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void
weierstrass_point_get_x_affine (field_elt x, ec_point_ptr P, ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void weierstrass_point_get_y_affine (field_elt y, ec_point_ptr P, ec_curve_srcptr E, uint8_t stack)
 * @brief Convert P->y to its affine representation
 * @param x     P->y in affine coordinates
 * @param P     Point
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void
weierstrass_point_get_y_affine (field_elt y, ec_point_ptr P, ec_curve_srcptr E, uint8_t stack);


/*******************************COMPARISON************************************/

/**
 * @fn     bool weierstrass_point_is_neutral (ec_point_srcptr P, ec_curve_srcptr E, uint8_t stack);
 * @brief  Test if P is the neutral element
 * @param  P     Point to test
 * @param  E     Elliptic curve
 * @param  stack Temporary memory stack to use
 * @return True if P is neutral, False otherwise
 */
bool 
weierstrass_point_is_neutral (ec_point_srcptr P, ec_curve_srcptr E, uint8_t stack);

/**
 * @fn     bool weierstrass_point_are_equal (ec_point_srcptr P1, ec_point_srcptr P2, ec_curve_srcptr E, uint8_t stack);
 * @brief  Test if P1 and P2 are equal on E (BUT do not test if they belong to the curve)
 * @param  P1    Point
 * @param  P2    Point
 * @param  E     Elliptic curve
 * @param  stack Temporary memory stack to use
 * @return True if P1 == P2, False otherwise
 */
bool
weierstrass_point_are_equal (ec_point_srcptr P1, ec_point_srcptr P2, 
        ec_curve_srcptr E, uint8_t stack);


/*******************************OPERATIONS************************************/

/**
 * @fn    void weierstrass_point_neg (ec_point_ptr P3, ec_point_srcptr P1, ec_curve_srcptr E);
 * @brief Set P3 to -P1
 * @param P3 Destination point
 * @param P1 Point
 * @param E  Elliptic curve
 */
void 
weierstrass_point_neg (ec_point_ptr P3, ec_point_srcptr P1, ec_curve_srcptr E);

/* UNIFIED => Resistant to SPA */

/**
 * @fn    void weierstrass_point_add_ZADDU (ec_point_ptr P3, ec_point_ptr P1, ec_point_srcptr P2, ec_curve_srcptr E, uint8_t stack);
 * @brief Co-Z point addition with update: Set P3 = (X3,Y3,Z3) to P1 + P2 using Co-Z addition and update P1 such that Z1 = Z3
 * @param P3    Destination point
 * @param P1    Point (X1, Y1, Z) 
 * @param P2    Point (X2, Y2, Z) 
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void 
weierstrass_point_add_ZADDU (ec_point_ptr P3, ec_point_ptr P1, ec_point_srcptr P2,
        ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void weierstrass_point_add_ZADDC (ec_point_ptr P3, ec_point_ptr P4, ec_point_srcptr P1, ec_point_srcptr P2, ec_curve_srcptr E, uint8_t stack);
 * @brief Conjugate Co-Z point addition: Set P3 to P1 + P2 and P4 to P1 - P2 such that Z3=Z4
 * @param P3    Destination point
 * @param P4    Destination point
 * @param P1    Point (X1, Y1, Z)
 * @param P2    Point (X2, Y2, Z)
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void 
weierstrass_point_add_ZADDC (ec_point_ptr P3, ec_point_ptr P4, ec_point_srcptr P1, 
        ec_point_srcptr P2, ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void weierstrass_point_add_ZDAU (ec_point_ptr P3, ec_point_srcptr P1, ec_point_ptr P2, ec_curve_srcptr E, uint8_t stack);
 * @brief Co-Z point doubling-addition with update: Set P3 to 2*P1 + P2 and update P2 such that Z2 = Z3
 * @param P3    Destination point
 * @param P1    Point (X1, Y1, Z2)
 * @param P2    Point (X2, Y2, Z2)
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void 
weierstrass_point_add_ZDAU (ec_point_ptr P3, ec_point_srcptr P1, ec_point_ptr P2,
        ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void weierstrass_point_DBLU (ec_point_ptr P3, ec_point_ptr P4, ec_point_srcptr P1, ec_curve_srcptr E, uint8_t stack);
 * @brief Co-Z point doubling with update: Set P3 to 2*P1 and P4 such that P1 = P4 and Z4 = Z3 and assume that P1->z==1
 * @param P3    Destination point
 * @param P4    Destination point
 * @param P1    Point with P1->z==1
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void 
weierstrass_point_DBLU (ec_point_ptr P3, ec_point_ptr P4, ec_point_srcptr P1, 
        ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void weierstrass_point_TPLU (ec_point_ptr P3, ec_point_ptr P4, ec_point_srcptr P1, ec_curve_srcptr E, uint8_t stack);
 * @brief Co-Z point tripling with update: Set P3 to 3*P1 and P4 such that P4 = P1 and Z4 = Z3
 * @param P3    Destination point
 * @param P4    Destination point
 * @param P1    Point 
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void 
weierstrass_point_TPLU (ec_point_ptr P3, ec_point_ptr P4, ec_point_srcptr P1, 
        ec_curve_srcptr E, uint8_t stack);

/**
 * @fn void weierstrass_Zmontgomery_Kim(ec_point_ptr P3, number_srcptr n, ec_point_srcptr P1,
ec_curve_srcptr E, uint8_t stack);
 * @brief       Perform the Montgomery ladder, taken from Kim et al's 2017 work, compute P3=[n]P1
 * it has the drawback to perform an inversion at the end of the computation to have affine coordinates
 * and its formula is not complete and does not work for n = -1 neither 0 see .c for details
 * @param P3    Destination point
 * @param n     Number element for the scalar multiplication (different from -1 mod #E)
 * @param P1    Input point
 * @param E     Elliptic curve on which the points belong
 * @param stack Temporary memory stack to use
 */
void
weierstrass_Zmontgomery_Kim(ec_point_ptr P3, number_srcptr n, ec_point_srcptr P1, 
        ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void weierstrass_Zmontgomery_ladder(ec_point_ptr P3, number_srcptr n, ec_point_srcptr P1, ec_curve_srcptr E, uint8_t stack);
 * @brief Set P3 to n * P1 using Montgomery ladder with Co-Z addition formula
 * @param P3    Destination point
 * @param P1    Point
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void
weierstrass_Zmontgomery_ladder(ec_point_ptr P3, number_srcptr n, ec_point_srcptr P1, 
        ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void weierstrass_Zjoye_mul(ec_point_ptr P3, number_srcptr n, ec_point_srcptr P1, ec_curve_srcptr E, uint8_t stack);
 * @brief Set P3 to n * P1 using Joye’s double-add algorithm with Co-Z addition
 *        formula.The formulae are not complete and does not work for n = -1 
 *        see .c for details.      
 * @param P3    Destination point
 * @param P1    Point
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void
weierstrass_Zjoye_mul(ec_point_ptr P3, number_srcptr n, ec_point_srcptr P1, 
        ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void weierstrass_point_add_unified (ec_point_ptr P3, ec_point_srcptr P1, ec_point_srcptr P2, ec_curve_srcptr E, uint8_t stack);
 * @brief Set P3 to P1 + P2, using unified formulae (protection against SPA)
 * @param P3    Destination point
 * @param P1    Point
 * @param P2    Point
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void 
weierstrass_point_add_unified (ec_point_ptr P3, ec_point_srcptr P1, ec_point_srcptr P2,
        ec_curve_srcptr E, uint8_t stack);

/* DEDICATED => Not resistant to SPA */

/**
 * @fn    void weierstrass_point_add_dedicated (ec_point_ptr P3, ec_point_srcptr P1, ec_point_srcptr P2, ec_curve_srcptr E, uint8_t stack);
 * @brief Set P3 to P1 + P2, using dedicated formulae (not protected against SPA, but faster)
 * @param P3    Destination point
 * @param P1    Point
 * @param P2    Point
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void 
weierstrass_point_add_dedicated (ec_point_ptr P3, ec_point_srcptr P1, ec_point_srcptr P2,
        ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void weierstrass_point_dbl_dedicated (ec_point_ptr P3, ec_point_srcptr P1, ec_curve_srcptr E, uint8_t stack);
 * @brief Set P3 to 2 * P1, using dedicated formulae (not protected against SPA, but faster)
 * @param P3    Destination point
 * @param P1    Point
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void 
weierstrass_point_dbl_dedicated (ec_point_ptr P3, ec_point_srcptr P1, ec_curve_srcptr E, uint8_t stack);


/* Miscellaneaous */

/**
 * @fn    void weierstrass_point_of_order_2 (ec_point_ptr dst, ec_curve_srcptr E, uint8_t stack);
 * @brief Set dst to one of the point of order 2 of the curve E assuming that at least one exist
 * @param dst   Destination point
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void
weierstrass_point_of_order_2 (ec_point_ptr dst, ec_curve_srcptr E, uint8_t stack);

/**
 * @fn    void weierstrass_points_of_order_2 (ec_point_ptr dst1, ec_point_ptr dst2, ec_point_ptr dst3, ec_curve_srcptr E, uint8_t stack);
 * @brief Set dst1, dst2, dst3 to the points of order 2 of the curve E assuming they are different otherwise they are all equal to the unique point of order 2
 * @param dst1  Destination point
 * @param dst2  Destination point
 * @param dst3  Destination point
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void 
weierstrass_points_of_order_2 (ec_point_ptr dst1, ec_point_ptr dst2, ec_point_ptr dst3, ec_curve_srcptr E, uint8_t stack);

#endif


