/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France) 

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell-ctr_drbg.c
  * \brief Definition of Counter mode Deterministic Random Bit Generator functions. Details can be found in "NIST Special Publication 800-90A, section 10.2.1
  */


#include "mphell-drbg_internal.h"

static uint8_t CONSTANT_DF[32] = 
{
    0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B,
    0x0C, 0x0D, 0x0E, 0x0F, 0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
    0x18, 0x19, 0x1A, 0x1B, 0x1C, 0x1D, 0x1E, 0x1F
};

/**
 * @fn    void bcc(uint8_t* res, const uint8_t* key, const uint8_t* data, const uint16_t len_data, const uint8_t outlen, const drbg_type type)
 * @brief Compress data using AES in counter mode
 * @param res      Destination, of size outlen bytes
 * @param key      AES key
 * @param data     Source data to compress
 * @param len_data Length of data, must be a multiple of outlen
 * @param outlen   Size (in bytes) of the output res
 * @param type     DRBG algorithm to use (can be DRBG_AES128, DRBG_AES192, DRBG_AES256)
 */
void bcc(uint8_t* res, const uint8_t* key, const uint8_t* data,
        const uint16_t len_data, const uint8_t outlen, const drbg_type type)
{
    uint8_t chaining_value[outlen];
    uint8_t input_block[outlen];
    uint8_t blocki[outlen];
    uint16_t i, n = 0; 
    memset(chaining_value, 0, outlen);
    memset(input_block, 0, outlen);
    memset(blocki, 0, outlen);

    /* chaining_value = 0^outlen */ 
    memset(chaining_value, 0, outlen);

    /* n = len(data) / outlen (len(data) is a multiple of outlen) */
    n = len_data / outlen;

    for(i = 1; i <= n; i++)
    {
        /* blocki = ith first block of outlen bytes of data */
        memcpy(blocki, data + (i - 1)*outlen, outlen);

        /* input_block = chaining_value ^ blocki */
        drbg_xor_data(input_block, chaining_value, blocki, outlen);

        /* chaining_value = Block_Encrypt(Key, input_block) */
        if(type == DRBG_AES128)
        {
            aes128_cipher(chaining_value, key, input_block);
        }
        else if(type == DRBG_AES192)
        {
            aes192_cipher(chaining_value, key, input_block);
        }
        else if(type == DRBG_AES256)
        {
            aes256_cipher(chaining_value, key, input_block);
        }
    }
    /* output_block = chaining_value */
    memcpy(res, chaining_value, outlen);
}

/**
 * @fn    void block_cipher_df(mphell_status ret, uint8_t* requested_bytes,const drbg_type type, const uint8_t keylen, const uint8_t outlen,const uint8_t* input_string, const uint32_t input_string_len, uint32_t no_of_bytes_to_return)
 * @brief Derivation function used to compress entropy into an entropy of size seedlen 
 * @param ret                   Set to MPHELL_ERROR_RANDOM if a problem occured, MPHELL_SUCCESS if everything worked !
 * @param requested_bytes       Destination array of bytes
 * @param type                  DRBG algorithm to use (can be DRBG_AES128, DRBG_AES192, DRBG_AES256)
 * @param keylen                Length of the AES key (in bytes)
 * @param outlen                Length of the AES cipher output
 * @param input_string          Entropy byte string to compress
 * @param input_string_len      Length of input_string (in bytes)
 * @param no_of_bytes_to_return Size (in bytes) of requested_bytes, must be <= 64
 */
void block_cipher_df(mphell_status ret, uint8_t* requested_bytes,const drbg_type type,
        const uint8_t keylen, const uint8_t outlen,const uint8_t* input_string,
        const uint32_t input_string_len, uint32_t no_of_bytes_to_return)
{
    /* If(no_of_bytes_to_return > max_number_of_bytes) then return an ERROR FLAG. 
     * For currently approved block cipher algorithms, it's 64 (512 bits). */
    if(no_of_bytes_to_return > 64)
    {
        ret->flag = MPHELL_ERROR_RANDOM;
        char *text = "block_cipher_df : no_of_bytes_to_return > 64";
        ret->info = malloc(strlen(text));
        memcpy(ret->info, text, strlen(text));
        return;
    }

    /* L = len(input_string) (as we work with bytes it's not necessary to 
     * divide by 8 */
    uint32_t L = input_string_len;

    /* N = number_of_bytes_to_return */ 
    uint32_t N = no_of_bytes_to_return;

    /* S = L || N || input_string || 0x80 
     * while(len(S) mod outlen != 0, S = S || 0x00) */
    uint16_t lenS = 8 + input_string_len + 1; 

    uint16_t diff = 0;
    while(lenS % outlen != 0)
    {
        lenS += 1;
        diff += 1;
    } 
    uint8_t S[lenS];

    S[0] = (uint8_t)(L >> 24);
    S[1] = (uint8_t)(L >> 16);
    S[2] = (uint8_t)(L >> 8);
    S[3] = (uint8_t) L;
    S[4] = (uint8_t)(N >> 24);
    S[5] = (uint8_t)(N >> 16);
    S[6] = (uint8_t)(N >> 8);
    S[7] = (uint8_t) N;
    memcpy(S + 8, input_string, input_string_len);
    S[8 + input_string_len] = 0x80;
    memset(S + 8 + input_string_len + 1, 0, diff);

    uint8_t tmp[keylen + outlen];
    uint32_t i = 0;

    /* K = leftmost keylen bytes of 0x00010203...1D1E1F */
    uint8_t K[keylen];
    memcpy(K, CONSTANT_DF, keylen); 

    uint16_t lentmp = 0;
    uint8_t IV[outlen];
    uint8_t data[outlen + lenS];

    memcpy(data + outlen, S, lenS);
    while(lentmp < keylen + outlen)
    {
        /* IV = i || 0^(outlen - len(i)) */
        IV[0] = (uint8_t)(i >> 24);
        IV[1] = (uint8_t)(i >> 16);
        IV[2] = (uint8_t)(i >> 8);
        IV[3] = (uint8_t)i;
        memset(IV + 4, 0, outlen - 4);

        /* temp = temp || BCC(K, (IV || S) */
        memcpy(data, IV, outlen);
        bcc(tmp + lentmp, K, data, outlen + lenS, outlen, type);
        lentmp += outlen;

        /* i = i + 1 */
        i += 1; 
    }

    /* K = leftmost(temp, keylen) */
    memcpy(K, tmp, keylen);

    /* X = select(temp, keylen + 1, keylen + outlen) */
    uint8_t X[outlen];
    memcpy(X, tmp + keylen, outlen);  

    /* temp = the null string */
    lentmp = 0;

    while(lentmp < no_of_bytes_to_return)
    {
        /* X = Block_Encrypt(K, X) */
        if(type == DRBG_AES128)
        {
            aes128_cipher(X, K, X);
        }
        else if(type == DRBG_AES192)
        {
            aes192_cipher(X, K, X);
        }
        else if(type == DRBG_AES256)
        {
            aes256_cipher(X, K, X);
        }
        else
        {
            ret->flag = MPHELL_ERROR_RANDOM;
            char *text = "block_cipher_df : wrong DRBG specified";
            ret->info = malloc(strlen(text));
            memcpy(ret->info, text, strlen(text));
            return;
        }
        
        /* temp = temp || X */
        memcpy(tmp + lentmp, X, outlen);
        lentmp += outlen;
    }

    /* requested_bytes = leftmost number_of_bytes_to_return of temp */
    memcpy(requested_bytes, tmp, no_of_bytes_to_return);
    ret->flag = MPHELL_SUCCESS;
}

/**
 * @fn    void ctr_drbg_update(drbg_internal_state working_state, uint8_t *provided_data, const drbg_type type)
 * @brief Updates the internal state of the DRBG (V and C_Key) using provided_data
 * @param working_state Current state of the DRBG
 * @param provided_data Data used to update the current state of the DRBG, must be exactly seedlen bits
 */
void ctr_drbg_update(drbg_internal_state working_state, uint8_t *provided_data, const drbg_type type)
{
    uint8_t seedlen = working_state->info->seedlen;
    uint8_t outlen = working_state->info->outlen;
    uint8_t keylen = seedlen - outlen;
    uint8_t tmp[seedlen];
    memset(tmp, 0, seedlen);            /* Avoid a lot of valgrind error */
    uint8_t output_block[outlen];
    memset(output_block, 0, outlen);    /* Avoid a lot of valgrind error */
    
    /* temp = Null */
    uint16_t lentmp = 0;

    while(lentmp < seedlen)
    {
        /* V = (V + 1) mod 2^outlen */
        drbg_incr_data(working_state->state->V, outlen);

        /* output_block = Block_Encrypt(Key, V) */
        if(type == DRBG_AES128)
        {
            aes128_cipher(output_block, working_state->state->C_Key,
               working_state->state->V);
        }
        else if(type == DRBG_AES192)
        {
            aes192_cipher(output_block, working_state->state->C_Key,
               working_state->state->V);
        }
        else if(type == DRBG_AES256)
        {
            aes256_cipher(output_block, working_state->state->C_Key,
               working_state->state->V);
        }

        /* tmp = tmp || output_block */
        /* we guarantee that we do not overflow on tmp */
        if(lentmp + outlen > seedlen)
        {
            uint16_t diff = seedlen - lentmp;
            memcpy(tmp + lentmp, output_block , diff);
            lentmp += diff;
        }
        else
        {
            memcpy(tmp + lentmp, output_block, outlen);
            lentmp += outlen;
        }
    }

    /* tmp = tmp ^ provided_data */
    drbg_xor_data(tmp, tmp, provided_data, seedlen);

    /* Key = Leftmost keylen bytes of temp */
    memcpy(working_state->state->C_Key, tmp, keylen);

    /* V = Rightmost bytes of temp */
    memcpy(working_state->state->V, tmp + keylen, outlen);
}

void ctr_drbg_instantiate(mphell_status ret, drbg_internal_state initial_state,
        const uint8_t security_strength, const uint8_t* entropy_input, const uint16_t entropy_length,
        const uint8_t* personalization_string, 
        const uint64_t pers_length, const drbg_type type)
{
    uint8_t seedlen = 0;
    uint8_t outlen = 0;
    uint8_t keylen = 0;

    switch(type)
    {
        case DRBG_AES128 :
            seedlen = 32;
            outlen = 16;
            keylen = 16;
            break;

        case DRBG_AES192 :
            seedlen = 40;
            outlen = 16;
            keylen = 24;
            break;

        case DRBG_AES256 :
            seedlen = 48;
            outlen = 16;
            keylen = 32;
            break;

        default:
            ret->flag = MPHELL_ERROR_RANDOM;
            char *text = "ctr_drbg_instantiate : Invalid DRBG type";
            ret->info = malloc(strlen(text));
            memcpy(ret->info, text, strlen(text));
            return;
            
    }

    uint64_t lenmaterial = entropy_length;

    if((personalization_string != NULL) && (pers_length>0))
    {
        lenmaterial += pers_length;
    }

    /* seed_material = entropy_input || personalization_string */
    uint8_t seed_material[lenmaterial];
    memcpy(seed_material, entropy_input, entropy_length);

    if((personalization_string != NULL) && (pers_length>0))
    {
        memcpy(seed_material + lenmaterial - pers_length,
                personalization_string, pers_length);
    }

    /* seed_material = Block_Cipher_df(seed_material, seedlen) */
    uint8_t seed_material2[seedlen];
    block_cipher_df(ret, seed_material2, type, keylen, outlen,
            seed_material, lenmaterial, seedlen);
    if(ret->flag != MPHELL_SUCCESS)
    {
        return;
    }

    initial_state->state->C_Key = malloc(keylen * sizeof(uint8_t));
    if(initial_state->state->C_Key == NULL)
    {
        ret->flag = MPHELL_ERROR_RANDOM;
        char *text = "ctr_drbg_instantiate : malloc failed";
        ret->info = malloc(strlen(text));
        memcpy(ret->info, text, strlen(text));
        return;
    }
    initial_state->state->V = malloc(outlen * sizeof(uint8_t));
    if(initial_state->state->V == NULL)
    {
        ret->flag = MPHELL_ERROR_RANDOM;
        char *text = "ctr_drbg_instantiate : malloc failed";
        ret->info = malloc(strlen(text));
        memcpy(ret->info, text, strlen(text));
        return;
    }

    /* Key = 0^keylen */
    memset(initial_state->state->C_Key, 0, keylen);

    /* V = 0^outlen */
    memset(initial_state->state->V, 0, outlen);
    
    /* store values */
    initial_state->info->seedlen = seedlen;
    initial_state->info->outlen = outlen;

    /* (Key, V) = CTR_DRBG_Update(seed_material, Key, V) */
    ctr_drbg_update(initial_state, seed_material2, type);
    
    /* reseed_counter = 1 */
    initial_state->state->reseed_counter = 1;

    ret->flag = MPHELL_SUCCESS;
}

void ctr_drbg_reseed(mphell_status ret, drbg_internal_state working_state,
        const uint8_t* entropy_input, const uint16_t entropy_length, const uint8_t* additional_input,
        const uint64_t add_length)
{
    uint64_t lenmaterial = entropy_length;
    
    if((additional_input != NULL) && (add_length > 0))
    {
        lenmaterial += add_length;
    }

    /* seed_material = entropy_input || additional_input */
    uint8_t seed_material[lenmaterial];
    memcpy(seed_material, entropy_input, entropy_length);
    
    if((additional_input != NULL) && (add_length > 0))
    {
        memcpy(seed_material + entropy_length, additional_input, 
                add_length);
    }

    uint8_t seedlen = working_state->info->seedlen;
    uint8_t outlen = working_state->info->outlen;

    /* seed_material = Block_Cipher_df(seed_material, seedlen) */
    uint8_t seed_material2[seedlen];
    block_cipher_df(ret, seed_material2, working_state->type,
               seedlen - outlen, outlen, seed_material, 
            lenmaterial, seedlen);

    if(ret->flag != MPHELL_SUCCESS)
    {
        return;
    }
    
    /* (Key, V) = CTR_DRBG_Update(seed_material, Key, V) */
    ctr_drbg_update(working_state, seed_material2, working_state->type);
    
    /* reseed_counter = 1 */
    working_state->state->reseed_counter = 1;
    
    ret->flag = MPHELL_SUCCESS;
}

void ctr_drbg_generate(mphell_status ret, uint8_t* returned_bytes, 
        drbg_internal_state working_state, const uint32_t requested_no_of_bytes,
        const uint8_t* additional_input, const uint64_t add_length)
{
    /* if reseed_counter > resed_interval the retuen an indication that a 
     * reseed is required */
    if(working_state->state->reseed_counter > working_state->info->reseed_interval)
    {
        memset(returned_bytes, 0, requested_no_of_bytes);
        ret->flag = MPHELL_RESEED_REQUIRED;
        return;
    }
    
    uint8_t seedlen = working_state->info->seedlen;
    uint8_t outlen = working_state->info->outlen;

    uint8_t add_input[seedlen];
    
    if((additional_input != NULL) && (add_length > 0))
    {
        /* additional_input = Block_Cipher_df(additionnal_input, seedlen) */
        block_cipher_df(ret, add_input, working_state->type,
               seedlen - outlen, outlen, additional_input, add_length, 
               seedlen);
        if(ret->flag != MPHELL_SUCCESS)
        {
            memset(returned_bytes, 0, requested_no_of_bytes);
            return ;
        }
        /* (Key, V) = CTR_DRBG_Update(additional_input, Key, V) */
        ctr_drbg_update(working_state, add_input, working_state->type);
    }
    else
    {
        /* additional_input = 0 ^seedlen */
        memset(add_input, 0, seedlen);
    }

    uint8_t tmp[requested_no_of_bytes];

    /* temp = Null */
    uint32_t lentmp = 0;

    drbg_type type = working_state->type;
    uint8_t output_block[outlen];
    while(lentmp < requested_no_of_bytes)
    {
        /* V = (V + 1) mod 2^outlen */
        drbg_incr_data(working_state->state->V, outlen);

        /* output_block = Block_Encrypt(Key, V) */

        if(type == DRBG_AES128)
        {
            aes128_cipher(output_block, working_state->state->C_Key,
               working_state->state->V);
        }
        else if(type == DRBG_AES192)
        {
            aes192_cipher(output_block, working_state->state->C_Key,
               working_state->state->V);
        }
        else if(type == DRBG_AES256)
        {
            aes256_cipher(output_block, working_state->state->C_Key,
               working_state->state->V);
        }
        else
        {
            ret->flag = MPHELL_ERROR_RANDOM;
            char *text = "ctr_drbg_generate : wrong DRBG specified";
            ret->info = malloc(strlen(text));
            memcpy(ret->info, text, strlen(text));
            return;
        }

        /* temp = temp || output_block */
        /* we guarantee that we do noyt overflow on temp */
        if(lentmp + outlen > requested_no_of_bytes)
        {
            uint32_t diff = requested_no_of_bytes - lentmp;
            memcpy(tmp + lentmp, output_block, diff);
            lentmp += diff;
        }
        else
        {
            memcpy(tmp + lentmp, output_block, outlen);
            lentmp += outlen;
        }
    }
    
    /* returned_bytes = lestmost requested_no_of_bytes of temp */ 
    memcpy(returned_bytes, tmp, requested_no_of_bytes);

    /* (Key, V) = CTR_DRBG_Update(additional_input, Key, V) */
    ctr_drbg_update(working_state, add_input, working_state->type);
    
    /* reseed_coutner = reseed_counter + 1 */
    working_state->state->reseed_counter += 1;

    ret->flag = MPHELL_SUCCESS;
}
