/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France) 

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell-drbg_common.c
  * \brief Definition of the Deterministic Random Bit Generator internal functions.
  */

#include "mphell-drbg_internal.h"

/* Utility functions */

uint32_t drbg_ceil(uint32_t n1, uint32_t n2)
{
    return ceil((float)n1 / n2);
}

/**
 * @fn     uint8_t drbg_add_byte_with_carry(uint8_t *b1, const uint8_t b2, uint8_t carry)
 * @brief  Set b1 to the lowest byte of b1 + b2 + carry
 * @param  b1    Destination (and source) byte
 * @param  b2    Source byte
 * @param  carry Can be 0 or 1
 * @return 1 if there is a carry (b1 + b2 + carry didn't fit in one byte)
 * @return 0 otherwise
 */
uint8_t drbg_add_byte_with_carry(uint8_t *b1, const uint8_t b2, uint8_t carry)
{
    uint16_t res = *b1 + b2 + carry;
    *b1 = (uint8_t) res;
    return (((uint8_t)(res >> 8))!=0);
}

/**
 * @fn     uint8_t drbg_add_byte(uint8_t *b1, const uint8_t b2)
 * @brief  Set b1 to the lowest byte of b1 + b2
 * @param  b1    Destination (and source) byte
 * @param  b2    Source byte
 * @return 1 if there is a carry (b1 + b2 didn't fit in one byte)
 * @return 0 otherwise
 */
uint8_t drbg_add_byte(uint8_t *b1, const uint8_t b2)
{
    uint16_t res = *b1 + b2;
    *b1 = (uint8_t) res;
    return ((uint8_t)(res >> 8));
}

void drbg_incr_data(uint8_t *data, const uint16_t data_len)
{
    MPHELL_ASSERT(data_len > 0, "data_len <= 0");
    int16_t i;
    uint8_t carry;
    carry = drbg_add_byte(data + data_len - 1, (uint8_t)1);
    for(i = data_len - 2; i >= 0; i--)
    {
        if(carry == 0)
        {
            break;
        }
        else
        {
            carry = drbg_add_byte(data + i, carry);
        }
    }
}

void drbg_add_data(uint8_t *data, const uint16_t data_len, const uint8_t *data1, 
        const uint16_t data1_len)
{
    MPHELL_ASSERT(data_len >= data1_len, "data_len < data1_len, consider to reverse data and data1");
    int16_t i;
    uint8_t carry;
    uint16_t diff = data_len - data1_len;

    carry = drbg_add_byte(data + data_len - 1, data1[data1_len - 1]);

    for(i = data1_len - 2; i >= 0; i--)
    {
         carry = drbg_add_byte_with_carry(data + i + diff, data1[i], carry);
    }
    if((carry == 1) && (diff > 0))
    {
        drbg_incr_data(data, diff);
    }
}

void drbg_add_int(uint8_t* data, const uint16_t data_len, const uint64_t n)
{
    MPHELL_ASSERT(data_len >= 8, "data_len < 8, cannot add integer of size 8 bytes");
    int16_t i;
    uint8_t carry;
    uint16_t diff = data_len - 8;

    carry = drbg_add_byte(data+ data_len - 1, (uint8_t)n);

    for(i = 6; i >= 0; i--)
    {
         carry = drbg_add_byte_with_carry(data + i + diff, (n >> (8 * (7-i))), carry);
    }
    if((carry == 1) && (diff > 0))
    {
        drbg_incr_data(data + diff -1, diff); 
    }
}

void drbg_xor_data(uint8_t* res, const uint8_t* data1, const uint8_t* data2,
        const uint16_t data_length)
{
    uint16_t i;
    for(i = 0; i < data_length; i++)
    {
        res[i] = data1[i] ^ data2[i];
    }
}


