/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France) 


 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell-edwards.c
  * \brief Definition of Edward ECC functions.
  *
  * Reference for the formulae: 
  * [HWCD08] Twisted Edwards Curves Revisited by    
  * Hisil, Wong, Carter and Dawson published in Asiacrypt volume 5350, pages  
  * 326-343.
  *
  * Reference for the conversion between weierstrass curve and Edwards Curve
  * [BL07] Faster addition and doubling on elliptic curves by Daniel J. 
  * Bernstein and Tanja Lange published in Asiacrypt volume 4833, pages 29-50
  * [BBJ+08]Twisted edwards curves by Bernstein, Birkner, Joye, Lange, Peters
  * published in International Conference on Cryptology in Africa, 2008 pages
  * 389-405  
  * [CS17] Montgomery curves and their arithmetic by Craig Costello and 
  * Benjamin Smith published in Journal of Cryptographic Engineering 2017
  * pages 1-14.
*/

#include "mphell-curve.h"

/************************************SETTERS**********************************/

void
edwards_compute_disc(ec_curve E, uint8_t stack)
{
    field_t *k = E->k;
    field_elt_sub(E->disc, E->a, E->b, k);                              /* disc = (a - d)                                   */
    field_elt_sqr(E->disc, E->disc, k, stack);
    field_elt_sqr(E->disc, E->disc, k, stack);                          /* disc = (a - d)^4                                 */
    field_elt_mul(E->disc, E->disc, E->a, k, stack);
    field_elt_mul(E->disc, E->disc, E->b, k, stack);                    /* disc = a.d.(a-d)^4                               */
    /* Since the j-invariant is equals to 16(a^2+14ad+d^2)^3/(ad(a-d)^4) [BBJ+07, §2]                                       */
    field_elt_set_ui(E->D, 0, true, k, stack);                          /* D has no use up to now and no definition known   */
}

#include "mphell-drbg_internal.h"
#include "mphell-sha1.h"

bool edwards_verify_random_generation(ec_curve E, const char * seed, uint8_t stack)
{
    /* Source : Working Draft AMERICAN NATIONAL STANDARD X9.62-1998, section A.3.4.2 */
    field_t *k = E->k;
    
    field_elt r1;
    field_elt_get_pool_elt(&r1, k, stack);
    
    number tmp;
    number_tmp_alloc(&tmp, k->size, stack);
    
    /* t = log2(p) */
    uint16_t t;
    switch(k->type)
    {
        case FP:
             t = number_log2(((fp_param)k->param)->p);
            break;
        case FP2:
        case FP3:
            t = number_log2(((fp_param)((fp2_param)k->param)->base)->p);
            break;
        default: 
            return false;
    }
    /* s = (t-1) / 160 */
    uint16_t s = (t-1)/160;
    /* h = t - 160*s */
    uint16_t h = t - (160*s);
    h --; /* set the leftmost bit of c0 to 0 */
    uint16_t hbytes = (h / 8) + (h%8 != 0);
    
    /* Alloc W' */
    
    uint8_t * Wprime = calloc(s*20+hbytes, sizeof(uint8_t));
    
    /* Step 1 and 2, compute W0: */
    
    uint8_t H[20];
    uint8_t seed_bin[20];
    /* Convert hexadecimal seed into binary seed */
    hex2bin(seed, seed_bin);
    sha1(H, seed_bin, 160);
    /* c0 = h rightmost bits of H */
    memcpy(Wprime, H+(20-hbytes), hbytes);
    uint8_t mask;
    switch (h%8)
    {
        case 0:
            mask = 0xff;
            break;
        case 1:
            mask = 0x1;
            break;
        case 2:
            mask = 0x3;
            break;
        case 3:
            mask = 0x7;
            break;
        case 4:
            mask = 0xf;
            break;
        case 5:
            mask = 0x1f;
            break;
        case 6:
            mask = 0x3f;
            break;
        case 7:
            mask = 0x7f;
            break;
    }
    
    Wprime[0]&=mask;
    
    /* Step 2 : done with h--*/
    
    /* Step 3 and 4 : */
    
    uint16_t i;
    for (i=0; i<s; i++)
    {
        drbg_incr_data(seed_bin, 20);
        sha1(H, seed_bin, 160);
        memcpy(Wprime+i*20+hbytes, H, 20);
    }
    
    /* Step 5 : Transform Wprime into a field element */
    
    /* Convert binary hash into hexadecimal string */
    unsigned char Wprime_hex[2*(s*20+hbytes)+1];
    bin2hex(Wprime, s*20+hbytes, Wprime_hex);
    switch(k->type)
    {
        case FP:
            field_elt_set_str(r1, (const char *)Wprime_hex, 16, false, k, stack);
            break;
        case FP2:
        case FP3:
            number_set_str(tmp, (const char *)Wprime_hex, 16);
            field_elt_set_number(r1, false, k, stack, 1, tmp);
            break;
        default:
            field_elt_set_one(r1, k);
    }
    
    /* Step 6 : Accept or reject */
    
    int8_t res = field_elt_cmp(r1, E->b, k);
    
    number_tmp_free(&tmp, k->size, stack);
    field_elt_relax_pool_elt(&r1, k, stack);
    free(Wprime);
    
    return (res == 0);
}

void edwards_curve_random_generation(fe_ptr d, char * seed_res, field_srcptr k, uint8_t stack)
{
    /* Source : Working Draft AMERICAN NATIONAL STANDARD X9.62-1998, section A.3.3.2 */
    bool test;
    number seed;
    number bound_h;
    number bound_l;
    number_tmp_alloc(&seed, bits_to_nblock(160), stack);
    number_tmp_alloc(&bound_h, bits_to_nblock(160), stack);
    number_tmp_alloc(&bound_l, bits_to_nblock(160), stack);
    number_set_str(bound_h, "ffffffffffffffffffffffffffffffffffffffff", 16); /* 2^161 - 1 */
    number_set_str(bound_l, "1000000000000000000000000000000000000000", 16); /* 2^160     */
    
    field_elt r1, tmp1;
    field_elt_get_pool_elt(&r1, (field_ptr)k, stack);
    field_elt_get_pool_elt(&tmp1, (field_ptr)k, stack);
    
    number tmp;
    number_tmp_alloc(&tmp, k->size, stack);
    
    char * seed_hex;
    
    /* t = log2(p) */
    uint16_t t;
    switch(k->type)
    {
        case FP:
             t = number_log2(((fp_param)k->param)->p);
            break;
        case FP2:
        case FP3:
            t = number_log2(((fp_param)((fp2_param)k->param)->base)->p);
            break;
        default: 
            t = 0;
    }
    /* s = (t-1) / 160 */
    uint16_t s = (t-1)/160;
    /* h = t - 160*s */
    uint16_t h = t - (160*s);
    h --; /* set the leftmost bit of c0 to 0 */
    uint16_t hbytes = (h / 8) + (h%8 != 0);
    
    /* Alloc W' */
    uint8_t * Wprime = calloc(s*20+hbytes, sizeof(uint8_t));

    do
    {
        /* Step 1 : Choose an arbitrary bit string SEED of bit length at least 160 bits */
        do
        {
            number_random1(seed, bound_h, stack);
        }
        while(number_cmp(seed, bound_l)<0);
        
        number_str(&seed_hex, seed, 16);
        
        /* Step 2 and 3, compute W0 :  */
        
        uint8_t H[20];
        uint8_t seed_bin[20];
        /* Convert hexadecimal seed into binary seed */
        hex2bin(seed_hex, seed_bin);
        sha1(H, seed_bin, 160);
        /* c0 = h rightmost bits of H */
        memcpy(Wprime, H+(20-hbytes), hbytes);
        uint8_t mask;
        switch (h%8)
        {
            case 0:
                mask = 0xff;
                break;
            case 1:
                mask = 0x1;
                break;
            case 2:
                mask = 0x3;
                break;
            case 3:
                mask = 0x7;
                break;
            case 4:
                mask = 0xf;
                break;
            case 5:
                mask = 0x1f;
                break;
            case 6:
                mask = 0x3f;
                break;
            case 7:
                mask = 0x7f;
                break;
        }
        
        Wprime[0]&=mask;
        
        /* Step 4 and 5 : */
        
        uint16_t i;
        for (i=0; i<s; i++)
        {
            drbg_incr_data(seed_bin, 20);
            sha1(H, seed_bin, 160);
            memcpy(Wprime+i*20+hbytes, H, 20);
        }
        
        /* Convert binary hash into hexadecimal string */
        unsigned char Wprime_hex[2*(s*20+hbytes)+1];
        bin2hex(Wprime, s*20+hbytes, Wprime_hex);
        
        /* Step 6 : Transform Wprime into a field element */
        
        switch(k->type)
        {
            case FP:
                field_elt_set_str(r1, (const char *)Wprime_hex, 16, false, k, stack);
                break;
            case FP2:
            case FP3:
                number_set_str(tmp, (const char *)Wprime_hex, 16);
                field_elt_set_number(r1, false, k, stack, 1, tmp);
                break;
            default:
                field_elt_set_one(r1, k);
        }
        
        /* Step 7: Choose a = r1 */
        
        field_elt_copy(d, r1, k);
        field_elt_set_one(tmp1, k);
        field_elt_sub(tmp1, tmp1, d, k); /* tmp1 = 1-d */
        field_elt_mul(tmp1, tmp1, d, k, stack); /* tmp1 = d(1-d) */
        
        test = field_elt_iszero(tmp1, k);
        if (test)
        {
            free(seed_hex);
        }
    }
    /* Step 8: verifie that the curve is elliptic */
    while(test);
    
    strcpy(seed_res, seed_hex);
    
    number_tmp_free(&tmp, k->size, stack);
    field_elt_relax_pool_elt(&r1, (field_ptr)k, stack);
    field_elt_relax_pool_elt(&tmp1, (field_ptr)k, stack);
    number_tmp_free(&seed, bits_to_nblock(160), stack);
    number_tmp_free(&bound_h, bits_to_nblock(160), stack);
    number_tmp_free(&bound_l, bits_to_nblock(160), stack);
    free(seed_hex);
    free(Wprime);
}

void 
edwards_point_set_neutral (ec_point_ptr dst, ec_curve_srcptr E, uint8_t stack)
{
    field_t *k = E->k;
    /* Naturally set in extended_edwards coordinate, t (and z) could be forget later */
    field_elt_set_ui(dst->x, 0, true, k, stack);
    field_elt_set_one(dst->y, k);
    field_elt_set_one(dst->z, k);
    field_elt_set_ui(dst->t, 0, true, k, stack);
}

void 
edwards_point_set_aff (ec_point_ptr P, fe_srcptr x, fe_srcptr y, field_srcptr k, uint8_t stack)
{
    /* Naturally set in extended_edwards coordinate, t could be forget later */
    field_elt_copy(P->x, x, k);
    field_elt_copy(P->y, y, k);
    field_elt_mul(P->t, P->x, P->y, k, stack);
    field_elt_set_one(P->z, k);
}

void 
edwards_point_set_aff_str (ec_point_ptr P, const char *str_x, const char *str_y,
        const bool is_reduced, const uint8_t base, field_srcptr k, uint8_t stack)
{
    /* Naturally set in extended_edwards coordinate, t could be forget later */
    field_elt_set_str(P->x, str_x, base, is_reduced, k, stack);
    field_elt_set_str(P->y, str_y, base, is_reduced, k, stack);
    field_elt_mul(P->t, P->x, P->y, k, stack);
    field_elt_set_one(P->z, k);
}

bool
edwards_belongs (ec_point_srcptr P, ec_curve_srcptr E, uint8_t stack)
{
    field_t *k = E->k;
    bool res;

    field_elt a, b, c, d, e;
    field_elt_get_pool_elt(&a, k, stack);
    field_elt_get_pool_elt(&b, k, stack);
    field_elt_get_pool_elt(&c, k, stack);
    field_elt_get_pool_elt(&d, k, stack);
    field_elt_get_pool_elt(&e, k, stack);    
    switch(E->f)
    {
        case PROJECTIVE :
            field_elt_sqr(a, P->x, k, stack);           /* a = x^2                  */
            field_elt_mul(d, a, E->a, k, stack);        /* d = a x^2                */
            field_elt_sqr(b, P->y, k, stack);           /* b = y^2                  */
            field_elt_add(d, d, b, k);                  /* d = a x^2 + y^2          */
            field_elt_mul(b, b, a, k, stack);           /* b = x^2 y^2              */
            field_elt_mul(b, b, E->b, k, stack);        /* b = d x^2 y^2            */
            field_elt_sqr(c, P->z, k, stack);           /* c = z^2                  */
            field_elt_mul(d, d, c, k, stack);           /* d = (a x^2 + y^2) z^2    */
            field_elt_sqr(c, c, k, stack);              /* c = z^4                  */
            field_elt_add(c, c, b, k);                  /* b = z^4 + d x^2 y^2      */
            res =  (field_elt_cmp(c, d, k) == 0);
            /* It is checked that (a x^2 + y^2) z^2 = z^4 + d x^2 y^2               */
            
            break;

        case EXTENDED_EDWARDS : /* The «natural» case */
            field_elt_sqr(a, P->x, k, stack);           /* a = x^2                  */
            field_elt_sqr(b, P->y, k, stack);           /* b = y^2                  */
            field_elt_add(c, P->x, P->y, k);            /* c = x + y                */
            field_elt_sqr(c, c, k, stack);              /* c = (x + y) ^2           */
            field_elt_sub(c, c, a, k);
            field_elt_sub(c, c, b, k);                  /* c = 2xy                  */
            field_elt_mul(a, a, E->a, k, stack);
            field_elt_add(a, a, b, k);                  /* a = ax^2 + y^2, keep a   */

            field_elt_sqr(b, P->t, k, stack);           /* b = t^2                  */
            field_elt_sqr(d, P->z, k, stack);           /* d = z^2                  */
            field_elt_add(e, P->t, P->z, k);            /* e = t + z                */
            field_elt_sqr(e, e, k, stack);
            field_elt_sub(e, e, b, k);
            field_elt_sub(e, e, d, k);                  /* e = 2tz                  */
            field_elt_mul(b, b, E->b, k, stack);
            field_elt_add(d, d, b, k);                  /* d = dt^2+z^2             */
            res =  (field_elt_cmp(a, d, k) == 0) && 
            (field_elt_cmp(c, e, k) == 0);
            /* It is checked that: a x^2 + y^2 = d t^2 + z^2 and that 2tz =2xy      */
                        
            break;
        default :
            MPHELL_ASSERT( (E->f == PROJECTIVE) || (E->f==EXTENDED_EDWARDS), "\nOnly Projective coordinates and Extended Edwards coordinates are supported for Twisted Edwards elliptic curve and thus for testing if a point belongs to it.\n");
            res = false;
            break;
    }
    
    field_elt_relax_pool_elt(&a, k, stack);
    field_elt_relax_pool_elt(&b, k, stack);
    field_elt_relax_pool_elt(&c, k, stack);
    field_elt_relax_pool_elt(&d, k, stack);
    field_elt_relax_pool_elt(&e, k, stack);

    return res;
}

void 
edwards_point_random (ec_point_ptr P, ec_curve_srcptr E, uint8_t stack)
{
    field_t *k = E->k;
    field_elt a, b, c, d;
    ec_point T;    
    field_elt_get_pool_elt(&a, k, stack);
    field_elt_get_pool_elt(&b, k, stack);
    field_elt_get_pool_elt(&c, k, stack);
    field_elt_get_pool_elt(&d, k, stack);
    ec_point_get_pool_elt(T, k, stack);
    do 
    {
      field_elt_random(a, k, stack);
      field_elt_sqr(b, a, k, stack);
      field_elt_dec(c, b, k);                  /* u^2 - 1      */
      field_elt_mul(b, b, E->b, k, stack);
      field_elt_sub(b, b, E->a, k);            /* du^2 - a     */
    } while ((field_elt_issquare(b, k, stack)==false) || (field_elt_issquare(c, k, stack)==false));
    
    MPHELL_ASSERT( field_elt_issquare(b, k, stack) && field_elt_issquare(c, k, stack), "The random elements choosen don't permit us to have squares \n");
    field_elt_sqrt(T->x, c, k, stack);
    field_elt_sqrt(T->z, b, k, stack);
    field_elt_mul(T->t, a, T->x, k, stack);
    field_elt_mul(T->y, a, T->z, k, stack);
    /* In the end one can check that a x^2 + y^2 = d t^2 + z^2 */
    switch(E->f)
    {
        case PROJECTIVE :
            field_elt_set_one(T->t,k);
            break;
        case EXTENDED_EDWARDS :
            break;     
        default :
            MPHELL_ASSERT( E->f== PROJECTIVE && E->f==EXTENDED_EDWARDS, "\nOnly Projective coordinates and Extended Edwards coordinates are supported for Twisted Edwards elliptic curve and thus for creating a random point.\n");
            break;
    }
    ec_point_copy(P,T, k);
    field_elt_relax_pool_elt(&a, k, stack);
    field_elt_relax_pool_elt(&b, k, stack);
    field_elt_relax_pool_elt(&c, k, stack);
    field_elt_relax_pool_elt(&d, k, stack);
    ec_point_relax_pool_elt(T, k, stack);
}

void
edwards_point_norm (ec_point_ptr P, ec_curve_srcptr E, uint8_t stack)
{
    field_t *k = E->k;
    field_elt u;
    field_elt_get_pool_elt(&u, k, stack);
    
    if(!edwards_point_is_neutral(P,E))
    {
        switch(E->f)
        {
            case PROJECTIVE :
                field_elt_inv(u, P->z, k, stack);               /* u = 1/Z      */
                field_elt_mul(P->x, P->x, u, k, stack);         /* X = X / Z    */
                field_elt_mul(P->y, P->y, u, k, stack);         /* Y = Y / Z    */
                field_elt_set_one(P->z, k);                     /* Z = 1        */
                field_elt_set_one(P->t, k);                     /* T = 1        */
                break;

             case EXTENDED_EDWARDS :
                field_elt_inv(u, P->z, k, stack);               /* u = 1/Z      */
                field_elt_mul(P->x, P->x, u, k, stack);         /* X = X / Z    */
                field_elt_mul(P->y, P->y, u, k, stack);         /* Y = Y / Z    */
                field_elt_set_one(P->z, k);                     /* Z = 1        */
                field_elt_mul(P->t, P->x, P->y, k, stack);      /* T = X Y / Z  */
                break;

            default :
                MPHELL_ASSERT( ((E->f == PROJECTIVE) && (E->f == EXTENDED_EDWARDS)), "\nOnly Projective coordinates and Extended Edwards coordinates are supported for Twisted Edwards elliptic curve and thus for normalizing a point.\n");
                break;
        }
    }
    field_elt_relax_pool_elt(&u, k, stack);
}

void
edwards_point_get_x_affine(field_elt x, ec_point_ptr P, ec_curve_srcptr E, uint8_t stack)
{
    field_t *k = E->k;
    field_elt u;
    field_elt_get_pool_elt(&u, k, stack);
    
    field_elt_inv(u, P->z, k, stack);               /* u = 1/Z      */
    field_elt_mul(x, P->x, u, k, stack);            /* x = X / Z    */
    
    field_elt_relax_pool_elt(&u, k, stack);
}

void
edwards_point_get_y_affine(field_elt y, ec_point_ptr P, ec_curve_srcptr E, uint8_t stack)
{
    field_t *k = E->k;
    field_elt u;
    field_elt_get_pool_elt(&u, k, stack);
    
    field_elt_inv(u, P->z, k, stack);               /* u = 1/Z      */
    field_elt_mul(y, P->y, u, k, stack);            /* y = Y / Z    */
    
    field_elt_relax_pool_elt(&u, k, stack);
}

/*******************************COMPARISON************************************/
bool 
edwards_point_is_neutral (ec_point_srcptr P, ec_curve_srcptr E)
{
    field_t *k = E->k;
    switch(E->f)
    {
        case PROJECTIVE :
            return ((field_elt_cmp(P->z, P->y, k)==0) && field_elt_iszero(P->x, k));
            break;
        case EXTENDED_EDWARDS :
            return (field_elt_iszero(P->x, k) && field_elt_iszero(P->t, k) && 
                   (field_elt_cmp(P->y, P->z, k) == 0));
            break;
        default :
            MPHELL_ASSERT( E->f== PROJECTIVE && E->f==EXTENDED_EDWARDS, "\nOnly Projective coordinates and Extended Edwards coordinates are supported for Twisted Edwards elliptic curve and thus for testing neutrality of a point.\n");
            return false;
            break;
    }
}

bool
edwards_point_are_equal (ec_point_srcptr P1, ec_point_srcptr P2, 
        ec_curve_srcptr E, uint8_t stack)
{
    field_t *k = E->k;

    field_elt a, b;
    field_elt_get_pool_elt(&a, k, stack);
    field_elt_get_pool_elt(&b, k, stack);
    bool res=true;
    switch(E->f)
    {
        case PROJECTIVE :
            if (field_elt_iszero(P1->y, k))
            {
                if(field_elt_iszero(P2->y, k))
                {
                    if(field_elt_cmp(P1->x, P2->x, k)==0)
                    {
                        res = true;
                        break;
                    }
                    else
                    { 
                        res = false;
                        break;
                    }
                }
                else
                {
                    res = false;
                    break;
                }
            }
        
            field_elt_mul(a, P1->x, P2->z, k, stack);           /* a = x1 * z2 */
            field_elt_mul(b, P1->z, P2->x, k, stack);           /* b = z1 * x2 */
            if(field_elt_cmp(a, b, k) != 0)
            {
                res=false;
                break;
            }
            field_elt_mul(a, P1->y, P2->z, k, stack);           /* a = y1 * z2 */
            field_elt_mul(b, P1->z, P2->y, k, stack);           /* b = z1 * y2 */

            res = (field_elt_cmp(b, a, k) == 0);
            break;

        case EXTENDED_EDWARDS :
#if MPHELL_USE_ASSERT == 1
            /* We test if we have points in the Extended Edwards coordinates system */
            field_elt_mul(a, P2->x, P2->y, k, stack);           /* a = x2 * y2 */
            field_elt_mul(b, P2->t, P2->z, k, stack);           /* b = z2 * t2 */
            MPHELL_ASSERT( field_elt_cmp(a, b, k)==0, "The point P2 is not a point in a system of Extended Edwards coordinates");
            field_elt_mul(a, P1->x, P1->y, k, stack);           /* a = x1 * y1 */
            field_elt_mul(b, P1->t, P1->z, k, stack);           /* b = z1 * t1 */
            MPHELL_ASSERT(field_elt_cmp(a, b, k)==0, "The point P1 is not a point in a system of Extended Edwards coordinates");
#endif
            field_elt_mul(a, P1->t, P2->t, k, stack);           /* a = t1 * t2 */
            if ((field_elt_iszero(a, k) && !field_elt_iszero(P1->t, k)) || (field_elt_iszero(a, k) && !field_elt_iszero(P2->t, k)) )
            {
                /* Easy case where one of the t is null whereas the other is not */
                res=false;
                break;
            }
            else
            {
                /* We ignore T and we work with the 3 remainings coordinates */
                field_elt_mul(a, P1->x, P2->z, k, stack);       /* a = x1 * z2 */
                field_elt_mul(b, P2->x, P1->z, k, stack);       /* b = x2 * z1 */
                if(field_elt_cmp(a, b, k)!=0)
                {
                    res=false;
                    break;
                }
                field_elt_mul(a, P1->y, P2->z, k, stack);       /* a = x1 * z2 */
                field_elt_mul(b, P2->y, P1->z, k, stack);       /* b = x2 * z1 */
                res=(field_elt_cmp(a, b, k)==0);
            }                
            break;      
        default :
            MPHELL_ASSERT( E->f== PROJECTIVE && E->f==EXTENDED_EDWARDS, "\nOnly Projective coordinates and Extended Edwards coordinates are supported for Twisted Edwards elliptic curve and thus for point comparison.\n");
            res = false;    
    
    }
    field_elt_relax_pool_elt(&a, k, stack);
    field_elt_relax_pool_elt(&b, k, stack);
    return res;
}

/*******************************OPERATIONS************************************/

void 
edwards_point_neg (ec_point_ptr P3, ec_point_srcptr P1, ec_curve_srcptr E)
{
    field_t *k = E->k;
    /* Using formulas from [HWCD08, 3] */
    field_elt_neg(P3->x, P1->x, k);
    field_elt_copy(P3->y, P1->y, k);
    field_elt_neg(P3->t, P1->t, k);
    field_elt_copy(P3->z, P1->z, k);
}

/* UNIFIED => Resistant to SPA, available only with EXTENDED_EDWARDS coordinates */

void 
edwards_point_add_unified (ec_point_ptr P3, ec_point_srcptr P1, ec_point_srcptr P2,
        ec_curve_srcptr E, uint8_t stack)
{
    /* Using formulas from [HWCD08, 3.1] */
    /* This formula can be viewed on the website http://www.hyperelliptic.org/EFD/g1p/auto-twisted-extended.html#addition-add-2008-hwcd and is called "add-2008-hwcd" */ 
    /* Costs 9 M 1* a 1* d 7 add */
    
    MPHELL_ASSERT(E->f==EXTENDED_EDWARDS, "\nOnly Extended Edwards coordinates are supported for Twisted Edwards elliptic curve with unified coordinates.\n");
    field_elt a, b, c, d, e, f;
    field_t *k = E->k;
  
    if(edwards_point_is_neutral(P1, E))
    {
        ec_point_copy(P3, P2, k);
        return;
    }
    if(edwards_point_is_neutral(P2, E))
    {
        ec_point_copy(P3, P1, k);
        return;
    }
    
    MPHELL_ASSERT(!field_elt_iszero(P1->z, k) && !field_elt_iszero(P2->z, k), "\nOne of the point is not regular it has its z-coordinate equals to 0.\n");
    /* Following formulas don't apply */ 

    field_elt_get_pool_elt(&a, k, stack);
    field_elt_get_pool_elt(&b, k, stack);
    field_elt_get_pool_elt(&c, k, stack);
    field_elt_get_pool_elt(&d, k, stack);
    field_elt_get_pool_elt(&e, k, stack);
    field_elt_get_pool_elt(&f, k, stack);

    field_elt_mul(a, P1->x, P2->x, k, stack);      /* A = x1 x2                */
    field_elt_mul(b, P1->y, P2->y, k, stack);      /* B = y1 y2                */
    field_elt_mul(c, P1->t, P2->t, k, stack); 
    field_elt_mul(c, c, E->b, k, stack);           /* C = d t1 t2              */
    field_elt_mul(d, P1->z, P2->z, k, stack);      /* D = z1 z2                */

    field_elt_add(e, P1->x, P1->y, k);
    field_elt_add(f, P2->x, P2->y, k);

    field_elt_mul(e, e, f, k, stack);
    field_elt_sub(e, e, a, k);
    field_elt_sub(e, e, b, k);                     /* E = (X1+Y1)(X2+Y2)-A-B   */

    field_elt_sub(f, d, c, k);                     /* F = D - C */
    field_elt_add(d, d, c, k);                     /* G = D + C (stored in D)  */

    field_elt_mul(a, a, E->a, k, stack);
    field_elt_sub(b, b, a, k);                     /* H = B - aA (stored in B) */
    
    field_elt_mul(P3->x, e, f, k, stack);          /* X= E * F                 */ 
    field_elt_mul(P3->y, d, b, k, stack);          /* Y= G * H                 */
    field_elt_mul(P3->t, e, b, k, stack);          /* T= E * H                 */
    field_elt_mul(P3->z, d, f, k, stack);          /* Z= G * F                 */

    field_elt_relax_pool_elt(&a, k, stack);
    field_elt_relax_pool_elt(&b, k, stack);
    field_elt_relax_pool_elt(&c, k, stack);
    field_elt_relax_pool_elt(&d, k, stack);
    field_elt_relax_pool_elt(&e, k, stack);
    field_elt_relax_pool_elt(&f, k, stack);
}

/* DEDICATED => Not resistant to SPA, but faster */

void 
edwards_point_add_dedicated (ec_point_ptr P3, ec_point_srcptr P1, ec_point_srcptr P2,
        ec_curve_srcptr E, uint8_t stack)
{
    /* Using formulae from [HWCD08, §3.2] for Extended Edwards coordinates */ 
    /* This formula can be viewed at this address http://www.hyperelliptic.org/EFD/g1p/auto-twisted-extended.html#addition-add-2008-hwcd-2  and is called "add-2008-hwcd-2" */
    /* Using formulae from [BBJ+08, §6] for Projective coordinates */
    /* This formula can be viewed on this website http://www.hyperelliptic.org/EFD/g1p/auto-twisted-projective.html#addition-add-2008-bbjlp where it is called "add-2008-bbjlp". */
    
    field_elt a, b, c, d, e, f;
    field_t *k = E->k;

    if(edwards_point_is_neutral(P1, E))
    {
        ec_point_copy(P3, P2, k);
        return;
    }
    if(edwards_point_is_neutral(P2, E))
    {
        ec_point_copy(P3, P1, k);
        return;
    }
    if(edwards_point_are_equal(P1, P2, E, stack))
    {
        edwards_point_dbl_dedicated(P3, P1, E, stack);
        return;
    }
    
    MPHELL_ASSERT(!field_elt_iszero(P1->z, k) && !field_elt_iszero(P2->z, k), "\nOne of the point is not regular it has its z-coordinate equals to 0.\n");
        
    field_elt_get_pool_elt(&a, k, stack);
    field_elt_get_pool_elt(&b, k, stack);
    field_elt_get_pool_elt(&c, k, stack);
    field_elt_get_pool_elt(&d, k, stack);
    field_elt_get_pool_elt(&e, k, stack);
    field_elt_get_pool_elt(&f, k, stack);
    
    switch(E->f)
    {
        case PROJECTIVE:
            /* Costs 10 M 1 S 1*a 1*d 7 add */
            field_elt_mul(a, P1->z, P2->z, k, stack);      /* A = z1 z2                        */
            field_elt_sqr(b, a, k, stack);                 /* B = A^2                          */
            field_elt_mul(c, P1->x, P2->x, k, stack);      /* C = x1 x2                        */
            field_elt_mul(d, P1->y, P2->y, k, stack);      /* D = y1 y2                        */
            field_elt_mul(e, E->b, c, k, stack);
            field_elt_mul(e, e, d, k, stack);              /* E = dC D                         */
            field_elt_sub(f, b, e, k);                     /* F = B - E                        */
            field_elt_add(P3->z, b, e, k);                 /* G = B + E (stored in P3->z)      */
            field_elt_add(P3->x, P1->x, P1->y, k); 
            field_elt_add(P3->y, P2->x, P2->y, k);
            field_elt_mul(P3->x, P3->y, P3->x, k, stack);  /* (x1 + y1) (x2 + y2)              */
            field_elt_sub(P3->x, P3->x, c, k);     
            field_elt_sub(P3->x, P3->x, d, k);
            field_elt_mul(P3->x, P3->x, f, k, stack);
            field_elt_mul(P3->x, P3->x, a, k, stack);      /* x3 = A F ((x1+y1)(x2+y2)-C-D)    */
            field_elt_mul(P3->y, E->a, c, k, stack);        
            field_elt_sub(P3->y, d, P3->y, k);       
            field_elt_mul(P3->y, P3->z, P3->y, k, stack);
            field_elt_mul(P3->y, a, P3->y, k, stack);      /* y3 = A G (D - a C)               */
            field_elt_mul(P3->z, P3->z,f, k, stack);       /* z3 = F G                         */
            field_elt_set_one(P3->t, k);                   /* T3 = 1                           */
            break;
    
        case EXTENDED_EDWARDS:
            /* Costs  9M 1*a 7 add */
            field_elt_mul(a, P1->x, P2->x, k, stack);      /* A = x1 x2                        */
            field_elt_mul(b, P1->y, P2->y, k, stack);      /* B = y1 y2                        */
            field_elt_mul(c, P1->z, P2->t, k, stack);      /* C = z1 t2                        */
            field_elt_mul(d, P1->t, P2->z, k, stack);      /* D = t1 z2                        */
            field_elt_add(e, d, c, k);                     /* E = D + C                        */
            field_elt_sub(c, d, c, k);                     /* H = D - C  (stored in C)         */
            field_elt_add(d, P2->x, P2->y, k);
            field_elt_sub(f, P1->x, P1->y, k);
            
            field_elt_mul(f, f, d, k, stack);
            field_elt_add(f, f, b, k);
            field_elt_sub(f, f, a, k);                     /* F = (x1 - y1)(x2 + y2) + B - A   */
            field_elt_mul(a, a, E->a, k, stack);
            field_elt_add(a, b, a, k);                     /* G = B + aA (stored in A)         */
                
            field_elt_mul(P3->x, e, f, k, stack);          /* X= E * F                         */ 
            field_elt_mul(P3->y, a, c, k, stack);          /* Y= G * H                         */
            field_elt_mul(P3->t, e, c, k, stack);          /* T= E * H                         */
            field_elt_mul(P3->z, a, f, k, stack);          /* Z= G * F                         */
            break;

            
            
        default :
            MPHELL_ASSERT( E->f== PROJECTIVE || E->f==EXTENDED_EDWARDS, "\nOnly Projective coordinates and Extended Edwards coordinates are supported for Twisted Edwards elliptic curve and thus for adding points on it.\n");
            break;
    }

    field_elt_relax_pool_elt(&a, k, stack);
    field_elt_relax_pool_elt(&b, k, stack);
    field_elt_relax_pool_elt(&c, k, stack);
    field_elt_relax_pool_elt(&d, k, stack);
    field_elt_relax_pool_elt(&e, k, stack);
    field_elt_relax_pool_elt(&f, k, stack);
 }

void 
edwards_point_dbl_dedicated (ec_point_ptr P3, ec_point_srcptr P1, ec_curve_srcptr E, uint8_t stack)
{
    /* Using formulas from [HWCD08, 3.3] */
    /* This formula can be viewed on the website http://www.hyperelliptic.org/EFD/g1p/auto-twisted-extended.html#doubling-dbl-2008-hwcd and is called "dbl-2008-hwcd" */
    /* Using formulae from [BBJ+08, §6] for Projective coordinates */
    /* This formula can be viewed on this website http://www.hyperelliptic.org/EFD/g1p/auto-twisted-projective.html#doubling-dbl-2008-bbjlp where it is called "dbl-2008-bbjlp" */

    field_elt a, b, c, d, e, f;
    field_t *k = E->k;

    field_elt_get_pool_elt(&a, k, stack);
    field_elt_get_pool_elt(&b, k, stack);
    field_elt_get_pool_elt(&c, k, stack);
    field_elt_get_pool_elt(&d, k, stack);
    field_elt_get_pool_elt(&e, k, stack);
    field_elt_get_pool_elt(&f, k, stack);

    MPHELL_ASSERT( !field_elt_iszero(P1->z, k), "This point is not regular it has its z-coordinate equals to 0"); 

    switch(E->f)
    {
        case PROJECTIVE:
            /*Cost 3M 4S 1*a 6 add 1 *2 */
            field_elt_add(a, P1->x, P1->y, k);  
            field_elt_sqr(b, a, k, stack);                  /* B = (X1 + Y1)^2              */
            field_elt_sqr(c, P1->x, k, stack);              /* C = (X1)^2                   */
            field_elt_sqr(d, P1->y, k, stack);              /* D = (Y1)^2                   */
            field_elt_mul(e, E->a, c, k, stack);            /* E = a.C                      */
            field_elt_add(f, e, d, k);                      /* F = E + D                    */
            field_elt_sqr(P3->x, P1->z, k, stack);          /* H = Z1^2  (stored in X3)     */
            field_elt_mul2(P3->x, P3->x, k);   
            field_elt_sub(P3->z, f, P3->x, k);              /* J = F - 2H (stored in z3)    */
            field_elt_sub(P3->x, b, c, k);
            field_elt_sub(P3->x, P3->x, d, k);
            field_elt_mul(P3->x, P3->x, P3->z, k, stack);   /* X3 = (B-C-D) J               */
            field_elt_sub(P3->y, e, d, k);
            field_elt_mul(P3->y, P3->y, f, k, stack);       /* Y3 = F (E-D)                 */
            field_elt_mul(P3->z, P3->z, f, k, stack);       /* Z3 = F J                     */
            field_elt_set_one(P3->t, k);                    /* T3 = 1                       */
            break;
    
        case EXTENDED_EDWARDS:
            /* Costs 4M 4 S 1 *a 6 add 1 *2  */
            field_elt_sqr(a, P1->x, k, stack);             /* A = X1^2                     */
            field_elt_sqr(b, P1->y, k, stack);             /* B = Y1^2                     */
            field_elt_sqr(c, P1->z, k, stack);     
            field_elt_mul2(c, c, k);                       /* C = 2Z1^2                    */
            field_elt_mul(d, E->a, a, k, stack);           /* D = aA */
            field_elt_add(e, P1->x, P1->y, k); 
            
            field_elt_sqr(e, e, k, stack);
            field_elt_sub(e, e, a, k);
            field_elt_sub(e, e, b, k);                     /* E = (X1 + Y1)^2 - A - B      */
            field_elt_add(a, d, b, k);                     /* G = D + B (stored in A)      */
            field_elt_sub(f, a, c, k);                     /* F = G - C                    */
            field_elt_sub(c, d, b, k);                     /* H = D - B (stored in C)      */
            field_elt_mul(P3->x, e, f, k, stack);          /* X = E * F                    */
            field_elt_mul(P3->y, a, c, k, stack);          /* Y = G * H                    */ 
            field_elt_mul(P3->t, e, c, k, stack);          /* T = E * H                    */
            field_elt_mul(P3->z, f, a, k, stack);          /* Z = F * G                    */
            break;
            
        default :
            MPHELL_ASSERT( E->f== PROJECTIVE || E->f==EXTENDED_EDWARDS, "\nOnly Projective coordinates and Extended Edwards coordinates are supported for Twisted Edwards elliptic curve and thus for doubling point on it !\n");
            break;
    }
    field_elt_relax_pool_elt(&a, k, stack);
    field_elt_relax_pool_elt(&b, k, stack);
    field_elt_relax_pool_elt(&c, k, stack);
    field_elt_relax_pool_elt(&d, k, stack);
    field_elt_relax_pool_elt(&e, k, stack);
    field_elt_relax_pool_elt(&f, k, stack);
}
