/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France) 

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell-hash_drbg.c
  * \brief Declaration of Hash Deterministic Random Bit Generator functions.
  */

#include "mphell-drbg_internal.h"

/**
 * @fn    void hash_df(uint8_t* requested_bytes, const drbg_type type, const uint8_t outlen, const uint8_t* input_string, uint32_t input_string_len, const uint32_t no_of_bytes_to_return)
 * @brief Derivation function used to compress entropy into an entropy of size seedlen 
 * @param requested_bytes       Destination array of bytes
 * @param type                  DRBG algorithm to use (can be DRBG_DRBG_SHA1)
 * @param outlen                Length of the HASH algorithm output cipher output
 * @param input_string          Entropy byte string to compress
 * @param input_string_len      Length of input_string (in bytes)
 * @param no_of_bytes_to_return Size (in bytes) of requested_bytes
 */
void hash_df(uint8_t* requested_bytes, const drbg_type type,
        const uint8_t outlen, const uint8_t* input_string, 
        uint32_t input_string_len, const uint32_t no_of_bytes_to_return)
{
    /* len = ceil(no_of_bytes_to_return / outlen) */
    uint32_t len = drbg_ceil(no_of_bytes_to_return, outlen);

    /* counter an 8-bit value representing 1 */
    uint8_t counter = 1;
    uint32_t i = 0;

    uint8_t tmp[outlen * len];
    uint8_t tmp2[5 + input_string_len];
    
    uint32_t no_of_bits_to_return = no_of_bytes_to_return*8;
    
    tmp2[1] = (uint8_t)(no_of_bits_to_return >> 24);
    tmp2[2] = (uint8_t)(no_of_bits_to_return >> 16);
    tmp2[3] = (uint8_t)(no_of_bits_to_return >> 8);
    tmp2[4] = (uint8_t)no_of_bits_to_return;
    memcpy(tmp2 + 5, input_string, input_string_len);
    for(i = 1; i <= len; i++)
    {
        /* temp = temp || Hash(counter||no_of_bits_to_return||input_string) */
        tmp2[0] = counter;
        
        switch(type)
        {
            case DRBG_SHA1 :
                sha1(tmp + outlen*(i-1), tmp2, (5 + input_string_len) * 8);
                break;
            case DRBG_SHA256 :
                sha256(tmp + outlen*(i-1), tmp2, (5 + input_string_len) * 8);
                break;
            default:
                return;
        }
        /* counter = coutner + 1*/
        counter = counter + 1;
    }
    /* requested_bytes = leftmost(temp, no_of_bytes_to_return) */
    memcpy(requested_bytes, tmp, no_of_bytes_to_return);
}

void hash_drbg_instantiate(mphell_status ret, drbg_internal_state initial_state,
        const uint8_t security_strength, const uint8_t* entropy_input, const uint16_t entropy_length,
        const uint8_t* personalization_string, 
        const uint64_t pers_length, const drbg_type type)
{
    uint8_t seedlen = 0;
    uint8_t outlen = 0;

    switch(type)
    {
        case DRBG_SHA1 :
            seedlen = 55;
            outlen = 20;
            break;
        case DRBG_SHA256 :
            seedlen = 55;
            outlen = 32;
            break;

        default:
            ret->flag = MPHELL_ERROR_RANDOM;
            char *text = "hash_drbg_instantiate : Invalid DRBG type";
            ret->info = malloc(strlen(text));
            memcpy(ret->info, text, strlen(text));
            return;
    }
    
    uint8_t seed[seedlen];
    uint32_t len = entropy_length ;
    
    if((personalization_string != NULL) && (pers_length>0))
    {
        len += pers_length;
    }

    /* seed_material = entropy_input || personalization_string */
    uint8_t seed_material[len];
    
    memcpy(seed_material, entropy_input, entropy_length);

    if((personalization_string != NULL) && (pers_length>0))
    {
        memcpy(seed_material + entropy_length, personalization_string, pers_length);
    }

    /* seed = Hash_df(seed_material, seedlen) */
    hash_df(seed, type, outlen, seed_material, len, seedlen);

    /* V = seed */
    initial_state->state->V = (uint8_t *)malloc(sizeof(uint8_t) * ((seedlen/8+(seedlen%8!=0))*8)); /* We reserve a multiple of 8 bits to avoid valgrind error with memcpy on raspberry */
    memcpy(initial_state->state->V, seed, seedlen);    

    /* C = Hash_df((0x00 || V), seedlen) */
    initial_state->state->C_Key = (uint8_t *)malloc(sizeof(uint8_t) * seedlen);
    uint8_t Vt[seedlen + 1];
    Vt[0] = 0x00;
    memcpy(Vt + 1, initial_state->state->V, seedlen);
    hash_df(initial_state->state->C_Key, type, outlen, Vt, seedlen + 1, seedlen);
    
    /* reseed_counter = 1 */
    initial_state->state->reseed_counter = 1;
        
    /* store values */
    initial_state->info->seedlen = seedlen;
    initial_state->info->outlen = outlen;
    
    ret->flag = MPHELL_SUCCESS;
}

void hash_drbg_reseed(mphell_status ret, drbg_internal_state working_state,
        const uint8_t* entropy_input, const uint16_t entropy_length, const uint8_t* additional_input,
        const uint64_t add_length)
{
    uint8_t seedlen = working_state->info->seedlen;
    uint8_t outlen = working_state->info->outlen;

    uint8_t seed[seedlen];
    uint64_t len = 1 + seedlen + entropy_length;
    
    if((additional_input != NULL) && (add_length > 0))
        len += add_length;

    /* seed_material = 0x01 || V || entropy_input || additional_input */
    uint8_t seed_material[len];
    seed_material[0] = 0x01;
    memcpy(seed_material + 1, working_state->state->V, seedlen); 
    memcpy(seed_material + 1 + seedlen, entropy_input, entropy_length);
    
    if((additional_input != NULL) && (add_length > 0))
    {
        memcpy(seed_material + 1 + seedlen + entropy_length, additional_input, add_length);
    }
    
    /* seed = Hash_df(seed_material, seedlen) */
    hash_df(seed, working_state->type, outlen, seed_material, len, seedlen);
        
    /* V = seed */
    memcpy(working_state->state->V, seed, seedlen);
    
    /* C = Hash_df((0x00 || V), seedlen)*/
    uint8_t Vt[seedlen + 1];
    Vt[0] = 0x00;
    memcpy(Vt + 1, working_state->state->V, seedlen);
    
    hash_df(working_state->state->C_Key, working_state->type , outlen, Vt, seedlen + 1, seedlen);
    
    /* reseed_counter = 1 */
    working_state->state->reseed_counter = 1;

    ret->flag = MPHELL_SUCCESS;
}

/**
 * @fn    void hashgen (uint8_t * returned_bytes, const uint32_t requested_no_of_bytes, const drbg_internal_state working_state)
 * @brief Generate random bytes using the internal secret value V and the cryptographic hash function
 * @param returned_bytes        Destination random bytes string
 * @param requested_no_of_bytes Number of requested random bytes
 * @param working_state         Current state of the DRBG (current value of V, C_key and reseed_counter)
 */
void hashgen (uint8_t * returned_bytes, const uint32_t requested_no_of_bytes,
        const drbg_internal_state working_state)
{
    uint8_t seedlen = working_state->info->seedlen;
    uint8_t outlen = working_state->info->outlen;

    /* m = ceil(requested_no_of_bytes / outlen) */
    uint32_t m = drbg_ceil(requested_no_of_bytes, outlen);

    /* data = V */
    uint8_t data[seedlen];
    memcpy(data, working_state->state->V, seedlen);
    uint8_t W[m * outlen];
    
    uint32_t i = 0;
    for(i = 1; i <= m; i++)
    {
        /* W = W || Hash(data) */
        switch(working_state->type)
        {
            case DRBG_SHA1 :
                sha1(W + outlen*(i-1), data, 8 * seedlen);
                break;
            case DRBG_SHA256 :
                sha256(W + outlen*(i-1), data, 8 * seedlen);
                break;
            default:
                return;
        }
        /* data = (data + 1) mod 2^seedlen */ 
        drbg_incr_data(data, seedlen);
    } 

    /* returned_bytes = Leftmost(W, requested_no_of_bytes) */
    memcpy(returned_bytes, W, requested_no_of_bytes);
}

void hash_drbg_generate(mphell_status ret, uint8_t* returned_bytes, 
        drbg_internal_state working_state, const uint32_t requested_no_of_bytes,
        const uint8_t* additional_input, const uint64_t add_length)
{
    uint8_t seedlen = working_state->info->seedlen;
    uint8_t outlen = working_state->info->outlen;

    /* If reseed_counter > ressed_interval, then return an indication that a 
     * reseed is required */
    if(working_state->state->reseed_counter > working_state->info->reseed_interval)
    {
        memset(returned_bytes, 0, requested_no_of_bytes);
        ret->flag = MPHELL_RESEED_REQUIRED;
        return;
    }
    
    uint8_t w[outlen];
    uint8_t* V = working_state->state->V;
    uint8_t *C = working_state->state->C_Key;

    if( (additional_input != NULL) && (add_length > 0)) 
    {
        /* w = Hash(0x02 || V || additional_input) */
        uint8_t tmp[1 + seedlen + add_length];
        tmp[0] = 0x02;
        memcpy(tmp + 1, V, seedlen);
        memcpy(tmp + 1 + seedlen, additional_input, add_length);
        if(working_state->type == DRBG_SHA1)
        {
            sha1(w, tmp, (1 + seedlen + add_length)*8);
        }
        else if(working_state->type == DRBG_SHA256)
        {
            sha256(w, tmp, (1 + seedlen + add_length)*8);
        }
        else
        {
            ret->flag = MPHELL_ERROR_RANDOM;
            char *text = "hash_drbg_generate : Invalid DRBG type";
            ret->info = malloc(strlen(text));
            memcpy(ret->info, text, strlen(text));
            return;
        }

        /* V = (V + w) mod 2^seedlen */
        drbg_add_data(V, seedlen, w, outlen);
    }
    
    /* (returned_bytes) = Hashgen(requested_no_of_bytes, V) */
    hashgen(returned_bytes, requested_no_of_bytes, working_state);

    /* H = Hash(0x03 || V) */
    uint8_t H[outlen];
    uint8_t tmp1[1 + seedlen];
    tmp1[0] = 0x03;
    memcpy(tmp1 + 1, V, seedlen);
    if(working_state->type == DRBG_SHA1)
    {
        sha1(H, tmp1, (1 + seedlen)*8);
    }
    else if(working_state->type == DRBG_SHA256)
    {
        sha256(H, tmp1, (1 + seedlen)*8);
    }
    else
    {
        ret->flag = MPHELL_ERROR_RANDOM;
        char *text = "hash_drbg_generate : Invalid DRBG type";
        ret->info = malloc(strlen(text));
        memcpy(ret->info, text, strlen(text));
        return;
    }
    
    /* V = (V + H + C + reseed_counter ) mod 2^seedlen */
    drbg_add_data(V, seedlen, H, outlen);
    drbg_add_data(V, seedlen, C, seedlen);
    drbg_add_int(V, seedlen, working_state->state->reseed_counter);

    /* reseed_counter = reseed_counter + 1 */
    working_state->state->reseed_counter += 1;
    
    ret->flag = MPHELL_SUCCESS;
}
