/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France) 

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell-init.c
  * \brief Definition of MPHELL initialisation function.
  */

#include <stdlib.h>
#include <time.h>

#include "mphell-init.h"

random_state_t *random_working_state; 
uint16_t random_security_strength;

void 
init_mphell(const uint16_t security_strength, const random_type type, 
        const entropy_type entropy)
{
    mphell_status ret;
    random_working_state = (random_state_t*)malloc(sizeof(random_state_t));
    /* Instantiate the DRBG with reseed_interval to 2^20 = 1048576, if an MPHELL DRBG is used */
    random_instantiate(ret, random_working_state, security_strength, NULL, 0, 
            type, entropy, 1048576);
    if(ret->flag != MPHELL_SUCCESS)
    {
        mphell_error_free(ret->info);
    }
    random_security_strength = security_strength;

    alloc_tmp_memory();
}

#if BLOCK_SIZE == 64

uint64_t 
rand64()
{
    mphell_status ret;
    uint64_t res;
    random_generate(ret, (uint8_t*)&res, random_working_state, 8, 
                random_security_strength, NULL, 0);
    if(ret->flag != MPHELL_SUCCESS)
    {
        mphell_error_free(ret->info);
    }
    return res;
}
#endif

uint32_t 
rand32()
{
    mphell_status ret;
    uint32_t res;
    random_generate(ret, (uint8_t*)&res, random_working_state, 4, 
                random_security_strength, NULL, 0);
    if(ret->flag != MPHELL_SUCCESS)
    {
        mphell_error_free(ret->info);
    }
    return res;
}

int 
mphell_rng(void * param, unsigned char * res, size_t n)
{
    mphell_status ret;
    random_generate(ret, (uint8_t*)res, random_working_state, n, 
                random_security_strength, NULL, 0);
    if(ret->flag != MPHELL_SUCCESS)
    {
        mphell_error_free(ret->info);
    }
    return 0;
}

void 
free_mphell()
{
    mphell_status ret;
    random_uninstantiate(ret, random_working_state);
    if(ret->flag != MPHELL_SUCCESS)
    {
        mphell_error_free(ret->info);
    }
    random_security_strength = 0;
    free(random_working_state);

    free_tmp_memory();
}
