/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France) 

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell-jacobi.c
  * \brief Definition of Jacobi ECC functions.
  */

#include "mphell-curve.h"


/************************************SETTERS**********************************/

void
jacobi_quartic_compute_disc(ec_curve E, uint8_t stack)
{
    field_t *k = E->k;
    field_elt t2, t1;
    field_elt_get_pool_elt(&t2, k, stack);
    field_elt_get_pool_elt(&t1, k, stack);
    
    field_elt_sqr(t1, E->a, k, stack);              /* t1 = a^2                 */
    field_elt_set_one(t2, k);                       /* t2 = 1                   */
    field_elt_sub(t1, t1, t2, k);                   /* t1 = a^2 - 1             */
    field_elt_sqr(t1, t1, k, stack);                /* t1 = (a^2 -1)^2          */
    field_elt_copy(E->D, t1, k);                    /* D = (a^2 -1)^2           */
    field_elt_set_ui(t2, 256, false, k, stack);
    field_elt_mul(E->disc, t1, t2, k, stack);       /* disc = (2^8).(a^2 - 1)^2 */
    
    field_elt_relax_pool_elt(&t1, k, stack);
    field_elt_relax_pool_elt(&t2, k, stack);
}

#include "mphell-drbg_internal.h"
#include "mphell-sha1.h"

bool jacobi_quartic_verify_random_generation(ec_curve E, const char * seed, uint8_t stack)
{
    /* Source : Working Draft AMERICAN NATIONAL STANDARD X9.62-1998, section A.3.4.2 */
    field_t *k = E->k;
    
    field_elt r1;
    field_elt_get_pool_elt(&r1, k, stack);
    
    number tmp;
    number_tmp_alloc(&tmp, k->size, stack);
    
    /* t = log2(p) */
    uint16_t t;
    switch(k->type)
    {
        case FP:
             t = number_log2(((fp_param)k->param)->p);
            break;
        case FP2:
        case FP3:
            t = number_log2(((fp_param)((fp2_param)k->param)->base)->p);
            break;
        default: 
            return false;
    }
    /* s = (t-1) / 160 */
    uint16_t s = (t-1)/160;
    /* h = t - 160*s */
    uint16_t h = t - (160*s);
    h --; /* set the leftmost bit of c0 to 0 */
    uint16_t hbytes = (h / 8) + (h%8 != 0);
    
    /* Alloc W' */
    
    uint8_t * Wprime = calloc(s*20+hbytes, sizeof(uint8_t));
    
    /* Step 1 and 2, compute W0: */
    
    uint8_t H[20];
    uint8_t seed_bin[20];
    /* Convert hexadecimal seed into binary seed */
    hex2bin(seed, seed_bin);
    sha1(H, seed_bin, 160);
    /* c0 = h rightmost bits of H */
    memcpy(Wprime, H+(20-hbytes), hbytes);
    uint8_t mask;
    switch (h%8)
    {
        case 0:
            mask = 0xff;
            break;
        case 1:
            mask = 0x1;
            break;
        case 2:
            mask = 0x3;
            break;
        case 3:
            mask = 0x7;
            break;
        case 4:
            mask = 0xf;
            break;
        case 5:
            mask = 0x1f;
            break;
        case 6:
            mask = 0x3f;
            break;
        case 7:
            mask = 0x7f;
            break;
    }
    
    Wprime[0]&=mask;
    
    /* Step 2 : done with h--*/
    
    /* Step 3 and 4 : */
    
    uint16_t i;
    for (i=0; i<s; i++)
    {
        drbg_incr_data(seed_bin, 20);
        sha1(H, seed_bin, 160);
        memcpy(Wprime+i*20+hbytes, H, 20);
    }
    
    /* Step 5 : Transform Wprime into a field element */
    
    /* Convert binary hash into hexadecimal string */
    unsigned char Wprime_hex[2*(s*20+hbytes)+1];
    bin2hex(Wprime, s*20+hbytes, Wprime_hex);
    switch(k->type)
    {
        case FP:
            field_elt_set_str(r1, (const char *)Wprime_hex, 16, false, k, stack);
            break;
        case FP2:
        case FP3:
            number_set_str(tmp, (const char *)Wprime_hex, 16);
            field_elt_set_number(r1, false, k, stack, 1, tmp);
            break;
        default:
            field_elt_set_one(r1, k);
    }
    
    /* Step 6 : Accept or reject */
    
    int8_t res = field_elt_cmp(r1, E->a, k);
    
    number_tmp_free(&tmp, k->size, stack);
    field_elt_relax_pool_elt(&r1, k, stack);
    free(Wprime);
    
    return (res == 0);
}

void jacobi_quartic_curve_random_generation(fe_ptr a, char * seed_res, field_srcptr k, uint8_t stack)
{
    /* Source : Working Draft AMERICAN NATIONAL STANDARD X9.62-1998, section A.3.3.2 */
    bool test;
    number seed;
    number bound_h;
    number bound_l;
    number_tmp_alloc(&seed, bits_to_nblock(160), stack);
    number_tmp_alloc(&bound_h, bits_to_nblock(160), stack);
    number_tmp_alloc(&bound_l, bits_to_nblock(160), stack);
    number_set_str(bound_h, "ffffffffffffffffffffffffffffffffffffffff", 16); /* 2^161 - 1 */
    number_set_str(bound_l, "1000000000000000000000000000000000000000", 16); /* 2^160     */
    
    field_elt r1;
    field_elt_get_pool_elt(&r1, (field_ptr)k, stack);
    
    number tmp;
    number_tmp_alloc(&tmp, k->size, stack);
    
    char * seed_hex;
    
    /* t = log2(p) */
    uint16_t t;
    switch(k->type)
    {
        case FP:
             t = number_log2(((fp_param)k->param)->p);
            break;
        case FP2:
        case FP3:
            t = number_log2(((fp_param)((fp2_param)k->param)->base)->p);
            break;
        default: 
            t = 0;
    }
    /* s = (t-1) / 160 */
    uint16_t s = (t-1)/160;
    /* h = t - 160*s */
    uint16_t h = t - (160*s);
    h --; /* set the leftmost bit of c0 to 0 */
    uint16_t hbytes = (h / 8) + (h%8 != 0);
    
    /* Alloc W' */
    uint8_t * Wprime = calloc(s*20+hbytes, sizeof(uint8_t));

    do
    {
        /* Step 1 : Choose an arbitrary bit string SEED of bit length at least 160 bits */
        do
        {
            number_random1(seed, bound_h, stack);
        }
        while(number_cmp(seed, bound_l)<0);
        
        number_str(&seed_hex, seed, 16);
        
        /* Step 2 and 3, compute W0 :  */
        
        uint8_t H[20];
        uint8_t seed_bin[20];
        /* Convert hexadecimal seed into binary seed */
        hex2bin(seed_hex, seed_bin);
        sha1(H, seed_bin, 160);
        /* c0 = h rightmost bits of H */
        memcpy(Wprime, H+(20-hbytes), hbytes);
        uint8_t mask;
        switch (h%8)
        {
            case 0:
                mask = 0xff;
                break;
            case 1:
                mask = 0x1;
                break;
            case 2:
                mask = 0x3;
                break;
            case 3:
                mask = 0x7;
                break;
            case 4:
                mask = 0xf;
                break;
            case 5:
                mask = 0x1f;
                break;
            case 6:
                mask = 0x3f;
                break;
            case 7:
                mask = 0x7f;
                break;
        }
        
        Wprime[0]&=mask;
        
        /* Step 4 and 5 : */
        
        uint16_t i;
        for (i=0; i<s; i++)
        {
            drbg_incr_data(seed_bin, 20);
            sha1(H, seed_bin, 160);
            memcpy(Wprime+i*20+hbytes, H, 20);
        }
        
        /* Convert binary hash into hexadecimal string */
        unsigned char Wprime_hex[2*(s*20+hbytes)+1];
        bin2hex(Wprime, s*20+hbytes, Wprime_hex);
        
        /* Step 6 : Transform Wprime into a field element */
        
        switch(k->type)
        {
            case FP:
                field_elt_set_str(r1, (const char *)Wprime_hex, 16, false, k, stack);
                break;
            case FP2:
            case FP3:
                number_set_str(tmp, (const char *)Wprime_hex, 16);
                field_elt_set_number(r1, false, k, stack, 1, tmp);
                break;
            default:
                field_elt_set_one(r1, k);
        }
        
        /* Step 7: Choose a = r1 */
        
        field_elt_copy(a, r1, k);
        
        test = field_elt_isone(a, k);
        if (test)
        {
            free(seed_hex);
        }
    }
    /* Step 8: verifie that the curve is elliptic */
    while(test);
    
    strcpy(seed_res, seed_hex);
    
    number_tmp_free(&tmp, k->size, stack);
    field_elt_relax_pool_elt(&r1, (field_ptr)k, stack);
    number_tmp_free(&seed, bits_to_nblock(160), stack);
    number_tmp_free(&bound_h, bits_to_nblock(160), stack);
    number_tmp_free(&bound_l, bits_to_nblock(160), stack);
    free(seed_hex);
    free(Wprime);
}

void 
jacobi_quartic_point_set_neutral (ec_point_ptr dst, ec_curve_srcptr E, uint8_t stack)
{
    field_t * k = E->k;
    field_elt_set_ui(dst->x, 0, true, k, stack);
    field_elt_set_one(dst->y, k);
    field_elt_set_ui(dst->t, 0, true, k, stack);
    field_elt_set_one(dst->z, k);
}

void 
jacobi_quartic_point_set_aff (ec_point_ptr P, fe_srcptr x, fe_srcptr y, field_srcptr k, uint8_t stack)
{
    field_elt_copy(P->x, x, k);
    field_elt_copy(P->y, y, k);
    field_elt_sqr(P->t, P->x, k, stack);
    field_elt_set_one(P->z, k);
}

void 
jacobi_quartic_point_set_aff_str (ec_point_ptr P, const char *str_x, const char *str_y,
        const bool is_reduced, const uint8_t base, field_srcptr k, uint8_t stack)
{
    field_elt_set_str(P->x, str_x, base, is_reduced, k, stack);
    field_elt_set_str(P->y, str_y, base, is_reduced, k, stack);
    field_elt_sqr(P->t, P->x, k, stack);
    field_elt_set_one(P->z, k);
}

bool
jacobi_quartic_belongs (ec_point_srcptr P, ec_curve_srcptr E, uint8_t stack)
{
    /* P = (X, Y, T, Z)                             */
    /* E: y^2 = z^2 + 2ax^2 + t^2; x^2 = zt; a != 1 (this is for the discriminant) */
    
    field_elt f, g, h, u;
    field_t *k = E->k;

    field_elt_get_pool_elt(&f, k, stack);
    field_elt_get_pool_elt(&g, k, stack);
    field_elt_get_pool_elt(&h, k, stack);
    field_elt_get_pool_elt(&u, k, stack);

    field_elt_sqr(f, P->z, k, stack);          /* f = Z^2                  */
    field_elt_sqr(g, P->t, k, stack);          /* g = T^2                  */
    field_elt_add(h, f, g, k);                 /* h = Z^2 + T^2            */

    field_elt_add(f, P->z, P->t, k);           /* f = Z + T                */
    field_elt_sqr(g, f, k, stack);             /* g = (Z+T)^2              */
    field_elt_sub(f, g, h, k);                 /* f = 2.Z.T                */

    field_elt_sqr(u, P->x, k, stack);          /* u = X^2                  */
    field_elt_add(g, u, u, k);                 /* g = 2.X^2                */

    if (field_elt_cmp(f, g, k) != 0)           /* 2*x^2 ?= 2*z*t           */
    {
        field_elt_relax_pool_elt(&f, k, stack);
        field_elt_relax_pool_elt(&g, k, stack);
        field_elt_relax_pool_elt(&h, k, stack);
        field_elt_relax_pool_elt(&u, k, stack);
        return false;
    }

    field_elt_mul(g, g, E->a, k, stack);       /* g = 2.a.X^2              */
    field_elt_add(f, h, g, k);                 /* f = Z^2 + 2.a.X^2 + T^2  */
    field_elt_sqr(h, P->y, k, stack);          /* h = Y^2                  */

    if (field_elt_cmp(h, f, k) != 0) 
    {
        field_elt_relax_pool_elt(&f, k, stack);
        field_elt_relax_pool_elt(&g, k, stack);
        field_elt_relax_pool_elt(&h, k, stack);
        field_elt_relax_pool_elt(&u, k, stack);
        return false;
    }
    
    field_elt_relax_pool_elt(&f, k, stack);
    field_elt_relax_pool_elt(&g, k, stack);
    field_elt_relax_pool_elt(&h, k, stack);
    field_elt_relax_pool_elt(&u, k, stack);

    return true;
}

void 
jacobi_quartic_point_random (ec_point_ptr P, ec_curve_srcptr E, uint8_t stack)
{
    field_elt f, g;
    field_t *k = E->k;

    field_elt_get_pool_elt(&f, k, stack);
    field_elt_get_pool_elt(&g, k, stack);

    do {
        field_elt_random(P->x, k, stack);
        field_elt_sqr(P->z, P->x, k, stack);
        field_elt_mul(g, P->z, E->a, k, stack); 
        field_elt_add(g, g, g, k);             /* g = 2*a*x^2          */
        field_elt_inc(f, g, k);
        field_elt_sqr(g, P->z, k, stack);
        field_elt_add(g, g, f, k);             /* g = z^2 + 2a x^2 + 1 */
    } while (!field_elt_issquare(g, k, stack));
    field_elt_sqrt(f, g, k, stack);
    
    
    field_elt_copy(P->y, f, k);  
    field_elt_copy(P->t, P->z, k); 
    field_elt_set_one(P->z,k); 

    field_elt_relax_pool_elt(&f, k, stack);
    field_elt_relax_pool_elt(&g, k, stack);
}

void
jacobi_quartic_point_norm (ec_point_ptr P, ec_curve_srcptr E, uint8_t stack)
{
    field_t *k = E->k;
    field_elt u;
    field_elt_get_pool_elt(&u, k, stack);

    field_elt_inv(u, P->z, k, stack);
    field_elt_mul(P->x, P->x, u, k, stack);
    field_elt_mul(P->y, P->y, u, k, stack);
    field_elt_sqr(P->t, P->x, k, stack);
    field_elt_set_one(P->z, k);

    field_elt_relax_pool_elt(&u, k, stack); 
}

void
jacobi_quartic_point_get_x_affine (field_elt x, ec_point_ptr P, ec_curve_srcptr E, uint8_t stack)
{
    field_t *k = E->k;
    field_elt u;
    field_elt_get_pool_elt(&u, k, stack);

    field_elt_inv(u, P->z, k, stack);
    field_elt_mul(x, P->x, u, k, stack);

    field_elt_relax_pool_elt(&u, k, stack); 
}

void
jacobi_quartic_point_get_y_affine (field_elt y, ec_point_ptr P, ec_curve_srcptr E, uint8_t stack)
{
    field_t *k = E->k;
    field_elt u;
    field_elt_get_pool_elt(&u, k, stack);

    field_elt_inv(u, P->z, k, stack);
    field_elt_mul(y, P->y, u, k, stack);

    field_elt_relax_pool_elt(&u, k, stack); 
}

/*******************************COMPARISON************************************/

bool
jacobi_quartic_point_is_neutral (ec_point_srcptr P, ec_curve_srcptr E)
{
    field_t *k = E->k;
    return (field_elt_iszero(P->x, k) && field_elt_iszero(P->t, k) &&
            (field_elt_cmp(P->y, P->z, k) == 0));
}

bool
jacobi_quartic_point_are_equal (ec_point_srcptr P1, ec_point_srcptr P2, 
        ec_curve_srcptr E, uint8_t stack)
{
    field_elt u, v;
    field_t *k = E->k;

    if (field_elt_iszero(P1->t, k))
    {
        if (!field_elt_iszero(P2->t, k))
        {
            return false;
        }
        if (field_elt_iszero(P1->y, k))
        {

            return field_elt_iszero(P2->y, k);
        }
        
        field_elt_get_pool_elt(&u, k, stack);
        field_elt_get_pool_elt(&v, k, stack);

        MPHELL_ASSERT_ALWAYS( field_elt_cmp(P1->x, P1->t, k)==0, "The point P1 is not a point in a system of Extended Homogenous Projective coordinates"); /* In this case x should be equals to 0 */
        MPHELL_ASSERT_ALWAYS( field_elt_cmp(P2->x, P2->t, k)==0, "The point P2 is not a point in a system of Extended Homogenous Projective coordinates"); /* In this case x should be equals to 0 */

        field_elt_mul(u, P1->z, P2->y, k, stack); 
        field_elt_mul(v, P1->y, P2->z, k, stack); 
        bool res = (field_elt_cmp(u, v, k) == 0);
        
        field_elt_relax_pool_elt(&u, k, stack);
        field_elt_relax_pool_elt(&v, k, stack);
        return res;
    } 
    else 
    {
        field_elt_get_pool_elt(&u, k, stack);
        field_elt_get_pool_elt(&v, k, stack);

        field_elt_mul(u, P1->x, P2->t, k, stack);
        field_elt_mul(v, P1->t, P2->x, k, stack);
        if (field_elt_cmp(u, v, k) != 0) 
        {
            field_elt_relax_pool_elt(&u, k, stack);
            field_elt_relax_pool_elt(&v, k, stack);
            return false; 
        }
        field_elt_mul(u, P1->y, P2->t, k, stack);
        field_elt_mul(v, P1->t, P2->y, k, stack);
        bool res = (field_elt_cmp(u, v, k) == 0);

        field_elt_sqr(u,P1->x, k, stack);
        field_elt_mul(v,P1->t,P1->z, k, stack); 
        MPHELL_ASSERT_ALWAYS( field_elt_cmp(u, v, k)==0, "The point P1 is not a point in a system of Extended Homogenous Projective coordinates");
        field_elt_sqr(v,P2->x, k, stack);
        field_elt_mul(u,P2->t,P2->z, k, stack);
        MPHELL_ASSERT_ALWAYS( field_elt_cmp(u, v, k)==0, "The point P2 is not a point in a system of Extended Homogenous Projective coordinates");

        field_elt_relax_pool_elt(&u, k, stack);
        field_elt_relax_pool_elt(&v, k, stack);
        return res;
    }
}

/*******************************OPERATIONS************************************/

void 
jacobi_quartic_point_neg (ec_point_ptr P3, ec_point_srcptr P1, ec_curve_srcptr E)
{
    field_t *k = E->k;
    field_elt_neg(P3->x, P1->x, k);
    field_elt_copy(P3->y, P1->y, k);
    field_elt_copy(P3->z, P1->z, k);
    field_elt_copy(P3->t, P1->t, k);
}

/* UNIFIED => Resistant to SPA */

void 
jacobi_quartic_point_add_unified (ec_point_ptr P3, ec_point_srcptr P1, ec_point_srcptr P2, ec_curve_srcptr E, uint8_t stack)
{
    /* P1 = (X1, Y1, T1, Z1)     */
    /* P2 = (X2, Y2, T2, Z2)     */
    /* P3 = (X3, Y3, T3, Z3)     */
    /* E: y^2 = z^2 + 2ax^2 + t^2; x^2 = zt; a != 1 */
    /*
    These formulas are derived from the paper [HKCD09] "Jacobi Quartic Curves Revisited" 
    written by Huseyin Hisil, Kenneth Koon-Ho Wong, Gary Carter and Ed Dawson published in 
    Australasian Conference on Information Security and Privacy 2009 (pp. 452-468). 
    Springer, Berlin, Heidelberg
    Costs 8 Mul 2 Sqr 1 *a 13 Add 1*2
    */
    
    field_t *k = E->k;

    if(jacobi_quartic_point_is_neutral(P1, E))
    {
        ec_point_copy(P3, P2, k);
    }
    else if(jacobi_quartic_point_is_neutral(P2, E))
    {
        ec_point_copy(P3, P1, k);
    }
    else
    {
        field_elt x, y, z, t;
        field_elt f, g, h;
        field_elt u3, v3;

        field_elt_get_pool_elt(&x, k, stack);
        field_elt_get_pool_elt(&y, k, stack);
        field_elt_get_pool_elt(&z, k, stack);
        field_elt_get_pool_elt(&t, k, stack);

        field_elt_get_pool_elt(&f, k, stack);
        field_elt_get_pool_elt(&g, k, stack);
        field_elt_get_pool_elt(&h, k, stack);

        field_elt_get_pool_elt(&u3, k, stack);
        field_elt_get_pool_elt(&v3, k, stack);

        field_elt_add(z, P1->x, P1->y, k);                 /* z = X1 + Y1                                                                              */
        field_elt_add(t, P2->x, P2->y, k);                 /* t = X2 + Y2                                                                              */
        field_elt_mul(v3, z, t, k, stack);                 /* v3 = (X1+Y1).(X2+Y2)                                                                     */
        field_elt_mul(x, P1->x, P2->x, k, stack);          /* x = X1.X2                                                                                */
        field_elt_mul(y, P1->y, P2->y, k, stack);          /* y = Y1.Y2                                                                                */
        field_elt_sub(v3, v3, x, k);                       /* v3 = (X1+Y1).(X2+Y2) - X1.X2                                                             */
        field_elt_sub(v3, v3, y, k);                       /* v3 = (X1+Y1).(X2+Y2) - X1.X2 - Y1.Y2 = X1.Y2 + X2.Y1                                     */
        
        field_elt_mul(z, P1->z, P2->z, k, stack);          /* z = Z1.Z2                                                                                */
        field_elt_mul(t, P1->t, P2->t, k, stack);          /* t = T1.T2                                                                                */
        field_elt_sub(u3, z, t, k);                        /* u3 = Z1.Z2 - T1.T2                                                                       */

        field_elt_add(z,z,t, k);                           /* z = Z1.Z2 + T1.T2                                                                        */ 

        field_elt_add(f, P1->z, P1->t, k);                 /* f = Z1 + T1                                                                              */
        field_elt_add(g, P2->z, P2->t, k);                 /* g = Z2 + T2                                                                              */
        field_elt_mul(h, f, g, k, stack);                  /* h = (Z1+T1).(Z2+T2)                                                                      */
        field_elt_sub(h,h,z, k);                           /* h = (Z1+T1).(Z2+T2) - Z1.Z2 - T1.T2 = Z1.T2 + Z2.T1                                      */
        
        field_elt_add(f, h, y, k);                         /* f = Z1.T2 + Z2.T1 + Y1.Y2                                                                */
        field_elt_mul2(y, x, k);                           /* y = 2.X1.X2                                                                              */
        field_elt_mul(h, y, E->a, k, stack);               /* h = 2.X1.X2.a                                                                            */
        field_elt_add(g, f, h, k);                         /* g = Z1.T2 + Z2.T1 + Y1.Y2 + 2.X1.X2.a                                                    */

        field_elt_add(f, y, z, k);                         /* f = 2.X1.X2 + Z1.Z2 + T1.T2                                                              */ 
        field_elt_mul(y, f, g, k, stack);                  /* y = (2.X1.X2 + Z1.Z2 + T1.T2).(Z1.T2 + Z2.T1 + Y1.Y2 + 2.X1.X2.a)                        */

        field_elt_sqr(P3->t, v3, k, stack);                /* T3 = (X1.Y2 + X2.Y1)^2                                                                   */ 
        field_elt_sqr(P3->z, u3, k, stack);                /* Z3 = (Z1.Z2 - T1.T2)^2                                                                   */ 
        field_elt_mul(P3->x, u3, v3, k, stack);            /* X3 = (Z1.Z2 - T1.T2)*(X1.Y2 + X2.Y1)                                                     */
        field_elt_sub(P3->y, y, P3->t, k);                 /* Y3 = (2.X1.X2 + Z1.Z2 + T1.T2).(Z1.T2 + Z2.T1 + Y1.Y2 + 2.X1.X2.a) - (X1.Y2 + X2.Y1)^2   */

        field_elt_relax_pool_elt(&x, k, stack);
        field_elt_relax_pool_elt(&y, k, stack);
        field_elt_relax_pool_elt(&z, k, stack);
        field_elt_relax_pool_elt(&t, k, stack);

        field_elt_relax_pool_elt(&f, k, stack);
        field_elt_relax_pool_elt(&g, k, stack);
        field_elt_relax_pool_elt(&h, k, stack);

        field_elt_relax_pool_elt(&u3, k, stack);
        field_elt_relax_pool_elt(&v3, k, stack);
    }
}


/* DEDICATED => Not resistant to SPA, but faster 
 * Only the doubling is implemented as dedicated because the dedicated addition is not faster than the unified addition 
 */

void 
jacobi_quartic_point_dbl_dedicated (ec_point_ptr P3, ec_point_srcptr P1, ec_curve_srcptr E, uint8_t stack)
{
    /* P1 = (X1, Y1, T1, Z1)                                        */
    /* P3 = (X3, Y3, T3, Z3)                                        */
    /* E: y^2 = z^2 + 2ax^2 + t^2; x^2 = zt; a != 1                 */
    
    /* X3 = 2.X1.Y1(2.Z1^2 + 2.a.X1^2 - Y1^2)                       */
    /* Y3 = 2.Y1^2(Y1^2 - 2.a.X1^2) - (2.Z1^2 + 2.a.X1^2 - Y1^2)^2  */
    /* Z3 = (2.Z1^2 + 2.a.X1^2 - Y1^2)^2                            */
    /* T3 = (2X1.Y1)^2                                              */
    
    /*
    These formulas are derived from the paper [HKCD09,§3.1] 
    Costs 3 Mul, 5 Sqr, 1 *a, 3 Add, 4*2                            */
    
    field_elt x12, y12, z12, x1y1, f1, f2, temp1;
    field_t *k = E->k;

    field_elt_get_pool_elt(&x12, k, stack);
    field_elt_get_pool_elt(&y12, k, stack);
    field_elt_get_pool_elt(&z12, k, stack);
    field_elt_get_pool_elt(&x1y1, k, stack);
    field_elt_get_pool_elt(&f1, k, stack);
    field_elt_get_pool_elt(&f2, k, stack);
    field_elt_get_pool_elt(&temp1, k, stack);
    
    field_elt_sqr(x12, P1->x, k, stack);           /* x12 = X1^2                                                   */
    field_elt_sqr(y12, P1->y, k, stack);           /* y12 = Y1^2                                                   */
    field_elt_sqr(z12, P1->z, k, stack);           /* z12 = Z1^2                                                   */
    field_elt_mul(x1y1, P1->x, P1->y, k, stack);   /* x1y1 = X1.Y1                                                 */
    field_elt_mul2(x1y1, x1y1, k);                 /* x1y1 = 2.X1.Y1                                               */
    
    field_elt_mul(f1, E->a, x12, k, stack);        /* f1 = a.X1^2                                                  */
    field_elt_mul2(f1, f1, k);                     /* f1 = 2.a.X1^2                                                */
    field_elt_sub(f1, f1, y12, k);                 /* f1 = 2.a.X1^2 - Y1^2                                         */
    
    field_elt_neg(f2, f1, k);                      /* f2 = Y1^2 - 2.a.X1^2                                         */
    
    field_elt_mul2(temp1, z12, k);                 /* temp1 = 2.Z1^2                                               */
    field_elt_add(f1, f1, temp1, k);               /* f1 = 2.Z1^2 + 2.a.X1^2 - Y1^2                                */
    
    field_elt_mul(P3->x, x1y1, f1, k, stack);      /* X3 = 2.X1.Y1(2.Z1^2 + 2.a.X1^2 - Y1^2)                       */
    
    field_elt_sqr(P3->z, f1, k, stack);            /* Z3 = (2.Z1^2 + 2.a.X1^2 - Y1^2)^2                            */
    
    field_elt_mul2(temp1, y12, k);                 /* temp1 = 2.Y1^2                                               */
    field_elt_mul(P3->y, temp1, f2, k, stack);     /* Y3 = 2.Y1^2(Y1^2 - 2.a.X1^2)                                 */
    field_elt_sub(P3->y, P3->y, P3->z, k);         /* Y3 = 2.Y1^2(Y1^2 - 2.a.X1^2) - (2.Z1^2 + 2.a.X1^2 - Y1^2)^2  */
    
    field_elt_sqr(P3->t, x1y1, k, stack);          /* T3 = (2X1.Y1)^2                                              */

    field_elt_relax_pool_elt(&x12, k, stack);
    field_elt_relax_pool_elt(&y12, k, stack);
    field_elt_relax_pool_elt(&z12, k, stack);
    field_elt_relax_pool_elt(&x1y1, k, stack);
    field_elt_relax_pool_elt(&f1, k, stack);
    field_elt_relax_pool_elt(&f2, k, stack);
    field_elt_relax_pool_elt(&temp1, k, stack);
}
