/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France) 

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell-random.c
  * \brief Definition of random functions.
  */

#include "mphell-random.h"

void
random_instantiate(mphell_status ret, random_state initial_state,
        const uint16_t requested_instantiation_security_strength,
        const uint8_t* personalization_string,
        const uint64_t pers_length, const random_type type,
        const entropy_type entropy, uint64_t reseed_interval)
{
    /* Get entropy from the source */
    
    uint8_t security_strength  = 0;
    uint8_t *entropy_input = NULL;
    char *text;
    
    switch(type)
    {
        case RANDOM_INTEL :
            /* Use its own entropy source */
            break;
        
        case STM32F4_RNG :
            /* Use its own entropy source */
            break;

        case RANDOM_AES128 :
        
        case RANDOM_AES192 :
        
        case RANDOM_AES256 :
        
        case RANDOM_SHA1 :
        
        case RANDOM_SHA256 :
            /* Set security_strength to the lowest security_strength greater than equal
             * to requested_instantiation_security_strength from the set
             *  {112, 128, 192, 256} */

            if(requested_instantiation_security_strength > 192)
            {
                security_strength = 256 / 8;
            }
            else if(requested_instantiation_security_strength > 128)
            {
                security_strength = 192 / 8;
            }
            else if(requested_instantiation_security_strength > 112)
            {
                security_strength = 128 / 8;
            }
            else
            {
                security_strength = 112 / 8;
            }
            
            /* We use 2*security_strength for entropy and 1*security_strength as nonce */
            entropy_input = malloc(3 * security_strength);

            get_entropy_input(ret, entropy_input, entropy, 3 * security_strength);
            
            if(ret->flag != MPHELL_SUCCESS)
            {
                ret->flag = MPHELL_ERROR_RANDOM;
                text = "random_instantiate : get_entropy_input failed"; 
                ret->info = malloc(strlen(text));
                memcpy(ret->info, text, strlen(text));
                if(entropy_input != NULL)
                {
                    free(entropy_input);
                }
                return;
            }
            break;
            
        default :
            ret->flag = MPHELL_ERROR_RANDOM;
            text = "random_instantiate : invalid type"; 
            ret->info = malloc(strlen(text));
            memcpy(ret->info, text, strlen(text));
            return;
    }
    
    /* Apply DRBG */
    
    switch(type)
    {
        case RANDOM_INTEL :
            initial_state->type = RANDOM_INTEL;
            ret->flag = MPHELL_SUCCESS; 
            break;
        
        case STM32F4_RNG :
            initial_state->type = STM32F4_RNG;
            ret->flag = MPHELL_SUCCESS; 
            break;

        case RANDOM_AES128 :
            drbg_instantiate (ret, &(initial_state->state_handle),
                        requested_instantiation_security_strength, 
                        DRBG_AES128, entropy, reseed_interval, entropy_input, 2 * security_strength, 
                        entropy_input + 2 * security_strength, security_strength, 
                        personalization_string, pers_length);
            if(ret->flag != MPHELL_SUCCESS)
            {
                initial_state->type = RANDOM_UNSET;
                ret->flag = MPHELL_ERROR_RANDOM;
            }
            else
            {
                initial_state->type = RANDOM_AES128;
                ret->flag = MPHELL_SUCCESS;
            }
            break;

        case RANDOM_AES192 :
            drbg_instantiate(ret, &(initial_state->state_handle),
                        requested_instantiation_security_strength, 
                        DRBG_AES192, entropy, reseed_interval, entropy_input, 2 * security_strength, 
                        entropy_input + 2 * security_strength, security_strength, 
                        personalization_string, pers_length);
            if(ret->flag != MPHELL_SUCCESS)
            {
                initial_state->type = RANDOM_UNSET;
                ret->flag = MPHELL_ERROR_RANDOM;
            }
            else
            {
                initial_state->type = RANDOM_AES192;
                ret->flag = MPHELL_SUCCESS;
            }
            break;

        case RANDOM_AES256 :
            drbg_instantiate(ret, &(initial_state->state_handle),
                        requested_instantiation_security_strength, 
                        DRBG_AES256, entropy, reseed_interval, entropy_input, 2 * security_strength, 
                        entropy_input + 2 * security_strength, security_strength, 
                        personalization_string, pers_length);
            if(ret->flag != MPHELL_SUCCESS)
            {
                initial_state->type = RANDOM_UNSET;
                ret->flag = MPHELL_ERROR_RANDOM;
            }
            else
            {
                initial_state->type = RANDOM_AES256;
                ret->flag = MPHELL_SUCCESS;
            }
            break;

        case RANDOM_SHA1 :
            drbg_instantiate (ret, &(initial_state->state_handle),
                        requested_instantiation_security_strength, 
                        DRBG_SHA1, entropy, reseed_interval, entropy_input, 2 * security_strength, 
                        entropy_input + 2 * security_strength, security_strength, 
                        personalization_string, pers_length);
            if(ret->flag != MPHELL_SUCCESS)
            {
                initial_state->type = RANDOM_UNSET;
                ret->flag = MPHELL_ERROR_RANDOM;
            }
            else
            {
                initial_state->type = RANDOM_SHA1;
                ret->flag = MPHELL_SUCCESS;
            }
            break;
        
        case RANDOM_SHA256 :
            drbg_instantiate (ret, &(initial_state->state_handle),
                        requested_instantiation_security_strength, 
                        DRBG_SHA256, entropy, reseed_interval, entropy_input, 2 * security_strength, 
                        entropy_input + 2 * security_strength, security_strength, 
                        personalization_string, pers_length);
            if(ret->flag != MPHELL_SUCCESS)
            {
                initial_state->type = RANDOM_UNSET;
                ret->flag = MPHELL_ERROR_RANDOM;
            }
            else
            {
                initial_state->type = RANDOM_SHA256;
                ret->flag = MPHELL_SUCCESS;
            }
            break;

        default :
            ret->flag = MPHELL_ERROR_RANDOM;
            text = "random_instantiate : invalid type"; 
            ret->info = malloc(strlen(text));
            memcpy(ret->info, text, strlen(text));
    }
    
    if(entropy_input != NULL)
    {
        free(entropy_input);
    }
}

void
random_reseed(mphell_status ret, random_state working_state, 
        const uint16_t entropy_length, const entropy_type entropy_src,
        const uint8_t* additional_input,
        const uint64_t add_length)
{
    /* Get entropy from the source */
    
    char *text;
    uint16_t length_bytes = 0;
    uint8_t* entropy_input = NULL;
    
    switch(working_state->type)
    {
        case RANDOM_INTEL :
            /* Use its own entropy source */
            break;
        
        case STM32F4_RNG :
            /* Use its own entropy source */
            break;

        case RANDOM_AES128 :
        
        case RANDOM_AES192 :
        
        case RANDOM_AES256 :
        
        case RANDOM_SHA1 :
        
        case RANDOM_SHA256 :
            length_bytes = entropy_length/8;
    
            entropy_input = malloc(length_bytes);
            get_entropy_input(ret, entropy_input, entropy_src,
                    length_bytes);
            
            if(ret->flag != MPHELL_SUCCESS)
            {
                ret->flag = MPHELL_ERROR_RANDOM;
                text = "random_reseed : get_entropy_input failed"; 
                ret->info = malloc(strlen(text));
                memcpy(ret->info, text, strlen(text));
                if(entropy_input != NULL)
                {
                    free(entropy_input);
                }
                return;
            }
            break;
            
        default :
            ret->flag = MPHELL_ERROR_RANDOM;
            text = "random_instantiate : invalid type"; 
            ret->info = malloc(strlen(text));
            memcpy(ret->info, text, strlen(text));
            return;
    }
    
    /* Apply DRBG */
    
    switch(working_state->type)
    {
        case RANDOM_INTEL :
            ret->flag = MPHELL_SUCCESS; 
            break;
        
        case STM32F4_RNG :
            ret->flag = MPHELL_SUCCESS; 
            break;

        case RANDOM_AES128 :

        case RANDOM_AES192 :

        case RANDOM_AES256 :

        case RANDOM_SHA1 :
        
        case RANDOM_SHA256 :
            drbg_reseed(ret, working_state->state_handle, 
                        entropy_input, length_bytes, 
                        additional_input, add_length);
            if(ret->flag != MPHELL_SUCCESS)
            {
                ret->flag = MPHELL_ERROR_RANDOM;
            }
            break;

        default :
            ret->flag = MPHELL_ERROR_RANDOM;
            text = "random_reseed : invalid type"; 
            ret->info = malloc(strlen(text));
            memcpy(ret->info, text, strlen(text));
    }
    if(entropy_input != NULL)
    {
        free(entropy_input);
    }
}

#if MPHELL_USE_RDRAND == 1

int RANDOM_INTEL_rdrand16_step(uint16_t *therand)
{ 
   uint16_t foo = 0;
   int cf_error_status = 0;
#if __x86_64__
   asm("rdrand %%ax;mov $1,%%edx;cmovae %%ax,%%dx;mov %%edx,%1;mov %%ax,\
           %0;":"=r"(foo),"=r"(cf_error_status)::"%ax","%dx");     
#endif
    *therand = foo;
   return cf_error_status;
}

#elif MPHELL_USE_RDRAND == 0

int RANDOM_INTEL_rdrand16_step(uint16_t *therand)
{ 
   /* Intel instruction RDRAND is not available, nb: gcc >= 4.8 is required */
   return 0;
}

#endif

/**
 * @fn    stm32f4_rng_error(mphell_status ret, uint8_t *data, const uint16_t length)
 * @brief If the user forget to write a correct function using the stm32f4 RNG to get random data, this one is used
 * @param ret    Set to MPHELL_ERROR_RANDOM
 * @param data   Byte array to fill with random data
 * @param length Length of data
 */
static void stm32f4_rng_error(mphell_status ret, uint8_t *data, const uint16_t length)
{
    /* You must redefine stm32f4_rng in your main STM32F4 application:                              */
    /* void (*stm32f4_rng)(mphell_status ret, uint8_t *data, const uint16_t length) = &your_implem; */
    ret->flag = MPHELL_ERROR_RANDOM;
}

void (*stm32f4_rng)(mphell_status ret, uint8_t *data, const uint16_t length) = &stm32f4_rng_error;

void
random_generate(mphell_status ret, uint8_t* returned_bytes, 
        random_state working_state,
        const uint32_t requested_number_of_bytes, 
        const uint16_t requested_security_strength,
        const uint8_t* additional_input, const uint64_t add_length)
{
    uint32_t nb = ceil((float)requested_number_of_bytes / 2);
    uint16_t tmp[nb];
    uint32_t i = 0;
    char *text;
    switch(working_state->type)
    {
        case RANDOM_INTEL :
            for(i = 0; i < nb; i++)
            {
                if(RANDOM_INTEL_rdrand16_step(tmp + i) != 1)
                {
                    ret->flag = MPHELL_ERROR_RANDOM;
                    text = "random_generate : RANDOM_INTEL_rdrand16_step\
                                   error";
                    ret->info = malloc(strlen(text));
                    memcpy(ret->info, text, strlen(text));
                    return;
                }
            }
            memcpy(returned_bytes, tmp, requested_number_of_bytes);
            ret->flag = MPHELL_SUCCESS; 
            break;
        
        case STM32F4_RNG :
            stm32f4_rng(ret, returned_bytes, requested_number_of_bytes);
            break;

        case RANDOM_AES128 :

        case RANDOM_AES192 :

        case RANDOM_AES256 :

        case RANDOM_SHA1 :
        
        case RANDOM_SHA256 :
            drbg_generate(ret, returned_bytes, working_state->state_handle, 
                        requested_number_of_bytes, requested_security_strength,
                        additional_input, add_length);
            if(ret->flag != MPHELL_SUCCESS)
            {
                ret->flag = MPHELL_ERROR_RANDOM;
            }
            break;

        default :
            ret->flag = MPHELL_ERROR_RANDOM;
            text = "random_generate : invalid type"; 
            ret->info = malloc(strlen(text));
            memcpy(ret->info, text, strlen(text));
    }
}

void
random_uninstantiate(mphell_status ret, random_state working_state)
{
    char *text;
    switch(working_state->type)
    {
        case RANDOM_INTEL :
            ret->flag = MPHELL_SUCCESS;
            break;
        
        case STM32F4_RNG :
            ret->flag = MPHELL_SUCCESS;
            break;

        case RANDOM_AES128 :

        case RANDOM_AES192 :

        case RANDOM_AES256 :

        case RANDOM_SHA1 :
        
        case RANDOM_SHA256 :
            drbg_uninstantiate(ret, working_state->state_handle);
            if(ret->flag != MPHELL_SUCCESS)
            {
                ret->flag = MPHELL_ERROR_RANDOM;
            }
            break;

        default :
            ret->flag = MPHELL_ERROR_RANDOM;
            text = "random_uninstantiate : invalid type"; 
            ret->info = malloc(strlen(text));
            memcpy(ret->info, text, strlen(text));
    }
}
