/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France) 

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell-sha1.c
  * \brief Definition of sha1 functions.
  */

#include "mphell-sha1.h"

/* SHA1 Constants */

#define SHA1_K0_19 0x5a827999
#define SHA1_K20_39 0x6ed9eba1
#define SHA1_K40_59 0x8f1bbcdc
#define SHA1_K60_79 0xca62c1d6


static uint32_t sha1_init_state[5] = 
{
    0x67452301, /* H0 */
    0xefcdab89, /* H1 */
    0x98badcfe, /* H2 */
    0x10325476, /* H3 */
    0xc3d2e1f0  /* H4 */
};

/**
 * @fn    void sha1_init(sha1_state *s);
 * @brief Initialise the sha1 state
 * @param s sha1 state
 */
void sha1_init(sha1_state *s)
{
    memcpy(s->state, sha1_init_state, 20);
    s->byte_count = 0;
    s->buffer_offset = 0;
}

/* SHA1 functions */

#define Ch(x, y, z) ((z ^ (x & (y ^ z))))
#define Parity(x, y, z) (x ^ y ^ z)
#define Maj(x,y,z) ((x & y) ^ (x & z) ^ (y & z))

/**
 * @fn     uint32_t sha1_rol(uint32_t n, uint8_t bits)
 * @brief  Rotation binaire (on the left) on 32-bits word
 * @param  n   32-bits word input
 * @param  bit Size of the rotation
 * @return The word after the rotation
 */
uint32_t 
sha1_rol(uint32_t n, uint8_t bits)
{
    return ((n << bits) | (n >> (32-bits)));
}

/* SHA1 Hash block loop */

/**
 * @fn    void sha1_hash_block(sha1_state *s)
 * @brief The main sha1 function, is apply each time the sha1 state buffer is full.
 * @param s sha1 state
 */
void 
sha1_hash_block(sha1_state *s)
{
    uint8_t i;
    uint32_t a, b, c, d, e, t;

    a = s->state[0];
    b = s->state[1];
    c = s->state[2];
    d = s->state[3];
    e = s->state[4];

    for( i = 0; i < 80; i++)
    {
        if (i >= 16)
        {
            t = s->buffer.w[(i + 13) & 15] ^ s->buffer.w[(i + 8) & 15] ^
                s->buffer.w[(i + 2) & 15] ^ s->buffer.w[i & 15];
            s->buffer.w[i & 15] = sha1_rol(t,1);
        }
        if (i < 20)
        {
            t = Ch(b, c, d) + SHA1_K0_19;
        }
        else if (i < 40)
        {
            t = Parity(b, c, d)+ SHA1_K20_39;
        }
        else if (i < 60)
        {
            t = Maj(b, c, d) + SHA1_K40_59;
        }
        else
        {
            t = Parity(b, c, d) + SHA1_K60_79;
        }

        t += sha1_rol(a, 5) + e + s->buffer.w[i & 15];
        e = d;
        d = c;
        c = sha1_rol(b, 30);
        b = a;
        a = t;
    } 

    s->state[0] += a;
    s->state[1] += b;
    s->state[2] += c;
    s->state[3] += d;
    s->state[4] += e;
}

/* SHA1 write data */

/**
 * @fn    void sha1_write_byte(sha1_state *s, uint8_t data)
 * @brief Write one data byte in the sha1 state buffer, apply the sha1 hash algorithm each time the buffer is full
 * @param s    sha1 state
 * @param data byte to write into the sha1 buffer
 */
void 
sha1_write_byte(sha1_state *s, uint8_t data)
{
    s->buffer.b[s->buffer_offset ^ 3] = data;
    (s->buffer_offset)++;
    if(s->buffer_offset == 64)
    {   
        sha1_hash_block(s);
        s->buffer_offset = 0;
    }   
}

/**
 * @fn    void sha1_write(sha1_state *s, const uint8_t* data, uint64_t len)
 * @brief Process a full message into the sha1 hash algorithm
 * @param s    sha1 state
 * @param data Message to process
 * @param len  Length of data (must be < 2^64)
 */
void 
sha1_write(sha1_state *s, const uint8_t* data, uint64_t len)
{
    int64_t data_len = len;
    while(data_len > 0)
    {
        ++(s->byte_count);
        sha1_write_byte(s, *data ++);
        data_len--;
    }
}

/* SHA1 padding and result */

/**
 * @fn    void sha1_padding_message(sha1_state *s)
 * @brief Add padding to finish to fill the sha1 buffer and apply the sha1 hash algorithm one last time
 * @param s sha1 state
 */
void 
sha1_padding_message(sha1_state *s)
{
    /* Pad with one byte beginning by 1 : 0x80              */
    sha1_write_byte(s, 0x80);
    
    /* Pad with 0x00 until buffer_offset*8 = 448            */
    while(s->buffer_offset != 56)
    {
        sha1_write_byte(s, 0x00);
    }
    
    /* Append message length (in bits) in the last 8 bytes. */
    sha1_write_byte(s, s->byte_count >> 53); 
    sha1_write_byte(s, s->byte_count >> 45); 
    sha1_write_byte(s, s->byte_count >> 37); 
    sha1_write_byte(s, s->byte_count >> 29); 
    sha1_write_byte(s, s->byte_count >> 21); 
    sha1_write_byte(s, s->byte_count >> 13); 
    sha1_write_byte(s, s->byte_count >> 5);
    sha1_write_byte(s, s->byte_count << 3);    
}

/**
 * @fn     uint8_t * sha1_result(sha1_state *s)
 * @brief  Compute the sha1 states (5 x 32-bits word = 160 bits hash), one time the full message has been processed by sha1_write.
 * @param  s sha1 state
 * @return The sha1 states (5 x 32-bits word = 160 bits hash)
 */
uint8_t * 
sha1_result(sha1_state *s)
{
    sha1_padding_message(s);
    
    /* Swap byte order back */
    uint8_t i;

    for (i = 0; i < 5; i++)
    {
        uint32_t a, b;
        a = s->state[i];
        b = a << 24;
        b |= (a << 8) & 0x00ff0000;
        b |= (a >> 8) & 0x0000ff00;
        b |= a >> 24;
        s->state[i]=b;
    }

    return ((uint8_t *) s->state);
}

void 
sha1(uint8_t *hashvalue, const uint8_t* data, const uint64_t data_len)
{
    sha1_state s;
    sha1_init(&s);
    sha1_write(&s, data, data_len/8);
    memcpy(hashvalue, sha1_result(&s), 20);
}
