/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France) 

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
 * \file  mphell-sha256.c
 * \brief Definition of sha256 functions.
 */

#include "mphell-sha256.h"

static uint32_t sha1_init_state[64] = {                                                           
    0x428a2f98,0x71374491,0xb5c0fbcf,0xe9b5dba5,0x3956c25b,0x59f111f1,0x923f82a4,0xab1c5ed5,
    0xd807aa98,0x12835b01,0x243185be,0x550c7dc3,0x72be5d74,0x80deb1fe,0x9bdc06a7,0xc19bf174,
    0xe49b69c1,0xefbe4786,0x0fc19dc6,0x240ca1cc,0x2de92c6f,0x4a7484aa,0x5cb0a9dc,0x76f988da,
    0x983e5152,0xa831c66d,0xb00327c8,0xbf597fc7,0xc6e00bf3,0xd5a79147,0x06ca6351,0x14292967,
    0x27b70a85,0x2e1b2138,0x4d2c6dfc,0x53380d13,0x650a7354,0x766a0abb,0x81c2c92e,0x92722c85,
    0xa2bfe8a1,0xa81a664b,0xc24b8b70,0xc76c51a3,0xd192e819,0xd6990624,0xf40e3585,0x106aa070,
    0x19a4c116,0x1e376c08,0x2748774c,0x34b0bcb5,0x391c0cb3,0x4ed8aa4a,0x5b9cca4f,0x682e6ff3,
    0x748f82ee,0x78a5636f,0x84c87814,0x8cc70208,0x90befffa,0xa4506ceb,0xbef9a3f7,0xc67178f2
};

/**
 * @fn    void sha256_transform(sha256_state *s, unsigned char data[])
 * @brief Called each time 64 bytes are written, modify the state structure.
 * @param s    Sha256 state structure to modify
 * @param data User data
 */
void 
sha256_transform(sha256_state *s, unsigned char data[])
{  
   uint32_t a,b,c,d,e,f,g,h,i,j,t1,t2,m[64];
      
   for (i=0,j=0; i < 16; ++i, j += 4)
      m[i] = (data[j] << 24) | (data[j+1] << 16) | (data[j+2] << 8) | (data[j+3]);
   for ( ; i < 64; ++i)
      m[i] = SIG1(m[i-2]) + m[i-7] + SIG0(m[i-15]) + m[i-16];

   a = s->state[0];
   b = s->state[1];
   c = s->state[2];
   d = s->state[3];
   e = s->state[4];
   f = s->state[5];
   g = s->state[6];
   h = s->state[7];
   
   for (i = 0; i < 64; ++i) {
      t1 = h + EP1(e) + CH(e,f,g) + sha1_init_state[i] + m[i];
      t2 = EP0(a) + MAJ(a,b,c);
      h = g;
      g = f;
      f = e;
      e = d + t1;
      d = c;
      c = b;
      b = a;
      a = t1 + t2;
   }
   
   s->state[0] += a;
   s->state[1] += b;
   s->state[2] += c;
   s->state[3] += d;
   s->state[4] += e;
   s->state[5] += f;
   s->state[6] += g;
   s->state[7] += h;
}  

/**
 * @fn    void sha256_init(sha256_state *s);
 * @brief Initialise a sha256_state structure
 * @param s Sha256 state structure to initialise
 */
void 
sha256_init(sha256_state *s)
{  
   s->datalen = 0; 
   s->bitlen[0] = 0; 
   s->bitlen[1] = 0; 
   s->state[0] = 0x6a09e667;
   s->state[1] = 0xbb67ae85;
   s->state[2] = 0x3c6ef372;
   s->state[3] = 0xa54ff53a;
   s->state[4] = 0x510e527f;
   s->state[5] = 0x9b05688c;
   s->state[6] = 0x1f83d9ab;
   s->state[7] = 0x5be0cd19;
}

/**
 * @fn    void sha256_write(sha256_state *s, const uint8_t* data, uint64_t len);
 * @brief Write user data (message) into the sha256 state structure
 * @param s    Sha256 state structure to modify
 * @param data user message
 * @param len  data length (in bytes)
 */
void sha256_write(sha256_state *s, const uint8_t* data, uint64_t len)
{  
   uint32_t i;
   
   for (i=0; i < len; ++i) { 
      s->data[s->datalen] = data[i]; 
      s->datalen++; 
      if (s->datalen == 64) { 
         sha256_transform(s,s->data);
         DBL_INT_ADD(s->bitlen[0],s->bitlen[1],512); 
         s->datalen = 0; 
      }  
   }  
}  

/**
 * @fn    void sha256_result(sha256_state *s, uint8_t* hash);
 * @brief Add padding, and perform a last computation on the sha256 states
 * @param s    Sha256 state structure
 * @param hash resulting hash
 */
void 
sha256_result(sha256_state *s, uint8_t* hash)
{  
   uint32_t i; 
   
   i = s->datalen; 
   
   /* Pad whatever data is left in the buffer. */ 
   if (s->datalen < 56) { 
      s->data[i++] = 0x80; 
      while (i < 56) 
         s->data[i++] = 0x00; 
   }  
   else { 
      s->data[i++] = 0x80; 
      while (i < 64) 
         s->data[i++] = 0x00; 
      sha256_transform(s,s->data);
      memset(s->data,0,56); 
   }  
   
   /* Append to the padding the total message's length in bits and transform. */
   DBL_INT_ADD(s->bitlen[0],s->bitlen[1],s->datalen * 8);
   s->data[63] = s->bitlen[0]; 
   s->data[62] = s->bitlen[0] >> 8; 
   s->data[61] = s->bitlen[0] >> 16; 
   s->data[60] = s->bitlen[0] >> 24; 
   s->data[59] = s->bitlen[1]; 
   s->data[58] = s->bitlen[1] >> 8; 
   s->data[57] = s->bitlen[1] >> 16;  
   s->data[56] = s->bitlen[1] >> 24; 
   sha256_transform(s,s->data);
   
   /* Since this implementation uses little endian byte ordering and SHA uses big endian,
      reverse all the bytes when copying the final state to the output hash. */
   for (i=0; i < 4; ++i) { 
      hash[i]    = (s->state[0] >> (24-i*8)) & 0x000000ff; 
      hash[i+4]  = (s->state[1] >> (24-i*8)) & 0x000000ff; 
      hash[i+8]  = (s->state[2] >> (24-i*8)) & 0x000000ff;
      hash[i+12] = (s->state[3] >> (24-i*8)) & 0x000000ff;
      hash[i+16] = (s->state[4] >> (24-i*8)) & 0x000000ff;
      hash[i+20] = (s->state[5] >> (24-i*8)) & 0x000000ff;
      hash[i+24] = (s->state[6] >> (24-i*8)) & 0x000000ff;
      hash[i+28] = (s->state[7] >> (24-i*8)) & 0x000000ff;
   }  
} 

void
sha256(uint8_t *hashvalue, const uint8_t* data, const uint64_t data_len)
{
    unsigned char hash[32];  

    sha256_state s;
    sha256_init(&s);
    sha256_write(&s, data, data_len/8);
    sha256_result(&s,hash);
    memcpy(hashvalue, hash, 32);
}
