/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France) 

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/
/**
  * \file  mphell-tmp.c
  * \brief Definition of internal MPHELL big_uint memory functions.
  */

#include "mphell-tmp.h"

#if MPHELL_USE_GMP == 1

mp_limb_t *Arith_stack1;
uint32_t ptr1;

#if MPHELL_USE_MULTITHREADING == 1
mp_limb_t *Arith_stack2;
uint32_t ptr2;
#endif

int size_stack1 = (int)INIT_SIZE_MAIN_STACK;
#if MPHELL_USE_MULTITHREADING == 1
int size_stack2 = (int)INIT_SIZE_OTHERS_STACKS;
#endif

void
alloc_tmp_memory()
{
    /* stack 1 */
    Arith_stack1 = (mp_limb_t *)malloc(8*size_stack1);
    MPHELL_ASSERT_ALWAYS(Arith_stack1 != NULL, "Allocation of temporary memory failed \n");
    ptr1 = 0;
    
#if MPHELL_USE_MULTITHREADING == 1
    /* stack 2 */
    Arith_stack2 = (mp_limb_t *)malloc(8*size_stack2);
    MPHELL_ASSERT_ALWAYS(Arith_stack2 != NULL, "Allocation of temporary memory failed \n");
    ptr2 = 0;
#endif
}

void
free_tmp_memory()
{
    /* stack 1 */
    free(Arith_stack1);
    ptr1 = 0;
    
#if MPHELL_USE_MULTITHREADING == 1
    /* stack 2 */
    free(Arith_stack2);
    ptr2 = 0;
#endif
}

void
gmp_tmp_alloc(__mpz_struct * t, const uint8_t size, uint8_t stack)
{
#if MPHELL_USE_MULTITHREADING == 1
    if(stack == STACK_1)
    {
        /* stack 1 */
        t->_mp_d = (mp_limb_t*)(Arith_stack1 + ptr1);
        ptr1 += size;
        MPHELL_ASSERT(ptr1 < size_stack1, "Allocated size not sufficient in stack 1\n");
    }
    else
    {
        /* stack 2 */
        t->_mp_d = (mp_limb_t*)(Arith_stack2 + ptr2);
        ptr2 += size;
        MPHELL_ASSERT(ptr2 < size_stack2, "Allocated size not sufficient in stack 2\n");
    }
#elif MPHELL_USE_MULTITHREADING == 0
    t->_mp_d = (mp_limb_t*)(Arith_stack1 + ptr1);
    ptr1 += size;
    MPHELL_ASSERT(ptr1 < size_stack1, "Allocated size not sufficient in stack 1\n");
#endif
    t->_mp_alloc = size;
    t->_mp_size = 0;
}

void
gmp_tmp_free(__mpz_struct * t, const uint8_t size, uint8_t stack)
{
#if MPHELL_USE_MULTITHREADING == 1
    if(stack == 1)
    {
        /* stack 1 */
        ptr1 -= size;
        MPHELL_ASSERT(ptr1 >= 0, "ptr1 is < 0 in stack 1\n");
    }
    else
    {
        /* stack 2 */
        ptr2 -= size;
        MPHELL_ASSERT(ptr2 >= 0, "ptr2 is < 0 in stack 2\n");
    }
#elif MPHELL_USE_MULTITHREADING == 0
    ptr1 -= size;
    MPHELL_ASSERT(ptr1 >= 0, "ptr1 is < 0 in stack 1\n");
#endif
}

#elif MPHELL_USE_IPP == 1

Ipp8u *Arith_stack1;
uint32_t ptr1;

#if MPHELL_USE_MULTITHREADING == 1
Ipp8u *Arith_stack2;
uint32_t ptr2;
#endif

int size_stack1 = (int)INIT_SIZE_MAIN_STACK;
#if MPHELL_USE_MULTITHREADING == 1
int size_stack2 = (int)INIT_SIZE_OTHERS_STACKS;
#endif

void
alloc_tmp_memory()
{
    /* stack 1 */
    Arith_stack1 = malloc(8*size_stack1);
    MPHELL_ASSERT_ALWAYS(Arith_stack1 != NULL, "Allocation of temporary memory failed \n");
    ptr1 = 0;
    
#if MPHELL_USE_MULTITHREADING == 1
    /* stack 2 */
    Arith_stack2 = malloc(8*size_stack2);
    MPHELL_ASSERT_ALWAYS(Arith_stack2 != NULL, "Allocation of temporary memory failed \n");
    ptr2 = 0;
#endif
}

void
free_tmp_memory()
{
    /* stack 1 */
    free(Arith_stack1);
    ptr1 = 0;
    
#if MPHELL_USE_MULTITHREADING == 1
    /* stack 2 */
    free(Arith_stack2);
    ptr2 = 0;
#endif
}

void
ippcp_bn_tmp_alloc(IppsBigNumState ** t, const uint8_t size, uint8_t stack)
{
#if MPHELL_USE_MULTITHREADING == 1
    if(stack == STACK_1)
    {
        /* stack 1 */
        int ctxSize; 
        ippsBigNumGetSize(size, &ctxSize); 
        *t = (IppsBigNumState*)(Arith_stack1 + ptr1);
        ptr1 += ctxSize;
        ippsBigNumInit(size, *t); /* Works with 32 bits blocks */
        MPHELL_ASSERT(ptr1 < size_stack1*8, "Allocated size not sufficient in stack 1\n");
    }
    else
    {
        /* stack 2 */
        int ctxSize; 
        ippsBigNumGetSize(size, &ctxSize); 
        *t = (IppsBigNumState*)(Arith_stack2 + ptr2);
        ptr2 += ctxSize;
        ippsBigNumInit(size, *t); /* Works with 32 bits blocks */
        MPHELL_ASSERT(ptr2 < size_stack2*8, "Allocated size not sufficient in stack 2\n");
    }
#elif MPHELL_USE_MULTITHREADING == 0
    int ctxSize; 
    ippsBigNumGetSize(size, &ctxSize); 
    *t = (IppsBigNumState*)(Arith_stack1 + ptr1);
    ptr1 += ctxSize;
    ippsBigNumInit(size, *t); /* Works with 32 bits blocks */
    MPHELL_ASSERT(ptr1 < size_stack1*8, "Allocated size not sufficient in stack 1\n");
#endif
}

void
ippcp_bn_tmp_free(IppsBigNumState ** t, const uint8_t size, uint8_t stack)
{
    int ctxSize; 
    ippsBigNumGetSize(size, &ctxSize);
#if MPHELL_USE_MULTITHREADING == 1
    if(stack == STACK_1)
    {
        /* stack 1 */
        ptr1 -= ctxSize;
        MPHELL_ASSERT(ptr1 >= 0, "ptr1 is < 0 in stack 1\n");
    }
    else
    {
        /* stack 2 */
        ptr2 -= ctxSize;
        MPHELL_ASSERT(ptr2 >= 0, "ptr2 is < 0 in stack 1\n");
    }
#elif MPHELL_USE_MULTITHREADING == 0
    ptr1 -= ctxSize;
    MPHELL_ASSERT(ptr1 >= 0, "ptr1 is < 0 in stack 1\n");
#endif
}

#elif MPHELL_USE_MBEDTLS == 1

mbedtls_mpi_uint *Arith_stack1;
uint32_t ptr1;

#if MPHELL_USE_MULTITHREADING == 1
mbedtls_mpi_uint *Arith_stack2;
uint32_t ptr2;
#endif

int size_stack1 = (int)INIT_SIZE_MAIN_STACK;
#if MPHELL_USE_MULTITHREADING == 1
int size_stack2 = (int)INIT_SIZE_OTHERS_STACKS;
#endif

void
alloc_tmp_memory()
{
    /* stack 1 */
    Arith_stack1 = (mbedtls_mpi_uint *)malloc(8*size_stack1);
    MPHELL_ASSERT_ALWAYS(Arith_stack1 != NULL, "Allocation of temporary memory failed \n");
    ptr1 = 0;
    
#if MPHELL_USE_MULTITHREADING == 1
    /* stack 2 */
    Arith_stack2 = (mbedtls_mpi_uint *)malloc(8*size_stack2);
    MPHELL_ASSERT_ALWAYS(Arith_stack2 != NULL, "Allocation of temporary memory failed \n");
    ptr2 = 0;
#endif
}

void
free_tmp_memory()
{
    /* stack 1 */
    free(Arith_stack1);
    ptr1 = 0;
    
#if MPHELL_USE_MULTITHREADING == 1
    /* stack 2 */
    free(Arith_stack2);
    ptr2 = 0;
#endif
}

void
mbedtls_tmp_alloc(mbedtls_mpi * t, const uint8_t size, uint8_t stack)
{
#if MPHELL_USE_MULTITHREADING == 1
    if(stack == STACK_1)
    {
        /* stack 1 */
        t->p = (mbedtls_mpi_uint *)(Arith_stack1 + ptr1);
        ptr1 += size;
        MPHELL_ASSERT(ptr1 < size_stack1, "Allocated size not sufficient in stack 1\n");
    }
    else
    {
        /* stack 2 */
        t->p = (mbedtls_mpi_uint *)(Arith_stack2 + ptr2);
        ptr2 += size;
        MPHELL_ASSERT(ptr2 < size_stack2, "Allocated size not sufficient in stack 2\n");
    }
#elif MPHELL_USE_MULTITHREADING == 0
    t->p = (mbedtls_mpi_uint *)(Arith_stack1 + ptr1);
    ptr1 += size;
    MPHELL_ASSERT(ptr1 < size_stack1, "Allocated size not sufficient in stack 1\n");
#endif
    t->n = size;
    t->s = 1;
}

void
mbedtls_tmp_free(mbedtls_mpi * t, const uint8_t size, uint8_t stack)
{
#if MPHELL_USE_MULTITHREADING == 1
    if(stack == 1)
    {
        /* stack 1 */
        ptr1 -= size;
        MPHELL_ASSERT(ptr1 >= 0, "ptr1 is < 0 in stack 1\n");
    }
    else
    {
        /* stack 2 */
        ptr2 -= size;
        MPHELL_ASSERT(ptr2 >= 0, "ptr2 is < 0 in stack 2\n");
    }
#elif MPHELL_USE_MULTITHREADING == 0
    ptr1 -= size;
    MPHELL_ASSERT(ptr1 >= 0, "ptr1 is < 0 in stack 1\n");
#endif
}

#endif

uint32_t get_stack_size(uint8_t stack)
{
#if MPHELL_USE_MULTITHREADING == 1
    if(stack == STACK_1)
    {
        /* stack 1 */
        return ptr1;
    }
    else
    {
        /* stack 2 */
        return ptr2;
    }
#elif MPHELL_USE_MULTITHREADING == 0
    return ptr1;
#endif
}

bool verify_stack_size(uint32_t prev_size, uint8_t stack)
{
#if MPHELL_USE_MULTITHREADING == 1
    if(stack == STACK_1)
    {
        /* stack 1 */
        return (ptr1==prev_size);
    }
    else
    {
        /* stack 2 */
        return (ptr2==prev_size);
    }
#elif MPHELL_USE_MULTITHREADING == 0
    return (ptr1==prev_size);
#endif
}

