/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France) 

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell-weierstrass.c
  * \brief Definition of Weierstrass ECC function.
  */

/* 
    Formulae:
    Refererences for different formulae are:
    -[CMO98]  Efficient elliptic curve exponentiation using mixed coordinates 
    by Cohen Henri,  Miyaji Atsuko, and Ono Takatoshi. International Conference
    on the Theory and Application of Cryptology and Information Security. 
    Springer, Berlin, Heidelberg, 1998.

    -[CC74] Sequences of numbers generated by addition in formal groups and new
    primality and factorization tests. by Chudnovsky, David V., and Gregory 
    V. Chudnovsky.  Advances in Applied Mathematics 7.4 (1986): 385-434.

    -[Ber01] A software implementation of NIST P-224 by Daniel J. Bernstein 
    slides available here: http://cr.yp.to/talks.html#2001.10.29

    -[BL] Explicit-formulas database Bernstein, Daniel J., Lange, Tania
    http://www.hyperelliptic.org/EFD/oldefd/jacobian.html 
    mentionned in [BL07]

    -[BL07] Faster addition and doubling on elliptic curves.by Bernstein, Daniel J. 
    and Lange, Tanja.  In : International Conference on the Theory and Application 
    of Cryptology and Information Security. Springer, Berlin, Heidelberg, 2007. 
    p. 29-50.

    
    CoZ formulae:
    References for CoZ formulae are:
    -[Mel07]  New point addition formulae for ECC applications by Meloni 
    Nicolas  published in International Workshop on the Arithmetic of Finite 
    Fields in 2007
    -[GJM10] CoZ addition formulae and binary ladder on elliptic curves by 
    Goundar, Joye and Miyaji published in International Workshop on 
    Cryptographic Hardware and Embedded Systems in 2010  
    
*/

#include <stdlib.h>
#include <string.h>

#include "mphell-curve.h"

/************************************SETTERS**********************************/

void
weierstrass_compute_disc(ec_curve E, uint8_t stack)
{
    field_t *k = E->k;
    /* disc = -16(4a3 + 27b2)   */
    /* D = -(4a3 + 27b2)        */
    field_elt tmp, tmp1;
    field_elt_get_pool_elt(&tmp, k, stack);
    field_elt_get_pool_elt(&tmp1, k, stack);
    
    field_elt_pow_ui(tmp, E->a, 3, k, stack);
    field_elt_set_ui(tmp1, 4, false, k, stack);
    field_elt_mul(tmp, tmp, tmp1, k, stack);

    field_elt_sqr(E->D, E->b, k, stack);
    field_elt_set_ui(tmp1, 27, false, k, stack);
    field_elt_mul(E->D, E->D, tmp1, k, stack);
    
    field_elt_add(E->D, E->D, tmp, k);
    field_elt_neg(E->D, E->D, k);

    field_elt_set_ui(tmp1, 16, false, k, stack);
    field_elt_mul(E->disc, E->D, tmp1, k, stack);

    field_elt_relax_pool_elt(&tmp, k, stack);
    field_elt_relax_pool_elt(&tmp1, k, stack);
}

#include "mphell-drbg_internal.h"
#include "mphell-sha1.h"

bool weierstrass_verify_random_generation(ec_curve E, const char * seed, uint8_t stack)
{
    /* Source : Working Draft AMERICAN NATIONAL STANDARD X9.62-1998, section A.3.4.2 */
    field_t *k = E->k;
    
    field_elt b2, a3, r1;
    field_elt_get_pool_elt(&b2, k, stack);
    field_elt_get_pool_elt(&a3, k, stack);
    field_elt_get_pool_elt(&r1, k, stack);
    
    number tmp;
    number_tmp_alloc(&tmp, k->size, stack);
    
    /* t = log2(p) */
    uint16_t t;
    switch(k->type)
    {
        case FP:
             t = number_log2(((fp_param)k->param)->p);
            break;
        case FP2:
        case FP3:
            t = number_log2(((fp_param)((fp2_param)k->param)->base)->p);
            break;
        default: 
            return false;
    }
    /* s = (t-1) / 160 */
    uint16_t s = (t-1)/160;
    /* h = t - 160*s */
    uint16_t h = t - (160*s);
    h --; /* set the leftmost bit of c0 to 0 */
    uint16_t hbytes = (h / 8) + (h%8 != 0);
    
    /* Alloc W' */
    
    uint8_t * Wprime = calloc(s*20+hbytes, sizeof(uint8_t));
    
    /* Step 1 and 2, compute W0: */
    
    uint8_t H[20];
    uint8_t seed_bin[20];
    /* Convert hexadecimal seed into binary seed */
    hex2bin(seed, seed_bin);
    sha1(H, seed_bin, 160);
    /* c0 = h rightmost bits of H */
    memcpy(Wprime, H+(20-hbytes), hbytes);
    uint8_t mask;
    switch (h%8)
    {
        case 0:
            mask = 0xff;
            break;
        case 1:
            mask = 0x1;
            break;
        case 2:
            mask = 0x3;
            break;
        case 3:
            mask = 0x7;
            break;
        case 4:
            mask = 0xf;
            break;
        case 5:
            mask = 0x1f;
            break;
        case 6:
            mask = 0x3f;
            break;
        case 7:
            mask = 0x7f;
            break;
    }
    
    Wprime[0]&=mask;
    
    /* Step 2 : done with h--*/
    
    /* Step 3 and 4 : */
    
    uint16_t i;
    for (i=0; i<s; i++)
    {
        drbg_incr_data(seed_bin, 20);
        sha1(H, seed_bin, 160);
        memcpy(Wprime+i*20+hbytes, H, 20);
    }
    
    /* Step 5 : Transform Wprime into a field element */
    
    /* Convert binary hash into hexadecimal string */
    unsigned char Wprime_hex[2*(s*20+hbytes)+1];
    bin2hex(Wprime, s*20+hbytes, Wprime_hex);
    
    switch(k->type)
    {
        case FP:
            field_elt_set_str(r1, (const char *)Wprime_hex, 16, false, k, stack);
            break;
        case FP2:
        case FP3:
            number_set_str(tmp, (const char *)Wprime_hex, 16);
            field_elt_set_number(r1, false, k, stack, 1, tmp);
            break;
        default:
            field_elt_set_one(r1, k);
    }
    
    /* Step 6 : Accept or reject */
    
    field_elt_sqr(b2, E->b, k, stack);
    field_elt_pow_ui(a3, E->a, 3, k, stack);
    field_elt_mul(r1, r1, b2, k, stack);
    int8_t res = field_elt_cmp(r1, a3, k);
    
    number_tmp_free(&tmp, k->size, stack);
    
    field_elt_relax_pool_elt(&a3, k, stack);
    field_elt_relax_pool_elt(&b2, k, stack);
    field_elt_relax_pool_elt(&r1, k, stack);
    free(Wprime);
    
    return (res == 0);
}

void weierstrass_curve_random_generation(fe_ptr a, fe_ptr b, char * seed_res, field_srcptr k, uint8_t stack)
{
    /* Source : Working Draft AMERICAN NATIONAL STANDARD X9.62-1998, section A.3.3.2 */
    bool test;
    number seed;
    number bound_h;
    number bound_l;
    number tmp;
    number_tmp_alloc(&tmp, k->size, stack);
    number_tmp_alloc(&seed, bits_to_nblock(160), stack);
    number_tmp_alloc(&bound_h, bits_to_nblock(160), stack);
    number_tmp_alloc(&bound_l, bits_to_nblock(160), stack);
    number_set_str(bound_h, "ffffffffffffffffffffffffffffffffffffffff", 16); /* 2^161 - 1 */
    number_set_str(bound_l, "1000000000000000000000000000000000000000", 16); /* 2^160     */
    
    field_elt b2, a3, r1, tmp1, tmp2;
    field_elt_get_pool_elt(&b2, (field_ptr)k, stack);
    field_elt_get_pool_elt(&a3, (field_ptr)k, stack);
    field_elt_get_pool_elt(&r1, (field_ptr)k, stack);
    field_elt_get_pool_elt(&tmp1, (field_ptr)k, stack);
    field_elt_get_pool_elt(&tmp2, (field_ptr)k, stack);
    
    char * seed_hex;
    
    /* t = log2(p) */
    uint16_t t;
    switch(k->type)
    {
        case FP:
             t = number_log2(((fp_param)k->param)->p);
            break;
        case FP2:
        case FP3:
            t = number_log2(((fp_param)((fp2_param)k->param)->base)->p);
            break;
        default: 
            t = 0;
    }
    /* s = (t-1) / 160 */
    uint16_t s = (t-1)/160;
    /* h = t - 160*s */
    uint16_t h = t - (160*s);
    h --; /* set the leftmost bit of c0 to 0 */
    uint16_t hbytes = (h / 8) + (h%8 != 0);
    
    /* Alloc W' */
    uint8_t * Wprime = calloc(s*20+hbytes, sizeof(uint8_t));

    do
    {
        /* Step 1 : Choose an arbitrary bit string SEED of bit length at least 160 bits */
        do
        {
            number_random1(seed, bound_h, stack);
        }
        while(number_cmp(seed, bound_l)<0);
        
        number_str(&seed_hex, seed, 16);
        
        /* Step 2 and 3, compute W0 :  */
        
        uint8_t H[20];
        uint8_t seed_bin[20];
        /* Convert hexadecimal seed into binary seed */
        hex2bin(seed_hex, seed_bin);
        sha1(H, seed_bin, 160);
        /* c0 = h rightmost bits of H */
        memcpy(Wprime, H+(20-hbytes), hbytes);
        uint8_t mask;
        switch (h%8)
        {
            case 0:
                mask = 0xff;
                break;
            case 1:
                mask = 0x1;
                break;
            case 2:
                mask = 0x3;
                break;
            case 3:
                mask = 0x7;
                break;
            case 4:
                mask = 0xf;
                break;
            case 5:
                mask = 0x1f;
                break;
            case 6:
                mask = 0x3f;
                break;
            case 7:
                mask = 0x7f;
                break;
        }
        
        Wprime[0]&=mask;
        
        /* Step 4 and 5 : */
        
        uint16_t i;
        for (i=0; i<s; i++)
        {
            drbg_incr_data(seed_bin, 20);
            sha1(H, seed_bin, 160);
            memcpy(Wprime+i*20+hbytes, H, 20);
        }
        
        /* Convert binary hash into hexadecimal string */
        unsigned char Wprime_hex[2*(s*20+hbytes)+1];
        bin2hex(Wprime, s*20+hbytes, Wprime_hex);
        
        /* Step 6 : Transform Wprime into a field element */
        
        switch(k->type)
        {
            case FP:
                field_elt_set_str(r1, (const char *)Wprime_hex, 16, false, k, stack);
                break;
            case FP2:
            case FP3:
                number_set_str(tmp, (const char *)Wprime_hex, 16);
                field_elt_set_number(r1, false, k, stack, 1, tmp);
                break;
            default:
                field_elt_set_one(r1, k);
        }
        
        
        /* Step 7: Choose a and b */
        
        do
        {
            field_elt_random(a, k, stack);
            field_elt_pow_ui(a3, a, 3, k, stack);
            field_elt_div(b2, a3, r1, k, stack);
        }
        while(!field_elt_issquare(b2, k, stack));
        
        field_elt_sqrt(b, b2, k, stack);
        field_elt_set_ui(tmp1, 4, false, k, stack);
        field_elt_set_ui(tmp2, 27, false,k, stack);
        field_elt_mul(tmp1, tmp1, a3, k, stack);          /* 4a^3         */
        field_elt_mul(tmp2, tmp2, b2, k, stack);          /* 27b^2        */
        field_elt_add(tmp1, tmp1, tmp2, k);               /* 4a^3 + 27b^2 */
        test = field_elt_iszero(tmp1, k);
        if (test)
        {
            free(seed_hex);
        }
    }
    /* Step 8: verifie that the curve is elliptic */
    while(test);
    
    strcpy(seed_res, seed_hex);
    number_tmp_free(&tmp, k->size, stack);
    number_tmp_free(&seed, bits_to_nblock(160), stack);
    number_tmp_free(&bound_h, bits_to_nblock(160), stack);
    number_tmp_free(&bound_l, bits_to_nblock(160), stack);
    field_elt_relax_pool_elt(&a3, (field_ptr)k, stack);
    field_elt_relax_pool_elt(&b2, (field_ptr)k, stack);
    field_elt_relax_pool_elt(&r1, (field_ptr)k, stack);
    field_elt_relax_pool_elt(&tmp1, (field_ptr)k, stack);
    field_elt_relax_pool_elt(&tmp2, (field_ptr)k, stack);
    free(seed_hex);
    free(Wprime);
}

void 
weierstrass_point_set_neutral (ec_point_ptr dst, ec_curve_srcptr E, uint8_t stack)
{
    field_t *k = E->k;
    
    switch(E->f)
    { 
        case PROJECTIVE :
            field_elt_set_ui(dst->x, 0, true, k, stack);
            field_elt_set_one(dst->y, k);
            field_elt_set_ui(dst->z, 0, true, k, stack);
            field_elt_set_one(dst->t, k);
            break;

        case JACOBIAN :
            field_elt_set_one(dst->x, k);
            field_elt_set_one(dst->y, k);
            field_elt_set_ui(dst->z, 0, true, k, stack);
            field_elt_set_one(dst->t, k);
            break;
        
        default :
            break;
    }
}

void 
weierstrass_point_set_aff (ec_point_ptr P, fe_srcptr x, fe_srcptr y, field_srcptr k)
{
    field_elt_copy(P->x, x, k);
    field_elt_copy(P->y, y, k);
    field_elt_set_one(P->z, k);
    field_elt_set_one(P->t, k);
}

void 
weierstrass_point_set_aff_str (ec_point_ptr P, const char *str_x, const char *str_y,
        const bool is_reduced, const uint8_t base, field_srcptr k, uint8_t stack)
{
    field_elt_set_str(P->x, str_x, base, is_reduced, k, stack);
    field_elt_set_str(P->y, str_y, base, is_reduced, k, stack);
    field_elt_set_one(P->z, k);
    field_elt_set_one(P->t, k);
}

bool
weierstrass_belongs (ec_point_srcptr P, ec_curve_srcptr E, uint8_t stack)
{
    /* P1 = (X1, Y1, Z1)     */
    /* E: y^2 = x^3 + ax + b */
    
    field_t *k = E->k;
    
    field_elt t1, t2, t3;
    field_elt_get_pool_elt(&t1, k, stack);
    field_elt_get_pool_elt(&t2, k, stack);
    field_elt_get_pool_elt(&t3, k, stack);
    
    switch(E->f)
    { 
        case PROJECTIVE :
    
            field_elt_sqr(t1, P->y, k, stack);         /* t1 = Y1^2                            */
            field_elt_mul(t1, t1, P->z, k, stack);     /* t1 = (Y1^2).Z1                       */

            field_elt_sqr(t2, P->z, k, stack);         /* t2 = Z1^2                            */
            field_elt_mul(t2, t2, E->a, k, stack);     /* t2 = (Z1^2).a                        */
            field_elt_mul(t2, t2, P->x, k, stack);     /* t2 = (Z1^2).a.X1                     */

            field_elt_sqr(t3, P->z, k, stack);         /* t3 = Z1^2                            */
            field_elt_mul(t3, t3, P->z, k, stack);     /* t3 = Z1^3                            */
            field_elt_mul(t3, t3, E->b, k, stack);     /* t3 = (Z1^3).b                        */

            field_elt_add(t2, t2, t3, k);              /* t2 = (Z1^2).a.X1 + (Z1^3).b          */
            
            field_elt_sqr(t3, P->x, k, stack);         /* t3 = X1^2                            */
            field_elt_mul(t3, t3, P->x, k, stack);     /* t3 = X1^3                            */
            field_elt_add(t2, t2, t3, k);              /* t2 = X1^3 + (Z1^2).a.X1 + (Z1^3).b   */
            break;
        
        case JACOBIAN :
        
            field_elt_sqr(t1, P->y, k, stack);         /* t1 = Y1^2                            */

            field_elt_sqr(t2, P->z, k, stack);         /* t2 = Z1^2                            */
            field_elt_sqr(t2, t2, k, stack);           /* t2 = Z1^4                            */
            field_elt_mul(t2, t2, E->a, k, stack);     /* t2 = (Z1^4).a                        */
            field_elt_mul(t2, t2, P->x, k, stack);     /* t2 = (Z1^4).a.X1                     */

            field_elt_sqr(t3, P->z, k, stack);         /* t3 = Z1^2                            */
            field_elt_mul(t3, t3, P->z, k, stack);     /* t3 = Z1^3                            */
            field_elt_sqr(t3, t3, k, stack);           /* t3 = Z1^6                            */
            field_elt_mul(t3, t3, E->b, k, stack);     /* t3 = (Z1^6).b                        */

            field_elt_add(t2, t2, t3, k);              /* t2 = (Z1^4).a.X1 + (Z1^6).b          */
            
            field_elt_sqr(t3, P->x, k, stack);         /* t3 = X1^2                            */
            field_elt_mul(t3, t3, P->x, k, stack);     /* t3 = X1^3                            */
            field_elt_add(t2, t2, t3, k);              /* t2 = X1^3 + (Z1^4).a.X1 + (Z1^6).b   */
            break;
        
        default :
            break;
    }

    bool res = (field_elt_cmp(t1, t2, k) == 0);

    field_elt_relax_pool_elt(&t1, k, stack);
    field_elt_relax_pool_elt(&t2, k, stack);
    field_elt_relax_pool_elt(&t3, k, stack);

    return res;
}

void 
weierstrass_point_random (ec_point_ptr P, ec_curve_srcptr E, uint8_t stack)
{
    field_t *k = E->k;

    field_elt t1, t2;
    field_elt_get_pool_elt(&t1, k, stack);
    field_elt_get_pool_elt(&t2, k, stack);

    do{
        field_elt_random(P->x, k, stack);
        field_elt_mul(t1, E->a, P->x, k, stack);       /* t1 = a.x             */
        field_elt_sqr(t2, P->x, k, stack);             /* t2 = x^2             */
        field_elt_mul(t2, t2, P->x, k, stack);         /* t2 = x^3             */
        field_elt_add(t1, t1, t2, k);                  /* t1 = x^3 + a.x       */
        field_elt_add(t1, t1, E->b, k);                /* t1 = x^3 + a.x + b   */
    }
    while(!field_elt_issquare(t1, k, stack));
    
    field_elt_sqrt(t2, t1, k, stack);
    weierstrass_point_set_aff(P, P->x, t2, k);

    field_elt_relax_pool_elt(&t1, k, stack);
    field_elt_relax_pool_elt(&t2, k, stack);
}

void
weierstrass_point_norm (ec_point_ptr P, ec_curve_srcptr E, uint8_t stack)
{
    field_t *k = E->k;
    field_elt u;
    field_elt_get_pool_elt(&u, k, stack);
    
    if(!weierstrass_point_is_neutral(P, E, stack))
    {
        switch(E->f)
        {
            case PROJECTIVE :
                field_elt_inv(u, P->z, k, stack);              /* u = 1/Z      */
                field_elt_mul(P->x, P->x, u, k, stack);        /* X = X / Z    */
                field_elt_mul(P->y, P->y, u, k, stack);        /* Y = Y / Z    */
                field_elt_set_one(P->z, k);                    /* Z = 1        */
                field_elt_set_one(P->t, k);                    /* T = 1        */
                break;

            case JACOBIAN :
                field_elt_inv(u, P->z, k, stack);              /* u = 1/Z      */
                field_elt_sqr(P->t, u, k, stack);              /* T = 1/Z^2    */
                field_elt_mul(P->x, P->x, P->t, k, stack);     /* X = X/Z^2    */
                field_elt_mul(u, u, P->t, k, stack);           /* u = 1/Z^3    */
                field_elt_mul(P->y, P->y, u, k, stack);        /* Y = Y/Z^3    */
                field_elt_set_one(P->z, k);                    /* Z = 1        */
                field_elt_set_one(P->t, k);                    /* T = 1        */
                break;

            default :
                break;
        }
    }
    else
    {
        /* Write neutral element under classical form */
        weierstrass_point_set_neutral(P, E, stack);
    }
    field_elt_relax_pool_elt(&u, k, stack);
}

void
weierstrass_point_get_x_affine (field_elt x, ec_point_ptr P, ec_curve_srcptr E, uint8_t stack)
{
    field_t *k = E->k;
    field_elt u;
    field_elt_get_pool_elt(&u, k, stack);
    
    switch(E->f)
    {
        case PROJECTIVE :
            field_elt_inv(u, P->z, k, stack);              /* u = 1/Z      */
            field_elt_mul(x, P->x, u, k, stack);           /* x = X / Z    */
            break;

        case JACOBIAN :
            field_elt_inv(u, P->z, k, stack);              /* u = 1/Z      */
            field_elt_sqr(u, u, k, stack);                 /* u = 1/Z^2    */
            field_elt_mul(x, P->x, u, k, stack);           /* x = X/Z^2    */
            break;

        default :
            break;
    }
    field_elt_relax_pool_elt(&u, k, stack);
}

void
weierstrass_point_get_y_affine (field_elt y, ec_point_ptr P, ec_curve_srcptr E, uint8_t stack)
{
    field_t *k = E->k;
    field_elt u;
    field_elt_get_pool_elt(&u, k, stack);
    
    switch(E->f)
    {
        case PROJECTIVE :
            field_elt_inv(u, P->z, k, stack);              /* u = 1/Z      */
            field_elt_mul(y, P->y, u, k, stack);           /* y = Y / Z    */
            break;

        case JACOBIAN :
            field_elt_inv(u, P->z, k, stack);              /* u = 1/Z      */
            field_elt_sqr(u, u, k, stack);                 /* u = 1/Z^2    */
            field_elt_mul(u, P->z, P->t, k, stack);        /* u = 1/Z^3    */
            field_elt_mul(y, P->y, u, k, stack);           /* y = Y/Z^3    */
            break;

        default :
            break;
    }
    field_elt_relax_pool_elt(&u, k, stack);
}

/*******************************COMPARISON************************************/

bool 
weierstrass_point_is_neutral (ec_point_srcptr P, ec_curve_srcptr E, uint8_t stack)
{
    field_t *k = E->k;
    switch(E->f)
    { 
        case PROJECTIVE :
            return (field_elt_iszero(P->x, k) && 
                    field_elt_isone(P->y, k) && field_elt_iszero(P->z, k));
            break;

        case JACOBIAN :
            /* The infinity point as representation (t^2, t^3, 0) */
            if(field_elt_iszero(P->z, k))
            {
                field_elt a, b;
                field_elt_get_pool_elt(&a, k, stack);
                field_elt_get_pool_elt(&b, k, stack);
                field_elt_sqr(a, P->y, k, stack);
                field_elt_pow_ui(b, P->x, (block)3, k, stack);
                if(field_elt_cmp(a, b, k)==0)
                {
                    field_elt_relax_pool_elt(&a, k, stack);
                    field_elt_relax_pool_elt(&b, k, stack);
                    return true;
                }
                else
                {
                    field_elt_relax_pool_elt(&a, k, stack);
                    field_elt_relax_pool_elt(&b, k, stack);
                    return false;
                }
            }
            else
            {
                return false;
            }
            break;
        
        default :
            return false;
    }
}

bool
weierstrass_point_are_equal (ec_point_srcptr P1, ec_point_srcptr P2, 
        ec_curve_srcptr E, uint8_t stack)
{
    field_t *k = E->k;
    field_elt u, v, tmp, tmp1;
    bool res = false;
    /* P1=(x1,y1,z1), P2=(x2,y2,z2) */
    if (field_elt_iszero(P1->z, k) || field_elt_iszero(P2->z, k))
    {
        return (weierstrass_point_is_neutral(P1, E, stack) && weierstrass_point_is_neutral(P2, E, stack));
    }
    
    switch(E->f)
    {
        case PROJECTIVE :
            if (field_elt_iszero(P1->y, k))
            {
                if(field_elt_iszero(P2->y, k))
                {
                    if(field_elt_cmp(P1->x, P2->x, k)==0)
                    {
                        res = true;
                        break;
                    }
                    else
                    { 
                        res = false;
                        break;
                    }
                }
                else
                {
                    res = false;
                    break;
                }
            }
        
            field_elt_get_pool_elt(&u, k, stack);
            field_elt_get_pool_elt(&v, k, stack);
            field_elt_mul(u, P1->x, P2->z, k, stack);   /* u = x1 * z2      */
            field_elt_mul(v, P1->z, P2->x, k, stack);   /* v = z1 * x2      */
            if(field_elt_cmp(u, v, k) != 0)
            {
                res = false;
                field_elt_relax_pool_elt(&u, k, stack);
                field_elt_relax_pool_elt(&v, k, stack);
                break;
            }
            field_elt_mul(u, P1->y, P2->z, k, stack);   /* u = y1 * z2      */
            field_elt_mul(v, P1->z, P2->y, k, stack);   /* v = z1 * y2      */

            res = (field_elt_cmp(u, v, k) == 0);
            field_elt_relax_pool_elt(&u, k, stack);
            field_elt_relax_pool_elt(&v, k, stack);
            break;

        case JACOBIAN :
            field_elt_get_pool_elt(&u, k, stack);
            field_elt_get_pool_elt(&v, k, stack);
            field_elt_get_pool_elt(&tmp, k, stack);
            field_elt_get_pool_elt(&tmp1, k, stack);

            field_elt_sqr(u, P1->z, k, stack);
            field_elt_sqr(v, P2->z, k, stack);

            field_elt_mul(tmp, P2->x, u, k, stack);     /* tmp =  x2 * z1^2 */
            field_elt_mul(tmp1, P1->x, v, k, stack);    /* tmp1 = x1 * z2^2 */
            if((field_elt_cmp(tmp, tmp1, k) != 0))
            {
                res = false;
                field_elt_relax_pool_elt(&u, k, stack);
                field_elt_relax_pool_elt(&v, k, stack);
                field_elt_relax_pool_elt(&tmp, k, stack);
                field_elt_relax_pool_elt(&tmp1, k, stack);
                break;
            }
            field_elt_mul(u, u, P1->z, k, stack);
            field_elt_mul(v, v, P2->z, k, stack);
            field_elt_mul(tmp, P2->y, u, k, stack);     /* tmp =  y2 * z1^3 */
            field_elt_mul(tmp1, P1->y, v, k, stack);    /* tmp1 = y1 * z2^3 */
            res = (field_elt_cmp(tmp, tmp1, k) == 0);

            field_elt_relax_pool_elt(&u, k, stack);
            field_elt_relax_pool_elt(&v, k, stack);
            field_elt_relax_pool_elt(&tmp, k, stack);
            field_elt_relax_pool_elt(&tmp1, k, stack);
            break;
        default :
            res = false;
    }
    return res;
}

/*******************************OPERATIONS************************************/

/** --> COZ */

void 
weierstrass_point_add_ZADDU (ec_point_ptr P3, ec_point_ptr P1, ec_point_srcptr P2,
        ec_curve_srcptr E, uint8_t stack)
{

    /* Those formulae are taken from [GJM10, §2.2, Algorithm 1]  */
    /* Costs 5M, 2S, 7Add  */
    field_t *k = E->k;
    field_elt tmp1, tmp3;

    field_elt_get_pool_elt(&tmp1, k, stack);
    field_elt_get_pool_elt(&tmp3, k, stack);
    
    field_elt_sub(tmp1, P1->x, P2->x, k);                   /* tmp1 = X1 - X2                   */
    field_elt_mul(P3->z, P2->z, tmp1, k, stack);            /* Z3 = Z1 * (X1 - X2)              */
    field_elt_sqr(tmp1, tmp1, k, stack);                    /* tmp1 = (X1 - X2)^2 = C           */
    
    field_elt_mul(P1->x, P1->x, tmp1, k, stack);            /* X1   = X1 * C = W1               */
    field_elt_mul(tmp3, P2->x, tmp1, k, stack);             /* tmp3 = X2 * C = W2               */
    
    field_elt_sub(P3->y, P1->y, P2->y, k);                  /* Y3 = Y1 - Y2                     */
    field_elt_sqr(tmp1, P3->y, k, stack);                   /* tmp1 = (Y1 - Y2)^2 = D           */
    
    field_elt_sub(P1->z, P1->x, tmp3, k);                   /* Z1 = W1 - W2                     */
    field_elt_mul(P1->y, P1->y, P1->z, k, stack);           /* Y1   = Y1 * (W1 - W2) = A1       */
    
    field_elt_sub(P3->x, tmp1, P1->x, k);                   /* X3 = D - W1                      */
    field_elt_sub(P3->x, P3->x, tmp3, k);                   /* X3 = D - W1 - W2                 */
    
    field_elt_sub(tmp3, P1->x, P3->x, k);                   /* tmp3 = W1 - X3                   */
    field_elt_mul(P3->y, P3->y, tmp3, k, stack);            /* Y3 = (Y1 - Y2) * (W1 - X3)       */
    field_elt_sub(P3->y, P3->y, P1->y, k);                  /* Y3 = (Y1 - Y2) * (W1 - X3) - A1  */

    field_elt_copy(P1->z, P3->z, k);                        /* Z1 = Z3                          */    
    field_elt_set_one(P3->t, k);                            /* T3 = 1                           */

    field_elt_relax_pool_elt(&tmp1, k, stack);
    field_elt_relax_pool_elt(&tmp3, k, stack);
}

void 
weierstrass_point_add_ZADDC (ec_point_ptr P3, ec_point_ptr P4, ec_point_srcptr P1, 
        ec_point_srcptr P2, ec_curve_srcptr E, uint8_t stack)
{
    /* Those formulae are taken from [GJM10,§4 Algorithm 6]  */
    /* Costs  6M, 3S, 11Add */
    field_t *k = E->k;
    field_elt tmp1, tmp6;

    field_elt_get_pool_elt(&tmp1, k, stack);
    field_elt_get_pool_elt(&tmp6, k, stack);
    
    field_elt_sub(tmp1, P1->x, P2->x, k);                   /* tmp1 = X1 - X2                   */
    field_elt_mul(P3->z, P1->z, tmp1, k, stack);            /* Z3 = Z * (X1 - X2)               */
    field_elt_sqr(tmp1, tmp1, k, stack);                    /* tmp1 = (X1 - X2)^2 = C           */
    
    field_elt_mul(P4->t, P1->x, tmp1, k, stack);             /* tmp2 = X1 * C = W1               */
    field_elt_mul(P4->x, P2->x, tmp1, k, stack);             /* tmp3 = X2 * C = W2               */
    field_elt_add(tmp1, P4->t, P4->x, k);                     /* tmp1 = W1 + W2                   */

    field_elt_sub(P4->z, P4->t, P4->x, k);                     /* tmp5 = W1 - W2                   */
    field_elt_mul(P4->z, P1->y, P4->z, k, stack);             /* tmp5 = Y1 * (W1 - W2) = A1       */
    
    field_elt_sub(P4->x, P1->y, P2->y, k);                   /* tmp3 = Y1 - Y2                   */
    field_elt_sqr(tmp6, P4->x, k, stack);                    /* tmp4 = (Y1 - Y2)^2 = D           */
    
    field_elt_sub(P3->x, tmp6, tmp1, k);                    /* X3 = D - W1 - W2                 */
    
    field_elt_add(P4->y, P1->y, P2->y, k);                   /* tmp4 = Y1 + Y2                   */
    
    field_elt_sub(tmp6, P4->t, P3->x, k);                    /* tmp6 = W1 - X3                   */
    field_elt_mul(P3->y, P4->x, tmp6, k, stack);             /* Y3 = (Y1 - Y2) * (W1 - X3)       */
    field_elt_sub(P3->y, P3->y, P4->z, k);                   /* Y3 = (Y1 - Y2) * (W1 - X3) - A1  */
    
    field_elt_set_one(P3->t, k);                            /* T3 = 1                           */
    
    field_elt_sqr(P4->x, P4->y, k, stack);                    /* tmp3 = (Y1 + Y2)^2 = D           */
    field_elt_sub(P4->x, P4->x, tmp1, k);                    /* X4 = D - W1 - W2                 */
    
    field_elt_sub(tmp6, P4->t, P4->x, k);                    /* tmp6 = W1 - X4                   */
    field_elt_mul(P4->y, P4->y, tmp6, k, stack);             /* Y4 = (Y1 + Y2) * (W1 - X4)       */
    field_elt_sub(P4->y, P4->y, P4->z, k);                   /* Y4 = (Y1 + Y2) * (W1 - X4) - A1  */
    
    field_elt_copy(P4->z, P3->z, k);                        /* Z4 = Z3                          */
    
    field_elt_set_one(P4->t, k);                            /* T4 = 1                           */

    field_elt_relax_pool_elt(&tmp1, k, stack);
    field_elt_relax_pool_elt(&tmp6, k, stack);
}

void 
weierstrass_point_add_ZDAU (ec_point_ptr P3, ec_point_srcptr P1, ec_point_ptr P2,
        ec_curve_srcptr E, uint8_t stack)
{
    /* Those formulae are taken from [GJM10,§4.4 and Algorithm 9]  */
    /* Costs 9 M + 7 S + 22 Add + 1*2 + 1*4 */
    field_t *k = E->k;
    field_elt c,cp,a1,w1,w1p,w2p,d,x3;

    field_elt_get_pool_elt(&cp, k, stack);
    field_elt_get_pool_elt(&w1p, k, stack);
    field_elt_get_pool_elt(&w2p, k, stack);
    field_elt_get_pool_elt(&d, k, stack);
    field_elt_get_pool_elt(&a1, k, stack);
    field_elt_get_pool_elt(&w1, k, stack);
    field_elt_get_pool_elt(&x3, k, stack);
    field_elt_get_pool_elt(&c, k, stack);

    field_elt_sub(P3->z, P1->x, P2->x, k);             
    field_elt_sqr(cp, P3->z, k, stack);                         /* C'  = (X1-X2)^2                                                  */
    field_elt_mul(w1p, P1->x, cp, k, stack);                    /* W1' = X1C'                                                       */
    field_elt_mul(w2p, P2->x, cp, k, stack);                    /* W2' = X2C'                                                       */
    field_elt_sub(a1, w1p, w2p, k);
    field_elt_mul(a1, P1->y, a1, k, stack);                     /* A1' = Y1(W1'-W2')                                                */
    field_elt_sub(P3->y, P1->y, P2->y, k);             
    field_elt_sqr(d, P3->y, k, stack);                          /* D'  = (Y1-Y2)^2                                                  */
    field_elt_sub(x3, d, w1p, k);
    field_elt_sub(x3, x3, w2p, k);                              /* X3' = D' - W1' -W2'                                              */
    field_elt_sub(P3->x, x3, w1p, k);
    field_elt_sqr(c, P3->x, k, stack);                          /* C   = (X3' - W1')^2                                              */
    field_elt_add(P3->z, P3->z,P3->x, k);
    field_elt_sqr(P3->z, P3->z, k, stack);
    field_elt_sub(P3->z, P3->z,cp, k);
    field_elt_sub(P3->z, P3->z,c, k);
    field_elt_mul(P3->z, P3->z,P2->z, k, stack);                /* Z3  =Z((X1-X2+X3'-W1')^2-C'-C)                                   */
    field_elt_sub(cp, P3->y,P3->x, k);   
    field_elt_sqr(cp, cp, k, stack);
    field_elt_sub(cp, cp, d, k);
    field_elt_sub(cp, cp, c, k);
    field_elt_mul2(a1, a1, k);
    field_elt_sub(cp, cp, a1, k);                               /* Y3' = [(Y1-Y2) - (X3'-W1')]^2 - D' - C - 2A1'  (stored in cp)    */
    field_elt_mul4(d, c, k);
    field_elt_mul(w1, x3, d, k, stack);                         /* W1  = 4X3'C                                                      */
    field_elt_mul(w2p, d, w1p, k, stack);                       /* W2  = 4W1'C                                                      */
    field_elt_sub(w1p, cp, a1, k);
    field_elt_sqr(d, w1p, k, stack);                            /* D   = (Y3' - 2A1')^2                                             */
    field_elt_add(P2->y, w1, w2p, k);
    field_elt_sub(P3->x, d, P2->y, k);                          /* X3  = D - W1 - W2                                                */
    field_elt_sub(c, w1, w2p, k);
    field_elt_mul(c, c, cp, k, stack);                          /* A1  = Y3'(W1-W2) (stored in c)                                   */
    field_elt_sub(d, w1, P3->x, k);
    field_elt_mul(P3->y, w1p, d, k, stack);                 
    field_elt_sub(P3->y, P3->y, c, k);                          /* Y3  = (Y3' - 2A1')(W1 - X3) - A1                                 */
    field_elt_add(w1p, cp, a1, k);               
    field_elt_sqr(d, w1p, k, stack);                            /* D   = (Y3' + 2A1')^2                                             */
    field_elt_sub(P2->x, d, P2->y, k);                          /* X2  = D - W1 - W2                                                */
    field_elt_sub(P2->y, w1, P2->x, k);
    field_elt_mul(P2->y, P2->y, w1p, k, stack);
    field_elt_sub(P2->y, P2->y, c, k);                          /* Y2  = (Y3'+ 2A1')(W1-X2)-A1                                      */
    field_elt_copy(P2->z, P3->z, k);                            /* Z2  = Z3 */
                                
    field_elt_set_one(P3->t, k);                                /* T3 = 1                                                           */

    field_elt_relax_pool_elt(&c, k, stack);
    field_elt_relax_pool_elt(&w1p, k, stack);
    field_elt_relax_pool_elt(&w2p, k, stack);
    field_elt_relax_pool_elt(&w1, k, stack);
    field_elt_relax_pool_elt(&a1, k, stack);
    field_elt_relax_pool_elt(&x3, k, stack);
    field_elt_relax_pool_elt(&cp, k, stack);
    field_elt_relax_pool_elt(&d, k, stack);
    
}

void 
weierstrass_point_DBLU (ec_point_ptr P3, ec_point_ptr P4, ec_point_srcptr P1, 
        ec_curve_srcptr E, uint8_t stack)
{
    field_t *k = E->k;
    field_elt tmp1, tmp2, tmp3, tmp4, tmp5, tmp6;

    /* Those formulae are taken from [GJM10,§4.3] for the case of P1->z==1 */
    /* Very similar to those of Jacobian case */
    /* Costs 1 M 5 S 6 Add 3*2  1*3 1*8 */

    /* It is supposed that Z1==1 */
    MPHELL_ASSERT( field_elt_isone(P1->z, k), "\nThis point P1 should have its Z coordinate equals to 1 to apply DBLU\n");

    field_elt_get_pool_elt(&tmp1, k, stack);
    field_elt_get_pool_elt(&tmp2, k, stack);
    field_elt_get_pool_elt(&tmp3, k, stack);
    field_elt_get_pool_elt(&tmp4, k, stack);
    field_elt_get_pool_elt(&tmp5, k, stack);
    field_elt_get_pool_elt(&tmp6, k, stack);

    field_elt_sqr(tmp1, P1->x, k, stack);                   /* B = X1^2                         */
    field_elt_sqr(tmp2, P1->y, k, stack);                   /* E = Y1^2                         */
    
    field_elt_sqr(tmp3, tmp2, k, stack);                    /* L = E^2                          */

    field_elt_add(tmp4, P1->x, tmp2, k);                    /* tmp4 = X1 + E                    */
    field_elt_sqr(tmp4, tmp4, k, stack);                    /* tmp4 = (X1 + E)^2                */
    field_elt_sub(tmp4, tmp4, tmp1, k);                     /* tmp4 = (X1 + E)^2 - B            */
    field_elt_sub(tmp4, tmp4, tmp3, k);                     /* tmp4 = (X1 + E)^2 - B - L        */
    field_elt_mul2(tmp4, tmp4, k);                          /* S = 2((X1 + E)^2 - B - L)        */
    
    field_elt_mul3(tmp5, tmp1, k, stack);                   /* tmp5 = 3 * B                     */
    field_elt_add(tmp5, tmp5, E->a, k);                     /* M = 3 * B + a                    */

    field_elt_sqr(P3->x, tmp5, k, stack);                   /* X3 = M^2                         */
    field_elt_mul2(tmp1, tmp4, k);                          /* tmp1 = 2.S                       */
    field_elt_sub(P3->x, P3->x, tmp1, k);                   /* X3 = M^2 - 2 * S                 */

    field_elt_sub(P3->y, tmp4, P3->x, k);                   /* Y3 = S - X3                      */
    field_elt_mul(P3->y, tmp5, P3->y, k, stack);            /* Y3 = M * (S - X3)                */
    field_elt_mul8(tmp3, tmp3, k);                          /* tmp3 = 8 * L                     */
    
    field_elt_sub(P3->y, P3->y, tmp3, k);                   /* Y3 = M * (S - X3) - 8 * L        */

    field_elt_mul2(P3->z, P1->y, k);                        /* Z3 = 2 * Y1                      */
    field_elt_set_one(P3->t, k);                            /* T3 = 1                           */
        
    field_elt_copy(P4->x, tmp4, k);                         /* X4 = S                           */
    field_elt_copy(P4->y, tmp3, k);                         /* Y4 = 8 * L                       */
    field_elt_copy(P4->z, P3->z, k);                        /* Z4 = Z3                          */
    field_elt_set_one(P4->t, k);                            /* T4 = 1                           */
       
    field_elt_relax_pool_elt(&tmp1, k, stack);
    field_elt_relax_pool_elt(&tmp2, k, stack);
    field_elt_relax_pool_elt(&tmp3, k, stack);
    field_elt_relax_pool_elt(&tmp4, k, stack);
    field_elt_relax_pool_elt(&tmp5, k, stack);
    field_elt_relax_pool_elt(&tmp6, k, stack);
}

void 
weierstrass_point_TPLU (ec_point_ptr P3, ec_point_ptr P4, ec_point_srcptr P1, 
        ec_curve_srcptr E, uint8_t stack)
{
    /* Those formulae are taken from [GJM10,§4.3]  */
    weierstrass_point_DBLU(P3, P4, P1, E, stack);
    weierstrass_point_add_ZADDU(P3, P4, P3, E, stack);
}

/** --> PROJECTIVE */

/**
 * @fn    void weierstrass_point_dbl_projective_dedicated (ec_point P3, ec_point_srcptr P1, ec_curve_srcptr E, uint8_t stack)
 * @brief Set P3 to 2 P1 using projective coordinate
 * @param P3    Resulting point 
 * @param P1    Source point (to double)
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void 
weierstrass_point_dbl_projective_dedicated (ec_point_ptr P3, ec_point_srcptr P1, ec_curve_srcptr E, uint8_t stack)
{
    /* P1 = (X1, Y1, Z1)     */
    /* P3 = (X3, Y3, Z3)     */
    /* E: y^2 = x^3 + ax + b */
    
    field_t *k = E->k;
    field_elt tmp1, tmp2, tmp3, tmp4, tmp5;

    field_elt_get_pool_elt(&tmp1, k, stack);
    field_elt_get_pool_elt(&tmp2, k, stack);
    field_elt_get_pool_elt(&tmp3, k, stack);
    field_elt_get_pool_elt(&tmp4, k, stack);
    field_elt_get_pool_elt(&tmp5, k, stack);

    if(E->ec_spec1 == true)
    {
        /* To improve efficiency for NIST curve */
        /* Formula taken from [BL07] viewable here: http://www.hyperelliptic.org/EFD/g1p/auto-shortw-projective-3.html#doubling-dbl-2007-bl-2 */
        /* Costs 7M 3S 5add 4*2 1*3 */

        field_elt_sub(tmp1, P1->x, P1->z, k);               /* tmp1 = X1 - Z1                                                   */
        field_elt_add(tmp2, P1->x, P1->z, k);               /* tmp2 = X1 + Z1                                                   */
        field_elt_mul(tmp1, tmp1, tmp2, k, stack);          /* tmp1 = (X1 - Z1).(X1 + Z1)                                       */
        field_elt_mul3(tmp2, tmp1, k, stack);               /* tmp2 = 3*(X1 - Z1).(X1 + Z1) = w                                 */
                                                                                                                                
        field_elt_mul(tmp1, P1->y, P1->z, k, stack);        /* tmp1 = Y1.Z1                                                     */
        field_elt_mul2(tmp1, tmp1, k);                      /* tmp1 = 2.Y1.Z1 = s                                               */ 
                                                                                                                                
        field_elt_mul(tmp3, P1->y, tmp1, k, stack);         /* tmp3 = Y1.s = R                                                  */ 
        field_elt_mul(tmp4, P1->x, tmp3, k, stack);         /* tmp4 = X1.R                                                      */
        field_elt_mul2(tmp4, tmp4, k);                      /* tmp4 = 2.X1.R = B                                                */ 
                                                                                                                                
        field_elt_sqr(P3->x, tmp2, k, stack);               /* X3 = w^2                                                         */
        field_elt_mul2(tmp5, tmp4, k);                                                                                  
        field_elt_sub(P3->x, P3->x, tmp5, k);               /* X3 = w^2 - 2.B = h                                               */
                                                                                                                                
        field_elt_sub(P3->y, tmp4, P3->x, k);               /* Y3 = B - h                                                       */
        field_elt_mul(P3->x, P3->x, tmp1, k, stack);        /* X3 = h.s                                                         */
                                                                                                                                
        field_elt_mul(P3->y, P3->y, tmp2, k, stack);        /* Y3 = (B - h).w                                                   */
        field_elt_sqr(tmp2, tmp3, k, stack);                /* tmp2 = R^2                                                       */
        field_elt_mul2(tmp2, tmp2, k);                                                                                   
        field_elt_sub(P3->y, P3->y, tmp2, k);               /* Y3 = (B - h).w - 2.R^2                                           */
                                                                                                                                
        field_elt_sqr(P3->z, tmp1, k, stack);               /* Z3 = s^2                                                         */
        field_elt_mul(P3->z, P3->z, tmp1, k, stack);        /* Z3 = s^3                                                         */
    }
    else
    {
        /*Formula taken from [BL07] viewable here: http://www.hyperelliptic.org/EFD/g1p/auto-shortw-projective.html#doubling-dbl-2007-bl */
        /* Costs 5M 6 S 1*a 7 add 3*2 1*3 */
        field_elt_sqr(tmp1, P1->x, k, stack);               /* tmp1 = X1^2                                                      */
        field_elt_sqr(tmp2, P1->z, k, stack);               /* tmp2 = Z1^2                                                      */
        field_elt_mul(tmp2, tmp2, E->a, k, stack);          /* tmp2 = a.Z1^2                                                    */
        field_elt_mul3(tmp3, tmp1, k, stack);               /* tmp3 = 3.X1^2                                                    */
        field_elt_add(tmp2, tmp2, tmp3, k);                 /* tmp2 = a.Z1^2 + 3.X1^2 = w                                       */
        field_elt_mul(P3->z, P1->y, P1->z, k, stack);       /* Z3 = Y1.Z1                                                       */
        field_elt_mul2(P3->z, P3->z, k);                    /* Z3 = 2.Y1.Z1 = s                                                 */
        
        field_elt_mul(tmp3, P1->y, P3->z, k, stack);        /* tmp3 = Y1.s = R                                                  */
       
        field_elt_add(P3->y, P1->x, tmp3, k);               /* Y3 = X1 + R                                                      */
        field_elt_sqr(P3->y, P3->y, k, stack);              /* Y3 = (X1 + R)^2                                                  */
        field_elt_sub(P3->y, P3->y, tmp1, k);               /* Y3 = (X1 + R)^2 - X1^2                                           */
        field_elt_sqr(tmp1, tmp3, k, stack);                /* tmp1 = R^2                                                       */
        field_elt_sub(P3->y, P3->y, tmp1, k);               /* Y3 = (X1 + R)^2 - X1^2 - R^2 = B                                 */
        
        field_elt_sqr(P3->x, tmp2, k, stack);               /* X3 = w^2                                                         */
        field_elt_mul2(tmp3, P3->y, k);
        field_elt_sub(P3->x, P3->x, tmp3, k);               /* X3 = w^2 - 2.B = h                                               */
     
        field_elt_sub(P3->y, P3->y, P3->x, k);              /* Y3 = B - h                                                       */
        field_elt_mul(P3->x, P3->x, P3->z, k, stack);       /* X3 = h.s                                                         */
        
        field_elt_mul(P3->y, P3->y, tmp2, k, stack);        /* Y3 = (B - h).w                                                   */
        field_elt_mul2(tmp3, tmp1, k);
        field_elt_sub(P3->y, P3->y, tmp3, k);               /* Y3 = (B - h).w - 2.R^2                                           */

        field_elt_sqr(tmp1, P3->z, k, stack);               /* tmp1 = s^2                                                       */
        field_elt_mul(P3->z, P3->z, tmp1, k, stack);        /* Z3 = s^3                                                         */
    }
    field_elt_set_one(P3->t, k);                            /* T3 = 1                                                           */

    field_elt_relax_pool_elt(&tmp1, k, stack);
    field_elt_relax_pool_elt(&tmp2, k, stack);
    field_elt_relax_pool_elt(&tmp3, k, stack);
    field_elt_relax_pool_elt(&tmp4, k, stack);
    field_elt_relax_pool_elt(&tmp5, k, stack);
}

/**
 * @fn    void weierstrass_point_add_projective_dedicated (ec_point P3, ec_point_srcptr P1, ec_point_srcptr P2, ec_curve_srcptr E, uint8_t stack)
 * @brief Set P3 to P1 + P2 using projective coordinate
 * @param P3    Resulting point 
 * @param P1    Source point
 * @param P2    Source point 
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void
weierstrass_point_add_projective_dedicated (ec_point_ptr P3, ec_point_srcptr P1, 
        ec_point_srcptr P2, ec_curve_srcptr E, uint8_t stack)
{
    /* P1 = (X1, Y1, Z1)     */
    /* P2 = (X2, Y2, Z2)     */
    /* P3 = (X3, Y3, Z3)     */
    /* E: y^2 = x^3 + ax + b */
    
    /* Formula taken from [CMO98, formula 3] viewable at this adress: http://www.hyperelliptic.org/EFD/g1p/auto-shortw-projective.html#addition-add-1998-cmo-2  */
    /* Cost 12 M 2 S 6 Add 1*2 */

    field_t *k = E->k;
    field_elt tmp1, tmp2, tmp3, tmp4, tmp5, tmp6;

    field_elt_get_pool_elt(&tmp1, k, stack);
    field_elt_get_pool_elt(&tmp2, k, stack);
    field_elt_get_pool_elt(&tmp3, k, stack);
    field_elt_get_pool_elt(&tmp4, k, stack);
    field_elt_get_pool_elt(&tmp5, k, stack);
    field_elt_get_pool_elt(&tmp6, k, stack);
    
    field_elt_mul(tmp1, P1->y, P2->z, k, stack);        /* tmp1 = Y1 * Z2                     */
    field_elt_mul(tmp2, P1->x, P2->z, k, stack);        /* tmp2 = X1 * Z2                     */
    field_elt_mul(tmp3, P1->z, P2->z, k, stack);        /* tmp3 = Z1 * Z2                     */

    field_elt_mul(tmp4, P2->y, P1->z, k, stack);        /* tmp4 = Y2 * Z1                     */
    field_elt_mul(tmp5, P2->x, P1->z, k, stack);        /* tmp5 = X2 * Z1                     */
    
    if(field_elt_cmp(tmp2, tmp5, k)==0)
    {
        /* Same x-coordinates (in affine) */
        if(field_elt_cmp(tmp1, tmp4, k)!=0)
        {
            /* Different y-coordinates (in affine) */
            weierstrass_point_set_neutral(P3, E, stack);
        }
        else
        {
            /* Same y-coordinates (in affine) */
            if(field_elt_iszero(P1->y, k))
            {
                /* y = 0 */
                weierstrass_point_set_neutral(P3, E, stack);
            }
            else
            {
                weierstrass_point_dbl_projective_dedicated(P3, P1, E, stack);
            }
        }
    }
    else
    {
        field_elt_sub(tmp4, tmp4, tmp1, k);             /* tmp4 = (Y2 * Z1) - tmp1  = u       */
        field_elt_sub(tmp5, tmp5, tmp2, k);             /* tmp5 = (X2 * Z1) - tmp2  = v       */

        field_elt_sqr(P3->x, tmp4, k, stack);           /* X3 = u^2                           */
        field_elt_mul(P3->x, P3->x, tmp3, k, stack);    /* X3 = u^2*Z1Z2                      */
        field_elt_sqr(P3->y, tmp5, k, stack);           /* Y3 = v^2                           */
        field_elt_mul(tmp6, P3->y, tmp5, k, stack);     /* tmp6 = tmp5^3 = v^3                */
        field_elt_sub(P3->x, P3->x, tmp6, k);           /* X3 = u^2*Z1Z2 - v^3                */

        field_elt_mul(P3->y, P3->y, tmp2, k, stack);    /* Y3 = tmp5^2 * tmp2 = v^2*X1Z2 = R  */
        field_elt_mul2(tmp2, P3->y, k);
        field_elt_sub(P3->x, P3->x, tmp2, k);           /* X3 = u^2*Z1Z2 - v^3 - 2R = A       */

        field_elt_sub(P3->y, P3->y, P3->x, k);          /* Y3 = R - A                         */
        field_elt_mul(P3->x, P3->x, tmp5, k, stack);    /* X3 = X3 * tmp5 = v * A             */

        field_elt_mul(P3->z, tmp6, tmp3, k, stack);     /* Z3 = tmp6 * tmp3 = v^3 * Z1Z2      */
        
        field_elt_mul(P3->y, P3->y, tmp4, k, stack);    /* Y3 = (R - A) * u                   */
        field_elt_mul(tmp1, tmp1, tmp6, k, stack);      /* tmp1 = Y1Z2 * v^3 */
        field_elt_sub(P3->y, P3->y, tmp1, k);           /* Y3 = (R - A) * u -  Y1Z2 * v^3     */
    }
    field_elt_set_one(P3->t, k);                        /* T3 = 1                             */

    field_elt_relax_pool_elt(&tmp1, k, stack);
    field_elt_relax_pool_elt(&tmp2, k, stack);
    field_elt_relax_pool_elt(&tmp3, k, stack);
    field_elt_relax_pool_elt(&tmp4, k, stack);
    field_elt_relax_pool_elt(&tmp5, k, stack);
    field_elt_relax_pool_elt(&tmp6, k, stack);
}

/**
 * @fn    void weierstrass_point_add_projective_unified (ec_point P3, ec_point_srcptr P1, ec_point_srcptr P2, ec_curve_srcptr E, uint8_t stack)
 * @brief Set P3 to P1 + P2 using projective coordinate
 * @param P3    Resulting point 
 * @param P1    Source point
 * @param P2    Source point 
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void
weierstrass_point_add_projective_unified (ec_point_ptr P3, ec_point_srcptr P1, 
        ec_point_srcptr P2, ec_curve_srcptr E, uint8_t stack)
{
    /* P1 = (X1, Y1, Z1)     */
    /* P2 = (X2, Y2, Z2)     */
    /* P3 = (X3, Y3, Z3)     */
    /* E: y^2 = x^3 + ax + b */
    
    /* Formula taken from [add-2007-bl] viewable at this adress: http://www.hyperelliptic.org/EFD/g1p/data/shortw/projective/addition/add-2007-bl  */
    /* Cost 11M + 6S + 1*a + 10add + 4*2 + 1*4 */

    field_t *k = E->k;
    field_elt U1, U2, S1, S2, ZZ, T, TT, M, R, F, L, LL, G, W;
    field_elt t0, t1, t2, t3;
    
    field_elt_get_pool_elt(&U1, k, stack);
    field_elt_get_pool_elt(&U2, k, stack);
    field_elt_get_pool_elt(&S1, k, stack);
    field_elt_get_pool_elt(&S2, k, stack);
    field_elt_get_pool_elt(&ZZ, k, stack);
    field_elt_get_pool_elt(&T, k, stack);
    field_elt_get_pool_elt(&TT, k, stack);
    field_elt_get_pool_elt(&M, k, stack);
    field_elt_get_pool_elt(&R, k, stack);
    field_elt_get_pool_elt(&F, k, stack);
    field_elt_get_pool_elt(&L, k, stack);
    field_elt_get_pool_elt(&LL, k, stack);
    field_elt_get_pool_elt(&G, k, stack);
    field_elt_get_pool_elt(&W, k, stack);
    
    field_elt_get_pool_elt(&t0, k, stack);
    field_elt_get_pool_elt(&t1, k, stack);
    field_elt_get_pool_elt(&t2, k, stack);
    field_elt_get_pool_elt(&t3, k, stack);
    
    field_elt_mul(U1, P1->x, P2->z, k, stack);      /* U1 = X1*Z2               */
    field_elt_mul(U2, P2->x, P1->z, k, stack);      /* U2 = X2*Z1               */
    field_elt_mul(S1, P1->y, P2->z, k, stack);      /* S1 = Y1*Z2               */
    field_elt_mul(S2, P2->y, P1->z, k, stack);      /* S2 = Y2*Z1               */
    
    if(field_elt_cmp(U1, U2, k)==0)
    {
        /* Same x-coordinates (in affine) */
        if(field_elt_cmp(S1, S2, k)!=0)
        {
            /* Different y-coordinates (in affine) */
            weierstrass_point_set_neutral(P3, E, stack);
            return;
        }
        else
        {
            /* Same y-coordinates (in affine) */
            if(field_elt_iszero(P1->y, k))
            {
                weierstrass_point_set_neutral(P3, E, stack);
                return;
            }
        }
    }
    
    field_elt_mul(ZZ, P1->z, P2->z, k, stack);      /* ZZ = Z1*Z2               */
    field_elt_add(T, U1, U2, k);                    /* T = U1+U2                */
    field_elt_sqr(TT, T, k, stack);                 /* TT = T2                  */
    field_elt_add(M, S1, S2, k);                    /* M = S1+S2                */
    field_elt_sqr(t0, ZZ, k, stack);                /* t0 = ZZ^2                */
    field_elt_mul(t1, E->a, t0, k, stack);          /* t1 = a*t0                */
    field_elt_mul(t2, U1, U2, k, stack);            /* t2 = U1*U2               */
    field_elt_sub(t3, TT, t2, k);                   /* t3 = TT-t2               */
    field_elt_add(R, t3, t1, k);                    /* R = t3+t1                */
    field_elt_mul(F, ZZ, M, k, stack);              /* F = ZZ*M                 */
    field_elt_mul(L, M, F, k, stack);               /* L = M*F                  */
    field_elt_sqr(LL, L, k, stack);                 /* LL = L^2                 */
    field_elt_add(t0, T, L, k);                     /* t0 = T+L                 */
    field_elt_sqr(t1, t0, k, stack);                /* t1 = t0^2                */
    field_elt_sub(t2, t1, TT, k);                   /* t2 = t1-TT               */
    field_elt_sub(G, t2, LL, k);                    /* G = t2-LL                */
    field_elt_sqr(t3, R, k, stack);                 /* t3 = R^2                 */
    field_elt_mul2(t0, t3, k);                      /* t0 = 2*t3                */
    field_elt_sub(W, t0, G, k);                     /* W = t0-G                 */
    field_elt_mul(t1, F, W, k, stack);              /* t1 = F*W                 */
    field_elt_mul2(P3->x, t1, k);                   /* X3 = 2*t1                */
    field_elt_mul2(t2, W, k);                       /* t2 = 2*W                 */
    field_elt_sub(t3, G, t2, k);                    /* t3 = G-t2                */
    field_elt_mul2(t0, LL, k);                      /* t0 = 2*LL                */
    field_elt_mul(t1, R, t3, k, stack);             /* t1 = R*t3                */
    field_elt_sub(P3->y, t1, t0, k);                /* Y3 = t1-t0               */
    field_elt_sqr(t2, F, k, stack);                 /* t2 = F^2                 */
    field_elt_mul(t3, F, t2, k, stack);             /* t3 = F*t2                */
    field_elt_mul4(P3->z, t3, k);                   /* Z3 = 4*t3                */
    
    field_elt_set_one(P3->t, k);                    /* T3 = 1                   */

    field_elt_relax_pool_elt(&U1, k, stack);
    field_elt_relax_pool_elt(&U2, k, stack);
    field_elt_relax_pool_elt(&S1, k, stack);
    field_elt_relax_pool_elt(&S2, k, stack);
    field_elt_relax_pool_elt(&ZZ, k, stack);
    field_elt_relax_pool_elt(&T, k, stack);
    field_elt_relax_pool_elt(&TT, k, stack);
    field_elt_relax_pool_elt(&M, k, stack);
    field_elt_relax_pool_elt(&R, k, stack);
    field_elt_relax_pool_elt(&F, k, stack);
    field_elt_relax_pool_elt(&L, k, stack);
    field_elt_relax_pool_elt(&LL, k, stack);
    field_elt_relax_pool_elt(&G, k, stack);
    field_elt_relax_pool_elt(&W, k, stack);

    field_elt_relax_pool_elt(&t0, k, stack);
    field_elt_relax_pool_elt(&t1, k, stack);
    field_elt_relax_pool_elt(&t2, k, stack);
    field_elt_relax_pool_elt(&t3, k, stack);
}

/** --> JACOBIAN */

/**
 * @fn    void weierstrass_point_dbl_jacobian_dedicated (ec_point P3, ec_point_srcptr P1, ec_curve_srcptr E, uint8_t stack)
 * @brief Set P3 to 2 P1 using jacobian coordinate with the 2007 Bernstein/Lange method
 * @param P3    Resulting point 
 * @param P1    Source point (to double)
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void 
weierstrass_point_dbl_jacobian_dedicated (ec_point_ptr P3, ec_point_srcptr P1, ec_curve_srcptr E, uint8_t stack)
{
    /* P1 = (X1, Y1, Z1)     */
    /* P3 = (X3, Y3, Z3)     */
    /* E: y^2 = x^3 + ax + b */
    
    /* Formulae adapted from the formulae used by Intel IPPCP */
    /* Costs 5 M 3 S 5 Add 4*2 1*3 when E->a == -3 mod p */
    /* Costs 7 M 3 S 4 Add 4*2 1*3 when E->a != -3 mod p */
    
    field_t *k = E->k;
    
    field_elt S, U, M;
    field_elt_get_pool_elt(&S, k, stack);
    field_elt_get_pool_elt(&U, k, stack);
    field_elt_get_pool_elt(&M, k, stack);
    
    field_elt_mul2(S, P1->y, k);                        /* S = 2*Y                                      */
    field_elt_sqr(U, P1->z, k, stack);                  /* U = Z^2                                      */

    field_elt_mul(M, S, P1->y, k, stack);               /* M = 2*Y^2                                    */
    field_elt_mul(P3->z, S, P1->z, k, stack);           /* Z3 = 2*Y*Z                                   */

    field_elt_sqr(P3->y, M, k, stack);                  /* Y3 = 4*Y^4                                   */
    field_elt_mul2(P3->y, P3->y, k);                    /* Y3 =  8*Y^4                                  */

    field_elt_mul(S, M, P1->x, k, stack);               /* S = 2*X*Y^2                                  */
    field_elt_mul2(S, S, k);                            /* S = 4*X*Y^2                                  */
    
    if(E->ec_spec1 == true)
    {
        field_elt_add(M, P1->x, U, k);                  /* M = 3*(X^2-Z^4)                              */
        field_elt_sub(U, P1->x, U, k);
        field_elt_mul(M, M, U, k, stack);
        field_elt_mul3(M, M, k, stack);
    }
    else 
    {
        field_elt_sqr(M, P1->x, k, stack);              /* M = 3*X^2                                    */
        field_elt_mul3(M, M, k, stack);
        field_elt_sqr(U, U, k, stack);                  /* M = 3*X^2+a*Z^4                              */
        field_elt_mul(U, U, E->a, k, stack);
        field_elt_add(M, M, U, k);
    }

    field_elt_mul2(U, S, k);                            /* U = 8*X*Y^2                                  */
    field_elt_sqr(P3->x, M, k, stack);                  /* X3 = M^2                                     */
    field_elt_sub(P3->x, P3->x, U, k);                  /* X3 = M^2 - U                                 */

    field_elt_sub(S, S, P3->x, k);                      /* S = 4*X*Y^2-X3                               */
    field_elt_mul(S, S, M, k, stack);                   /* S = M*(4*X*Y^2-X3)                           */ 
    field_elt_sub(P3->y, S, P3->y, k);                  /* Y3 = M*(4*X*Y^2-X3) - 8*Y^4                  */

    field_elt_set_one(P3->t, k);                        /* T3 = 1                                       */

    field_elt_relax_pool_elt(&S, k, stack);
    field_elt_relax_pool_elt(&U, k, stack);
    field_elt_relax_pool_elt(&M, k, stack);
}

/**
 * @fn    void weierstrass_point_add_jacobian_dedicated (ec_point P3, ec_point_srcptr P1, ec_point_srcptr P2, ec_curve_srcptr E, uint8_t stack)
 * @brief Set P3 to P1 + P2 using jacobian coordinate with the  2007 Bernstein/Lange method
 * @param P3    Resulting point 
 * @param P1    Source point
 * @param P2    Source point 
 * @param E     Elliptic curve
 * @param stack Temporary memory stack to use
 */
void
weierstrass_point_add_jacobian_dedicated (ec_point_ptr P3, ec_point_srcptr P1, ec_point_srcptr P2, ec_curve_srcptr E, uint8_t stack)
{
    /* P1 = (X1, Y1, Z1)     */
    /* P2 = (X2, Y2, Z2)     */
    /* P3 = (X3, Y3, Z3)     */
    /* E: y^2 = x^3 + ax + b */

    /* Formulae adapted from the formulae used by Intel IPPCP */
    /* Costs 12 M 4 S 6 Add 1*2 */

    field_t *k = E->k;
    field_elt U1, U2, S1, S2, H, R;
    field_elt_get_pool_elt(&U1, k, stack);
    field_elt_get_pool_elt(&U2, k, stack);
    field_elt_get_pool_elt(&S1, k, stack);
    field_elt_get_pool_elt(&S2, k, stack);
    field_elt_get_pool_elt(&H, k, stack);
    field_elt_get_pool_elt(&R, k, stack);
    
    field_elt_mul(S1, P1->y, P2->z, k, stack);          /* S1 = Y1*Z2                       */
    field_elt_sqr(U1, P2->z, k, stack);                 /* U1 = Z2^2                        */

    field_elt_mul(S2, P2->y, P1->z, k, stack);          /* S2 = Y2*Z1                       */
    field_elt_sqr(U2, P1->z, k, stack);                 /* U2 = Z1^2                        */

    field_elt_mul(S1, S1, U1, k, stack);                /* S1 = Y1*Z2^3                     */
    field_elt_mul(S2, S2, U2, k, stack);                /* S2 = Y2*Z1^3                     */

    field_elt_mul(U1, P1->x, U1, k, stack);             /* U1 = X1*Z2^2                     */
    field_elt_mul(U2, P2->x, U2, k, stack);             /* U2 = X2*Z1^2                     */

    field_elt_sub(R, S2, S1, k);                        /* R = S2 - S1                      */
    field_elt_sub(H, U2, U1, k);                        /* H = U2 - U1                      */

    if(field_elt_iszero(H, k))
    {
        /* Same x-coordinates (in affine) */
        if(!field_elt_iszero(R, k))
        {
            /* Different y-coordinates (in affine) */
            weierstrass_point_set_neutral(P3, E, stack);
        }
        else
        {
            /* Same y-coordinates (in affine) */
            if(field_elt_iszero(P1->y, k))
            {
                /* y = 0 */
                weierstrass_point_set_neutral(P3, E, stack);
            }
            else
            {
                weierstrass_point_dbl_jacobian_dedicated(P3, P1, E, stack);
            }
        }
    }
    else
    {
        field_elt_mul(P3->z, P1->z, P2->z, k, stack);   /* Z3 = Z1*Z2                       */
        field_elt_sqr(U2, H, k, stack);                 /* U2 = H^2                         */
        field_elt_mul(P3->z, P3->z, H, k, stack);       /* Z3 = (Z1*Z2)*H                   */
        field_elt_sqr(S2, R, k, stack);                 /* S2 = R^2                         */
        field_elt_mul(H, H, U2, k, stack);              /* H = H^3                          */

        field_elt_mul(U1, U1, U2, k, stack);            /* U1 = U1*H^2                      */
        field_elt_sub(P3->x, S2, H, k);                 /* X3 = R^2 - H^3                   */
        field_elt_mul2(U2, U1, k);                      /* U2 = 2*U1*H^2                    */
        field_elt_mul(S1, S1, H, k, stack);             /* S1 = S1*H^3                      */
        field_elt_sub(P3->x, P3->x, U2, k);             /* X3 = (R^2 - H^3) - 2*U1*H^2      */

        field_elt_sub(P3->y, U1, P3->x, k);             /* Y3 = R*(U1*H^2 - X3) - S1*H^3    */
        field_elt_mul(P3->y, P3->y, R, k, stack);
        field_elt_sub(P3->y, P3->y, S1, k);
        field_elt_set_one(P3->t, k);                    /* T3 = 1                           */
    }
    
    field_elt_relax_pool_elt(&U1, k, stack);
    field_elt_relax_pool_elt(&U2, k, stack);
    field_elt_relax_pool_elt(&S1, k, stack);
    field_elt_relax_pool_elt(&S2, k, stack);
    field_elt_relax_pool_elt(&H, k, stack);
    field_elt_relax_pool_elt(&R, k, stack);
}

/** --> COZ multiplication */

void
weierstrass_Zmontgomery_ladder(ec_point_ptr P3, number_srcptr n, ec_point_srcptr P1, 
        ec_curve_srcptr E, uint8_t stack)
{
    /* The Montgomery ladder permits to have at each ladder R0-R1=+/-P, 
    *  therefore if we want to avoid to have in input of ZADDU with z 
    *  coordinate=0 we must avoid that R0=R1 or R0=-R1 in input of ZADDC
    *  we have to solve the following equation +/-[k]P = [k +/-1] P
    *  which simplifies to 2k = +/-1 mod p with p the order of P
    *  Therefore we obtain that the case k = p-1/2 must be avoided
    *  which forbids the input n=p-1 for this function. 
    */
    ec_point_set_neutral(P3, E, stack);
    if(!number_iszero(n))
    {
        field_t *k = E->k;
        ec_point R0, R1;
        ec_point_get_pool_elt(R0, k, stack);
        ec_point_get_pool_elt(R1, k, stack);

        ec_point_copy(R0, P1, k);
        weierstrass_point_DBLU(R1, R0, R0, E, stack);
        char n_str[k->bit_size+4];
        number_to_bit_string(n_str, n);

        uint32_t i;
        
        for(i = 1; i < strlen(n_str); i++)
        {
            if(n_str[i] == '0')
            {
                weierstrass_point_add_ZADDC(R1, R0, R0, R1, E, stack);
                weierstrass_point_add_ZADDU(R0, R1, R0, E, stack);
            }
            else
            {
                weierstrass_point_add_ZADDC(R0, R1, R1, R0, E, stack);
                weierstrass_point_add_ZADDU(R1, R0, R1, E, stack);
            }
        }

        ec_point_copy(P3, R0, k);
        ec_point_relax_pool_elt(R0, k, stack);
        ec_point_relax_pool_elt(R1, k, stack);
    }
}


void setup_Kim(fe_ptr X1, fe_ptr X2, fe_ptr K, fe_ptr L, fe_ptr A, fe_ptr S, fe_ptr T, fe_srcptr x0, fe_srcptr y0, bool bit, ec_curve_srcptr E, uint8_t stack)
{
    /* From the work Speeding up Elliptic Curve Scalar Multiplication without Precomputation by Kim, Choe, Kim, Kim and Hong 2017 */ 
    /* Section 4 algorithm setup*/  
    /* Costs 8M + 7S + 15A */

    field_t *k = E->k;    
    field_elt T7;
    field_elt_get_pool_elt(&T7, k, stack);

    field_elt_random(K,k,stack);
    while(field_elt_iszero(K,k))
    {
        field_elt_random(K,k,stack);
    }
    field_elt_sqr(L ,K, k, stack);            /* T3 = T2^2 = Z^2                              */
    field_elt_mul(A, x0, L, k, stack);        /* T4 = x0*T3 = x0*Z^2                          */
    field_elt_mul(S, y0, L, k, stack);        /* T5 = y0*T3 = y0*Z^2                          */
    field_elt_mul(X2, S, K, k, stack);        /* T1 = T5*T2 = y0*Z^3                          */
    field_elt_sqr(T7 ,X2, k, stack);            /* T7 = T1^2 = y0^2*Z^6                         */
    field_elt_mul(X1, A, T7, k, stack);        /* T0 = T4*T7 = x0*y0^2*Z^8                     */
    field_elt_mul2(X1, X1, k);
    field_elt_mul2(X1, X1, k);                  /* T0 = 4*T0 = 4*x0*y0^2*Z^8 = X1 =S            */
    field_elt_sqr(T, T7, k, stack);            /* T6 = T7^2 = y0^4*Z^12                        */
    field_elt_mul2(T, T, k);
    field_elt_mul2(T, T, k);
    field_elt_mul2(T, T, k);                  /* T6 = 8*T6 = 8*y0^4*Z^12= N^~ =T              */
    field_elt_sqr(S, A, k, stack);            /* T5 = T4^2 = x0^2*Z^4                         */
    field_elt_mul2(A, S, k);                  
    field_elt_add(A, A, S, k);               /* T4 = 3*T5 = 3*x0^2*Z^4                       */
    field_elt_sqr(S, L, k, stack);            /* T5 = T3^2 = Z^4                              */
    field_elt_mul(T7, S, E->a, k, stack);      /* T7 = T5*a = a*Z^4                            */
    field_elt_add(T7, A, T7, k);               /* T7 = T4 + T7 = 3*x0^2*Z^4 + a*Z^4            */
    field_elt_sqr(X2, T7, k, stack);            /* T1 = T7^2 = (3*x0^2*Z^4 + a*Z^4)^2           */
    field_elt_sub(X2, X2, X1, k);               /* T1 = T1-T0=(3*x0^2*Z^4 + a*Z^4)^2-X1         */
    field_elt_sub(X2, X2, X1, k);               /* T1 = T1-T0=(3*x0^2*Z^4 + a*Z^4)^2-2X1 = X2   */
    field_elt_sub(K, X1, X2, k);               /* T2 = T0-T1= X1 -X2                           */
    field_elt_mul(A, T7, K, k, stack);        /* T4 = T7*T2 = (3*x0^2*Z^4 + a*Z^4)(X1-X2)     */
    field_elt_sub(S, T, A, k);               /* T5 = T6 - T4 = N                             */
    field_elt_mul(L, S, T, k, stack);        /* T3 = T5+*T6 = N * N^~                        */
    field_elt_mul2(L, L, k);                  /* T3 = 2*N * N^~ = L                           */
    if(bit)
    {
        field_elt_mul(A, S, K, k, stack);    /* T4 = (X1-X2)*N                               */
    }
    else
    {
        field_elt_mul(A, T, K, k, stack);    /* T4 = N^~(X1-X2)                              */
    }    
    field_elt_mul2(A, A, k);                  /* T4 = A = 2(X1-X2)*N / 2 N^~(X1-X2)           */
    if(bit)
    {    
        field_elt_sqr(K, S, k, stack);        /* K = T5^2 = N^2                              */
    }
    else
    {
        field_elt_sqr(K, T, k, stack);        /* T2 = T6^2 =N^~^2                             */
    }
    field_elt_mul2(K, K ,k);                  /* T2 = K                                       */
    
    field_elt_copy(S ,X1 ,k);

    field_elt_relax_pool_elt(&T7, k, stack);	
}

void update_Kim(fe_ptr X1, fe_ptr X2, fe_ptr K, fe_ptr L, fe_ptr A, fe_ptr S, fe_ptr T, bool prev, bool new, ec_curve_srcptr E, uint8_t stack)
{
    /* From the work Speeding up Elliptic Curve Scalar Multiplication without Precomputation by Kim, Choe, Kim, Kim and Hong 2017 */
    /* Algorithm from section 4 8M +4S +15.5A */
    field_t *k = E->k;    
    field_elt T7;
    field_elt_get_pool_elt(&T7, k, stack);

    if(prev) 
    {
        /* We work like if u=1 in the formula of section 4 */
        
        field_elt_sub(X1, X2, X1, k);               /* T0 = T1-T0           */
        field_elt_sub(X2, S, X2, k);               /* T1 = T5-T1           */
        field_elt_sqr(T7, X1, k, stack);            /* T7 = T0^2            */
        field_elt_mul(X1, X2, T7, k, stack);        /* T0 = T1*T7           */
        field_elt_sqr(X2, A, k, stack);            /* T1 = T4^2            */
        field_elt_mul(T7, X2, S, k, stack);        /* T7 = T1*T5           */
        field_elt_copy(S, T7, k);                  /* T5 = T7              */
        field_elt_mul(T7, A, T, k, stack);        /* T7 = T4*T6           */
        field_elt_mul(T, X2, T7, k, stack);        /* T6 = T1*T7           */
        field_elt_mul(A, K, L, k, stack);        /* T4 = T2*T3           */
        field_elt_add(K, K, L, k);               /* T2 = T2+T3           */
        field_elt_add(L, X1, K, k);               /* T3 = T0+T2           */
        field_elt_mul2(A, A, k);                  
        field_elt_mul2(A, A, k);                  /* T4 = 4*T4            */
        field_elt_add(X1, A, S, k);               /* T0 = T4+T5           */
        field_elt_sqr(A, L, k, stack);            /* T4 = T3^2            */
        field_elt_sub(A, A, S, k);               /* T4 = T4 - T5         */
        field_elt_sub(X2, A, X1, k);               /* T1 = T4 - T0         */
        if(new)                                    /* v = 1                */
        {
            field_elt_sub(A, X2, S, k);           /* T4 = T1 -T5          */
        }
        else
        {   
            field_elt_sub(A, X1, S, k);           /* T4 = T0 - T5         */
        }     
        field_elt_sub(T7, X2, X1, k);               /* T7 = T1 - T0         */
        field_elt_mul(K, L, A, k, stack);        /* T2 = T3 * T4         */
        field_elt_sub(K, K, T, k);               /* T2 = T2 - T6         */
        field_elt_mul(A, K, T7, k, stack);        /* T4 = T2 * T7         */
        field_elt_mul2(A, A, k);                  /* T4 = 2*T4            */
        field_elt_sqr(T7, K, k, stack);            /* T7 = T2^2            */
        field_elt_mul2(K, T7, k);                  /* T2 = 2*T7            */
        field_elt_mul(T7, L, A, k, stack);        /* T7 = T3 * T4         */
        field_elt_neg(L, T7, k);                   /* T3 =-T7              */
        if(!new)
        {
            field_elt_sub(L, K, L, k);
        }
        else
        {
            field_elt_sub(L, K, T7, k);
        }          
    }
    else
    {
        /* We work like if u=0 in the formula of section 4*/
        
        field_elt_sub(X2, X1, X2, k);               /* T1 = T0-T1           */
        field_elt_sub(X1, S, X1, k);                /* T0 = T5-T0           */
        field_elt_sqr(T7, X2, k, stack);            /* T7 = T1^2            */
        field_elt_mul(X2, X1, T7, k, stack);        /* T1 = T0*T7           */
        field_elt_sqr(X1, A, k, stack);             /* T0 = T4^2            */
        field_elt_mul(T7, X1, S, k, stack);         /* T7 = T0*T5           */
        field_elt_copy(S, T7, k);                   /* T5 = T7              */
        field_elt_mul(T7, A, T, k, stack);          /* T7 = T4*T6           */
        field_elt_mul(T, X1, T7, k, stack);         /* T6 = T0*T7           */
        field_elt_mul(A, K, L, k, stack);           /* T4 = T2*T3           */
        field_elt_add(K, K, L, k);                  /* T2 = T2+T3           */
        field_elt_add(L, X2, K, k);                 /* T3 = T1*T2           */
        field_elt_mul2(A, A, k);                  
        field_elt_mul2(A, A, k);                    /* T4 = 4*T4            */
        field_elt_add(X2, A, S, k);                 /* T1 = T4+T5           */
        field_elt_sqr(A, L, k, stack);              /* T4 = T3^2            */
        field_elt_sub(A, A, S, k);                  /* T4 = T4 - T5         */
        field_elt_sub(X1, A, X2, k);                /* T0 = T4 - T1         */
        if(new)                                     /* v = 1                */
        {
            field_elt_sub(A, S, X2, k);             /* T4 = T5 -T1          */
        }
        else
        {   
            field_elt_sub(A, S, X1, k);             /* T4 = T5 - T0         */
        }     
        field_elt_sub(T7, X2, X1, k);               /* T7 = T1 - T0         */
        field_elt_mul(K, L, A, k, stack);           /* T2 = T3 * T4         */
        field_elt_sub(K, K, T, k);                  /* T2 = T2 - T6         */
        field_elt_mul(A, K, T7, k, stack);          /* T4 = T2 * T7         */
        field_elt_mul2(A, A, k);                    /* T4 = 2*T4            */
        field_elt_sqr(T7, K, k, stack);             /* T7 = T2^2            */
        field_elt_mul2(K, T7, k);                   /* T2 = 2*T7            */
        field_elt_mul(T7, L, A, k, stack);          /* T7 = T3 * T4         */
        field_elt_neg(L, T7, k);                    /* T3 =-T7              */
        if(new)
        {
            field_elt_sub(L, K, L, k);
        }
        else
        {
            field_elt_sub(L, K, T7, k);
        }    
    }
    
    field_elt_relax_pool_elt(&T7, k, stack);
}

void recovery_Kim(fe_ptr P3x, fe_ptr P3y, fe_ptr X1, fe_ptr X2, fe_ptr K, fe_ptr L, fe_ptr A, fe_ptr S, fe_ptr T, fe_srcptr x0, fe_srcptr y0, field_ptr k, bool safe, uint8_t stack)
{
    /* From the work Speeding up Elliptic Curve Scalar Multiplication without Precomputation by Kim, Choe, Kim, Kim and Hong 2017*/
    /* Algorithm from section 4 costs M + S + A */
    field_elt T7;
    field_elt_get_pool_elt(&T7, k, stack);
  

    field_elt_sub(X2, X1, X2, k);               /* T1 = T0 - T1 = X1-X2             */
    field_elt_mul(K, L, X2, k, stack);          /* T2 = T3 * T1 = K * (X1-X2)       */
    field_elt_mul(T7, A, T, k, stack);          /* T7 = T4 * T6 = A * T             */
    field_elt_mul(T, T7, S, k, stack);          /* T6 = T7 * T5 = A * T *S          */
    
    if(safe)
    {
        /* FLT */
        number tmp; number_tmp_alloc(&tmp, k->size, stack);
        field_get_size(tmp,k);
        number_dec(tmp, tmp);
        number_dec(tmp, tmp);
        field_elt_pow_number(X2, T, tmp, k, stack);
        number_tmp_free(&tmp, k->size, stack);
    }
    else
    {    
        field_elt_inv(X2, T, k, stack);         /* T1 = 1 / T6 = 1 /(A*T*S)         */       
    }
    field_elt_mul(L, X2, T7, k, stack);         /* T3 = T1 * T7 = (1)/(S)           */
    field_elt_mul(T7, L, X1, k, stack);         /* T7 = T3 * T0 = X1/(S)            */
    field_elt_mul(X1, T7, x0, k, stack);        /* T0 = T7 * x0 = x0X1/(S)          */
    field_elt_mul(A, S, X2, k, stack);          /* T4 = T5 * T1 = 1/(A*T)           */
    field_elt_mul(L, A, K, k, stack);           /* T3 = T4 * T2 = K(X1-X2)/(AT)     */
    field_elt_mul(X2, L, y0, k, stack);         /* T1 = T3 * y0 = y0K(X1-X2)/(AT)   */

    field_elt_copy(P3x, X1, k);
    field_elt_copy(P3y, X2, k);
    
    field_elt_relax_pool_elt(&T7, k, stack);
}

void
weierstrass_Zmontgomery_Kim(ec_point_ptr P3, number_srcptr n, ec_point_srcptr P1, 
        ec_curve_srcptr E, uint8_t stack)
{
    /* From the work Speeding up Elliptic Curve Scalar Multiplication without Precomputation by Kim, Choe, Kim, Kim and Hong 2017*/
    /* As in weierstrass_Zmontgomery_ladder
    *  The Montgomery ladder permits to have at each ladder R0-R1=+/-P, 
    *  therefore if we want to avoid to have in input of ZADDU with z 
    *  coordinate=0 we must avoid that R0=R1 or R0=-R1 in input of ZADDC
    *  we have to solve the following equation +/-[k]P = [k +/-1] P
    *  which simplifies to 2k = +/-1 mod p with p the order of P
    *  Therefore we obtain that the case k = p-1/2 must be avoided
    *  which forbids the input n=p-1 for this function. 
    *  In addition here it is not permitted to obtain a final result of the
    *  neutral point since recovery_Kim will divide the values by the 
    *  Z-coordinate (equals to 0 in neutral point)
    */
    /* Algorithm 2  section 4*/
    ec_point_set_neutral(P3, E, stack);
    if(!number_iszero(n))
    {
        field_t *k = E->k;
        field_elt X1, X2, K, L, A, S, T;
        field_elt_get_pool_elt(&X1, k, stack);
        field_elt_get_pool_elt(&X2, k, stack);
        field_elt_get_pool_elt(&K, k, stack);
        field_elt_get_pool_elt(&L, k, stack);
        field_elt_get_pool_elt(&A, k, stack);
        field_elt_get_pool_elt(&S, k, stack);
        field_elt_get_pool_elt(&T, k, stack);

        char n_str[k->bit_size+4];
        number_to_bit_string(n_str, n);

        setup_Kim(X1, X2, K, L, A, S, T, P1->x, P1->y, n_str[1]=='1', E, stack);
      
        uint32_t i;
        //printf("\nstrlen(n_str):%ld\n", strlen(n_str));
        for(i = 1; i < strlen(n_str)-1; i++)
        {
            if(n_str[i] == '0')
            {
                if(n_str[i+1]== '0')
                {
                    update_Kim(X1, X2, K, L, A, S, T, false, false, E, stack);	                        
                }
                else
                {
                    update_Kim(X1, X2, K, L, A, S, T, false, true, E, stack);        
                }
            }
            else
            {
                if(n_str[i+1]== '0')
                {
                    update_Kim(X1, X2, K, L, A, S, T, true, false, E, stack);	                        
                }
                else
                {
                    update_Kim(X1, X2, K, L, A, S, T, true, true, E, stack);        
                }                       
            }
        }
        if(n_str[strlen(n_str)-1]=='0')
        {
            update_Kim(X1, X2, K, L, A, S, T, false, true, E, stack);                  
        }
        else
        {
            update_Kim(X1, X2, K, L, A, S, T, true, true, E, stack);
        }
        
        recovery_Kim(P3->x, P3->y, X1, X2, K, L, A, S, T, P1->x, P1->y, E->k,false, stack);
        field_elt_set_one(P3->z, k);      

        field_elt_relax_pool_elt(&X1, k, stack);
        field_elt_relax_pool_elt(&X1, k, stack);
        field_elt_relax_pool_elt(&K, k, stack);
        field_elt_relax_pool_elt(&L, k, stack);
        field_elt_relax_pool_elt(&A, k, stack);
        field_elt_relax_pool_elt(&S, k, stack);
        field_elt_relax_pool_elt(&T, k, stack);
    }
}


void
weierstrass_Zjoye_mul(ec_point_ptr P3, number_srcptr n, ec_point_srcptr P1, 
        ec_curve_srcptr E, uint8_t stack)
{
    ec_point_set_neutral(P3, E, stack);
    if(!number_iszero(n))
    {
        /* From Co-Z Addition Formulæ and Binary Ladders on Elliptic Curves, algorithm 10 */
        field_t *k = E->k;
        ec_point R0, R1;
        ec_point_get_pool_elt(R0, k, stack);
        ec_point_get_pool_elt(R1, k, stack);

        int32_t i;
        char n_str[k->bit_size+4];
        number_to_bit_string(n_str, n);

        if(n_str[strlen(n_str)-1] == '1')
        {
            if(n_str[strlen(n_str)-2] == '0')
            {
                ec_point_copy(R0, P1, k);
                weierstrass_point_TPLU(R1, R0, R0, E, stack);
            }
            else
            {
                ec_point_copy(R1, P1, k);
                weierstrass_point_TPLU(R0, R1, R1, E, stack);
            }

            for(i = strlen(n_str) - 3; i >=0; i--)
            {
                if(n_str[i] == '0')
                {
                    weierstrass_point_add_ZDAU(R1, R1, R0, E, stack);
                }
                else
                {
                    weierstrass_point_add_ZDAU(R0, R0, R1, E, stack);
                }
            }
            ec_point_copy(P3, R0, k);
        }
        else
        {
            if(n_str[strlen(n_str)-2] == '0')
            {
                ec_point_copy(R0, P1, k);
                weierstrass_point_TPLU(R1, R0, R0, E, stack);
            }
            else
            {
                ec_point_copy(R1, P1, k);
                weierstrass_point_TPLU(R0, R1, R1, E, stack);
            }

            for(i = strlen(n_str) - 3; i >=0; i--)
            {
                if(n_str[i] == '0')
                {
                    weierstrass_point_add_ZDAU(R1, R1, R0, E, stack);
                }
                else
                {
                    weierstrass_point_add_ZDAU(R0, R0, R1, E, stack);
                }
            }
            /* Remove 1 x P1 that we add */
            weierstrass_point_neg(R1, P1, E);
            weierstrass_point_add_jacobian_dedicated(P3, R0, R1, E, stack);
        }
        ec_point_relax_pool_elt(R0, k, stack);
        ec_point_relax_pool_elt(R1, k, stack);
    }
}

/** --> Interface */

void 
weierstrass_point_add_dedicated (ec_point_ptr P3, ec_point_srcptr P1, ec_point_srcptr P2,
        ec_curve_srcptr E, uint8_t stack)
{
    if(field_elt_iszero(P1->z, E->k))
    {
        ec_point_copy(P3, P2, E->k);
    }
    else if(field_elt_iszero(P2->z, E->k))
    {
        ec_point_copy(P3, P1, E->k);
    }
    else
    {
        /* P1 and P2 != infinity */
        switch(E->f)
        {
            case PROJECTIVE : 
                weierstrass_point_add_projective_dedicated(P3, P1, P2, E, stack);
                break;

            case JACOBIAN :
                weierstrass_point_add_jacobian_dedicated(P3, P1, P2, E, stack);
                break;
            default :
                break;
        }
    }
}

void 
weierstrass_point_add_unified (ec_point_ptr P3, ec_point_srcptr P1, ec_point_srcptr P2,
        ec_curve_srcptr E, uint8_t stack)
{
    if(field_elt_iszero(P1->z, E->k))
    {
        ec_point_copy(P3, P2, E->k);
    }
    else if(field_elt_iszero(P2->z, E->k))
    {
        ec_point_copy(P3, P1, E->k);
    }
    else
    {
        /* P1 and P2 != infinity */
        switch(E->f)
        {
            case PROJECTIVE : 
                weierstrass_point_add_projective_unified(P3, P1, P2, E, stack);
                break;

            case JACOBIAN :
                /* Not implemented because the COZ arithmetic is anyhow faster in this case */
                break;
            default :
                break;
        }
    }
}

void 
weierstrass_point_dbl_dedicated (ec_point_ptr P3, ec_point_srcptr P1, ec_curve_srcptr E, uint8_t stack)
{
    if(field_elt_iszero(P1->y, E->k) || field_elt_iszero(P1->z, E->k))
    {
        /* Then 2 P1 = point at the infinity */
        weierstrass_point_set_neutral(P3, E, stack);
    }
    else
    {
        /* P1 != infinity, 2*P1 != infinity */
        switch(E->f)
        {
            case PROJECTIVE : 
                weierstrass_point_dbl_projective_dedicated(P3, P1, E, stack);
                break;

            case JACOBIAN :
                weierstrass_point_dbl_jacobian_dedicated(P3, P1, E, stack);
                break;
            default :
                break;
        }
    }
}

void 
weierstrass_point_neg (ec_point_ptr P3, ec_point_srcptr P1, ec_curve_srcptr E)
{
    field_t *k = E->k;
    field_elt_copy(P3->x, P1->x, k);
    field_elt_neg(P3->y, P1->y, k);
    field_elt_copy(P3->z, P1->z, k);
    field_elt_copy(P3->t, P1->t, k);
}

/** --> Miscellaneaous */

void 
weierstrass_point_of_order_2 (ec_point_ptr dst, ec_curve_srcptr E, uint8_t stack)
{
    ec_point Q;
    field_elt c, d, e, f, g, c2, d2, e2, f2, non_residue;  
    field_t *k = E->k;
    field k2;
    int i=0;
    ec_point_get_pool_elt(Q, k, stack);
    field_elt_get_pool_elt(&c, k, stack);
    field_elt_get_pool_elt(&d, k, stack);
    field_elt_get_pool_elt(&e, k, stack);
    field_elt_get_pool_elt(&f, k, stack);
    field_elt_get_pool_elt(&g, k, stack);
    field_elt_get_pool_elt(&non_residue, k, stack);    
    
    field_find_nonsquare(non_residue, k, stack);
    field_alloc(k2, FP2, E->k->size, k);  
    field_create(k2, "", stack, 2, k, non_residue);

    field_elt_get_pool_elt(&c2, k2, stack);
    field_elt_get_pool_elt(&d2, k2, stack);
    field_elt_get_pool_elt(&e2, k2, stack);
    field_elt_get_pool_elt(&f2, k2, stack);
                   

    MPHELL_ASSERT( E->type == WEIERSTRASS, "This curve is not a Weierstrass Curve \n");    
    /* First we compute the abscissa of a point of order 2 of the Curve */
    /* To obtain such a point we solve the cubic equation x^3 + E->a*x + E->b */
    /* We first use Cardano's formulas */
    field_elt_set_one(c, k);
    field_elt_inc(c, c, k);
    field_elt_div(d, E->b, c, k, stack);                        /* d = B / 2                                                        */
    field_elt_inc(c, c, k);
    field_elt_div(c, E->a, c, k, stack);                        /* c = A / 3                                                        */
    field_elt_sqr(e, c, k, stack);                              /* e = (A / 3)^2                                                    */
    field_elt_mul(c, c, e, k, stack);                           /* c = (A / 3)^3                                                    */
    field_elt_sqr(f, d, k, stack);                              /* f = (B/2)^2                                                      */
    field_elt_add(c, c, f, k);                                  /* c = (A / 3)^3 + (B/2)^2                                          */

    if(field_elt_issquare(c, k, stack)==false)
    {
        /* Cardano's formulas do not apply in k */
        /* We search for a simple root of the cubic equation x^3 + A x + B */
        switch(k->type)
        {
            case FP:
                /* We have to create first the quadratic extension FP2 of FP in which we will do the intermediate computations */
                field_elt_set_ui(e, 0, true, k, stack);
                field_elt_set_fp_elts(c2, k2, 2, c, e);
                field_elt_set_fp_elts(d2, k2, 2, d, e);
                field_elt_sqrt(c2, c2, k2, stack);              /* c = ((A / 3)^3 + (B/2)^2) ^(1/2)                                 */
                
                field_elt_neg(d2, d2, k2);                      /* d = -B / 2 */
                field_elt_sub(e2, d2, c2, k2);                  /* e = -B / 2 - ((A / 3)^3 + (B/2)^2) ^(1/2)                        */
                field_elt_add(f2, d2, c2, k2);                  /* f = -B / 2 + ((A / 3)^3 + (B/2)^2) ^(1/2)                        */
                                
                MPHELL_ASSERT( field_elt_ispower_ui(e2, 3, k2, stack), "This curve do not seem to have points of order 2 defined  \n");
                MPHELL_ASSERT( field_elt_ispower_ui(f2, 3, k2, stack), "This curve do not seem to have points of order 2 defined  \n");
                field_elt_cube_root(e2, e2, k2, stack);         /* S = (-B / 2 - ((A / 3)^3 + (B/2)^2) ^(1/2))^(1/3) (stored in e)  */
                field_elt_cube_root(f2, f2, k2, stack);         /* T = (-B / 2 + ((A / 3)^3 + (B/2)^2) ^(1/2))^(1/3) (stored in f)  */
                field_elt_add(c2, e2, f2, k2);                  /* c = S + T                                                        */    
                field_elt_set_one(d, k);
                field_elt_dec(d, d, k);
                field_elt_get_fp_elt(c, c2, 2, k2);
                field_elt_set_one(d2, k2);
                field_elt_unity_nth_root(d2, 3, k2, stack);                 
                if(!field_elt_iszero(c, k))
                {
                    /* In this case we search for a solution defined in FP */    
                    field_elt_mul(f2, f2, d2, k2, stack);
                    field_elt_add(c2, e2, f2, k2);
                    field_elt_get_fp_elt(c, c2, 2, k2);
                    if(!field_elt_iszero(c, k))
                    {
                        field_elt_mul(f2, f2, d2, k2, stack);
                        field_elt_add(c2, e2, f2, k2);
                    }
                }
                
                /*We update (choice of cubic root) the value of c2 to get a point of order 2 */
                do
                {
                    i=i+1;
                    field_elt_get_fp_elt(c, c2, 1, k2);
                    field_elt_sqr(e, c, k, stack);
                    field_elt_add(e, E->a, e, k);
                    field_elt_mul(e, e, c, k, stack);
                    field_elt_add(e, e, E->b, k);               /* e = c^3 + a c + b                                                */
                    field_elt_mul(c2, c2, d2, k2, stack);
                }while(!field_elt_iszero(e, k) && i<3);
                MPHELL_ASSERT( field_elt_iszero(e, k), "\nThere is no point of order 2 defined in FP  \n");
                break;
            case FP2:
            case FP3:
                MPHELL_ASSERT( k->type==FP, "We have to work in an a quadratic extension and it is not implemented for non FP field type  \n");
                break;
        }
    }
    else
    {
        /* Cardano's formula apply we use them  */
        field_elt_sqrt(c, c, k, stack);                         /* c = ((A / 3)^3 + (B/2)^2) ^(1/2)                                 */
        field_elt_neg(d, d, k);                                 /* d = -B / 2                                                       */
        field_elt_sub(e, d, c, k);                              /* e = -B / 2 - ((A / 3)^3 + (B/2)^2) ^(1/2)                        */
        field_elt_add(f, d, c, k);                              /* f = -B / 2 + ((A / 3)^3 + (B/2)^2) ^(1/2) */
        MPHELL_ASSERT(field_elt_ispower_ui(e, 3, k, stack), "This curve do not seem to have points of order 2 defined  \n");
        MPHELL_ASSERT(field_elt_ispower_ui(f, 3, k, stack), "This curve do not seem to have points of order 2 defined  \n");
        field_elt_cube_root(e, e, k, stack);                    /* S = (-B / 2 - ((A / 3)^3 + (B/2)^2) ^(1/2))^(1/3) (stored in e)  */
        field_elt_cube_root(f, f, k, stack);                    /* T = (-B / 2 + ((A / 3)^3 + (B/2)^2) ^(1/2))^(1/3) (stored in f)  */
        field_elt_unity_nth_root(d, 3, k, stack); 
        /* In this do while loop we update (choice of cubic root) the value of c to get a point of order 2 */
        
        do
        {
            i=i+1;
            field_elt_add(c, e,f, k);                           /* c = S + T                                                        */
            field_elt_sqr(g, c, k, stack);
            field_elt_add(g, E->a, g, k);
            field_elt_mul(g, g, c, k, stack);
            field_elt_add(g, g, E->b, k);                       /* g = c^3 + a c + b                                                */
            field_elt_mul(e, e, d, k, stack);
        }while(i<3 && !field_elt_iszero(g, k)); 
        field_elt_set_one(d, k);
        field_elt_dec(d, d, k);    
        
    }

    weierstrass_point_set_aff(Q, c, d, k);
    /* We verify that this point computed belongs to the curve*/    
    MPHELL_ASSERT(weierstrass_belongs(Q, E, stack),"The Point of order 2 computed does not belong to the Weierstrass curve E ! \n");
    ec_point_copy(dst, Q, k);

    field_elt_relax_pool_elt(&non_residue, k, stack); 
    field_elt_relax_pool_elt(&c, k, stack);
    field_elt_relax_pool_elt(&d, k, stack);
    field_elt_relax_pool_elt(&e, k, stack);
    field_elt_relax_pool_elt(&f, k, stack);
    field_elt_relax_pool_elt(&g, k, stack);
    field_elt_relax_pool_elt(&c2, k2, stack);
    field_elt_relax_pool_elt(&d2, k2, stack);
    field_elt_relax_pool_elt(&e2, k2, stack);
    field_elt_relax_pool_elt(&f2, k2, stack);
    field_free(k2);
    ec_point_relax_pool_elt(Q, k, stack);
}

void 
weierstrass_points_of_order_2 (ec_point_ptr dst1, ec_point_ptr dst2, ec_point_ptr dst3, ec_curve_srcptr E, uint8_t stack)
{
    ec_point Q;
    field_elt c, d, e, f, g, c2, d2, e2, f2, non_residue;  
    field_t *k = E->k;
    field k2;
    int i=0;
    ec_point_get_pool_elt(Q, k, stack);
    field_elt_get_pool_elt(&c, k, stack);
    field_elt_get_pool_elt(&d, k, stack);
    field_elt_get_pool_elt(&e, k, stack);
    field_elt_get_pool_elt(&f, k, stack);
    field_elt_get_pool_elt(&g, k, stack);
    field_elt_get_pool_elt(&non_residue, k, stack);
    
    field_find_nonsquare(non_residue, k, stack);
    field_alloc(k2, FP2, E->k->size, k);  
    field_create(k2, "", stack, 2, k, non_residue);

    field_elt_get_pool_elt(&c2, k2, stack);
    field_elt_get_pool_elt(&d2, k2, stack);
    field_elt_get_pool_elt(&e2, k2, stack);
    field_elt_get_pool_elt(&f2, k2, stack);

    MPHELL_ASSERT(E->type == WEIERSTRASS, "This curve is not a Weierstrass Curve \n");    
    /* First we compute the abscissa of a point of order 2 of the Curve */
    /* To obtain such a point we solve the cubic equation x^3 + E->a*x + E->b */
    /* We first use Cardano's formulas */
    field_elt_set_one(c, k);
    field_elt_inc(c, c, k);
    field_elt_div(d, E->b, c, k, stack);                        /* d = B / 2                                                        */
    field_elt_inc(c, c, k);
    field_elt_div(c, E->a, c, k, stack);                        /* c = A / 3                                                        */
    field_elt_sqr(e, c, k, stack);                              /* e = (A / 3)^2                                                    */
    field_elt_mul(c, c, e, k, stack);                           /* c = (A / 3)^3                                                    */
    field_elt_sqr(f, d, k, stack);                              /* f = (B/2)^2                                                      */
    field_elt_add(c, c, f, k);                                  /* c = (A / 3)^3 + (B/2)^2                                          */
    
    if(field_elt_issquare(c, k, stack)==false)
    {
        /* Cardano's formulas do not apply in k */
        /* We search for a simple root of the cubic equation x^3 + A x + B */
        switch(k->type)
        {
            case FP:
                /* We have to create first the quadratic extension FP2 of FP in which we will do the intermediate computations */
                field_elt_set_ui(e, 0, true, k, stack);
                field_elt_set_fp_elts(c2, k2, 2, c, e);
                field_elt_set_fp_elts(d2, k2, 2, d, e);

                field_elt_sqrt(c2, c2, k2, stack);              /* c = ((A / 3)^3 + (B/2)^2) ^(1/2)                                 */           
                field_elt_neg(d2, d2, k2);                      /* d = -B / 2 */
                field_elt_sub(e2, d2, c2, k2);                  /* e = -B / 2 - ((A / 3)^3 + (B/2)^2) ^(1/2)                        */
                field_elt_add(f2, d2, c2, k2);                  /* f = -B / 2 + ((A / 3)^3 + (B/2)^2) ^(1/2)                        */
      
                MPHELL_ASSERT( field_elt_ispower_ui(e2, 3, k2, stack), "This curve do not seem to have points of order 2 defined  \n");
                MPHELL_ASSERT( field_elt_ispower_ui(f2, 3, k2, stack), "This curve do not seem to have points of order 2 defined  \n");
                field_elt_cube_root(e2, e2, k2, stack);         /* S = (-B / 2 - ((A / 3)^3 + (B/2)^2) ^(1/2))^(1/3) (stored in e)  */
                field_elt_cube_root(f2, f2, k2, stack);         /* T = (-B / 2 + ((A / 3)^3 + (B/2)^2) ^(1/2))^(1/3) (stored in f)  */
                field_elt_add(c2, e2, f2, k2);                  /* c = S + T                                                        */
                
                field_elt_get_fp_elt(c, e2, 2, k2);
                field_elt_get_fp_elt(d, f2, 2, k2);
                field_elt_set_one(d2, k2);
                field_elt_unity_nth_root(d2, 3, k2, stack); 
                while(field_elt_cmp(c, d, k)!=0 && i<9)
                {
                    i=i+1;
                    if(i%3!=0)
                    {
                        field_elt_mul(e2, e2, d2, k2, stack);
                    }
                    else
                    {
                        field_elt_mul(f2, f2, d2, k2, stack);
                    }
                    field_elt_add(c2, e2, f2, k2);                  /* c = S + T                                                        */
                    field_elt_get_fp_elt(c, e2, 2, k2);
                    field_elt_get_fp_elt(d, f2, 2, k2);
                }                    
                i=0;
                field_elt_set_one(d, k);
                field_elt_dec(d, d, k); 
                field_elt_get_fp_elt(c, c2, 2, k2);            

                /*We update (choice of cubic root) the value of c2 to get a point of order 2 */
                do
                {
                    i=i+1;
                    field_elt_get_fp_elt(c, c2, 1, k2);
                    field_elt_sqr(e, c, k, stack);
                    field_elt_add(e, E->a, e, k);
                    field_elt_mul(e, e, c, k, stack);
                    field_elt_add(e, e, E->b, k);               /* e = c^3 + a c + b                                                */
                    field_elt_mul(c2, c2, d2, k2, stack);
                }while(!field_elt_iszero(e, k) && i<3);
                MPHELL_ASSERT( field_elt_iszero(e, k), "\nThere is no point of order 2 defined in FP  \n");
                break;
            case FP2:
            case FP3:
                MPHELL_ASSERT(k->type==FP, "We have to work in an a quadratic extension and it is not implemented for non FP field type  \n");
                break;
        }
    }
    else
    {
        /* Cardano's formula apply we use them  */
        field_elt_sqrt(c, c, k, stack);                         /* c = ((A / 3)^3 + (B/2)^2) ^(1/2)                                 */
        field_elt_neg(d, d, k);                                 /* d = -B / 2                                                       */
        field_elt_sub(e, d, c, k);                              /* e = -B / 2 - ((A / 3)^3 + (B/2)^2) ^(1/2)                        */
        field_elt_add(f, d, c, k);                              /* f = -B / 2 + ((A / 3)^3 + (B/2)^2) ^(1/2)                        */
        MPHELL_ASSERT(field_elt_ispower_ui(e, 3, k, stack), "This curve do not seem to have points of order 2 defined  \n");
        MPHELL_ASSERT(field_elt_ispower_ui(f, 3, k, stack), "This curve do not seem to have points of order 2 defined  \n");
        field_elt_cube_root(e, e, k, stack);                    /* S = (-B / 2 - ((A / 3)^3 + (B/2)^2) ^(1/2))^(1/3) (stored in e)  */
        field_elt_cube_root(f, f, k, stack);                    /* T = (-B / 2 + ((A / 3)^3 + (B/2)^2) ^(1/2))^(1/3) (stored in f)  */
        field_elt_unity_nth_root(d, 3, k, stack); 
        /* In this do while loop we update (choice of cubic root) the value of c to get a point of order 2 */
        do 
        {
            i=i+1;
            field_elt_add(c, e, f, k);                          /* c = S + T                                                        */
            field_elt_sqr(g, c, k, stack);
            field_elt_add(g, E->a, g, k);
            field_elt_mul(g, g, c, k, stack);
            field_elt_add(g, g, E->b, k);                       /* g = c^3 + a c + b                                                */
            field_elt_mul(e, e, d, k, stack);
        }while(i<3 && !field_elt_iszero(g, k)); 
        field_elt_set_one(d, k);
        field_elt_dec(d, d, k);        
    }
    weierstrass_point_set_aff(Q, c, d, k);
    /* We verify that this point computed belongs to the curve*/    
    MPHELL_ASSERT(weierstrass_belongs(Q, E, stack),"The Point of order 2 computed does not belong to the Weierstrass curve E ! \n");

    ec_point_copy(dst1, Q, k);
    /* In this case we compute the two others points of order 2 assuming they exist */
    /* Fist we compute the discriminant of the quadratic equation */
    field_elt_sqr(e, c, k, stack);                              /* e = alpha^2                                                      */
    field_elt_add(d, e, E->a, k);                               /* d = alpha^2 + a                                                  */
    field_elt_add(g, d, d, k);
    field_elt_add(g, g, g, k);        
    field_elt_sub(f, e, g, k);                                  /* f = alpha^2 - 4 d (discriminant)                                 */
    field_elt_set_one(d, k);
    field_elt_inc(d, d, k);
    if(field_elt_issquare(f, k, stack))
    {
        /* In this case we compute the 3 points of order 2 */
        field_elt_sqrt(f, f, k, stack);                         /* f = (alpha^2 - 4 d)^(1/2)                                        */
        field_elt_add(e, f, c, k);
        field_elt_neg(e, e, k);                  
        field_elt_div(e, e, d, k, stack);                       /* e =- ( (alpha^2 - 4 d)^(1/2) + alpha ) / 2                       */
        field_elt_sub(c, f, c, k);
        field_elt_div(c, c, d, k, stack);                       /* c = ( (alpha^2 - 4 d)^(1/2) - alpha ) / 2                        */
        field_elt_set_one(d, k);
        field_elt_dec(d, d, k);
        weierstrass_point_set_aff(Q, c, d, k);       
        MPHELL_ASSERT(weierstrass_belongs(Q, E, stack),"The Point of order 2 computed does not belong to the Weierstrass curve E ! \n");
        ec_point_copy(dst2, Q, k);
        weierstrass_point_set_aff(Q, e, d, k);      
        MPHELL_ASSERT(weierstrass_belongs(Q, E, stack),"The Point of order 2 computed does not belong to the Weierstrass curve E ! \n");
        ec_point_copy(dst3, Q, k);    
    }
    else
    {
        /* Otherwise we just return 3 times the same point */
        ec_point_copy(dst2, Q, k);
        ec_point_copy(dst3, Q, k);
    }

    /* Now we free the memory */
    field_elt_relax_pool_elt(&non_residue, k, stack); 
    field_elt_relax_pool_elt(&c, k, stack);
    field_elt_relax_pool_elt(&d, k, stack);
    field_elt_relax_pool_elt(&e, k, stack);
    field_elt_relax_pool_elt(&f, k, stack);
    field_elt_relax_pool_elt(&g, k, stack);
    field_elt_relax_pool_elt(&c2, k2, stack);
    field_elt_relax_pool_elt(&d2, k2, stack);
    field_elt_relax_pool_elt(&e2, k2, stack);
    field_elt_relax_pool_elt(&f2, k2, stack);
    field_free(k2);
    ec_point_relax_pool_elt(Q, k, stack);
}
