/* 
  MPHELL-3.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France)

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell_perf_ecdsa.c
  * \brief Benchmark for ecdsa signature and verification (with precomputation) using mphell.
  */

#include <stdio.h>
#include <time.h>
#include "mphell.h"

/**
 * @struct  ecdsa_sig
 * @brief   Define an ECDSA signature
 */
struct ecdsa_sig 
{
    number *r; /*!< First signature number, belongs to [1, n-1] */
    number *s; /*!< Second signature number, belongs to [1, n-1] */
};

/**
 * @typedef  ecdsa_sig_t
 * @brief    Define the ECDSA signature type
 */
typedef struct ecdsa_sig ecdsa_sig_t;

/**
 * @typedef ecdsa_sig
 * @brief   Pointer on an allocated ecdsa_sig_t structure.
 */
typedef ecdsa_sig_t ecdsa_sig[1];

#define PRECOMP_WIN_SIZE 9
#define PRECOMP_SIZE 256
#define HASH_SIZE 64

/**
 * @struct  ecdsa_precomp
 * @brief   Precomputation structure
 */
struct ecdsa_precomp 
{
    ec_point prec_G[2*PRECOMP_SIZE]; /*!< store 0G, G, 2G, 3G, 4G, 5G, 6G, 7G, ..., (2^(31)-1) G where G is the base point */
    ec_point prec_K[2*PRECOMP_SIZE]; /*!< store 0Q, Q, 2Q, 3Q, 4Q, 5Q, 6Q, 7Q, ..., (2^(31)-1) Q where Q is the public key */
};

/**
 * @typedef ecdsa_precomp_t
 * @brief   Precomputation structure type
 */
typedef struct ecdsa_precomp ecdsa_precomp_t;

/**
 * @typedef ecdsa_precomp
 * @brief   Pointer on an allocated ecdsa_precomp_t structure.
 */
typedef ecdsa_precomp_t ecdsa_precomp[1];

/**
 * @fn     void ecdsa_sign_alloc(ecdsa_sig *sig, uint8_t size);
 * @brief  Allocate a signature structure
 * @param  sig Signature
 */
void
ecdsa_sign_alloc(ecdsa_sig *sig, uint8_t size)
{
    (*sig)->r = (number*)malloc(sizeof(number));
    (*sig)->s = (number*)malloc(sizeof(number));
    number_init(((*sig)->r), size);
    number_init(((*sig)->s), size);
    number_set_ui(*((*sig)->r), (block)0);
    number_set_ui(*((*sig)->s), (block)0);    
}

/**
 * @fn     void ecdsa_sign_free(ecdsa_sig *sig);
 * @brief  Free a used signature structure
 * @param  sig  Signature
 * @param  size Size (in block) of the signature
 */
void 
ecdsa_sign_free(ecdsa_sig *sig)
{
    if ((*sig)->r)
    {
        number_free((*sig)->r);
        free((*sig)->r);
    }
    if ((*sig)->s)
    {
        number_free((*sig)->s);
        free((*sig)->s);
    }
}

/**
 * @fn     void ecdsa_precal(ecdsa_precomp *precomp, ec_point *key, ec_curve_ptr E);
 * @brief  Compute 1*G, 2*G, 3*G, 4*G, 5*G, 6*G, 7*G and 1*Q, 2*Q, 3*Q, 4*Q, 5*Q, 6*Q, 7*Q where G is the base point and Q the public key
 * @param  precomp Pointer on an allocated ecdsa_precomp structure
 * @param  key     Public key
 * @param  E       Elliptic curve
 */
void
ecdsa_precal(ecdsa_precomp *precomp, ec_point *key, ec_curve_ptr E)
{
    ec_set_fast_dedicated_coordinates(E);
    uint16_t i;
    uint16_t pr = 2*PRECOMP_SIZE;
    
    /* Allocate */
    for(i = 0; i < pr; i++)
    {
        ec_point_alloc(((*precomp)->prec_G)[i], E->k);
        ec_point_alloc(((*precomp)->prec_K)[i], E->k);
        ec_point_init(((*precomp)->prec_G)[i], E->k);
        ec_point_init(((*precomp)->prec_K)[i], E->k);
    }
    ec_point_set_neutral(((*precomp)->prec_K)[0], E, STACK_1);
    ec_point_set_neutral(((*precomp)->prec_G)[0], E, STACK_1);
    
    ec_point_copy(((*precomp)->prec_G)[1], E->G, E->k);      /* G[0] = G     */
    ec_point_copy(((*precomp)->prec_K)[1], *key, E->k);      /* K[0] = K     */
    
    for(i = 2; i < pr; i++)
    {
        /* tab[i]=(i)* P1 */
        ec_point_add(((*precomp)->prec_G)[i], E->G, ((*precomp)->prec_G)[i-1], E, STACK_1);
        ec_point_add(((*precomp)->prec_K)[i], *key, ((*precomp)->prec_K)[i-1], E, STACK_1);
    }
}

/**
 * @fn     void ecdsa_precal_free(ecdsa_precomp *precomp, ec_curve_ptr E);
 * @brief  Free memory of the ecdsa_precomp structure
 * @param  precomp Pointer on an allocated ecdsa_precomp structure
 */
void
ecdsa_precal_free(ecdsa_precomp *precomp, ec_curve_ptr E)
{
    uint16_t i;
    uint16_t pr = 2*PRECOMP_SIZE;
    for(i = 0; i < pr; i++)
    {
        ec_point_free(((*precomp)->prec_G)[i], E->k);
        ec_point_free(((*precomp)->prec_K)[i], E->k);
    }
}

/**
 * @fn     void ecdsa_sign(unsigned char *hash, ecdsa_sig *sig, number_ptr priv_key, ec_curve *curve);
 * @brief  Sign "hash" with ECDSA algorithm using the EC "curve" and the private key "key"
 * @param  hash     H(m) where H is a cryptographic hash function and m the message to sign. Must be written under hexadecimal form. ex: 64 hexadecimal characters for sha256.
 * @param  sig      Destination signature. Pointer on an allocated ecdsa_sig structure.
 * @param  priv_key Private key
 * @param  curve    Elliptic curve
 */
void
ecdsa_sign(unsigned char *hash, ecdsa_sig *sig, number_ptr priv_key, ec_curve *curve)
{
    ec_set_fast_unified_coordinates(*curve);
    number *rr = (*sig)->r;
    number *ss = (*sig)->s;
    
    uint8_t size = (*curve)->k->size;
    
    number k; number_tmp_alloc(&k, size, STACK_1);
    number h; number_tmp_alloc(&h, (strlen((char *)hash)/(BLOCK_SIZE/4)), STACK_1);
    number t; number_tmp_alloc(&t, size, STACK_1);
    number tt; number_tmp_alloc(&tt, 2*size+1, STACK_1);
        
    number abs;
    number_tmp_alloc(&abs, size, STACK_1);
        
    ec_point x;
    ec_point_get_pool_elt(x, (*curve)->k, STACK_1);
    
    field_elt xx;
    field_elt_get_pool_elt(&xx, (*curve)->k, STACK_1);
    
    bool newk = true;
    number_set_str(h , (char*)hash, 16);
    
    while(newk)
    {
        /* Choose k randomly in [1, n-1] */
        number_random1(k, (*curve)->n, STACK_1);
            
        /* Compute (x,y) = kG  */
        ec_point_mul_unified(x, k, (*curve)->G, *curve, STACK_1);
        ec_point_get_x_affine(xx, x, *curve, STACK_1);
            
        /* r = x mod n */
        field_elt_get_number(abs, xx, 1, (*curve)->k, STACK_1);
        number_mod(*rr, abs, (*curve)->n);
                
        /* if r == 0 --> new k */
        if(!number_iszero(*rr)) 
        {
            /* s = k^-1 (hash + r*priv_key) mod n */
            number_mul(tt, *rr, priv_key);
            number_add(tt, tt, h);
            number_mod(tt, tt, (*curve)->n);            /* To keep the multiplication small */
            number_invmod(t, k, (*curve)->n);
            number_mul(tt, tt, t);
            number_mod(*ss, tt, (*curve)->n);

            /* if s == 0 --> new k */
            if(!number_iszero(*ss))
            {
                newk = false;
            } 
        }
    }
    (*sig)->r = rr;
    (*sig)->s = ss;
    
    /* Free memory */
    ec_point_relax_pool_elt(x, (*curve)->k, STACK_1);
    field_elt_relax_pool_elt(&xx, (*curve)->k, STACK_1);
    number_tmp_free(&tt, 2*size+1, STACK_1);
    number_tmp_free(&t, size, STACK_1);
    number_tmp_free(&h, (strlen((char *)hash)/(BLOCK_SIZE/4)), STACK_1);
    number_tmp_free(&k, size, STACK_1);
    number_tmp_free(&abs, size, STACK_1);
}

/**
 * @fn     int8_t ecdsa_pub_key_validation(ec_point *pub_key, ec_curve *curve);
 * @brief  Verify that the public key is valid (id est check taht pub_key belongs to the curve, that pub_key is not the neutral element and that pub_key is of ordre n) 
 * @param  hash    H(m) where H is a cryptographic hash function and m the message to sign. Must be written under hexadecimal form. ex: 64 hexadecimal characters for sha256.
 * @param  pub_key Public key
 * @param  curve   Elliptic curve
 * @return 1 on success, 0 on failure
 */
int8_t 
ecdsa_pub_key_validation(ec_point *pub_key, ec_curve *curve)
{    
    if(!ec_belongs(*pub_key, *curve, STACK_1))
    {
        /* pubkey does not belong  to the curve */
        return false;   
    }   
    if(ec_point_is_neutral(*pub_key, *curve, STACK_1))
    {
        /* pubkey is the neutral element */
        return false;
    }
    /* We finally test the order of the curve */
    ec_point x;
    ec_point_get_pool_elt(x, (*curve)->k, STACK_1);    
    ec_point_mul(x, (*curve)->n, *pub_key, *curve, STACK_1);
    if(!ec_point_is_neutral(x, *curve, STACK_1))
    {
        /* pubkey is the neutral element */
        return false;
    }
    ec_point_relax_pool_elt(x, (*curve)->k, STACK_1);
    
    return true;
}

/**
 * @fn     int8_t ecdsa_verify(unsigned char *hash, ecdsa_sig *sig, ec_point *pub_key, ec_curve *curve, ecdsa_precomp precomp);
 * @brief  Verify the signature "sig" of hash "hash" using public key "pub_key" and the EC "curve"
 * @param  hash    H(m) where H is a cryptographic hash function and m the message to sign. Must be written under hexadecimal form. ex: 64 hexadecimal characters for sha256.
 * @param  sig     Signature of hash
 * @param  pub_key Public key
 * @param  curve   Elliptic curve
 * @param  precomp Precomputation structure, used in the wNAF multiplication point method
 * @return 1 on success, 0 on failure
 */
int8_t 
ecdsa_verify(unsigned char *hash, ecdsa_sig *sig, ec_point *pub_key, ec_curve *curve, ecdsa_precomp * precomp)
{
    ec_set_fast_dedicated_coordinates(*curve);

    if( number_iszero(*((*sig)->r)) || number_cmp(*((*sig)->r),(*curve)->n)>=0 ||  number_iszero(*((*sig)->s)) || number_cmp(*((*sig)->s),(*curve)->n)>=0)
    {
        return false;
    } 
    uint8_t size = (*curve)->k->size;
    
    number w; number_tmp_alloc(&w, size, STACK_1);
    number u1; number_tmp_alloc(&u1, size, STACK_1);
    number u2; number_tmp_alloc(&u2, size, STACK_1);
    number tt2; number_tmp_alloc(&tt2, 2*size, STACK_1);
    
    number abs;
    number_tmp_alloc(&abs, size, STACK_1);
    
    ec_point x;
    ec_point_get_pool_elt(x, (*curve)->k, STACK_1);
    number h; number_tmp_alloc(&h, (strlen((char *)hash)/(BLOCK_SIZE/4)), STACK_1);
    number_set_str(h , (char*)hash, 16);
    
    field_elt xx;
    field_elt_get_pool_elt(&xx, (*curve)->k, STACK_1);

    /* w = s^-1 mod n */
    number_invmod(w, *((*sig)->s), (*curve)->n);
    
    /* u1 = hash * w mod n */
    number_mod(u1, h, (*curve)->n);
    number_mul(tt2, u1, w);
    number_mod(u1, tt2, (*curve)->n);

    /* u2 = r*w mod n */
    number_mul(tt2, *((*sig)->r), w);
    number_mod(u2, tt2, (*curve)->n);

    /* x = u1*G + u2*Q */
    ec_point_2mul_with_precomp(x, u1, (*precomp)->prec_G, u2, (*precomp)->prec_K, PRECOMP_WIN_SIZE, *curve, STACK_1);
    ec_point_get_x_affine(xx, x, *curve, STACK_1);
    
    /* Get the result, by comparing x value with r */
    field_elt_get_number(abs, xx, 1, (*curve)->k, STACK_1);
    number_mod(w, abs, (*curve)->n);
    bool result = number_cmp(w, *((*sig)->r)) == 0;

    /* Free memory */
    number_tmp_free(&h, (strlen((char *)hash)/(BLOCK_SIZE/4)), STACK_1);
    ec_point_relax_pool_elt(x, (*curve)->k, STACK_1);
    field_elt_relax_pool_elt(&xx, (*curve)->k, STACK_1);
    number_tmp_free(&tt2, 2*size, STACK_1);
    number_tmp_free(&u2, size, STACK_1);
    number_tmp_free(&u1, size, STACK_1);
    number_tmp_free(&w, size, STACK_1);
    number_tmp_free(&abs, size, STACK_1);

    return result;
}

/**
 * @struct  eddsa_sig
 * @brief   Define an EdDSA signature
 */
struct eddsa_sig 
{
    ec_point R; /*!< First signature point, belongs to the signature curve */
    number *s; /*!< Second signature number, belongs to [1, n-1] */
};

/**
 * @typedef  eddsa_sig_t
 * @brief    Define the EdDSA signature type
 */
typedef struct eddsa_sig eddsa_sig_t;

/**
 * @typedef eddsa_sig
 * @brief   Pointer on an allocated eddsa_sig_t structure.
 */
typedef eddsa_sig_t eddsa_sig[1];

/**
 * @fn     void eddsa_sign_alloc(eddsa_sig *sig, uint8_t size);
 * @brief  Allocate a signature structure
 * @param  sig Signature
 * @param  E Elliptic Curve used for the signature
 * @param  size Size (in block) of the signature
 */
void
eddsa_sign_alloc(eddsa_sig *sig, ec_curve_srcptr E, uint8_t size)
{
    (*sig)->s = (number*)malloc(sizeof(number));
    number_init((*sig)->s, size);
    number_set_ui(*((*sig)->s), (block)0);    

    ec_point_alloc((*sig)->R, E->k);
    ec_point_init((*sig)->R, E->k);
    ec_point_set_neutral((*sig)->R, E, STACK_1);
}

/**
 * @fn     void eddsa_sign_free(eddsa_sig *sig, ec_curve_srcptr E);
 * @brief  Free a used signature structure
 * @param  sig  Signature
 */
void 
eddsa_sign_free(eddsa_sig *sig, ec_curve_srcptr E)
{
    if ((*sig)->R)
    {
        ec_point_free((*sig)->R, E->k);
    }
    if ((*sig)->s)
    {
        number_free((*sig)->s);
        free((*sig)->s);
    }
}

/**
 * @fn     void eddsa_sign(unsigned char *hash, eddsa_sig *sig, number priv_key, ec_curve *curve);
 * @brief  Sign "hash" with EdDSA algorithm using the EC "curve" and the private key "key"
 * @param  hash     H(m) where H is a cryptographic hash function and m the message to sign. Must be written under hexadecimal form. ex: 64 hexadecimal characters for sha256.
 * @param  sig      Destination signature. Pointer on an allocated ecdsa_sig structure.
 * @param  priv_key Private key
 * @param  curve    Elliptic curve
 */
void
eddsa_sign(unsigned char *hash, eddsa_sig *sig, number priv_key, ec_point *pub_key, ec_curve *curve)
{
    number *ss = (*sig)->s;
    
    uint32_t i;
    char hashvalue[129];  
    char hashvaluebis[129];  
    char str[97];
    uint8_t * message_bin;
    char * strx;
    char * stry;
    char strH[321]; /* 2* 128 (representation of point in hexa) + 64 (size of the message) +1 */
    uint8_t hashbin[64];  
    int64_t len_bin;    
    number r, s, test, test2;
    number_init(&r, bits_to_nblock(256));
    number_init(&test, bits_to_nblock(256));
    number_init(&s, bits_to_nblock(256));
    number_init(&test2, bits_to_nblock(512));

    /* Computation of the public key */
    number_str(&strx, priv_key, 16);
    message_bin = (uint8_t *)calloc(strlen(strx)*4, sizeof(uint8_t));
    len_bin = hex2bin(strx, message_bin);
    
    /* Computation of the hash of the private key */
    sha512(hashbin, message_bin, len_bin*8);
    /* sha512 is used since we need to have hashes of size 512 bits, another
     * hash function could have been used under this condition, we choosed it
     * because sha512 is implemented already in MPHELL 
     */
    free(message_bin);
    
    /* Computation of s=2^n+\sum_{i=c}^{n-1}2^ih_i (h_i are bits of the hash)*/
    bin2hex(hashbin, 64, (unsigned char *)hashvalue);

    /* Computation of h_{bi}|M  with i>=b (here 256)*/
    for(i=0;i<32;i++)
    {
        str[i]=hashvalue[32+i];
    }
    for(i=32;i<96;i++)
    {
        str[i]=hash[i];
    }
    str[96]= '\0';
    message_bin =  (uint8_t *)calloc(strlen(str)*4, sizeof(uint8_t));
    len_bin= hex2bin(str, message_bin);

    /* r = H(h_{bi}|M)  */
    sha512(hashbin, message_bin, len_bin*8);
    free(message_bin);
    bin2hex(hashbin, 64, (unsigned char *)hashvaluebis);
    number_set_str(test2, hashvaluebis, 16);

    /* r mod l  (in [EdDSA15] l represents E->n)*/
    number_mod(r,test2,(*curve)->n);

    /* Computation of R=rB (in [EdDSA15] B represents E->G)*/
    ec_point_mul_unified((*sig)->R, r, (*curve)->G, (*curve), STACK_1); 
    /* Remove this for optimisation */
    ec_point_norm((*sig)->R, (*curve), STACK_1);  
    
    free(strx);

    field_elt_str(&strx, ((*sig)->R)->x, 16, false, (*curve)->k, STACK_1); 
    field_elt_str(&stry, ((*sig)->R)->y, 16, false, (*curve)->k, STACK_1);
    for(i=0;i<strlen(strx);i++)
    {
        strH[i]=strx[i];
    }
    for(i=strlen(strx);i<64;i++)
    {
        strH[i]='0';
    }
    for(i=0;i<strlen(stry);i++)
    {
        strH[i+64]=stry[i];
    }
    for(i=strlen(stry);i<64;i++)
    {
        strH[i+64]='0';
    }
    /* We have to compute again s and A */    

    /* Here we need to compute again the public key A*/
    number_set_str(test, "1", 16);
    number_lshift(test, test, 256);
    
    /* Computation of s=2^n+\sum_{i=c}^{n-1}2^ih_i */
    number_set_str(test2, hashvalue, 16);
    //number_rshift(test2, test2, 256);
    char *test3;
    number_str(&test3, test2, 2);
    test3[0]='0';
    test3[1]='1';
    test3[253]='0';
    test3[254]='0';
    test3[255]='0';
    test3[256]='\0';
    
    number_set_str(test2, test3, 2);
    free(test3);
    //number_rshift(test2, test2, 258);
    //number_lshift(test2, test2, 2);
    //if(number_cmp(test2, test)==-1)
    //{
        //number_add(test2,test2,test);
    //}    
    /* Computation of s mod l */
    number_mod(s, test2, (*curve)->n);

    /* Computation of A=sB */
    /*
    ec_point p_key;
    ec_point_get_pool_elt(p_key, (*curve)->k, STACK_1);
    ec_point_mul(p_key, s, (*curve)->G, (*curve), STACK_1);
    ec_point_norm(p_key, (*curve), STACK_1);  
    */
    
    free(strx);
    free(stry);

    field_elt_str(&strx, (*pub_key)->x, 16,false, (*curve)->k, STACK_1);
    field_elt_str(&stry, (*pub_key)->y, 16,false, (*curve)->k, STACK_1);

    for(i=0;i<strlen(strx);i++)
    {
        strH[i+128]=strx[i];
    }
    for(i=strlen(strx);i<64;i++)
    {
        strH[i+128]='0';
    }
    for(i=0;i<strlen(stry);i++)
    {
        strH[i+192]=stry[i];
    }
    for(i=strlen(stry);i<64;i++)
    {
        strH[i+192]='0';
    }
    for(i=0;i<64;i++)
    {
        strH[i+256]=hash[i];
    }
    strH[320]= '\0';
    
    free(strx);
    free(stry);

    message_bin =  (uint8_t *)calloc(strlen(strH)*4, sizeof(uint8_t));
    len_bin= hex2bin(strH, message_bin);

    /* Computation of the hash of R|A|M */
    sha512(hashbin,message_bin,len_bin*8);
    free(message_bin);
    bin2hex(hashbin, 64, (unsigned char *)hashvalue);
    number_set_str(test2,hashvalue,16);
    
    /* Computation of H(R,A,M) mod l */
    number_mod(test,test2,(*curve)->n);
    
    /* Computation of H(R,A,M)*s mod l */
    number_mul(test2,test,s);
    number_mod(test,test2,(*curve)->n);
    
    /* Computation of r + H(R,A,M)*s mod l */
    number_add(test2,test,r);
    number_mod(*ss,test2,(*curve)->n);

    /* We set the signature */
    (*sig)->s = ss;
    
    /* Free memory */
    //ec_point_relax_pool_elt(p_key, (*curve)->k, STACK_1);
    number_free(&r);
    number_free(&s);
    number_free(&test);
    number_free(&test2);
}
/**
 *  For the hash several functions can be used, the interested reader can refer to [FIPS-180] 
 *    
 */

/**
 * @fn     int8_t eddsa_verify(unsigned char *hash, eddsa_sig *sig, ec_point *pub_key, ec_curve *curve);
 * @brief  Verify the signature "sig" of hash "hash" using public key "pub_key" and the EC "curve"
 * @param  hash    H(m) where H is a cryptographic hash function and m the message to sign. Must be written under hexadecimal form. ex: 64 hexadecimal characters for sha256.
 * @param  sig     Signature of hash
 * @param  pub_key Public key
 * @param  curve   Elliptic curve
 * @return 1 on success, 0 on failure
 */
int8_t
eddsa_verify(unsigned char *hash, eddsa_sig *sig, ec_point *pub_key, ec_curve *curve, ecdsa_precomp precomp)
{
    if(!edwards_belongs((*sig)->R, (*curve), STACK_1))
    {
        /* Just check that R belongs to the curve */
        return false;
    }
    /* Just check that s \in {0, ... , q-1} */
    if( number_cmp(*((*sig)->s),(*curve)->n)>=0)
    {      
        return false;
    } 

    uint8_t hashbin[64];
    uint8_t * message_bin;
    unsigned char hashvalue[129];
    int64_t len_bin;
    uint32_t i;   
    char * strx;
    char * stry;  
    char strH[321]; /* 2* 128 (representation of point in hexadecimal) + 64 (size of the message to sign) +1 */
    number test, test2;
    number_init(&test, bits_to_nblock(256));
    number_init(&test2, bits_to_nblock(512));

    /* First compute R|A|M */
    field_elt_str(&strx, ((*sig)->R)->x, 16,false, (*curve)->k, STACK_1);
    field_elt_str(&stry, ((*sig)->R)->y, 16,false, (*curve)->k, STACK_1);
    for(i=0;i<strlen(strx);i++)
    {
        strH[i]=strx[i];
    }
    for(i=strlen(strx);i<64;i++)
    {
        strH[i]='0';
    }
    for(i=0;i<strlen(stry);i++)
    {
        strH[i+64]=stry[i];
    }
    for(i=strlen(stry);i<64;i++)
    {
        strH[i+64]='0';
    }
    
    free(strx);
    free(stry);
    field_elt_str(&strx, (*pub_key)->x, 16,false, (*curve)->k, STACK_1);
    field_elt_str(&stry, (*pub_key)->y, 16,false, (*curve)->k, STACK_1);

    for(i=0;i<strlen(strx);i++)
    {
        strH[i+128]=strx[i];
    }
    for(i=strlen(strx);i<64;i++)
    {
        strH[i+128]='0';
    }
    for(i=0;i<strlen(stry);i++)
    {
        strH[i+192]=stry[i];
    }
    for(i=strlen(stry);i<64;i++)
    {
        strH[i+192]='0';
    }
    for(i=0;i<64;i++)
    {
        strH[i+256]=hash[i];
    }
    strH[320]= '\0';
    free(strx);
    free(stry);
    

    message_bin =  (uint8_t *)calloc(strlen(strH)*4, sizeof(uint8_t));
    len_bin= hex2bin(strH, message_bin);
    
    /* Computation of the hash of R,A,M */
    sha512(hashbin,message_bin,len_bin*8);
    /* sha512 is used since we need to have hashes of size 512 bits, another
     * hash function could have been used under this condition, we choosed it
     * because sha512 is implemented already in MPHELL 
     */
    free(message_bin);
    bin2hex(hashbin,64,hashvalue);
    

    /* Computation of the values optimised */
    number_set_str(test2,(char *)hashvalue,16);
    /* Computation of -H(R,A,M) mod l (in [EdDSA15] l represents E->n)*/
    number_mod(test,test2,(*curve)->n);
    number_sub(test2,(*curve)->n,test);
    number_mod(test,test2,(*curve)->n);

    /* Multiplication of -H(R,A,M) by the cofactor 2**c of the curve */
    number_mul(test2,(*curve)->h,test);
    number_mod(test,test2,(*curve)->n);

    /* x = (2**c)S*B - 2**c*H(R,A,M)*A (in [EdDSA15] B represents E->G, A is publickey)*/
    ec_point x;
    ec_point_get_pool_elt(x, (*curve)->k, STACK_1);
    number_mul(test2,(*curve)->h,*((*sig)->s));
    number_mod(test2,test2,(*curve)->n);
    ec_point_2mul_with_precomp(x, test2, (precomp)->prec_G, test, (precomp)->prec_K, PRECOMP_WIN_SIZE, *curve, STACK_1);
     
    /* y = 2**c*R */
    ec_point y;
    ec_point_get_pool_elt(y, (*curve)->k, STACK_1);
    ec_point_mul(y, (*curve)->h, (*sig)->R, *curve, STACK_1);
    /*Get the result by comparing x with y */        
    bool result = ec_point_are_equal(x, y, *curve, STACK_1);

    /* Free memory */
    ec_point_relax_pool_elt(x, (*curve)->k, STACK_1);
    ec_point_relax_pool_elt(y, (*curve)->k, STACK_1);
    number_free(&test);
    number_free(&test2);

    return result;
}

/**
 * @fn     int8_t eddsa_pub_key_validation(ec_point *pub_key, ec_curve *curve);
 * @brief  Verify that the public key is valid (id est check taht pub_key belongs to the curve, that pub_key is not the neutral element and that pub_key is of ordre n) 
 * @param  hash    H(m) where H is a cryptographic hash function and m the message to sign. Must be written under hexadecimal form. ex: 64 hexadecimal characters for sha256.
 * @param  pub_key Public key
 * @param  curve   Elliptic curve
 * @return 1 on success, 0 on failure
 */
int8_t 
eddsa_pub_key_validation(ec_point *pub_key, ec_curve *curve)
{
    
    if(!ec_belongs(*pub_key, *curve, STACK_1))
    {
        /* pubkey does not belong  to the curve */
        return false;   
    }   
    if(ec_point_is_neutral(*pub_key, *curve, STACK_1))
    {
        /* pubkey is the neutral element */
        return false;
    }
    /* We finally test the order of the curve */
    ec_point x;
    ec_point_get_pool_elt(x, (*curve)->k, STACK_1);    
    ec_point_mul(x, (*curve)->n, *pub_key, *curve, STACK_1);
    if(!ec_point_is_neutral(x, *curve, STACK_1))
    {
        /* pubkey is the neutral element */
        return false;
    }
    ec_point_relax_pool_elt(x, (*curve)->k, STACK_1);
    
    return true;
}


int main()
{
    /* Init Mphell */
    init_mphell(256, RANDOM_AES256, DEVURANDOM);
    uint32_t i;
    int nb_iteration = 10000;
    
    uint8_t hashbin[HASH_SIZE/2];
    unsigned char hashval[nb_iteration][HASH_SIZE+1];
    uint8_t * message_bin;
    int32_t len_bin;
    uint8_t message[200];
    
    struct timespec start, end;
    
    /* Number of digests to sign */
    printf("Vector test size %d \n", nb_iteration);
    printf("Curve         \t Signature \t\t Verification \n");
    
    /* Common to all curves */
    field k;
    ec_curve E;
    ecdsa_sig *lib_sig = (ecdsa_sig*)malloc(sizeof(ecdsa_sig)*nb_iteration);
    ec_point pub_key;
    number priv_key;
    ecdsa_precomp precomp;
    int err = 0;
    int ret = 0;
    
    /* Create digest */
    for(i = 0; i < nb_iteration; i++)
    {
        sprintf((char *)message, "This is the %d message to sign and verify on stm32F4", i);
        len_bin=strlen((char *)message);
        message_bin = (uint8_t *)calloc(len_bin, sizeof(uint8_t));
        /* Convert ascii message into binary message */
        string2bin((char *)message, message_bin);
        /* Hash binary message */
        sha256(hashbin, message_bin, len_bin*8);
        free(message_bin);
        /* Convert binary hash into hexadecimal string */
        bin2hex(hashbin, HASH_SIZE/2, hashval[i]);
    }
    
    /* Curve Nist 256 */
    
    /* Create elliptic curve */
    field_alloc(k, FP, bits_to_nblock(256), NULL);
    ec_alloc(E, k);
    ec_use_curve(E, k, NIST_256, JACOBIAN, STACK_1);
    printf("Secp256 \t");
    
    /* Allocate nb_iteration ecdsa_sig structure */
    for (i = 0; i < nb_iteration; i++)
    {
        ecdsa_sign_alloc(&(lib_sig[i]), bits_to_nblock(256));
    }
    
    /* Create public key */
    ec_point_alloc(pub_key, k);
    ec_point_init(pub_key, k);
    ec_point_set_aff_str(pub_key, "CBE2BCC92CDA44BB7851833F41684935DAD703C9AFB5D209F2922F8F739BA2D7", "8B620F0C41F572563648CC9C1C782F71C952B87782B61431640505AA317FCC07", false, 16, WEIERSTRASS, k, STACK_1);

    /* Get private key  */
    number_init(&priv_key,bits_to_nblock(256));
    number_set_str(priv_key, "5A23E08CF6E463BFEDFBA98F0CCE0B3B0A18EFED7DD31834577721A96D10BB8B", 16);
    
    /* Precomputation (to speed up verification) */
    ecdsa_precal(&precomp, &pub_key, E);

    /* Sign each digest */
    clock_gettime(CLOCK_MONOTONIC_RAW, &start);
    for(i = 0; i < nb_iteration; i++)
    {
        ecdsa_sign((unsigned char *)hashval[i], &(lib_sig[i]), priv_key, &E);
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end);
    printf(" %lu.%02u \t\t\t", get_s(&start, &end), get_ns(&start, &end));

    /* Verify signatures */
    clock_gettime(CLOCK_MONOTONIC_RAW, &start);
    for(i = 0; i < nb_iteration; i++)
    {
        ret = ecdsa_verify((unsigned char *)hashval[i] , &(lib_sig[i]), &pub_key, &E, &precomp);
        if( ret < 1 )
        {
            err++;
        }
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end);
    printf(" %lu.%02u \n", get_s(&start, &end), get_ns(&start, &end));

    if( err > 0 )
    {
        printf("\nERROR :%d\n", err);
    }

    /* Free allocated space */
    number_free(&priv_key);
    ec_point_free(pub_key, k);
    field_free(k);
    ec_free(E);
    ecdsa_precal_free(&precomp, E);
    for(i = 0; i < nb_iteration; i++)
    {
        ecdsa_sign_free(&(lib_sig[i]));
    }
    
    /* Curve Nist 384 */
    
    /* Create elliptic curve */
    field_alloc(k, FP, bits_to_nblock(384), NULL);
    ec_alloc(E, k);
    ec_use_curve(E, k, NIST_384, JACOBIAN, STACK_1);
    printf("Secp384 \t");
    
    /* Allocate nb_iteration ecdsa_sig structure */
    for (i = 0; i < nb_iteration; i++)
    {
        ecdsa_sign_alloc(&(lib_sig[i]), bits_to_nblock(384));
    }
    
    /* Create public key */
    ec_point_alloc(pub_key, k);
    ec_point_init(pub_key, k);
    ec_point_set_aff_str(pub_key, "A91822E77B104A4EAA65E62AA61EE458DE9157B2B4D06B95D93828BE114377AE104D4FBC1E86F7C4592B7EC2220EB963", "BAD603D61F7C04078725209EE80D16A764121567A9E02F11CA2330B7B8D3009A6976080B1F9199A1C98AF906647289D8", false, 16, WEIERSTRASS, k, STACK_1);

    /* Get private key  */
    number_init(&priv_key,bits_to_nblock(384));
    number_set_str(priv_key, "3B3D1632092F7397B8B9E697A3DCA19B7070BF7B9491B499EB97D296740413E74B9452718FF4F6AEB0151EA72B8D179B", 16);
    
    /* Precomputation (to speed up verification) */
    ecdsa_precal(&precomp, &pub_key, E);

    /* Sign each digest */
    clock_gettime(CLOCK_MONOTONIC_RAW, &start);
    for(i = 0; i < nb_iteration; i++)
    {
        ecdsa_sign((unsigned char *)hashval[i], &(lib_sig[i]), priv_key, &E);
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end);
    printf(" %lu.%02u \t\t\t", get_s(&start, &end), get_ns(&start, &end));

    /* Verify signatures */
    clock_gettime(CLOCK_MONOTONIC_RAW, &start);
    for(i = 0; i < nb_iteration; i++)
    {
        ret = ecdsa_verify((unsigned char *)hashval[i] , &(lib_sig[i]), &pub_key, &E, &precomp);
        if( ret < 1 )
        {
            err++;
        }
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end);
    printf(" %lu.%02u \n", get_s(&start, &end), get_ns(&start, &end));

    if( err > 0 )
    {
        printf("\nERROR :%d\n", err);
    }

    /* Free allocated space */
    number_free(&priv_key);
    ec_point_free(pub_key, k);
    field_free(k);
    ec_free(E);
    ecdsa_precal_free(&precomp, E);
    for(i = 0; i < nb_iteration; i++)
    {
        ecdsa_sign_free(&(lib_sig[i]));
    }
    
    /* Curve Nist 521 */
    
    /* Create elliptic curve */
    field_alloc(k, FP, bits_to_nblock(521), NULL);
    ec_alloc(E, k);
    ec_use_curve(E, k, NIST_521, JACOBIAN, STACK_1);
    printf("Secp521 \t");
    
    /* Allocate nb_iteration ecdsa_sig structure */
    for (i = 0; i < nb_iteration; i++)
    {
        ecdsa_sign_alloc(&(lib_sig[i]), bits_to_nblock(521));
    }
    
    /* Create public key */
    ec_point_alloc(pub_key, k);
    ec_point_init(pub_key, k);
    ec_point_set_aff_str(pub_key, "1FF819660EFE86ABA77F2D8E8FCFC862746BA785ADFD628AFF99D1AF9B79B40E48C01D124468D8CBB2C8A547D8CC0DE02F7FF9CB655089E48D5A5A1CBC22A0CA7F", "01F4D49238F42E7557A66BACC38DB1B7CB384591CECB386106888C4920E353DECD17966F7E14BF65992BC39F2CB6D4F94F7D076FC60FD3B2E25FEF9E6CD8AB6FC21A", false, 16, WEIERSTRASS, k, STACK_1);

    /* Get private key  */
    number_init(&priv_key,bits_to_nblock(521));
    number_set_str(priv_key, "C16FE10D6C89449622A9150ED3C640EE55274309193C03AD88DF2968623E939480D9D78B9FDD9EF54FA78885C451FACC51FB631B0AC0473B10812E2D8D6FACEE2A", 16);
    
    /* Precomputation (to speed up verification) */
    ecdsa_precal(&precomp, &pub_key, E);

    /* Sign each digest */
    clock_gettime(CLOCK_MONOTONIC_RAW, &start);
    for(i = 0; i < nb_iteration; i++)
    {
        ecdsa_sign((unsigned char *)hashval[i], &(lib_sig[i]), priv_key, &E);
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end);
    printf(" %lu.%02u \t\t\t", get_s(&start, &end), get_ns(&start, &end));

    /* Verify signatures */
    clock_gettime(CLOCK_MONOTONIC_RAW, &start);
    for(i = 0; i < nb_iteration; i++)
    {
        ret = ecdsa_verify((unsigned char *)hashval[i] , &(lib_sig[i]), &pub_key, &E, &precomp);
        if( ret < 1 )
        {
            err++;
        }
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end);
    printf(" %lu.%02u \n", get_s(&start, &end), get_ns(&start, &end));

    if( err > 0 )
    {
        printf("\nERROR :%d\n", err);
    }

    /* Free allocated space */
    number_free(&priv_key);
    ec_point_free(pub_key, k);
    field_free(k);
    ec_free(E);
    ecdsa_precal_free(&precomp, E);
    for(i = 0; i < nb_iteration; i++)
    {
        ecdsa_sign_free(&(lib_sig[i]));
    }
    
    /* Curve Brainpool 256 */
    
    /* Create elliptic curve */
    field_alloc(k, FP, bits_to_nblock(256), NULL);
    ec_alloc(E, k);
    ec_use_curve(E, k, BRAINPOOL_256, JACOBIAN, STACK_1);
    printf("BP256 \t\t");
    
    /* Allocate nb_iteration ecdsa_sig structure */
    for (i = 0; i < nb_iteration; i++)
    {
        ecdsa_sign_alloc(&(lib_sig[i]), bits_to_nblock(256));
    }
    
    /* Create public key */
    ec_point_alloc(pub_key, k);
    ec_point_init(pub_key, k);
    ec_point_set_aff_str(pub_key, "2B868B1C85C91B2AB534A51055B33FC7F572C5D1A5A8A5ACD27A493AF714F64C", "6CE0528C96A9A42681DE971559E004B0B155BF10131285B456EC1C5B5EAB260B", false, 16, WEIERSTRASS, k, STACK_1);

    /* Get private key  */
    number_init(&priv_key,bits_to_nblock(256));
    number_set_str(priv_key, "208D988D2C615D812EF7864B1BB982ACB19BBDF081DF5709C359AD577965BCEE", 16);
    
    /* Precomputation (to speed up verification) */
    ecdsa_precal(&precomp, &pub_key, E);

    /* Sign each digest */
    clock_gettime(CLOCK_MONOTONIC_RAW, &start);
    for(i = 0; i < nb_iteration; i++)
    {
        ecdsa_sign((unsigned char *)hashval[i], &(lib_sig[i]), priv_key, &E);
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end);
    printf(" %lu.%02u \t\t\t", get_s(&start, &end), get_ns(&start, &end));

    /* Verify signatures */
    clock_gettime(CLOCK_MONOTONIC_RAW, &start);
    for(i = 0; i < nb_iteration; i++)
    {
        ret = ecdsa_verify((unsigned char *)hashval[i] , &(lib_sig[i]), &pub_key, &E, &precomp);
        if( ret < 1 )
        {
            err++;
        }
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end);
    printf(" %lu.%02u \n", get_s(&start, &end), get_ns(&start, &end));

    if( err > 0 )
    {
        printf("\nERROR :%d\n", err);
    }

    /* Free allocated space */
    number_free(&priv_key);
    ec_point_free(pub_key, k);
    field_free(k);
    ec_free(E);
    ecdsa_precal_free(&precomp, E);
    for(i = 0; i < nb_iteration; i++)
    {
        ecdsa_sign_free(&(lib_sig[i]));
    }
    
    /* Curve Brainpool 384 */
    
    /* Create elliptic curve */
    field_alloc(k, FP, bits_to_nblock(384), NULL);
    ec_alloc(E, k);
    ec_use_curve(E, k, BRAINPOOL_384, JACOBIAN, STACK_1);
    printf("BP384 \t\t");
    
    /* Allocate nb_iteration ecdsa_sig structure */
    for (i = 0; i < nb_iteration; i++)
    {
        ecdsa_sign_alloc(&(lib_sig[i]), bits_to_nblock(384));
    }
    
    /* Create public key */
    ec_point_alloc(pub_key, k);
    ec_point_init(pub_key, k);
    ec_point_set_aff_str(pub_key, "62CD2BCDF81438A020CC0D8E0F4BB0260526BB2E3CFA317A2B83DD41C076A053008D8339F7C38B82E4B9488112DDAF04", "75FECEA32C8F20A0D0465F5C7145DCED81E1AFB3CA8334B196970AE0ABB7935ED58309507FAD771809DC76FD79DE9531", false, 16, WEIERSTRASS, k, STACK_1);

    /* Get private key  */
    number_init(&priv_key,bits_to_nblock(384));
    number_set_str(priv_key, "46EFD31C5221FB4611492EAC41B10C3CB5BC749546C89C65B0CC208A983AFB540DFAD3D7CC02BD25B45D2388ABB35A89", 16);
    
    /* Precomputation (to speed up verification) */
    ecdsa_precal(&precomp, &pub_key, E);

    /* Sign each digest */
    clock_gettime(CLOCK_MONOTONIC_RAW, &start);
    for(i = 0; i < nb_iteration; i++)
    {
        ecdsa_sign((unsigned char *)hashval[i], &(lib_sig[i]), priv_key, &E);
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end);
    printf(" %lu.%02u \t\t\t", get_s(&start, &end), get_ns(&start, &end));

    /* Verify signatures */
    clock_gettime(CLOCK_MONOTONIC_RAW, &start);
    for(i = 0; i < nb_iteration; i++)
    {
        ret = ecdsa_verify((unsigned char *)hashval[i] , &(lib_sig[i]), &pub_key, &E, &precomp);
        if( ret < 1 )
        {
            err++;
        }
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end);
    printf(" %lu.%02u \n", get_s(&start, &end), get_ns(&start, &end));

    if( err > 0 )
    {
        printf("\nERROR :%d\n", err);
    }

    /* Free allocated space */
    number_free(&priv_key);
    ec_point_free(pub_key, k);
    field_free(k);
    ec_free(E);
    ecdsa_precal_free(&precomp, E);
    for(i = 0; i < nb_iteration; i++)
    {
        ecdsa_sign_free(&(lib_sig[i]));
    }
    
    /* Curve Brainpool 512 */
    
    /* Create elliptic curve */
    field_alloc(k, FP, bits_to_nblock(512), NULL);
    ec_alloc(E, k);
    ec_use_curve(E, k, BRAINPOOL_512, JACOBIAN, STACK_1);
    printf("BP512 \t\t");
    
    /* Allocate nb_iteration ecdsa_sig structure */
    for (i = 0; i < nb_iteration; i++)
    {
        ecdsa_sign_alloc(&(lib_sig[i]), bits_to_nblock(512));
    }
    
    /* Create public key */
    ec_point_alloc(pub_key, k);
    ec_point_init(pub_key, k);
    ec_point_set_aff_str(pub_key, "8B5BAF98F24CDA60EDD7A1D025D11B347BBD3239405AF8E2A8AE1F20454C50705EFF652D90BD9139EF6D4D400D1D76161168C7557773A4F78F9C9B4347799396", "5CDD48F398FFF30A1A38F797EFF4538461604BC6CDF781F75ACC63865063315BBBD9606839A7861327E2133F5DB00BDA831CFB5F166E2DF5123DE53EDED7F6BD", false, 16, WEIERSTRASS, k, STACK_1);

    /* Get private key  */
    number_init(&priv_key,bits_to_nblock(512));
    number_set_str(priv_key, "A1AC65BF25D7B0DFC25B0943578D2591D2C587B6CE09052FAC0E06B03728A335955224E905647E2C003B69124D5FF17E8759E1F950926BD83780A89C42DA38AD", 16);
    
    /* Precomputation (to speed up verification) */
    ecdsa_precal(&precomp, &pub_key, E);

    /* Sign each digest */
    clock_gettime(CLOCK_MONOTONIC_RAW, &start);
    for(i = 0; i < nb_iteration; i++)
    {
        ecdsa_sign((unsigned char *)hashval[i], &(lib_sig[i]), priv_key, &E);
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end);
    printf(" %lu.%02u \t\t\t", get_s(&start, &end), get_ns(&start, &end));

    /* Verify signatures */
    clock_gettime(CLOCK_MONOTONIC_RAW, &start);
    for(i = 0; i < nb_iteration; i++)
    {
        ret = ecdsa_verify((unsigned char *)hashval[i] , &(lib_sig[i]), &pub_key, &E, &precomp);
        if( ret < 1 )
        {
            err++;
        }
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end);
    printf(" %lu.%02u \n", get_s(&start, &end), get_ns(&start, &end));

    if( err > 0 )
    {
        printf("\nERROR :%d\n", err);
    }

    /* Free allocated space */
    number_free(&priv_key);
    ec_point_free(pub_key, k);
    field_free(k);
    ec_free(E);
    ecdsa_precal_free(&precomp, E);
    for(i = 0; i < nb_iteration; i++)
    {
        ecdsa_sign_free(&(lib_sig[i]));
    }
    
    /* Curve FR 256 */
    
    /* Create elliptic curve */
    field_alloc(k, FP, bits_to_nblock(256), NULL);
    ec_alloc(E, k);
    ec_use_curve(E, k, FR_256, JACOBIAN, STACK_1);
    printf("Prime256 \t");
    
    /* Allocate nb_iteration ecdsa_sig structure */
    for (i = 0; i < nb_iteration; i++)
    {
        ecdsa_sign_alloc(&(lib_sig[i]), bits_to_nblock(256));
    }
    
    /* Create public key */
    ec_point_alloc(pub_key, k);
    ec_point_init(pub_key, k);
    ec_point_set_aff_str(pub_key, "D4AB04EE8F45C249D67BA2089F358F7879A9A72FFD09D9FFC044F2484EE884A3", "AD18C9B96661D9FDA3274099D9E12ACEEAE71F137BEBD97C58D1A518B4212BCD", false, 16, WEIERSTRASS, k, STACK_1);

    /* Get private key  */
    number_init(&priv_key,bits_to_nblock(256));
    number_set_str(priv_key, "DE64EABE6EBD26C2F583E8F577FD54DFCCF4E82BDDFEDC37FC221F869BB55E05", 16);
    
    /* Precomputation (to speed up verification) */
    ecdsa_precal(&precomp, &pub_key, E);

    /* Sign each digest */
    clock_gettime(CLOCK_MONOTONIC_RAW, &start);
    for(i = 0; i < nb_iteration; i++)
    {
        ecdsa_sign((unsigned char *)hashval[i], &(lib_sig[i]), priv_key, &E);
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end);
    printf(" %lu.%02u \t\t\t", get_s(&start, &end), get_ns(&start, &end));

    /* Verify signatures */
    clock_gettime(CLOCK_MONOTONIC_RAW, &start);
    for(i = 0; i < nb_iteration; i++)
    {
        ret = ecdsa_verify((unsigned char *)hashval[i] , &(lib_sig[i]), &pub_key, &E, &precomp);
        if( ret < 1 )
        {
            err++;
        }
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end);
    printf(" %lu.%02u \n", get_s(&start, &end), get_ns(&start, &end));

    if( err > 0 )
    {
        printf("\nERROR :%d\n", err);
    }

    /* Free allocated space */
    number_free(&priv_key);
    ec_point_free(pub_key, k);
    field_free(k);
    ec_free(E);
    ecdsa_precal_free(&precomp, E);
    for(i = 0; i < nb_iteration; i++)
    {
        ecdsa_sign_free(&(lib_sig[i]));
    }
    
    
    /* Curve ED 25519 */
    
    /* Create elliptic curve */
    field_alloc(k, FP, bits_to_nblock(256), NULL);
    ec_alloc(E, k);
    ec_use_curve(E, k, ED_25519, EXTENDED_EDWARDS, STACK_1);
    printf("Ed25519 \t");
    eddsa_sig *lib_sig_ed = (eddsa_sig*)malloc(sizeof(eddsa_sig)*nb_iteration);
    
    unsigned char hashvalue[129];  
    char * str;   
    uint8_t hashbin_2[64];  
    //int64_t len_bin;  
    number test, test2;
    number_init(&test, bits_to_nblock(257));
    number_init(&test2, bits_to_nblock(512));
    number_init(&priv_key, bits_to_nblock(256));
    number_set_str(test,"1",16);
    number_lshift(test,test,256);    
    number_dec(test, test);    
    /* The private key is a number of b bits with b the size of the security 
     * desired.
     * Here the private key does not need to come from a random value between 
     * 0 and the size of the group generated by G:E->n.
     */
    number_random1(priv_key, test, STACK_1);


    /* Computation of the public key */
    number_inc(test, test);
    number_str(&str,priv_key,16);
    message_bin =  (uint8_t *)calloc(strlen(str)*4, sizeof(uint8_t));
    len_bin= hex2bin(str, message_bin);
    /* Computation of the hash of the private key */
    sha512(hashbin_2,message_bin,len_bin*8);
    /* sha512 is used since we need to have hashes of size 512 bits, another
     * hash function could have been used under this condition, we choosed it
     * because sha512 is implemented already in MPHELL 
     */
    free(str);
    free(message_bin);
    /* Computation of s=2^n+\sum_{i=c}^{n-1}2^ih_i (h_i are bits of the hash)*/
    bin2hex(hashbin_2, 64,(unsigned char *)hashvalue);
    
    number_set_str(test2, (const char *)hashvalue, 16);
    //number_rshift(test2, test2, 256);
    
    char *test3;
    number_str(&test3, test2, 2);
    test3[0]='0';
    test3[1]='1';
    test3[253]='0';
    test3[254]='0';
    test3[255]='0';
    test3[256]='\0';
    
    number_set_str(test2, test3, 2);
    free(test3);    
    /* Computation of s mod l (in [EdDSA15] l represents E->n)*/
    number_mod(test,test2,E->n);
    
    /* Computation of A=sB (in [EdDSA15] B represents E->G)*/
    
    ec_point_alloc(pub_key, E->k);
    ec_point_init(pub_key, E->k);
    ec_point_mul_unified(pub_key, test, E->G, E, STACK_1);
    
    /* Free memory done for the test */
    number_free(&test);
    number_free(&test2);

    /* Print the public key A */
    ec_point_norm(pub_key, E, STACK_1);  

    /* This test is optional.*/
    ret = eddsa_pub_key_validation(&pub_key, &E);
    if( ret < 1 )
    {
        printf("ERROR\n");
    }

    /* Allocate nb_iteration ecdsa_sig structure */
    for (i = 0; i < nb_iteration; i++)
    {
        eddsa_sign_alloc(&(lib_sig_ed[i]), E, bits_to_nblock(256));
    }

    /* Precomputation (to speed up verification) */
    ecdsa_precal(&precomp, &pub_key, E);
    
    /* Sign the digest */
    
    clock_gettime(CLOCK_MONOTONIC_RAW, &start);    
    for(i = 0; i < nb_iteration; i++)
    {
        eddsa_sign((unsigned char *)hashval[i], &(lib_sig_ed[i]), priv_key, &pub_key, &E);
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end);
    printf(" %lu.%02u \t\t\t", get_s(&start, &end), get_ns(&start, &end));
    
    /* Verify the signature */
    err=0;

    clock_gettime(CLOCK_MONOTONIC_RAW, &start);
    for(i = 0; i < nb_iteration; i++)
    {
        ret = eddsa_verify((unsigned char *)hashval[i], &(lib_sig_ed[i]), &pub_key, &E, precomp);
        if( ret < 1 )
        {
            err++;
        }
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end);
    printf(" %lu.%02u \n", get_s(&start, &end), get_ns(&start, &end));
    
    if( err > 0 )
    {
        printf("\nERROR :%d\n", err);
    }

    /* Free allocated space */
    number_free(&priv_key);
    ec_point_free(pub_key, k);
    ecdsa_precal_free(&precomp, E);
    for(i = 0; i < nb_iteration; i++)
    {
        eddsa_sign_free(&(lib_sig_ed[i]), E);
    }
    free(lib_sig_ed);
    ec_free(E);
    field_free(k);
    
    /* Curve JQ_256_3 */
    
    /* Create elliptic curve */
    field_alloc(k, FP, bits_to_nblock(256), NULL);
    ec_alloc(E, k);
    ec_use_curve(E, k, JQ_256_3, EXTENDED_HOMOGENEOUS_PROJECTIVE, STACK_1);
    printf("JQ_256_3 \t");
    
    /* Allocate nb_iteration ecdsa_sig structure */
    for (i = 0; i < nb_iteration; i++)
    {
        ecdsa_sign_alloc(&(lib_sig[i]), bits_to_nblock(256));
    }

    /* Get private key  */
    number_init(&priv_key,bits_to_nblock(256));
    number_set_str(priv_key, "068F45E01ECFBC6E3CACE3595B631899966F1189967BF3FED84A6B495F703855", 16);
    
    /* Create public key */
    ec_point_alloc(pub_key, k);
    ec_point_init(pub_key, k);
    ec_point_mul(pub_key, priv_key, E->G, E, STACK_1);
    
    /* Precomputation (to speed up verification) */
    ecdsa_precal(&precomp, &pub_key, E);

    /* Sign each digest */
    clock_gettime(CLOCK_MONOTONIC_RAW, &start);
    for(i = 0; i < nb_iteration; i++)
    {
        ecdsa_sign((unsigned char *)hashval[i], &(lib_sig[i]), priv_key, &E);
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end);
    printf(" %lu.%02u \t\t\t", get_s(&start, &end), get_ns(&start, &end));

    /* Verify signatures */
    clock_gettime(CLOCK_MONOTONIC_RAW, &start);
    for(i = 0; i < nb_iteration; i++)
    {
        ret = ecdsa_verify((unsigned char *)hashval[i] , &(lib_sig[i]), &pub_key, &E, &precomp);
        if( ret < 1 )
        {
            err++;
        }
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end);
    printf(" %lu.%02u \n", get_s(&start, &end), get_ns(&start, &end));

    if( err > 0 )
    {
        printf("\nERROR :%d\n", err);
    }

    /* Free allocated space */
    number_free(&priv_key);
    ec_point_free(pub_key, k);
    field_free(k);
    ec_free(E);
    ecdsa_precal_free(&precomp, E);
    for(i = 0; i < nb_iteration; i++)
    {
        ecdsa_sign_free(&(lib_sig[i]));
    }

    free(lib_sig);
    free_mphell();
    
    /* Return */
    if(err == 0)
    {
        printf("Test OK \n");
        return 0;
    }
    else
    {
        printf("%d errors for %d signatures \n", err, nb_iteration);
        return -1;
    }
}
