/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France)

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell_perf_fp.c
  * \brief Benchmark for the prime field functions
  */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <string.h>

#include "mphell.h"

int main()
{
    init_mphell(128, RANDOM_AES128, DEVURANDOM);
    int i;
    int nb_iteration = 20000;
    
    struct timespec start, end;
    
    fp_elt x[nb_iteration];
    fp_elt y[nb_iteration];
    fp_elt dst;
    
    printf("Enter TEST:\n");
    
    /* 192 bits */
    
    fp_param k3;
    k3 = (fp_param_t *)malloc(sizeof(fp_param_t));
    fp_alloc(k3, bits_to_nblock(192));
    number p;
    number_init(&p, bits_to_nblock(192));
    number_set_str(p, "fffffffffffffffffffffffffffffffeffffffffffffffff", 16);
    fp_create(k3, p, ARBITRARY, STACK_1);
    number_free(&p);
    
    for (i=0; i<nb_iteration; i++)
    {
        fp_elt_alloc(&x[i], k3);
        fp_elt_init(x[i], k3);
    }
    for (i=0; i<nb_iteration; i++)
    {
        fp_elt_alloc(&y[i], k3);
        fp_elt_init(y[i], k3);
    }
    fp_elt_alloc(&dst, k3);
    fp_elt_init(dst, k3);
    
    for(i=0; i<nb_iteration; i++)
    {
        /* Set element random */
        
        fp_elt_random(x[i], k3, STACK_1);
        fp_elt_random(y[i], k3, STACK_1);
    }
    
    /* Add */
    
    clock_gettime(CLOCK_MONOTONIC_RAW, &start);
    for (i=0; i<nb_iteration; i++)
    {
        fp_elt_add(dst, x[i], y[i], k3);
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end);
    printf("Add 192 bits | %lu s %02u ns \n", get_s(&start, &end), get_ns(&start, &end));
    
    /* Sub */
    
    clock_gettime(CLOCK_MONOTONIC_RAW, &start);
    for (i=0; i<nb_iteration; i++)
    {
        fp_elt_sub(dst, x[i], y[i], k3);
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end);
    printf("Sub 192 bits | %lu s %02u ns \n", get_s(&start, &end), get_ns(&start, &end));
    
    /* Mul */
    
    clock_gettime(CLOCK_MONOTONIC_RAW, &start);
    for (i=0; i<nb_iteration; i++)
    {
        fp_elt_mul(dst, x[i], y[i], k3, STACK_1);
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end);
    printf("Mul 192 bits | %lu s %02u ns \n", get_s(&start, &end), get_ns(&start, &end));
    
    /* Inv */
    
    clock_gettime(CLOCK_MONOTONIC_RAW, &start);
    for (i=0; i<nb_iteration; i++)
    {
        fp_elt_inv(dst, x[i], k3, STACK_1);
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end);
    printf("Inv 192 bits | %lu s %02u ns \n\n", get_s(&start, &end), get_ns(&start, &end));
    
    /* Free allocated field elts */
    
    for (i=0; i<nb_iteration; i++)
    {
        fp_elt_free(&x[i]);
        fp_elt_free(&y[i]);
    }
    fp_elt_free(&dst);
    
    /* Free allocated field */
    
    fp_free(k3);
    free(k3);
    
    /* 256 bits */
    
    fp_param k4;
    k4 = (fp_param_t *)malloc(sizeof(fp_param_t));
    fp_alloc(k4, bits_to_nblock(256));
    number_init(&p, bits_to_nblock(256));
    number_set_str(p, "57896044618658097711785492504343953926634992332820282019728792003956564819949", 10);
    fp_create(k4, p, ARBITRARY, STACK_1);
    number_free(&p);
    
    for (i=0; i<nb_iteration; i++)
    {
        fp_elt_alloc(&x[i], k4);
        fp_elt_init(x[i], k4);
    }
    for (i=0; i<nb_iteration; i++)
    {
        fp_elt_alloc(&y[i], k4);
        fp_elt_init(y[i], k4);
    }
    fp_elt_alloc(&dst, k4);
    fp_elt_init(dst, k4);
    
    for(i=0; i<nb_iteration; i++)
    {
        /* Set element random */
        fp_elt_random(x[i], k4, STACK_1);
        fp_elt_random(y[i], k4, STACK_1);
    }
    
    /* Add */
    
    clock_gettime(CLOCK_MONOTONIC_RAW, &start);
    for (i=0; i<nb_iteration; i++)
    {
        fp_elt_add(dst, x[i], y[i], k4);
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end);
    printf("Add 256 bits | %lu s %02u ns \n", get_s(&start, &end), get_ns(&start, &end));
    
    /* Sub */
    
    clock_gettime(CLOCK_MONOTONIC_RAW, &start);
    for (i=0; i<nb_iteration; i++)
    {
        fp_elt_sub(dst, x[i], y[i], k4);
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end);
    printf("Sub 256 bits | %lu s %02u ns \n", get_s(&start, &end), get_ns(&start, &end));
    
    /* Mul */
    
    clock_gettime(CLOCK_MONOTONIC_RAW, &start);
    for (i=0; i<nb_iteration; i++)
    {
        fp_elt_mul(dst, x[i], y[i], k4, STACK_1);
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end);
    printf("Mul 256 bits | %lu s %02u ns \n", get_s(&start, &end), get_ns(&start, &end));
    
     /* Inv */
    
    clock_gettime(CLOCK_MONOTONIC_RAW, &start);
    for (i=0; i<nb_iteration; i++)
    {
        fp_elt_inv(dst, x[i], k4, STACK_1);
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end);
    printf("Inv 256 bits | %lu s %02u ns \n\n", get_s(&start, &end), get_ns(&start, &end));
    
    /* Free allocated field elts */
    
    for (i=0; i<nb_iteration; i++)
    {
        fp_elt_free(&x[i]);
        fp_elt_free(&y[i]);
    }
    fp_elt_free(&dst);
    
    /* Free allocated field */
    
    fp_free(k4);
    free(k4);

    /* 384 bits */
    
    fp_param k6;
    k6 = (fp_param_t *)malloc(sizeof(fp_param_t));
    fp_alloc(k6, bits_to_nblock(384));
    number_init(&p, bits_to_nblock(384));
    number_set_str(p, "fffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffeffffffff0000000000000000ffffffff", 16);
    fp_create(k6, p, ARBITRARY, STACK_1);
    number_free(&p);
    
    for (i=0; i<nb_iteration; i++)
    {
        fp_elt_alloc(&x[i], k6);
        fp_elt_init(x[i], k6);
    }
    for (i=0; i<nb_iteration; i++)
    {
        fp_elt_alloc(&y[i], k6);
        fp_elt_init(y[i], k6);
    }
    fp_elt_alloc(&dst, k6);
    fp_elt_init(dst, k6);
    
    for(i=0; i<nb_iteration; i++)
    {
        /* Set element random */
        fp_elt_random(x[i], k6, STACK_1);
        fp_elt_random(y[i], k6, STACK_1);
    }
    
    /* Add */
    
    clock_gettime(CLOCK_MONOTONIC_RAW, &start);
    for (i=0; i<nb_iteration; i++)
    {
        fp_elt_add(dst, x[i], y[i], k6);
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end);
    printf("Add 384 bits | %lu s %02u ns \n", get_s(&start, &end), get_ns(&start, &end));
    
    /* Sub */
    
    clock_gettime(CLOCK_MONOTONIC_RAW, &start);
    for (i=0; i<nb_iteration; i++)
    {
        fp_elt_sub(dst, x[i], y[i], k6);
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end);
    printf("Sub 384 bits | %lu s %02u ns \n", get_s(&start, &end), get_ns(&start, &end));
    
    /* Mul */
    
    clock_gettime(CLOCK_MONOTONIC_RAW, &start);
    for (i=0; i<nb_iteration; i++)
    {
        fp_elt_mul(dst, x[i], y[i], k6, STACK_1);
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end);
    printf("Mul 384 bits | %lu s %02u ns \n", get_s(&start, &end), get_ns(&start, &end));
    
    /* Inv */
    
    clock_gettime(CLOCK_MONOTONIC_RAW, &start);
    for (i=0; i<nb_iteration; i++)
    {
        fp_elt_inv(dst, x[i], k6, STACK_1);
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end);
    printf("Inv 384 bits | %lu s %02u ns \n\n", get_s(&start, &end), get_ns(&start, &end));
    
    /* Free allocated field elts */
    
    for (i=0; i<nb_iteration; i++)
    {
        fp_elt_free(&x[i]);
        fp_elt_free(&y[i]);
    }
    fp_elt_free(&dst);
    
    /* Free allocated field */
    
    fp_free(k6);
    free(k6);

    /* 512 bits */
    
    fp_param k8;
    k8 = (fp_param_t *)malloc(sizeof(fp_param_t));
    fp_alloc(k8, bits_to_nblock(512));
    number_init(&p, bits_to_nblock(512));
    number_set_str(p, "AADD9DB8DBE9C48B3FD4E6AE33C9FC07CB308DB3B3C9D20ED6639CCA703308717D4D9B009BC66842AECDA12AE6A380E62881FF2F2D82C68528AA6056583A48F3", 16);
    fp_create(k8, p, ARBITRARY, STACK_1);
    number_free(&p);
    
    for (i=0; i<nb_iteration; i++)
    {
        fp_elt_alloc(&(x[i]), k8);
        fp_elt_init(x[i], k8);
    }
    for (i=0; i<nb_iteration; i++)
    {
        fp_elt_alloc(&(y[i]), k8);
        fp_elt_init(y[i], k8);
    }
    fp_elt_alloc(&dst, k8);
    fp_elt_init(dst, k8);
    
    for(i=0; i<nb_iteration; i++)
    {
        /* Set element random */
        fp_elt_random(x[i], k8, STACK_1);
        fp_elt_random(y[i], k8, STACK_1);
    }
    
    /* Add */
    
    clock_gettime(CLOCK_MONOTONIC_RAW, &start);
    for (i=0; i<nb_iteration; i++)
    {
        fp_elt_add(dst, x[i], y[i], k8);
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end);
    printf("Add 512 bits | %lu s %02u ns \n", get_s(&start, &end), get_ns(&start, &end));
    
    /* Sub */
    
    clock_gettime(CLOCK_MONOTONIC_RAW, &start);
    for (i=0; i<nb_iteration; i++)
    {
        fp_elt_sub(dst, x[i], y[i], k8);
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end);
    printf("Sub 512 bits | %lu s %02u ns \n", get_s(&start, &end), get_ns(&start, &end));
    
    /* Mul */
    
    clock_gettime(CLOCK_MONOTONIC_RAW, &start);
    for (i=0; i<nb_iteration; i++)
    {
        fp_elt_mul(dst, x[i], y[i], k8, STACK_1);
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end);
    printf("Mul 512 bits | %lu s %02u ns \n", get_s(&start, &end), get_ns(&start, &end));
    
    /* Inv */
    
    clock_gettime(CLOCK_MONOTONIC_RAW, &start);
    for (i=0; i<nb_iteration; i++)
    {
        fp_elt_inv(dst, x[i], k8, STACK_1);
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end);
    printf("Inv 512 bits | %lu s %02u ns \n\n", get_s(&start, &end), get_ns(&start, &end));
    
    /* Free allocated field elts */
    
    for (i=0; i<nb_iteration; i++)
    {
        fp_elt_free(&x[i]);
        fp_elt_free(&y[i]);
    }
    fp_elt_free(&dst);
    
    /* Free allocated field */
    
    fp_free(k8);
    free(k8);
    
    free_mphell();
}
