/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France)

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell_perf_jacobi_quartic_arithmetic.c
  * \brief Preformance tests for the functions: ec_point_mul_naive in Extended Homogeneous Projective coordinates
  */

/** 
 * Should be compiled with the file mphell_test_jacobi_quartic.c
 * For this example DDATADIR should contain the relative path, an example of 
 * compilation is: gcc mphell_perf_jacobi_quartic_arithmetic.c 
 * mphell_test_jacobi_quartic.c  -lgmp -I../include -I/usr/local/lib 
 * -Wl,-rpath=/usr/local/lib -lmphell -g -DDATADIR='"../data"'  -o 
 * mphell-perf-jacobi_quartic-arithmetic
 */

#include "mphell_test_jacobi_quartic.h"

bool 
test_curve_jacobi_quartic_read_points(data_t *data)
{
    uint8_t n = data->k->size;
    uint32_t i = 0;
    ec_point Res;
    number n1;
    FILE *filep;
    char test_file[200];
    sprintf(test_file, DATADIR "/vector_test_elliptic_curves/%s_vector_test.txt", data->E->id_curve);
    filep = fopen(test_file,"r");
    if(filep == NULL)
    {
        fprintf(stderr, "Error : no curve data file found \n");
        exit(EXIT_FAILURE);
    }
    number_init(&n1, n);
    ec_point_alloc(Res, data->k);
    ec_point_init(Res, data->k);
    for(i = 0; i < data->nb_points; i++)
    {
        data->E->f = PROJECTIVE;
        test_curve_jacobi_quartic_read_point(filep, data->P1[i], data);
        test_curve_jacobi_quartic_read_point(filep, data->P3[i], data);
        
        if(!jacobi_quartic_belongs(data->P1[i], data->E, STACK_1))
        {
            printf("P1 does not belongs to E\n");
            return false;
        }
        if(!jacobi_quartic_belongs(data->P3[i], data->E, STACK_1))
        {
            printf("P3 does not belongs to E\n");
            return false;
        }  

        /*P1 + P2*/
        test_curve_jacobi_quartic_read_point(filep, Res, data);

        /*P1 - P2*/
        test_curve_jacobi_quartic_read_point(filep, Res, data);

        /*2 * P1*/
        test_curve_jacobi_quartic_read_point(filep, Res, data);

        /*n1 * P1*/
        test_curve_jacobi_quartic_read_data(filep, n1, data);
        test_curve_jacobi_quartic_read_point(filep, Res, data);
    }
    fclose(filep);
    ec_point_free(Res, data->k);
    number_free(&n1);
    return true;
}

void
test_curve_jacobi_quartic_perf(data_t *data)
{
    printf("Time on curve %s for %d points \n", data->E->id_curve, data->nb_points); 
    printf("Operation   | Extended Homogeneous Projective Coordinates \n");
    uint32_t i;
    number m;
    number_init(&m, data->k->size);
    
    number_sub_ui(m, data->E->n, 100);    /* m= q -100 */ 
    clock_t start, end;
    double elapsed1;

    /************************************ ADD  *******************************/

    data->E->f = PROJECTIVE;
    start = clock();
    for(i = 0; i < data->nb_points; i++)
    {
        ec_point_add(data->P3[i], data->P1[i], data->P2[i], data->E, STACK_1);
    }
    end = clock();
    elapsed1 = ((double) (end - start)) / CLOCKS_PER_SEC;
    
    printf("Add         | %f        \n", elapsed1);

    /************************************ DBL  *******************************/

    data->E->f = PROJECTIVE;
    start = clock();
    for(i = 0; i < data->nb_points; i++)
    {
        ec_point_dbl(data->P3[i], data->P1[i], data->E, STACK_1);
    }
    end = clock();
    elapsed1 = ((double) (end - start)) / CLOCKS_PER_SEC;

    printf("Dbl         | %f        \n", elapsed1);

    /************************************ SUB  *******************************/

    data->E->f = PROJECTIVE;
    start = clock();
    for(i = 0; i < data->nb_points; i++)
    {
        ec_point_add(data->P3[i], data->P1[i], data->P2[i], data->E, STACK_1);
    }
    end = clock();
    elapsed1 = ((double) (end - start)) / CLOCKS_PER_SEC;

    printf("Sub         | %f        \n", elapsed1);


    /************************************ MUL  *******************************/

    data->E->f = PROJECTIVE;
    start = clock();
    for(i = 0; i < data->nb_points; i++)
    {
        ec_point_mul(data->P3[i], m, data->P1[i], data->E, STACK_1);
    }
    end = clock();
    elapsed1 = ((double) (end - start)) / CLOCKS_PER_SEC;
    
    printf("Mul         | %f      \n", elapsed1);
    
    /******************************* MUL UNIFIED *****************************/

    data->E->f = PROJECTIVE;
    start = clock();
    for(i = 0; i < data->nb_points; i++)
    {
        ec_point_mul_unified(data->P3[i], m, data->P1[i], data->E, STACK_1);
    }
    end = clock();
    elapsed1 = ((double) (end - start)) / CLOCKS_PER_SEC;
    
    printf("Mul Unified | %f      \n", elapsed1);
    number_free(&m);
}

bool
test_curve_jacobi_quartic(const char *id_curve, const uint16_t nbits, 
        const uint32_t num_point, data_t *data)
{
    bool res;
    test_curve_jacobi_quartic_alloc(nbits, num_point, data);
    test_curve_jacobi_quartic_read_silent(id_curve, data, true);
    test_curve_jacobi_quartic_init(data);
    res = test_curve_jacobi_quartic_read_points(data);    
    test_curve_jacobi_quartic_perf(data);
    test_curve_jacobi_quartic_free(data);
    return res;
}

int main()
{
    data_t data;
    init_mphell(128, RANDOM_AES128, DEVURANDOM);
    int nb_points=100;
    if (!test_curve_jacobi_quartic("jq256_3", 256, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    if (!test_curve_jacobi_quartic("jacobi_quartic_p256", 256, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    free_mphell();
    return 0;
}
