/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France)

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell_perf_weierstrass_arithmetic.c
  * \brief Benchmark for the functions: ec_point_add, ec_point_sub, ec_point_dbl, ec_point_mul in Projective coordinates
  */

#include "mphell_test_weierstrass.h"

bool 
test_curve_weierstrass_read_points(data_t *data)
{
    uint8_t n = data->k->size;
    uint32_t i = 0;
    ec_point Res;
    number n1;
    FILE *filep;
    char test_file[200];
    sprintf(test_file, DATADIR "/vector_test_elliptic_curves/%s_vector_test.txt", data->E->id_curve);
    filep = fopen(test_file,"r");
    if(filep == NULL)
    {
        fprintf(stderr, "Error : no curve data file found \n");
        exit(EXIT_FAILURE);
    }
    number_init(&n1, n);
    ec_point_alloc(Res, data->k);
    ec_point_init(Res, data->k);
    
    for(i = 0; i < data->nb_points; i++)
    {
        data->E->f = PROJECTIVE;
        test_curve_weierstrass_read_point(filep, data->P1[i], data);
        test_curve_weierstrass_read_point(filep, data->P2[i], data);
        
        if(!weierstrass_belongs(data->P1[i], data->E, STACK_1))
        {
            printf("P1 does not belongs to E\n");
            return false;
        }
        if(!weierstrass_belongs(data->P2[i], data->E, STACK_1))
        {
            printf("P2 does not belongs to E\n");
            return false;
        }

        /*P1 + P2*/
        test_curve_weierstrass_read_point(filep, Res, data);

        /*P1 - P2*/
        test_curve_weierstrass_read_point(filep, Res, data);

        /*2 * P1*/
        test_curve_weierstrass_read_point(filep, Res, data);

        /*n1 * P1*/
        test_curve_weierstrass_read_data(filep, n1, data);
        test_curve_weierstrass_read_point(filep, Res, data);
    }

    fclose(filep);
    ec_point_free(Res, data->k);
    number_free(&n1);
    return true;
}

void
test_curve_weierstrass_perf(data_t *data)
{
    printf("Time on curve %s for %d points \n", data->E->id_curve, data->nb_points); 
    printf("Operation   | Projective           | Jacobian           \n");
    uint32_t i;
    number n;
    number_init(&n, data->k->size);
    number_sub_ui(n, data->E->n, 10000);

    struct timespec start1, end1, start2, end2;

    /************************************ ADD  *******************************/
    
    int nb_point = data->nb_points;
    int nb_test = 100;

    data->E->f = PROJECTIVE;
    clock_gettime(CLOCK_MONOTONIC_RAW, &start1);
    for(i = 0; i < nb_test; i++)
    {
        ec_point_add(data->P3[i%nb_point], data->P1[i%nb_point], data->P2[i%nb_point], data->E, STACK_1);
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end1);
    
    data->E->f = JACOBIAN;
    
    clock_gettime(CLOCK_MONOTONIC_RAW, &start2);
    for(i = 0; i < nb_test; i++)
    {
        ec_point_add(data->P3[i%nb_point], data->P1[i%nb_point], data->P2[i%nb_point], data->E, STACK_1);
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end2);
    printf("Add         | %lu s %02u ns         | %lu s %02u ns       \n", get_s(&start1, &end1), get_ns(&start1, &end1), get_s(&start2, &end2), get_ns(&start2, &end2));

    /************************************ DBL  *******************************/

    data->E->f = PROJECTIVE;
    clock_gettime(CLOCK_MONOTONIC_RAW, &start1);
    for(i = 0; i < nb_test; i++)
    {
        ec_point_dbl(data->P3[i%nb_point], data->P1[i%nb_point], data->E, STACK_1);
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end1);

    data->E->f = JACOBIAN;
    clock_gettime(CLOCK_MONOTONIC_RAW, &start2);
    for(i = 0; i < nb_test; i++)
    {
        ec_point_dbl(data->P3[i%nb_point], data->P1[i%nb_point], data->E, STACK_1);
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end2);
    printf("Dbl         | %lu s %02u ns         | %lu s %02u ns       \n", get_s(&start1, &end1), get_ns(&start1, &end1), get_s(&start2, &end2), get_ns(&start2, &end2));
    
    /************************************ SUB  *******************************/

    data->E->f = PROJECTIVE;
    clock_gettime(CLOCK_MONOTONIC_RAW, &start1);
    for(i = 0; i < nb_test; i++)
    {
        ec_point_add(data->P3[i%nb_point], data->P1[i%nb_point], data->P2[i%nb_point], data->E, STACK_1);
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end1);

    data->E->f = JACOBIAN;
    clock_gettime(CLOCK_MONOTONIC_RAW, &start2);
    for(i = 0; i < nb_test; i++)
    {
        ec_point_add(data->P3[i%nb_point], data->P1[i%nb_point], data->P2[i%nb_point], data->E, STACK_1);
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end2);
    printf("Sub         | %lu s %02u ns         | %lu s %02u ns       \n", get_s(&start1, &end1), get_ns(&start1, &end1), get_s(&start2, &end2), get_ns(&start2, &end2));

    /************************************ MUL  *******************************/

    data->E->f = PROJECTIVE;
    clock_gettime(CLOCK_MONOTONIC_RAW, &start1);
    for(i = 0; i < nb_test; i++)
    {
        ec_point_mul(data->P3[i%nb_point], n, data->P1[i%nb_point], data->E, STACK_1);
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end1);

    data->E->f = JACOBIAN;
    clock_gettime(CLOCK_MONOTONIC_RAW, &start2);
    for(i = 0; i < nb_test; i++)
    {
        ec_point_mul(data->P3[i%nb_point], n, data->P1[i%nb_point], data->E, STACK_1);
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end2);
    printf("Mul         | %lu s %02u ns         | %lu s %02u ns       \n", get_s(&start1, &end1), get_ns(&start1, &end1), get_s(&start2, &end2), get_ns(&start2, &end2));
    
    
    /******************************* MUL UNIFIED *****************************/

    data->E->f = PROJECTIVE;
    clock_gettime(CLOCK_MONOTONIC_RAW, &start1);
    for(i = 0; i < nb_test; i++)
    {
        ec_point_mul_unified(data->P3[i%nb_point], n, data->P1[i%nb_point], data->E, STACK_1);
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end1);

    data->E->f = JACOBIAN;
    clock_gettime(CLOCK_MONOTONIC_RAW, &start2);
    for(i = 0; i < nb_test; i++)
    {
        ec_point_mul_unified(data->P3[i%nb_point], n, data->P1[i%nb_point], data->E, STACK_1);
    }
    clock_gettime(CLOCK_MONOTONIC_RAW, &end2);
    printf("Mul Unified | %lu s %02u ns         | %lu s %02u ns       \n", get_s(&start1, &end1), get_ns(&start1, &end1), get_s(&start2, &end2), get_ns(&start2, &end2));
    
    ///************************************ MUL Montgomery *******************************/

    //data->E->f = JACOBIAN;
    //clock_gettime(CLOCK_MONOTONIC_RAW, &start2);
    //for(i = 0; i < nb_test; i++)
    //{
        //weierstrass_Zmontgomery_ladder(data->P3[i%nb_point], n, data->P1[i%nb_point], data->E, STACK_1);
    //}
    //clock_gettime(CLOCK_MONOTONIC_RAW, &end2);
    //printf("Mul Montg   |                      | %lu s %02u ns       \n", get_s(&start2, &end2), get_ns(&start2, &end2));
    
    ///************************************ MUL Joye *******************************/

    //data->E->f = JACOBIAN;
    //clock_gettime(CLOCK_MONOTONIC_RAW, &start2);
    //for(i = 0; i < nb_test; i++)
    //{
        //weierstrass_Zjoye_mul(data->P3[i%nb_point], n, data->P1[i%nb_point], data->E, STACK_1);
    //}
    //clock_gettime(CLOCK_MONOTONIC_RAW, &end2);
    //printf("Mul Joye    |                      | %lu s %02u ns       \n", get_s(&start2, &end2), get_ns(&start2, &end2));
    number_free(&n);
}


bool
test_curve_weierstrass(const ec_known_curve id_curve, const uint16_t nbits, 
        const uint32_t num_point, data_t *data)
{
    bool res;
    test_curve_weierstrass_alloc(nbits, num_point, data);
    ec_use_curve(data->E, data->k, id_curve, PROJECTIVE, STACK_1);
    test_curve_weierstrass_init(data);
    res = test_curve_weierstrass_read_points(data);
    test_curve_weierstrass_perf(data);
    test_curve_weierstrass_free(data);
    return res;
}

int main()
{
    data_t data;
    init_mphell(128, RANDOM_AES128, DEVURANDOM);
    int nb_points=100;
    //if (!test_curve_weierstrass(NIST_192, 192, nb_points, &data))
    //{
        //free_mphell();
        //return -1;
    //}
    //if (!test_curve_weierstrass(NIST_224, 224, nb_points, &data))
    //{
        //free_mphell();
        //return -1;
    //}
    if (!test_curve_weierstrass(NIST_256, 256, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    if (!test_curve_weierstrass(NIST_384, 384, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    if (!test_curve_weierstrass(NIST_521, 521, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    //if (!test_curve_weierstrass(BRAINPOOL_160, 160, nb_points, &data))
    //{
        //free_mphell();
        //return -1;
    //}
    //if (!test_curve_weierstrass(BRAINPOOL_192, 192, nb_points, &data))
    //{
        //free_mphell();
        //return -1;
    //}
    //if (!test_curve_weierstrass(BRAINPOOL_224, 224, nb_points, &data))
    //{
        //free_mphell();
        //return -1;
    ////}
    if (!test_curve_weierstrass(BRAINPOOL_256, 256, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    //if (!test_curve_weierstrass(BRAINPOOL_320, 320, nb_points, &data))
    //{
        //free_mphell();
        //return -1;
    //}
    if (!test_curve_weierstrass(BRAINPOOL_384, 384, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    if (!test_curve_weierstrass(BRAINPOOL_512, 512, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    if (!test_curve_weierstrass(FR_256, 256, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    free_mphell();
    return 0;
}
