/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France)

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell_test_ec_random.c
  * \brief Self check test for the function ec_random over FP field
  */

#include "mphell.h"
#include <stdio.h>
#include <string.h>

int main()
{
    init_mphell(128, RANDOM_AES128, DEVURANDOM);
    
    int nb_iteration = 10;
    int i, j;
    bool test = true;
    bool res1, res2;
    
    number p, n;
    field k;
    ec_curve E;
    
    ec_point P1[10];
    ec_point P2[10];
    ec_point P3[10];
    
    number_init(&p, bits_to_nblock(512));
    number_init(&n, bits_to_nblock(512));
    field_alloc(k, FP, bits_to_nblock(512), NULL);
    
    number_set_str(p, "6703903964971298549787012499102923063739682910296196688861780721860882015036773488400937149083451713845015929093243025426876941405973284973216824503042159", 10); 
    
    field_create(k, "k_512", STACK_1, 1, p);
    
    ec_alloc(E, k);
    ec_init(E, k);
    
    for(i=0; i< 10; i++)
    {
        ec_point_alloc(P1[i], k);
        ec_point_alloc(P2[i], k);
        ec_point_alloc(P3[i], k);
        ec_point_init(P1[i], k);
        ec_point_init(P2[i], k);
        ec_point_init(P3[i], k);
    }
    
    char seed[41];

    char bad_seed[41]="ffffffffffffffffffffffffffffffffffffffff";
    
    /* Create random Weietrass curve and test them */
    for(i=0; i<nb_iteration; i++)
    {
        /* Create random curve */
        ec_random(E, "randomCurveWeiertrass", seed, k, WEIERSTRASS, PROJECTIVE, STACK_1);
        
        /* Test the seed given with the generated curve */
        res1 = ec_verify_random_generation(E, seed, STACK_1);
        res2 = ec_verify_random_generation(E, bad_seed, STACK_1);
        strcpy(bad_seed, seed); /* Set the previous good seed in bad seed */
        if((res1==true) && (res2==false))
        {
            printf("Weietrass: E was generated from seed:%s\n", seed);
        }
        else
        {
            test = false;
            printf("Error random Weiertrass\n");
        }
        
        /* Test base arithmetic */
        for(j=0; j< 10; j++)
        {
            ec_point_random(P1[j], E, STACK_1);
            ec_point_random(P2[j], E, STACK_1);
            
            /* Add */
            
            ec_point_add(P3[j], P1[j], P2[j], E, STACK_1);
            
            /* Dbl */
            
            ec_point_dbl(P3[j], P1[j], E, STACK_1);
            
            /* Mul */
            
            number_random(n, bits_to_nblock(512));
            ec_point_mul(P3[j], n, P1[j], E, STACK_1);
        }
    }
    
    /* Create random Jacobi Quartic curve and test them */
    for(i=0; i<nb_iteration; i++)
    {
        /* Create random curve */
        ec_random(E, "randomCurveJacobiQuartic", seed, k, JACOBI_QUARTIC, PROJECTIVE, STACK_1);
        
        /* Test the seed given with the generated curve */
        res1 = ec_verify_random_generation(E, seed, STACK_1);
        res2 = ec_verify_random_generation(E, bad_seed, STACK_1);
        strcpy(bad_seed, seed); /* Set the previous good seed in bad seed */
        if((res1==true) && (res2==false))
        {
            printf("Jacobi Quartic: E was generated from seed:%s\n", seed);
        }
        else
        {
            test = false;
            printf("Error random Jacobi Quartic \n");
        }
        
        /* Test base arithmetic */
        for(j=0; j< 10; j++)
        {
            ec_point_random(P1[j], E, STACK_1);
            ec_point_random(P2[j], E, STACK_1);
            
            /* Add */
            
            ec_point_add(P3[j], P1[j], P2[j], E, STACK_1);
            
            /* Dbl */
            
            ec_point_dbl(P3[j], P1[j], E, STACK_1);
            
            /* Mul */
            
            number_random(n, bits_to_nblock(512));
            ec_point_mul(P3[j], n, P1[j], E, STACK_1);
        }
        
    }
    
    /* Create random Edwards curve and test them */
    for(i=0; i<nb_iteration; i++)
    {
        /* Create random curve */
        ec_random(E, "randomCurveEdwards", seed, k, EDWARDS, EXTENDED_EDWARDS, STACK_1);
        
        /* Test the seed given with the generated curve */
        res1 = ec_verify_random_generation(E, seed, STACK_1);
        res2 = ec_verify_random_generation(E, bad_seed, STACK_1);
        strcpy(bad_seed, seed); /* Set the previous good seed in bad seed */
        if((res1==true) && (res2==false))
        {
            printf("Edward: E was generated from seed:%s\n", seed);
        }
        else
        {
            test = false;
            printf("Error random Edward \n");
        }
        
        /* Test base arithmetic */
        for(j=0; j< 10; j++)
        {
            ec_point_random(P1[j], E, STACK_1);
            ec_point_random(P2[j], E, STACK_1);
            
            /* Add */
            
            ec_point_add(P3[j], P1[j], P2[j], E, STACK_1);
            
            /* Dbl */
            
            ec_point_dbl(P3[j], P1[j], E, STACK_1);
            
            /* Mul */
            
            number_random(n, bits_to_nblock(512));
            ec_point_mul(P3[j], n, P1[j], E, STACK_1);
        }
    }
    
    /* Free allocated memory */
    
    for(i=0; i< 10; i++)
    {
        ec_point_free(P1[i], k);
        ec_point_free(P2[i], k);
        ec_point_free(P3[i], k);
    }
    
    number_free(&p);
    number_free(&n);
    ec_free(E);
    field_free(k);
    free_mphell();
    
    /* Result */
    
    if (test)
    {
        printf("SUCCESS \n");
        return 0;
    }
    else
    {
        printf("ERROR \n");
        return -1;
    }
}
