/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France)

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell_test_ec_random_fp3.c
  * \brief Self check test for the function ec_random over FP3 field
  */

#include "mphell.h"
#include <stdio.h>
#include <string.h>

int main()
{
    init_mphell(128, RANDOM_AES128, DEVURANDOM);
    
    int nb_iteration = 1;
    int i, j;
    bool test = true;
    bool res1, res2;
    
    number p, n;
    field k, k3;
    ec_curve E;
    
    ec_point P1[10];
    ec_point P2[10];
    ec_point P3[10];
    
    number_init(&p, bits_to_nblock(384));
    number_init(&n, bits_to_nblock(384));
    field_alloc(k, FP, bits_to_nblock(384), NULL);
    
    number_set_str(p, "8CB91E82A3386D280F5D6F7E50E641DF152F7109ED5456B412B1DA197FB71123ACD3A729901D1A71874700133107EC53", 16);
    field_create(k, "k_384", STACK_1, 1, p);
    
    /* Create finite field GF(p^3) of characteristic p and order p^3 */
    field_elt non_residue;
    field_elt_alloc(&non_residue, k);
    field_elt_init(non_residue, k);
    /* Fix the irreducible polynomial P for the test: P = X^3 - 621f1701f33bf1abbee217f8f17520cbb28d366f424c1c2569c37ee91b7508dc1dfa5fb4041a4153e9b62a45b755533e */
    field_elt_set_str(non_residue, "621f1701f33bf1abbee217f8f17520cbb28d366f424c1c2569c37ee91b7508dc1dfa5fb4041a4153e9b62a45b755533e", 16, false, k, STACK_1);
    field_alloc(k3, FP3, bits_to_nblock(384), k);
    field_create(k3, "", STACK_1, 2, k, non_residue);
    ec_alloc(E, k3);
    ec_init(E, k3);
    
    for(i=0; i< 10; i++)
    {
        ec_point_alloc(P1[i], k3);
        ec_point_alloc(P2[i], k3);
        ec_point_alloc(P3[i], k3);
        ec_point_init(P1[i], k3);
        ec_point_init(P2[i], k3);
        ec_point_init(P3[i], k3);
    }
    
    char seed[41];

    char bad_seed[41]="ffffffffffffffffffffffffffffffffffffffff";
    
    /* Create random Weietrass curve and test them */
    for(i=0; i<nb_iteration; i++)
    {
        /* Create random curve */
        ec_random(E, "randomCurveWeiertrassFP3", seed, k3, WEIERSTRASS, PROJECTIVE, STACK_1);
        
        /* Test the seed given with the generated curve */
        res1 = ec_verify_random_generation(E, seed, STACK_1);
        res2 = ec_verify_random_generation(E, bad_seed, STACK_1);
        strcpy(bad_seed, seed); /* Set the previous good seed in bad seed */
        if((res1==true) && (res2==false))
        {
            printf("Weietrass: E was generated from seed:%s\n", seed);
        }
        else
        {
            test = false;
            printf("Error random Weiertrass\n");
        }
        
        /* Test base arithmetic */
        for(j=0; j< 10; j++)
        {
            ec_point_random(P1[j], E, STACK_1);
            ec_point_random(P2[j], E, STACK_1);
            
            /* Add */
            
            ec_point_add(P3[j], P1[j], P2[j], E, STACK_1);
            
            /* Dbl */
            
            ec_point_dbl(P3[j], P1[j], E, STACK_1);
            
            /* Mul */
            
            number_random(n, bits_to_nblock(384));
            ec_point_mul(P3[j], n, P1[j], E, STACK_1);
        }
    }
    
    /* Create random Jacobi Quartic curve and test them */
    for(i=0; i<nb_iteration; i++)
    {
        /* Create random curve */
        ec_random(E, "randomCurveJacobiQuarticFP3", seed, k3, JACOBI_QUARTIC, PROJECTIVE, STACK_1);
        
        /* Test the seed given with the generated curve */
        res1 = ec_verify_random_generation(E, seed, STACK_1);
        res2 = ec_verify_random_generation(E, bad_seed, STACK_1);
        strcpy(bad_seed, seed); /* Set the previous good seed in bad seed */
        if((res1==true) && (res2==false))
        {
            printf("Jacobi Quartic: E was generated from seed:%s\n", seed);
        }
        else
        {
            test = false;
            printf("Error random Jacobi Quartic \n");
        }
        
        /* Test base arithmetic */
        for(j=0; j< 10; j++)
        {
            ec_point_random(P1[j], E, STACK_1);
            ec_point_random(P2[j], E, STACK_1);
            
            /* Add */
            
            ec_point_add(P3[j], P1[j], P2[j], E, STACK_1);
            
            /* Dbl */
            
            ec_point_dbl(P3[j], P1[j], E, STACK_1);
            
            /* Mul */
            
            number_random(n, bits_to_nblock(384));
            ec_point_mul(P3[j], n, P1[j], E, STACK_1);
        }
        
    }
    
    /* Create random Edwards curve and test them */
    for(i=0; i<nb_iteration; i++)
    {
        /* Create random curve */
        ec_random(E, "randomCurveEdwardsFP3", seed, k3, EDWARDS, EXTENDED_EDWARDS, STACK_1);
        
        /* Test the seed given with the generated curve */
        res1 = ec_verify_random_generation(E, seed, STACK_1);
        res2 = ec_verify_random_generation(E, bad_seed, STACK_1);
        strcpy(bad_seed, seed); /* Set the previous good seed in bad seed */
        if((res1==true) && (res2==false))
        {
            printf("Edward: E was generated from seed:%s\n", seed);
        }
        else
        {
            test = false;
            printf("Error random Edward \n");
        }
        
        /* Test base arithmetic */
        for(j=0; j< 10; j++)
        {
            ec_point_random(P1[j], E, STACK_1);
            ec_point_random(P2[j], E, STACK_1);
            
            /* Add */
            
            ec_point_add(P3[j], P1[j], P2[j], E, STACK_1);
            
            /* Dbl */
            
            ec_point_dbl(P3[j], P1[j], E, STACK_1);
            
            /* Mul */
            
            number_random(n, bits_to_nblock(384));
            ec_point_mul(P3[j], n, P1[j], E, STACK_1);
        }
    }
    
    /* Free allocated memory */
    
    for(i=0; i< 10; i++)
    {
        ec_point_free(P1[i], k3);
        ec_point_free(P2[i], k3);
        ec_point_free(P3[i], k3);
    }
    
    number_free(&p);
    number_free(&n);
    field_elt_free(&non_residue, k);
    ec_free(E);
    field_free(k3);
    field_free(k);
    free_mphell();
    
    /* Result */
    
    if (test)
    {
        printf("SUCCESS \n");
        return 0;
    }
    else
    {
        printf("ERROR \n");
        return -1;
    }
}
