/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France)

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell_test_edwards_arithmetic_extended_edwards.c
  * \brief Self check tests for the functions: edwards_point_add_unified, edwards_point_add, edwards_point_dbl edwards_point_neg ec_point_mul in Extended Edwards coordinates
  */

#include "mphell_test_weierstrass.h"
#include "mphell_test_edwards.h"

bool 
test_curve_edwards_arithmetic_extended_edwards(data_t *data)
{
    printf("Validity tests on curve %s with %d points: ", data->E->id_curve, data->nb_points);
    uint8_t n = data->k->size;
    uint32_t i = 0;
    uint8_t j = 0;
    ec_point Res;
    number n1;
    uint8_t check[data->nb_points][9];
    FILE *filep;
    char test_file[200];

    for(i = 0; i < data->nb_points; i++)
    {
        for(j = 0; j < 9; j++)
        {
            check[i][j] = (uint8_t)0;
        }
    }
    sprintf(test_file, DATADIR "/vector_test_elliptic_curves/%s_vector_test.txt", data->E->id_curve);
    filep = fopen(test_file,"r");
    if(filep == NULL)
    {
        printf(" \n Chemin:    %s \n", test_file);
        fprintf(stderr, "Error : no curve data1 file found \n");        
        exit(EXIT_FAILURE);
    }
    number_init(&n1, n);
    ec_point_alloc(Res, data->k);
    ec_point_init(Res, data->k);
    
    for(i = 0; i < data->nb_points; i++)
    {        
        test_curve_edwards_read_point(filep, data->P1[i], data);    /* We read P1  */
        test_curve_edwards_read_point(filep, Res, data);            /* We read -P1 */
        test_curve_edwards_read_point(filep, data->P2[i], data);    /* We read P2  */
        
        if(!edwards_belongs(data->P1[i], data->E, STACK_1))
        {
            printf("P1 does not belongs to E\n");
            return false;
        }
        
        if(!edwards_belongs(data->P2[i], data->E, STACK_1))
        {
            printf("P2 does not belongs to E\n");
            return false;
        }
        
        /* P1 + P2                  */
        test_curve_edwards_read_point(filep, Res, data);    /* We read P1 + P2 */
        
        /* --> PROJECTIVE           */
        data->E->f = PROJECTIVE;
        
        /* ----> DEDICATED          */
        ec_point_add(data->P3[i], data->P1[i], data->P2[i], data->E, STACK_1);
        if(!edwards_belongs(data->P3[i], data->E, STACK_1))
        {
            printf("P3=P1+P2 (dedicated addition) in Projective coordinate does not belongs to E\n");
            return false;
        }
        check[i][0] = (uint8_t)edwards_point_are_equal(data->P3[i], Res, data->E, STACK_1);
        
        /* --> EXTENDED_EDWARDS     */
        data->E->f = EXTENDED_EDWARDS;
        
        /* ----> UNIFIED            */
        ec_point_add_unified(data->P3[i], data->P1[i], data->P2[i], data->E, STACK_1);
        if(!edwards_belongs(data->P3[i], data->E, STACK_1))
        {
            printf("P3=P1+P2 (unified addition) in Extended Edwards coordinate does not belongs to E\n");
            return false;
        }
        check[i][1] = (uint8_t)edwards_point_are_equal(data->P3[i], Res, data->E, STACK_1);
        
        /* ----> DEDICATED          */
        ec_point_add(data->P3[i], data->P1[i], data->P2[i], data->E, STACK_1);
        if(!edwards_belongs(data->P3[i], data->E, STACK_1))
        {
            printf("P3=P1+P2 (dedicated addition) in Extended Edwards does not belongs to E\n");
            return false;
        }
        check[i][2] = (uint8_t)edwards_point_are_equal(data->P3[i], Res, data->E, STACK_1);


        /* 2 * P1                   */
        test_curve_edwards_read_point(filep, Res, data);            /* We read 2 * P1 */
        
        /* --> PROJECTIVE           */
        data->E->f = PROJECTIVE;
        
        /* ----> DEDICATED          */
        ec_point_dbl(data->P3[i], data->P1[i], data->E, STACK_1);
        if(!edwards_belongs(data->P3[i], data->E, STACK_1))
        {
            printf("P3=2*P1 (dedicated doubling) in Projective coordinate does not belongs to E\n");
            return false;
        }
        check[i][3] = (uint8_t)edwards_point_are_equal(data->P3[i], Res, data->E, STACK_1);
        
        /* --> EXTENDED_EDWARDS     */
        data->E->f = EXTENDED_EDWARDS;
        
        /* ----> UNIFIED            */
        ec_point_dbl_unified(data->P3[i], data->P1[i], data->E, STACK_1);
        if(!edwards_belongs(data->P3[i], data->E, STACK_1))
        {
            printf("P3=2*P1 (unified doubling) in Extended Edwards coordinate does not belongs to E\n");
            return false;
        }
        check[i][4] = (uint8_t)edwards_point_are_equal(data->P3[i], Res, data->E, STACK_1);
        
        /* ----> DEDICATED          */
        ec_point_dbl(data->P3[i], data->P1[i], data->E, STACK_1);
        if(!edwards_belongs(data->P3[i], data->E, STACK_1))
        {
            printf("P3=2*P1 (dedicated doubling) in Extended Edwards coordinate does not belongs to E\n");
            return false;
        }
        check[i][5] = (uint8_t)edwards_point_are_equal(data->P3[i], Res, data->E, STACK_1);
        

        /* n1 * P1                  */
        test_curve_weierstrass_read_data(filep, n1, data);      /* We read n1       */
        test_curve_edwards_read_point(filep, Res, data);        /* We read n1 * P1  */
        
        /* --> PROJECTIVE           */
        data->E->f = PROJECTIVE;
        
        /* ----> DEDICATED          */
        ec_point_mul(data->P3[i], n1, data->P1[i], data->E, STACK_1);
        if(!edwards_belongs(data->P3[i], data->E, STACK_1))
        {
            printf("P3=n.P1 (dedicated multiplication) in Projective coordinate does not belongs to E\n");
            return false;
        }
        check[i][6] = edwards_point_are_equal(data->P3[i], Res, data->E, STACK_1);
        
        /* --> EXTENDED_EDWARDS     */
        data->E->f = EXTENDED_EDWARDS;
        
        /* ----> UNIFIED            */
        ec_point_mul_unified(data->P3[i], n1, data->P1[i], data->E, STACK_1);
        if(!edwards_belongs(data->P3[i], data->E, STACK_1))
        {
            printf("P3=n.P1 (unified multiplication) in Extended Edwards coordinate does not belongs to E\n");
            return false;
        }
        check[i][7] = edwards_point_are_equal(data->P3[i], Res, data->E, STACK_1);
        
        /* ----> DEDICATED          */
        ec_point_mul(data->P3[i], n1, data->P1[i], data->E, STACK_1);
        if(!edwards_belongs(data->P3[i], data->E, STACK_1))
        {
            printf("P3=n.P1 (dedicated multiplication) in Extended Edwards coordinate does not belongs to E\n");
            return false;
        }
        check[i][8] = edwards_point_are_equal(data->P3[i], Res, data->E, STACK_1);
    }
    
    /* ----> DEDICATED          */
    
    number_set_ui(n1, 0);
    
    ec_point_mul(data->P3[0], n1, data->P1[0], data->E, STACK_1);
    
    if(!edwards_belongs(data->P3[0], data->E, STACK_1))
    {
        printf("P3=n.P1 in Extended Edwards coordinate, dedicated does not belongs to E\n");
        printf("P3[0] = \n"); ec_point_print(data->P3[0], 16, true, data->k, STACK_1); printf("\n");
        return false;
    }
    if(!edwards_point_is_neutral(data->P3[0], data->E))
    {
        printf("P3=0.P1 in Extended Edwards coordinate, dedicated is not neutral E\n");
        printf("P3[0] = \n"); ec_point_print(data->P3[0], 16, true, data->k, STACK_1); printf("\n");
        return false;
    }
            
    /* ----> UNIFIED            */
    
    ec_point_mul_unified(data->P3[0], n1, data->P1[0], data->E, STACK_1);
    
    if(!edwards_belongs(data->P3[0], data->E, STACK_1))
    {
        printf("P3=n.P1 in Extended Edwards coordinate, unified does not belongs to E\n");
        printf("P3[0] = \n"); ec_point_print(data->P3[0], 16, true, data->k, STACK_1); printf("\n");
        return false;
    }
    if(!edwards_point_is_neutral(data->P3[0], data->E))
    {
        printf("P3=0.P1 in Extended Edwards coordinate, unified is not neutral E\n");
        printf("P3[0] = \n"); ec_point_print(data->P3[0], 16, true, data->k, STACK_1); printf("\n");
        return false;
    }

    fclose(filep);
    ec_point_free(Res, data->k);
    number_free(&n1);
    
    for(i = 0; i < data->nb_points; i++)
    {
        for(j = 0; j < 9; j++)
        {
            if(check[i][j] == false)
            {
                printf("ERROR\n");
                printf("i = %d, j= %d \n", i , j);
                return false;
            }
        }
    }
    printf("SUCCESS\n");
    return true;
}

bool
test_curve_edwards(const ec_known_curve id_curve, const uint16_t nbits, 
        const uint32_t num_point, data_t *data)
{
    bool res;
    test_curve_weierstrass_alloc(nbits, num_point, data);       /* We can re use this function */
    ec_use_curve(data->E, data->k, id_curve, EXTENDED_EDWARDS, STACK_1);
    test_curve_weierstrass_init(data);
    res=test_curve_edwards_arithmetic_extended_edwards(data);
    test_curve_weierstrass_free(data);                          /* We can re use this function */
    return res;
}

int main()
{
    data_t data;
    init_mphell(128, RANDOM_AES128, DEVURANDOM);
    int nb_points=100;
    if (!test_curve_edwards(ED_25519, 256, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    free_mphell();
    return 0;
}
