/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France)

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell_test_jacobi_quartic.c
  * \brief Self check tests for Jacobi Quartic Elliptic Curves.
  */


#include "mphell_test_jacobi_quartic.h"

void 
test_curve_jacobi_quartic_alloc(const uint16_t nbits, const uint32_t num_point, data_t *data)
{
    uint8_t n = bits_to_nblock(nbits);
    data->nb_points = num_point;
    data->nbits = nbits;
    
    field_alloc(data->k, FP, n, NULL);
    ec_alloc(data->E, data->k);
    ec_alloc(data->E1, data->k);
    data->P1 = (ec_point*)malloc(sizeof(ec_point) * num_point);
    data->P2 = (ec_point*)malloc(sizeof(ec_point) * num_point);
    data->P3 = (ec_point*)malloc(sizeof(ec_point) * num_point);

    uint32_t i;
    for(i = 0; i < num_point; i++)
    {
        ec_point_alloc(data->P1[i], data->k);
        ec_point_alloc(data->P2[i], data->k);
        ec_point_alloc(data->P3[i], data->k);
    }

    ec_alloc(data->Ew, data->k);
    data->P1w = (ec_point*)malloc(sizeof(ec_point) * num_point);
    data->P2w = (ec_point*)malloc(sizeof(ec_point) * num_point);
    data->P3w = (ec_point*)malloc(sizeof(ec_point) * num_point);

    for(i = 0; i < num_point; i++)
    {
        ec_point_alloc(data->P1w[i], data->k);
        ec_point_alloc(data->P2w[i], data->k);
        ec_point_alloc(data->P3w[i], data->k);
    }
}

void 
test_curve_jacobi_quartic_init(data_t *data)
{
    uint32_t i;
    for(i = 0; i < data->nb_points; i++)
    {
        ec_point_init(data->P1[i], data->k);
        ec_point_init(data->P2[i], data->k);
        ec_point_init(data->P3[i], data->k);
    }

    for(i = 0; i < data->nb_points; i++)
    {
        ec_point_init(data->P1w[i], data->k);
        ec_point_init(data->P2w[i], data->k);
        ec_point_init(data->P3w[i], data->k);
    }
}

void 
test_curve_jacobi_quartic_free(data_t *data)
{
    ec_free(data->E);
    ec_free(data->E1);
    uint64_t i;
    for(i = 0; i < data->nb_points; i++)
    {
        ec_point_free(data->P1[i], data->k);
        ec_point_free(data->P2[i], data->k);
        ec_point_free(data->P3[i], data->k);
    }
    free(data->P1);
    free(data->P2);
    free(data->P3);

    ec_free(data->Ew);
    for(i = 0; i < data->nb_points; i++)
    {
        ec_point_free(data->P1w[i], data->k);
        ec_point_free(data->P2w[i], data->k);
        ec_point_free(data->P3w[i], data->k);
    }
    free(data->P1w);
    free(data->P2w);
    free(data->P3w);
    field_free(data->k);
}

void
test_curve_jacobi_quartic_read_silent(const char *id_curve, data_t *data, bool boolean)
{
    char* curve_file = malloc(sizeof(char) * 200);
    sprintf(curve_file, DATADIR "/elliptic_curves/%s.txt", id_curve);
    FILE *fp = NULL;
    char line[160];
    fp = fopen(curve_file, "r");
    free(curve_file);

    if(fp == NULL)
    {
        fprintf(stderr, "Error : no test data file found \n");
        exit(EXIT_FAILURE);
    }
    char *str = (char*)malloc(sizeof(char) * 200);
    number tmp;
    number_init(&tmp, bits_to_nblock(data->nbits));

    while(fgets(line, sizeof(line), fp) != NULL)
    { 
        if(strncmp(line, "p = ", 4)== 0)
        {
            strcpy(str, line + 4);
            number_set_str(tmp, str, 16);
            field_create(data->k, "", STACK_1, 1, tmp);
            /* Initialise the curve, now that k is set up */
            ec_init(data->E, data->k);
            ec_init(data->E1, data->k);
            ec_init(data->Ew, data->k);
        }
        if(strncmp(line, "a = ", 4)== 0)
        {
            strcpy(str, line + 4);
            number_set_str(tmp, str, 16); 
            field_elt_set_number(data->E->a, false, data->k, STACK_1, 1, tmp);
        }
        if(strncmp(line, "b = ", 4)== 0)
        {
            strcpy(str, line + 4);
            number_set_str(tmp, str, 16);
            field_elt_set_number(data->E->b, false, data->k, 1, STACK_1, tmp);
        }
        if(strncmp(line, "xG = ", 5)== 0)
        {
            strcpy(str, line + 5);
            number_set_str(tmp, str, 16);
            field_elt_set_number(data->E->G->x, false, data->k, STACK_1, 1, tmp);
        }
        if(strncmp(line, "yG = ", 5)== 0)
        {
            strcpy(str, line + 5);
            number_set_str(tmp, str, 16);
            field_elt_set_number(data->E->G->y, false, data->k, STACK_1, 1, tmp);
        }
        if(strncmp(line, "zG = ", 5)== 0)
        {
            strcpy(str, line + 5);
            number_set_str(tmp, str, 16);
            field_elt_set_number(data->E->G->z, false, data->k, STACK_1, 1, tmp);
        }
        if(strncmp(line, "tG = ", 5)== 0)
        {
            strcpy(str, line + 5);
            number_set_str(tmp, str, 16);
            field_elt_set_number(data->E->G->t, false, data->k, STACK_1, 1, tmp);
        }
        if(strncmp(line, "n = ", 4) == 0)
        {
            strcpy(str, line + 4);
            number_set_str(data->E->n, str, 16);
            break;
        }
    }
    data->E->k = data->k;
    data->E->type = JACOBI_QUARTIC;
    strcpy(data->E->id_curve, id_curve);
    ec_compute_disc(data->E, STACK_1);
    
    if(!boolean)
    {
        printf("\n\n");
        printf("Jacobian curve :\n");
        ec_curve_print(data->E, 16, STACK_1);
        printf("\n\n");
    }
    /* Convert E into a Weierstrass elliptic curve */
    
    strcpy(data->Ew->id_curve, id_curve);
    jacobi_quartic_to_weierstrass(data->Ew, data->E, STACK_1);

    if(!boolean)
    {
        printf("Weierstrass curve, using jacobi_quartic_to_weierstrass:\n");
        ec_curve_print(data->Ew, 16, STACK_1);
        printf("\n\n");
    }      

    field_elt t1, t2;
    field_elt_alloc(&t1, data->k);
    field_elt_init(t1, data->k);
    field_elt_alloc(&t2, data->k);
    field_elt_init(t2, data->k);
    field_elt_set_ui(t1, 3, false, data->k, STACK_1);        /* t1 = 3                                           */
    field_elt_div(t1, data->E->a, t1, data->k, STACK_1);              /* t1 = a / 3                                       */
    field_elt_set_ui(t2, 4, false, data->k, STACK_1);        /* t2 = 4                                           */
    field_elt_mul(t1, t1, t2, data->k, STACK_1);                      /* t1 = (4/3).a                                     */
    field_elt_neg(t2, t1, data->k);                          /* t2 = -(4/3).a                                    */
    
    /* Convert Ew into a Jacobi Quartic elliptic curve */
    
    strcpy(data->E1->id_curve, id_curve); 
    weierstrass_to_jacobi_quartic(data->E1, data->Ew, false, t1, STACK_1);

    MPHELL_ASSERT_ALWAYS( field_elt_cmp(t1, t2, data->k)==0, "Error : miscomputation of theta \n");

    field_elt_copy(t1,data->E1->a, data->k);    

    weierstrass_to_jacobi_quartic(data->E1, data->Ew, true, t2, STACK_1);

    MPHELL_ASSERT_ALWAYS( field_elt_cmp(t1, data->E1->a, data->k)==0, "Error : conversion with pre computed theta \n");

    field_elt_free(&t1, data->k);
    field_elt_free(&t2, data->k);
    if(!boolean)
    {
        printf("Jacobian curve, using weierstrass_to_jacobi_quartic :\n");
        ec_curve_print(data->E1, 16, STACK_1);

        printf("\n\n");
    }

    fclose(fp);
    free(str);
    number_free(&tmp);
}

void
test_curve_jacobi_quartic_read(const char *id_curve, data_t *data)
{
    char* curve_file = malloc(sizeof(char) * 200);
    sprintf(curve_file, DATADIR "/elliptic_curves/%s.txt", id_curve);
    FILE *fp = NULL;
    char line[160];
    fp = fopen(curve_file, "r");
    free(curve_file);

    if(fp == NULL)
    {
        fprintf(stderr, "Error : no test data file found \n");
        exit(EXIT_FAILURE);
    }
    char *str = (char*)malloc(sizeof(char) * 200);
    number tmp;
    number_init(&tmp, bits_to_nblock(data->nbits));

    while(fgets(line, sizeof(line), fp) != NULL)
    { 
        if(strncmp(line, "p = ", 4)== 0)
        {
            strcpy(str, line + 4);
            number_set_str(tmp, str, 16);
            field_create(data->k, "", STACK_1, 1, tmp);
            /* Initialise the curve, now that k is set up */
            ec_init(data->E, data->k);
            ec_init(data->E1, data->k);
            ec_init(data->Ew, data->k);
        }
        if(strncmp(line, "a = ", 4)== 0)
        {
            strcpy(str, line + 4);
            number_set_str(tmp, str, 16); 
            field_elt_set_number(data->E->a, false, data->k, STACK_1, 1, tmp);
        }
        if(strncmp(line, "b = ", 4)== 0)
        {
            strcpy(str, line + 4);
            number_set_str(tmp, str, 16);
            field_elt_set_number(data->E->b, false, data->k, STACK_1, 1, tmp);
        }
        if(strncmp(line, "xG = ", 5)== 0)
        {
            strcpy(str, line + 5);
            number_set_str(tmp, str, 16);
            field_elt_set_number(data->E->G->x, false, data->k, STACK_1, 1, tmp);
        }
        if(strncmp(line, "yG = ", 5)== 0)
        {
            strcpy(str, line + 5);
            number_set_str(tmp, str, 16);
            field_elt_set_number(data->E->G->y, false, data->k, STACK_1, 1, tmp);
        }
        if(strncmp(line, "zG = ", 5)== 0)
        {
            strcpy(str, line + 5);
            number_set_str(tmp, str, 16);
            field_elt_set_number(data->E->G->z, false, data->k, STACK_1, 1, tmp);
        }
        if(strncmp(line, "tG = ", 5)== 0)
        {
            strcpy(str, line + 5);
            number_set_str(tmp, str, 16);
            field_elt_set_number(data->E->G->t, false, data->k, STACK_1, 1, tmp);
        }
        if(strncmp(line, "n = ", 4) == 0)
        {
            strcpy(str, line + 4);
            number_set_str(data->E->n, str, 16);
            break;
        }
    }
   
    data->E->k = data->k;
    data->E->type = JACOBI_QUARTIC;
    strcpy(data->E->id_curve, id_curve);
    ec_compute_disc(data->E, STACK_1);
    
    printf("\n\n");
    printf("Jacobian curve :\n");
    ec_curve_print(data->E, 16, STACK_1);
    printf("\n\n");
    
    /* Convert E into a Weierstrass elliptic curve */
    
    jacobi_quartic_to_weierstrass(data->Ew, data->E, STACK_1);
    
    printf("Weierstrass curve, using jacobi_quartic_to_weierstrass:\n");
    ec_curve_print(data->Ew, 16, STACK_1);
    printf("\n\n");
    
    field_elt t1, t2;
    field_elt_alloc(&t1, data->k);
    field_elt_init(t1, data->k);
    field_elt_alloc(&t2, data->k);
    field_elt_init(t2, data->k);
    field_elt_set_ui(t1, 3, false, data->k, STACK_1);        /* t1 = 3                                           */
    field_elt_div(t1, data->E->a, t1, data->k, STACK_1);              /* t1 = a / 3                                       */
    field_elt_set_ui(t2, 4, false, data->k, STACK_1);        /* t2 = 4                                           */
    field_elt_mul(t1, t1, t2, data->k, STACK_1);                      /* t1 = (4/3).a                                     */
    field_elt_neg(t2, t1, data->k);                          /* t2 = -(4/3).a                                    */
    
    /* Convert Ew into a Jacobi Quartic elliptic curve */
    
    weierstrass_to_jacobi_quartic(data->E1, data->Ew, false, t1, STACK_1);

    MPHELL_ASSERT_ALWAYS( field_elt_cmp(t1, t2, data->k)==0, "Error : miscomputation of theta \n");

    field_elt_copy(t1,data->E1->a, data->k);    

    weierstrass_to_jacobi_quartic(data->E1, data->Ew, true, t2, STACK_1);

    MPHELL_ASSERT_ALWAYS( field_elt_cmp(t1, data->E1->a, data->k)==0, "Error : conversion with pre computed theta \n");

    field_elt_free(&t1, data->k);
    field_elt_free(&t2, data->k);
    
    printf("Jacobian curve, using weierstrass_to_jacobi_quartic :\n");
    ec_curve_print(data->E1, 16, STACK_1);

    printf("\n\n");
    
    fclose(fp);
    free(str);
    number_free(&tmp);
}

void
test_curve_jacobi_quartic_read_point(FILE *filep, ec_point P, data_t *data)
{
    char line[180];

    if(filep == NULL)
    {
        fprintf(stderr, "Error : no test data file found \n");
        exit(EXIT_FAILURE);
    }

    char *str = malloc(sizeof(char) * 170);
    field_elt tmp, tmp1;
    field_elt_alloc(&tmp, data->k);
    field_elt_init(tmp, data->k);
    field_elt_alloc(&tmp1, data->k);
    field_elt_init(tmp1, data->k);

    while(fgets(line, sizeof(line), filep) != NULL)
    { 
        if(strncmp(line, "x = ", 4) == 0)
        {
            strcpy(str, line + 4);
            /* Remove last character introduced by fgets */
            str[strlen(str)-1]='\0';
            
            if(str[strlen(str)-2] == '\\')
            {
                if(fgets(line, sizeof(line), filep) == NULL)
                {
                    exit(EXIT_FAILURE);
                }
                strcpy(str + strlen(str)-2, line);
            }
            field_elt_set_str(tmp, str, 10, false, data->k, STACK_1);
        }
        if(strncmp(line, "y = ", 4) == 0)
        {
            strcpy(str, line + 4);
            /* Remove last character introduced by fgets */
            str[strlen(str)-1]='\0';
            
            if(str[strlen(str)-2] == '\\')
            {
                if(fgets(line, sizeof(line), filep) == NULL)
                {
                    exit(EXIT_FAILURE);
                }
                strcpy(str + strlen(str)-2, line);
            }
            field_elt_set_str(tmp1, str, 10, false, data->k, STACK_1);
            ec_point_set_aff(P, tmp, tmp1, JACOBI_QUARTIC, data->k, STACK_1);
            break;
        }
    }
    free(str);
    field_elt_free(&tmp, data->k);
    field_elt_free(&tmp1, data->k);
}


void
test_curve_jacobi_quartic_read_data(FILE *filep, number n, data_t *data)
{
    char line[170];
    if(filep == NULL)
    {
        fprintf(stderr, "Error : no curve data file found \n");
        exit(EXIT_FAILURE);
    }
    char *str = malloc(sizeof(char) * 160);
    while(fgets(line, sizeof(line), filep) != NULL)
    { 
        if(strncmp(line, "k = ", 4) == 0)
        {
            strcpy(str, line + 4);
            /* Remove last character introduced by fgets */
            str[strlen(str)-1]='\0';
            if(str[strlen(str)-2] == '\\')
            {
                if(fgets(line, sizeof(line), filep) == NULL)
                {
                    exit(EXIT_FAILURE);
                }
                strcpy(str + strlen(str)-2, line);
            }
            number_set_str(n, str, 10);
            break;
        }
    }
    free(str);
}
