/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France)

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell_test_jacobi_quartic.h
  * \brief Self check tests for Jacobi Quartic Elliptic Curves.
  */


#ifndef MPHELL_TEST_JACOBI_H
#define MPHELL_TEST_JACOBI_H

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include "mphell.h"

/**
 * @struct data
 * @brief  datas for the tests
 */
struct data_s
{
    field k;        /*!< Base field                     */
    ec_curve E;     /*!< Jacobi Quartic Elliptic curve  */
    ec_curve Ew;    /*!< Weierstrass ellippic curve     */
    ec_curve E1;    /*!< Jacobi Quartic Elliptic curve  */
    
    int nb_points;  /*!< Size of the points arrays      */
    
    uint16_t nbits; /*!< Size in bits of the base field */

    ec_point *P1;   /*!< Array of elliptic curve points */
    ec_point *P2;   /*!< Array of elliptic curve points */
    ec_point *P3;   /*!< Array of elliptic curve points */
    
    ec_point *P1w;  /*!< Array of elliptic curve points */
    ec_point *P2w;  /*!< Array of elliptic curve points */
    ec_point *P3w;  /*!< Array of elliptic curve points */
};

/**
 * @typedef data_test
 * @brief   data for the tests
 */
typedef struct data_s data_t;

/**
 * @fn    void test_curve_jacobi_quartic_alloc(const uint16_t nbits, const uint32_t num_point, data_t *data);
 * @brief Allocate the data structure
 * @param nbits Size in bits of the base field
 * @param num_point Size of the elliptic curve points array
 * @param data Data for the tests
 */
void 
test_curve_jacobi_quartic_alloc(const uint16_t nbits, const uint32_t num_point, data_t *data);

/**
 * @fn    void test_curve_jacobi_quartic_init(data_t *data);
 * @brief Initialise the data structure, the field data->k must have been created
 * @param data Data for the tests
 */
void 
test_curve_jacobi_quartic_init(data_t *data);

/**
 * @fn    void test_curve_jacobi_quartic_free(data_t *data);
 * @brief Free the data structure
 * @param data Data for the tests
 */
void 
test_curve_jacobi_quartic_free(data_t *data);

/**
 * @fn    void test_curve_jacobi_quartic_read(const char *id_curve, data_t *data);
 * @brief Set E to the curve defined by the file id_curve.txt
 * @param id_curve Name of the curve, read the file data/elliptic_curves/id_curve.txt and do some test of conversion
 * @param data     Data for the tests
 */
void
test_curve_jacobi_quartic_read(const char *id_curve, data_t *data);

/**
 * @fn    void test_curve_jacobi_quartic_read_silent(const char *id_curve, data_t *data);
 * @brief Set E to the curve defined by the file id_curve.txt
 * @param id_curve  Name of the curve, read the file data/elliptic_curves/id_curve.txt
 * @param data      Data for the tests
 * @param boolean   A boolean saying if we want the function to print some tests of conversion
 */
void
test_curve_jacobi_quartic_read_silent(const char *id_curve, data_t *data, bool boolean);

/**
 * @fn    void test_curve_jacobi_quartic_read_point(FILE *filep, ec_point P, data_t *data);
 * @brief Read a point P=(x,y) in a vector test file
 * @param filep Name of the vector test file
 * @param P Destination point
 * @param data     Data for the tests
 */
void
test_curve_jacobi_quartic_read_point(FILE *filep, ec_point P, data_t *data);

/**
 * @fn    void test_curve_jacobi_quartic_read_data(FILE *filep, number n, data_t *data);
 * @brief Read a number in a vector test file
 * @param filep Name of the vector test file
 * @param n Destination number
 * @param data     Data for the tests
 */
void
test_curve_jacobi_quartic_read_data(FILE *filep, number n, data_t *data);


#endif
