/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France)

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell_test_jacobi_quartic_conversion.c
  * \brief Self check tests for the functions: 
    jacobi_quartic_point_to_weierstrass_point and 
    weierstrass_point_to_jacobi_quartic_point
  */

#include "mphell_test_jacobi_quartic.h"

bool 
test_curve_jacobi_quartic_read_points(data_t *data)
{
    printf("Validity tests on curve %s with %d points: ", data->E->id_curve, data->nb_points);
    uint8_t n = data->k->size;
    uint32_t i = 0;
    ec_point Res;
    number n1;
    FILE *filep;
    char test_file[200];

    sprintf(test_file, DATADIR "/vector_test_elliptic_curves/%s_vector_test.txt", data->E->id_curve);
    filep = fopen(test_file,"r");
    if(filep == NULL)
    {
        fprintf(stderr, "Error : no curve data file found \n");
        exit(EXIT_FAILURE);
    }
    number_init(&n1, n);
    ec_point_alloc(Res, data->k);
    ec_point_init(Res, data->k);
    
    for(i = 0; i < data->nb_points; i++)
    {
        data->E->f = PROJECTIVE;
        test_curve_jacobi_quartic_read_point(filep, data->P1[i], data);
        test_curve_jacobi_quartic_read_point(filep, data->P2[i], data);
        
        jacobi_quartic_point_to_weierstrass_point(data->P1w[i], data->P1[i], data->E, STACK_1);
        jacobi_quartic_point_to_weierstrass_point(data->P2w[i], data->P2[i], data->E, STACK_1);
        
        if(!jacobi_quartic_belongs(data->P1[i], data->E, STACK_1))
        {
            printf("P1 does not belongs to E\n");
            return false;
        }
        if(!jacobi_quartic_belongs(data->P2[i], data->E, STACK_1))
        {
            printf("P2 does not belongs to E\n");
            return false;
        }
        
        /*P1 + P2*/
        test_curve_jacobi_quartic_read_point(filep, Res, data);

        /*P1 - P2*/
        test_curve_jacobi_quartic_read_point(filep, Res, data);

        /*2 * P1*/
        test_curve_jacobi_quartic_read_point(filep, Res, data);

        /*n1 * P1*/
        test_curve_jacobi_quartic_read_data(filep, n1, data);
        test_curve_jacobi_quartic_read_point(filep, Res, data);
    }

    fclose(filep);
    ec_point_free(Res, data->k);
    number_free(&n1);
    return true;
}

bool
test_curve_jacobi_quartic_conversion(data_t * data)
{
    printf("Test on curve %s\n", data->E->id_curve);     
    uint64_t i;
    printf("General operations: \n");
    
    ec_point tmp;
    ec_point_alloc(tmp, data->k);
    ec_point_init(tmp, data->k);
    
    bool res = true;
    int res1 = 0;
    int res2 = 0;
    
    /************************************ ADD  *******************************/

    for(i = 0; i < data->nb_points; i++)
    {
        ec_point_add(data->P3[i], data->P1[i], data->P2[i], data->E, STACK_1);
        if(!jacobi_quartic_belongs(data->P3[i], data->E, STACK_1))
        {
            printf("ERROR: add, jacobi_quartic_belongs\n");
            res = false;
        }
    }

    data->Ew->f = PROJECTIVE;
    for(i = 0; i < data->nb_points; i++)
    {
        ec_point_add(data->P3w[i], data->P1w[i], data->P2w[i], data->Ew, STACK_1);
        if(!weierstrass_belongs(data->P3w[i], data->Ew, STACK_1))
        {
            printf("ERROR: add, weierstrass_belongs \n");
            res = false;
        }
    }
    
    for(i = 0; i < data->nb_points; i++)
    {
        jacobi_quartic_point_to_weierstrass_point(tmp, data->P3[i], data->E, STACK_1);
        if(!ec_point_are_equal(tmp, data->P3w[i], data->Ew, STACK_1))
        {
            printf("ERROR: jacobi_quartic_point_to_weierstrass_point, add, point %d, on jacobi quartic \n", (int)i);
            res1++;
        }
        weierstrass_point_to_jacobi_quartic_point(tmp, data->P3w[i], data->E, STACK_1);
        ec_point_norm(tmp, data->E, STACK_1);
        ec_point_norm(data->P3[i], data->E, STACK_1);
        if(!ec_point_are_equal(tmp, data->P3[i], data->E, STACK_1)!=0)
        {
            printf("ERROR: weierstrass_point_to_jacobi_quartic_point, add, point %d, on weierstrass \n", (int)i);
            res2++;
        }
    }
    
    if(res1==0)
    {
        printf("Add points on jacobi quartic, OK\n");
    }
    else
    {
        res = false;
    }
    if(res2==0)
    {
        printf("Add points on weierstrass, OK\n");
    }
    else
    {
        res = false;
    }

    /************************************ DBL  *******************************/
    
    res1=0;
    res2=0;

    for(i = 0; i < data->nb_points; i++)
    {
        ec_point_dbl(data->P3[i], data->P1[i], data->E, STACK_1);
        if(!jacobi_quartic_belongs(data->P3[i], data->E, STACK_1))
        {
            printf("ERROR: double, jacobi_quartic_belongs\n");
            res = false;
        }
    }

    data->Ew->f = PROJECTIVE;
    for(i = 0; i < data->nb_points; i++)
    {
        ec_point_dbl(data->P3w[i], data->P1w[i],data->Ew, STACK_1);
        if(!weierstrass_belongs(data->P3w[i], data->Ew, STACK_1))
        {
            printf("ERROR: double, weierstrass_belongs \n");
            res = false;
        }
    }
    
    for(i = 0; i < data->nb_points; i++)
    {
        jacobi_quartic_point_to_weierstrass_point(tmp, data->P3[i], data->E, STACK_1);
        if(!ec_point_are_equal(tmp, data->P3w[i], data->Ew, STACK_1))
        {
            printf("ERROR: jacobi_quartic_point_to_weierstrass_point, double, point %d, on jacobi quartic \n", (int)i);
            res1++;
        }
        weierstrass_point_to_jacobi_quartic_point(tmp, data->P3w[i], data->E, STACK_1);
        ec_point_norm(tmp, data->E, STACK_1);
        ec_point_norm(data->P3[i], data->E, STACK_1);
        if(!ec_point_are_equal(tmp, data->P3[i], data->E, STACK_1)!=0)
        {
            printf("ERROR: weierstrass_point_to_jacobi_quartic_point, double, point %d, on weierstrass \n", (int)i);
            res2++;
        }
    }
    
    if(res1==0)
    {
        printf("Double point on jacobi quartic, OK\n");
    }
    else
    {
        res = false;
    }
    if(res2==0)
    {
        printf("Double point on weierstrass, OK\n");
    }
    else
    {
        res = false;
    }

    /************************************ SUB  *******************************/
    
    res1=0;
    res2=0;

    for(i = 0; i < data->nb_points; i++)
    {
        ec_point_sub(data->P3[i], data->P1[i], data->P2[i], data->E, STACK_1);
        if(!jacobi_quartic_belongs(data->P3[i], data->E, STACK_1))
        {
            printf("ERROR: sub, jacobi_quartic_belongs\n");
            res = false;
        }
    }

    data->Ew->f = PROJECTIVE;
    for(i = 0; i < data->nb_points; i++)
    {
        ec_point_sub(data->P3w[i], data->P1w[i], data->P2w[i], data->Ew, STACK_1);
        if(!weierstrass_belongs(data->P3w[i], data->Ew, STACK_1))
        {
            printf("ERROR: sub, weierstrass_belongs \n");
            res = false;
        }
    }
    
    for(i = 0; i < data->nb_points; i++)
    {
        jacobi_quartic_point_to_weierstrass_point(tmp, data->P3[i], data->E, STACK_1);
        if(!ec_point_are_equal(tmp, data->P3w[i], data->Ew, STACK_1))
        {
            printf("ERROR: jacobi_quartic_point_to_weierstrass_point, sub, point %d, on jacobi quartic \n", (int)i);
            res1++;
        }
        weierstrass_point_to_jacobi_quartic_point(tmp, data->P3w[i], data->E, STACK_1);
        ec_point_norm(tmp, data->E, STACK_1);
        ec_point_norm(data->P3[i], data->E, STACK_1);
        if(!ec_point_are_equal(tmp, data->P3[i], data->E, STACK_1)!=0)
        {
            printf("ERROR: weierstrass_point_to_jacobi_quartic_point, sub, point %d, on weierstrass \n", (int)i);
            res2++;
        }
    }
    
    if(res1==0)
    {
        printf("Sub points on jacobi quartic, OK\n");
    }
    else
    {
        res = false;
    }
    
    if(res2==0)
    {
        printf("Sub points on weierstrass, OK\n");
    }
    else
    {
        res = false;
    }

    /************************************ MUL  *******************************/
    
    res1=0;
    res2=0;
    
    number m;
    number_init(&m, data->k->size);
    number_sub_ui(m, data->E->n, 100);

    for(i = 0; i < data->nb_points; i++)
    {
        ec_point_mul(data->P3[i], m, data->P2[i], data->E, STACK_1);
        if(!jacobi_quartic_belongs(data->P3[i], data->E, STACK_1))
        {
            printf("ERROR: ec_point_mul_naive, jacobi_quartic_belongs\n");
            res = false;
        }
    }

    data->Ew->f = PROJECTIVE;
    for(i = 0; i < data->nb_points; i++)
    {
        ec_point_mul(data->P3w[i], m, data->P2w[i], data->Ew, STACK_1);
        if(!weierstrass_belongs(data->P3w[i], data->Ew, STACK_1))
        {
            printf("ERROR: ec_point_mul_naive, weierstrass_belongs \n");
            res = false;
        }
    }
    
    for(i = 0; i < data->nb_points; i++)
    {
        jacobi_quartic_point_to_weierstrass_point(tmp, data->P3[i], data->E, STACK_1);
        if(!ec_point_are_equal(tmp, data->P3w[i], data->Ew, STACK_1))
        {
            printf("ERROR: jacobi_quartic_point_to_weierstrass_point, ec_point_mul_naive, point %d, on jacobi quartic \n", (int)i);
            res1++;
        }
        weierstrass_point_to_jacobi_quartic_point(tmp, data->P3w[i], data->E, STACK_1);
        ec_point_norm(tmp, data->E, STACK_1);
        ec_point_norm(data->P3[i], data->E, STACK_1);
        if(!ec_point_are_equal(tmp, data->P3[i], data->E, STACK_1)!=0)
        {
            printf("ERROR: weierstrass_point_to_jacobi_quartic_point, ec_point_mul_naive, point %d, on weierstrass \n", (int)i);
            res2++;
        }
    }
    if(res1==0)
    {
        printf("Multiply points on jacobi quartic, OK\n");
    }
    else
    {
        res = false;
    }
    if(res2==0)
    {
        printf("Multiply points on weierstrass, OK\n");
    }
    else
    {
        res = false;
    }
    
    ec_point_free(tmp, data->k);
    number_free(&m);
    return res;
}

bool
test_curve_jacobi_quartic(const char *id_curve, const uint16_t nbits, 
        const uint32_t num_point, data_t *data)
{
    bool res;
    test_curve_jacobi_quartic_alloc(nbits, num_point, data);
    test_curve_jacobi_quartic_read(id_curve, data);
    test_curve_jacobi_quartic_init(data);
    test_curve_jacobi_quartic_read_points(data);
    res=test_curve_jacobi_quartic_conversion(data);
    test_curve_jacobi_quartic_free(data);
    return res;
}

int main()
{
    data_t data;
    init_mphell(128, RANDOM_AES128, DEVURANDOM);
    int nb_points=100;
    if (!test_curve_jacobi_quartic("jacobi_quartic_p256", 256, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    else if (!test_curve_jacobi_quartic("jq256_3", 256, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    free_mphell();
    return 0;
}
