/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France)

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell_test_weierstrass.c
  * \brief Self check tests for Weierstrass Elliptic Curves.
  */


#include "mphell_test_weierstrass.h"

void 
test_curve_weierstrass_alloc(const uint16_t nbits, const uint32_t num_point, data_t *data)
{
    uint8_t n = bits_to_nblock(nbits);
    field_alloc(data->k, FP, n, NULL);
    ec_alloc(data->E, data->k);
    data->nb_points = num_point;
    data->nbits = nbits;
    
    data->P1 = (ec_point*)malloc(sizeof(ec_point) * data->nb_points);
    data->P2 = (ec_point*)malloc(sizeof(ec_point) * data->nb_points);
    data->P3 = (ec_point*)malloc(sizeof(ec_point) * data->nb_points);

    uint32_t i;
    for(i = 0; i < data->nb_points; i++)
    {
        ec_point_alloc(data->P1[i], data->k);
        ec_point_alloc(data->P2[i], data->k);
        ec_point_alloc(data->P3[i], data->k);
    }
}

void 
test_curve_weierstrass_init(data_t *data)
{
    uint32_t i;
    for(i = 0; i < data->nb_points; i++)
    {
        ec_point_init(data->P1[i], data->k);
        ec_point_init(data->P2[i], data->k);
        ec_point_init(data->P3[i], data->k);
    }
}

void 
test_curve_weierstrass_free(data_t *data)
{
    ec_free(data->E);
    uint32_t i;
    for(i = 0; i < data->nb_points; i++)
    {
        ec_point_free(data->P1[i], data->k);
        ec_point_free(data->P2[i], data->k);
        ec_point_free(data->P3[i], data->k);
    }
    field_free(data->k);
    free(data->P1);
    free(data->P2);
    free(data->P3);
}

void
test_curve_weierstrass_read(const char *id_curve, const ec_formula f, data_t *data)
{
    char* curve_file = (char*)malloc(sizeof(char) * 200);
    sprintf(curve_file, DATADIR "/elliptic_curves/%s.txt", id_curve);
    FILE *fp = NULL;
    char line[160];
    fp = fopen(curve_file, "r");
    free(curve_file);

    if(fp == NULL)
    {
        fprintf(stderr, "Error : no curve data file found \n");
        exit(EXIT_FAILURE);
    }
    char *str = (char*)malloc(sizeof(char) * 200);
    char *str1 = (char*)malloc(sizeof(char) * 200);
    number tmp;
    number_init(&tmp, bits_to_nblock(data->nbits));

    while(fgets(line, sizeof(line), fp) != NULL)
    { 
        if(strncmp(line, "p = ", 4)== 0)
        {
            strcpy(str, line + 4);
            number_set_str(tmp, str, 16);
            field_create(data->k, "", STACK_1, 1, tmp);
            /* Initialise the curve, now that k is set up */
            ec_init(data->E, data->k);
        }
        if(strncmp(line, "a = ", 4)== 0)
        {
            strcpy(str, line + 4);
            number_set_str(tmp, str, 16);
            field_elt_set_number(data->E->a, false, data->k, STACK_1, 1, tmp);
        }
        if(strncmp(line, "b = ", 4)== 0)
        {
            strcpy(str, line + 4);
            number_set_str(tmp, str, 16);
            field_elt_set_number(data->E->b, false, data->k, STACK_1, 1, tmp);
        }
        if(strncmp(line, "xG = ", 5)== 0)
        {
            strcpy(str, line + 5);
        }
        if(strncmp(line, "yG = ", 5)== 0)
        {
            strcpy(str1, line + 5);
            ec_point_set_aff_str(data->E->G, str, str1, false, 16, WEIERSTRASS, data->k, STACK_1);
        }
        if(strncmp(line, "h = ", 4) == 0)
        {
            strcpy(str, line + 4);
            number_set_str(data->E->h, str, 16);
        }
        if(strncmp(line, "n = ", 4) == 0)
        {
            strcpy(str, line + 4);
            number_set_str(data->E->n, str, 16);
            break;
        }
    }
    data->E->k = data->k;
    data->E->type = WEIERSTRASS;
    data->E->f = f;
    strcpy(data->E->id_curve, id_curve);
    ec_compute_disc(data->E, STACK_1);
    ec_test_spec(data->E, STACK_1);
    fclose(fp);
    free(str);
    free(str1);
    number_free(&tmp);
}

void
test_curve_weierstrass_read_point(FILE *filep, ec_point P, data_t *data)
{
    char line[180];

    if(filep == NULL)
    {
        fprintf(stderr, "Error : no test data file found \n");
        exit(EXIT_FAILURE);
    }

    char *str = malloc(sizeof(char) * 170);
    field_elt tmp, tmp1;
    field_elt_alloc(&tmp, data->k);
    field_elt_init(tmp, data->k);
    field_elt_alloc(&tmp1, data->k);
    field_elt_init(tmp1, data->k);
    while(fgets(line, sizeof(line), filep) != NULL)
    { 
        if(strncmp(line, "x = ", 4) == 0)
        {
            strcpy(str, line + 4);
            if(str[strlen(str)-2] == '\\')
            {
                if(fgets(line, sizeof(line), filep) == NULL)
                {
                    exit(EXIT_FAILURE);
                }
                strcpy(str + strlen(str)-2, line);
            }
            field_elt_set_str(tmp, str, 16, false, data->k, STACK_1);
        }
        if(strncmp(line, "y = ", 4) == 0)
        {
            strcpy(str, line + 4);
            if(str[strlen(str)-2] == '\\')
            {
                if(fgets(line, sizeof(line), filep) == NULL)
                {
                    exit(EXIT_FAILURE);
                }
                strcpy(str + strlen(str)-2, line);
            }
            field_elt_set_str(tmp1, str, 16, false, data->k, STACK_1);
            ec_point_set_aff(P, tmp, tmp1, WEIERSTRASS, data->k, STACK_1);
            break;
        }
    }
    free(str);
    field_elt_free(&tmp, data->k);
    field_elt_free(&tmp1, data->k);
}

void
test_curve_weierstrass_read_data(FILE *filep, number n, data_t *data)
{
    char line[170];
    if(filep == NULL)
    {
        fprintf(stderr, "Error : no curve data file found \n");
        exit(EXIT_FAILURE);
    }
    char *str = malloc(sizeof(char) * 160);
    while(fgets(line, sizeof(line), filep) != NULL)
    { 
        if(strncmp(line, "k = ", 4) == 0)
        {
            strcpy(str, line + 4);
            if(str[strlen(str)-2] == '\\')
            {
                if(fgets(line, sizeof(line), filep) == NULL)
                {
                    exit(EXIT_FAILURE);
                }
                strcpy(str + strlen(str)-2, line);
            }
            number_set_str(n, str, 16);
            break;
        }
    }
    free(str);
}
