/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France)

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell_test_weierstrass_ZDAU.c
  * \brief Self check tests for the functions: weierstrass_point_add_ZDAU in Co-Z coordinates
  */

#include "mphell_test_weierstrass.h"

bool 
test_curve_weierstrass_ZDAU(data_t *data)
{
    printf("Validity tests on curve %s with %d points: ", data->E->id_curve, data->nb_points);
    uint8_t n = data->k->size;
    uint32_t i = 0;
    uint32_t j = 0;
    ec_point Res;
    ec_point Old;
    number n1;
    uint8_t check[data->nb_points][5];
    FILE *filep;
    char test_file[200];

    for(i = 0; i < data->nb_points; i++)
    {
        for (j = 0; j < 5; j++)
        {
            check[i][j] = (uint8_t)0;
        }
    }
    sprintf(test_file, DATADIR "/vector_test_elliptic_curves/%s_vector_test.txt", data->E->id_curve);
    filep = fopen(test_file,"r");
    if(filep == NULL)
    {
        fprintf(stderr, "Error : no curve data file found \n");
        exit(EXIT_FAILURE);
    }
    number_init(&n1, n);
    ec_point_alloc(Res, data->k);
    ec_point_alloc(Old, data->k);
    ec_point_init(Res, data->k);
    ec_point_init(Old, data->k);
    
    for(i = 0; i < data->nb_points; i++)
    {
        test_curve_weierstrass_read_point(filep, data->P1[i], data);
        test_curve_weierstrass_read_point(filep, data->P2[i], data);
        
        if(!weierstrass_belongs(data->P1[i], data->E, STACK_1))
        {
            printf("P1 does not belongs to E\n");
            return false;
        }
        if(!weierstrass_belongs(data->P2[i], data->E, STACK_1))
        {
            printf("P2 does not belongs to E\n");
            return false;
        }
        
        /* P1 + P2 */
        test_curve_weierstrass_read_point(filep, Res, data);
        ec_point_copy(Old, data->P2[i], data->k);
        /* P3= 2*P1+P2 and P3->z==P2->z */
        weierstrass_point_add_ZDAU(data->P3[i], data->P1[i], data->P2[i], data->E, STACK_1);
        if(!weierstrass_belongs(data->P3[i], data->E, STACK_1))
        {
            printf("P3=2*P1+P2 in JACOBIAN coordinate does not belongs to E\n");
            return false;
        }
        ec_point_dbl(Res, data->P1[i], data->E, STACK_1);               /* Res = 2 * P1         */
        ec_point_add(Res, Res, data->P2[i], data->E, STACK_1);          /* Res = (2 * P1) + P2  */
        /* We test if P3 == 2*P1+P2 */        
        check[i][0] = (uint8_t)ec_point_are_equal(data->P3[i], Res, data->E, STACK_1);

        /* We test if P3 and P2 have the same Z coordinate */
        check[i][1] = (uint8_t)(field_elt_cmp(data->P3[i]->z, data->P2[i]->z, data->k)==0);
        
        /* We test that P2 is still P2 */
        check[i][2] = (uint8_t)weierstrass_point_are_equal(Old,data->P2[i],data->E, STACK_1);
        
        /*  We test with the same point in input and output*/        
        weierstrass_point_add_ZDAU(data->P1[i], data->P1[i], Old, data->E, STACK_1);
        check[i][3] = (uint8_t)ec_point_are_equal(data->P1[i], Res, data->E, STACK_1);
        
        /* We test again if we have the same Z coordinate */      
        check[i][4] = (uint8_t)(field_elt_cmp(data->P1[i]->z, Old->z, data->k)==0);

        /*P1 - P2*/
        test_curve_weierstrass_read_point(filep, Res, data);

        /*2 * P1*/
        test_curve_weierstrass_read_point(filep, Res, data);

        /*n1 * P1*/
        test_curve_weierstrass_read_data(filep, n1, data);
        test_curve_weierstrass_read_point(filep, Res, data);
    }

    fclose(filep);
    ec_point_free(Res, data->k);
    ec_point_free(Old, data->k);
    number_free(&n1);
    for(i = 0; i < data->nb_points; i++)
    {
        for (j = 0; j < 5; j++)
        {
            if(check[i][j] == false)
            {
                printf("ERROR\n");
                printf("i: %d, j: %d \n", i, j);
                return false;
            }
        }
    }
    printf("SUCCESS\n");
    return true;
}

bool
test_curve_weierstrass(const ec_known_curve id_curve, const uint16_t nbits, 
        const uint32_t num_point, data_t *data)
{
    bool res;
    test_curve_weierstrass_alloc(nbits, num_point, data);
    ec_use_curve(data->E, data->k, id_curve, JACOBIAN, STACK_1);
    test_curve_weierstrass_init(data);
    res=test_curve_weierstrass_ZDAU(data);
    test_curve_weierstrass_free(data);
    return res;
}

int main()
{
    data_t data;
    init_mphell(256, RANDOM_AES256, DEVURANDOM);
    int nb_points=100;
    if (!test_curve_weierstrass(NIST_192, 192, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    if (!test_curve_weierstrass(NIST_224, 224, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    if (!test_curve_weierstrass(NIST_256, 256, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    if (!test_curve_weierstrass(NIST_384, 384, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    if (!test_curve_weierstrass(NIST_521, 521, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    if (!test_curve_weierstrass(BRAINPOOL_160, 160, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    if (!test_curve_weierstrass(BRAINPOOL_192, 192, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    if (!test_curve_weierstrass(BRAINPOOL_224, 224, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    if (!test_curve_weierstrass(BRAINPOOL_256, 256, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    if (!test_curve_weierstrass(BRAINPOOL_320, 320, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    if (!test_curve_weierstrass(BRAINPOOL_384, 384, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    if (!test_curve_weierstrass(BRAINPOOL_512, 512, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    if (!test_curve_weierstrass(FR_256, 256, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    free_mphell();
    return 0;
}

