/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France)

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell_test_weierstrass_Zmontgomery_Kim.c
  * \brief Self check tests for the functions:  in CoZ Jacobian coordinates
  * \ From the work of Speeding up Elliptic Curve Scalar Multiplication without Precomputation by Kim, Choe, Kim, Kim and Hong 2017
  */

#include "mphell_test_weierstrass.h"

bool 
test_curve_weierstrass_Zmontgomery_Kim_ladder(data_t *data)
{
    printf("Validity tests on curve %s with %d points: ", data->E->id_curve, data->nb_points);
    uint8_t n = data->k->size;
    uint32_t i = 0;
    ec_point Res;
    number n1;
    uint8_t check[data->nb_points];
    FILE *filep;
    char test_file[200];

    for(i = 0; i < data->nb_points; i++)
    {
        check[i] = (uint8_t)0;
    }
    sprintf(test_file, DATADIR "/vector_test_elliptic_curves/%s_vector_test.txt", data->E->id_curve);
    filep = fopen(test_file,"r");
    if(filep == NULL)
    {
        fprintf(stderr, "Error : no curve data file found \n");
        exit(EXIT_FAILURE);
    }
    number_init(&n1, n);
    ec_point_alloc(Res, data->k);
    ec_point_init(Res, data->k);
    
    for(i = 0; i < data->nb_points; i++)
    {
        test_curve_weierstrass_read_point(filep, data->P1[i], data);
        test_curve_weierstrass_read_point(filep, data->P2[i], data);
        
        if(!weierstrass_belongs(data->P1[i], data->E, STACK_1))
        {
            printf("P1 does not belongs to E\n");
            return false;
        }
        if(!weierstrass_belongs(data->P2[i], data->E, STACK_1))
        {
            printf("P2 does not belongs to E\n");
            return false;
        }
        

        /*P1 + P2*/
        test_curve_weierstrass_read_point(filep, Res, data);

        /*P1 - P2*/
        test_curve_weierstrass_read_point(filep, Res, data);

        /*2 * P1*/
        test_curve_weierstrass_read_point(filep, Res, data);

        /*n1 * P1*/
        test_curve_weierstrass_read_data(filep, n1, data);
        test_curve_weierstrass_read_point(filep, Res, data);
        weierstrass_Zmontgomery_Kim(data->P3[i], n1, data->P1[i], data->E, STACK_1);
        if(!weierstrass_belongs(data->P3[i], data->E, STACK_1))
        {
            printf("\nP3=n1.P1 in Jacobian coordinate using Montgomery ladder with Co-Z addition does not belongs to E bis\n");
            ec_point_print(data->P3[i], 16, true, data->k, STACK_1);
            printf("\nP3 above, P1 below:\n");
            ec_point_print(data->P1[i], 16, true, data->k, STACK_1);
            printf("\nAnd now Res:\n");
            ec_point_print(Res, 16, true, data->k, STACK_1);
            printf("\nAnd now n1:\n");
            number_print(n1,2);
            return false;
        }
        check[i] = (uint8_t)ec_point_are_equal(data->P3[i], Res, data->E, STACK_1);
    }

    fclose(filep);
    ec_point_free(Res, data->k);
    number_free(&n1);
    for(i = 0; i < data->nb_points; i++)
    {
        if(check[i] == false)
        {
            printf("ERROR\n");
            printf("i: %d \n", i);
            ec_point_print(data->P3[i], 16, true, data->k, STACK_1);
            return false;
        }
    }
    printf("SUCCESS\n");
    return true;
}

bool
test_curve_weierstrass(const ec_known_curve id_curve, const uint16_t nbits, 
        const uint32_t num_point, data_t *data)
{
    bool res;
    test_curve_weierstrass_alloc(nbits, num_point, data);
    ec_use_curve(data->E, data->k, id_curve, JACOBIAN, STACK_1);
    test_curve_weierstrass_init(data);
    res=test_curve_weierstrass_Zmontgomery_Kim_ladder(data);
    test_curve_weierstrass_free(data);
    return res;
}

int main()
{
    data_t data;
    init_mphell(256, RANDOM_AES256, DEVURANDOM);
    int nb_points=100;
    if (!test_curve_weierstrass(NIST_192, 192, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    if (!test_curve_weierstrass(NIST_224, 224, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    if (!test_curve_weierstrass(NIST_256, 256, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    if (!test_curve_weierstrass(NIST_384, 384, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    if (!test_curve_weierstrass(NIST_521, 521, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    if (!test_curve_weierstrass(BRAINPOOL_160, 160, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    if (!test_curve_weierstrass(BRAINPOOL_192, 192, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    if (!test_curve_weierstrass(BRAINPOOL_224, 224, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    if (!test_curve_weierstrass(BRAINPOOL_256, 256, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    if (!test_curve_weierstrass(BRAINPOOL_320, 320, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    if (!test_curve_weierstrass(BRAINPOOL_384, 384, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    if (!test_curve_weierstrass(BRAINPOOL_512, 512, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    if (!test_curve_weierstrass(FR_256, 256, nb_points, &data))
    {
        free_mphell();
        return -1;
    }
    free_mphell();
    return 0;
}

