/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France)

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell_tuto_conversion_edwards_to_weierstrass.c
  * \brief Tutorial for Edwards elliptic curve conversion to weierstrass curve
  */

#include <stdio.h>
#include "mphell/mphell.h"

int main()
{
    /* Initialise MPHELL with 256 bits of security strength for the entropy, RANDOM_AES256 as DRBG and DEVURANDOM as entropy source */
    
    init_mphell(256, RANDOM_AES256, DEVURANDOM);
    
    /* Allocate a field of size 4*block_SIZE = 4*64 = 256 on 64 bits architecture */
    
    field k;
    field_alloc(k, FP, bits_to_nblock(256), NULL);
       
    /* Allocate temporary memory */
    
    number p;
    field_elt a, b;
    ec_point G, res, res_ed;
    number n, h;
    
    number_init(&p, bits_to_nblock(256));

    /* Set the number p from a string in base 16 */
    
    number_set_str(p, "7fffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffed", 16);
    
    /* Create the field of characteristic p */
    
    field_create(k, "", STACK_1, 1, p);

    ec_point_alloc(res, k);
    ec_point_init(res, k);
    ec_point_alloc(res_ed, k);
    ec_point_init(res_ed, k);
    field_elt_alloc(&a, k);
    field_elt_init(a, k);
    field_elt_alloc(&b, k);
    field_elt_init(b, k);
    ec_point_alloc(G, k);
    ec_point_init(G, k);
    number_init(&n, bits_to_nblock(256));
    number_init(&h, bits_to_nblock(256));

    
    
    field_elt_set_str(a, "178bccfe3009fb7a6adc25a5981cc87fcb86f5d89097a446249d137d8676d379", 16, false, k, STACK_1);
    field_elt_set_str(b, "55f4c3f0ed6ed8e9501feca7866d4e8d70d9960f148207bf7d618f19d7fe4169", 16, false, k, STACK_1);
    ec_point_set_aff_str(G, "460ed8dffbd2fc64813be575cc7034ba059bfb51e03357653b07c4f6b00f8859", "361e0dc66944371de8e1860980b5c14414d5bf32c315a5a5374e38f87126c0c1", false, 16, WEIERSTRASS, k, STACK_1);
    number_set_str(h, "1", 16);
    number_set_str(n, "1", 16);
    
    /* Allocate curves */
    
    ec_curve E, E_ed;
    ec_alloc(E, k);
    ec_init(E, k);
    ec_alloc(E_ed, k);
    ec_init(E_ed, k);
    
    /* Create curve */
    
    ec_create (E_ed, "RandP25519_rank2_Edwards", k, a, b, G, h, n, EDWARDS, EXTENDED_EDWARDS, STACK_1);

    printf("The edwards departing curve \nE_ed: \n"); ec_curve_print(E_ed, 16, STACK_1); printf("\n");

    /* Creation of a random point on the elliptic curve */
    
    edwards_point_random(res_ed, E_ed, STACK_1);
    printf("random point, res_ed = "); ec_point_print(res_ed, 16, true, k, STACK_1); printf("\n");
    printf("res_ed belongs to E_ed : %d\n\n", edwards_belongs(res_ed, E_ed, STACK_1));

    /* Conversion to the isomorphic weierstrass curve */
    
    edwards_to_weierstrass(E, E_ed, STACK_1);     /* Here there is only one isomorphic weierstrass curve */

    printf("The Weierstrass curve isomorphic to E_ed calculated \n"); 
    printf("E: \n"); ec_curve_print(E, 16, STACK_1); printf("\n");

    /* Conversion of a point on a Weierstrass Elliptic curve */
    
    edwards_point_to_weierstrass_point (res, res_ed, E_ed, STACK_1);
    printf("Weierstrass random point, res = "); ec_point_print(res, 16, true, k, STACK_1); printf("\n");
    printf("res belongs to E : %d\n\n", weierstrass_belongs(res, E, STACK_1));
    
    /* Free allocated memory */
    
    field_elt_free(&a, k);
    field_elt_free(&b, k);
    ec_point_free(G, k);
    number_free(&n);
    number_free(&h);
    ec_point_free(res, k);
    ec_point_free(res_ed, k);
    number_free(&p);
    field_free(k);
    ec_free(E_ed); 
    ec_free(E);
    
    free_mphell();

    return 0;
}
