/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France)

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell_tuto_conversion_jacobi_quartic_to_weierstrass.c
  * \brief Tutorial for Jacobi Quartic elliptic curve conversion to weierstrass curve
  */

#include <stdio.h>
#include "mphell/mphell.h"

int main()
{
    /* Initialise MPHELL with 256 bits of security strength for the entropy, RANDOM_AES256 as DRBG and DEVURANDOM as entropy source */
    
    init_mphell(256, RANDOM_AES256, DEVURANDOM);
    
    /* Allocate a field of size 6*block_SIZE = 6*64 = 384 on 64 bits architecture */
    
    field k;
    field_alloc(k, FP, bits_to_nblock(384), NULL);
    
    /* Allocate temorary memory */
    
    number p;
    field_elt a, b;
    ec_point G, res, res_jq;
    number n, h;
    
    number_init(&p, bits_to_nblock(384));

    /* Set the number p from a string in base 16 */
    
    number_set_str(p, "fffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffeffffffff0000000000000000ffffffff", 16);

    /* Create the field of characteristic p */
    
    field_create(k, "", STACK_1, 1, p);

    ec_point_alloc(res, k);
    ec_point_init(res, k);
    ec_point_alloc(res_jq, k);
    ec_point_init(res_jq, k);
    field_elt_alloc(&a, k);
    field_elt_init(a, k);
    field_elt_alloc(&b, k);
    field_elt_init(b, k);
    ec_point_alloc(G, k);
    ec_point_init(G, k);
    number_init(&n, bits_to_nblock(384));
    number_init(&h, bits_to_nblock(384));

    
    
    field_elt_set_str(a, "a3aeb56058fa59051926a1559fe934bf64545ee7dee231a8d65e7211702ac9bf9801cfdae54ad7a050e20e612d4c080", 16, false, k, STACK_1);
    field_elt_set_str(b, "1", 16, false, k, STACK_1);
    ec_point_set_str(G, "1541e4e6fec9431b561ad9d815306961cbda834477c2884b93bb5fb3a4c865d1638d6b1e1eb5d91afea443da57ec3c95", "e55245714090febf2f55c3dee76267c5aa27edb22a95ffa0328adecc2690ffef38c5cb63423401492f8e75bd4753f0b6","1","a7c8a5c9364fb8f2142801d10e8321c885c9fe7b88a7c76eabc474f635b090ac8001e9efcfc308ff06960a29869ebf8", false, 16, k, STACK_1);
    number_set_str(h, "15ad3c423d6df1572016199efaeee94ca5242a87a4", 16);
    number_set_str(n, "bcf569c46c00fa8b5f4c7d41b55e14295468dac4097918b25e24499", 16);
    
    /* Allocate curves */
    
    ec_curve E, E_jq;
    ec_alloc (E, k);
    ec_init(E, k);
    ec_alloc (E_jq, k);
    ec_init(E_jq, k);
    
    /* Create curve */
    
    ec_create (E_jq, "RandP384JQ", k, a, b, G, h, n, JACOBI_QUARTIC, PROJECTIVE, STACK_1);
    
    printf("The Jacobi Quartic departing curve \nE_jq: \n"); ec_curve_print(E_jq, 16, STACK_1); printf("\n");

    /* Creation of a random point on the Jacobi Quartic elliptic curve */
    
    jacobi_quartic_point_random(res_jq, E_jq, STACK_1);
    printf("random point, res_jq = "); ec_point_print(res_jq, 16, true, k, STACK_1); printf("\n");
    printf("res_jq belongs to E_jq : %d\n\n", jacobi_quartic_belongs(res_jq, E_jq, STACK_1));

    /* Conversion to the Weierstrass Elliptic curve */
    
    jacobi_quartic_to_weierstrass(E, E_jq, STACK_1);

    printf("The Weierstrass curve isomorphic to E_jq \n"); 
    printf("E: \n"); ec_curve_print(E, 16, STACK_1); printf("\n");

    /* Conversion of a point on a Weierstrass Elliptic curve*/
    
    jacobi_quartic_point_to_weierstrass_point (res, res_jq, E_jq, STACK_1);

    printf("Weierstrass random point, res = "); ec_point_print(res, 16, true, k, STACK_1); printf("\n");
    printf("res belongs to E : %d\n\n", weierstrass_belongs(res, E, STACK_1));
    
    /* Free allocated memory */
    
    field_elt_free(&a, k);
    field_elt_free(&b, k);
    ec_point_free(G, k);
    number_free(&n);
    number_free(&h);
    ec_point_free(res, k);
    ec_point_free(res_jq, k);
    number_free(&p);
    field_free(k);
    ec_free(E_jq); 
    ec_free(E);
    
    free_mphell();

    return 0;
}
