/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France)

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell_tuto_conversion_weierstrass_to_edwards.c
  * \brief Tutorial for Weierstrass elliptic curve conversion to Edwards curve
  */

#include <stdio.h>
#include "mphell/mphell.h"

int main()
{
    /* Initialise MPHELL with 256 bits of security strength for the entropy, RANDOM_AES256 as DRBG and DEVURANDOM as entropy source */
    
    init_mphell(256, RANDOM_AES256, DEVURANDOM);
    
    /* Allocate a field of size 3*block_SIZE = 4*64 = 256 on 64 bits architecture */
    
    field k;
    field_alloc(k, FP, bits_to_nblock(256), NULL);
       
    /* Allocate temporary memory */
    
    number p;
    field_elt a, b;
    ec_point G, res, res_ed;
    number n, h;
    field_elt alpha,beta;

    number_init(&p, bits_to_nblock(256));

    /* Set the number p from a string in base 16 */
    
    number_set_str(p, "7fffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffed", 16);
    
    /* Create the field of characteristic p */
    
    field_create(k, "", STACK_1, 1, p);
    
    ec_point_alloc(res, k);
    ec_point_init(res, k);
    ec_point_alloc(res_ed, k);
    ec_point_init(res_ed, k);

    field_elt_alloc(&a, k);
    field_elt_init(a, k);
    field_elt_alloc(&b, k);
    field_elt_init(b, k);
    ec_point_alloc(G, k);
    ec_point_init(G, k);
    number_init(&n, bits_to_nblock(256));
    number_init(&h, bits_to_nblock(256));
    field_elt_alloc(&alpha, k);
    field_elt_init(alpha, k);
    field_elt_alloc(&beta, k);
    field_elt_init(beta, k);

    
    
    field_elt_set_str(a, "48ef6d7eaa12562382bde7e49077db6ef4567d64201ef50ed140ae822742f2fa", 16, false, k, STACK_1);
    field_elt_set_str(b, "243c1342c1b1b6bf31fd865230a01c7cfee6616d8a2f62dc3065af2c897faf06", 16, false, k, STACK_1);
    ec_point_set_aff_str(G, "7020bedabe6fd4b4598e27db377d002d05b4090d711b769d6f79c13e2816ff18", "4a6a25cec9a22817414df1e9eaa2b5771129328c30fb28b2d85c5c44a37397fe", false, 16, WEIERSTRASS, k, STACK_1);
    number_set_str(h, "65eb9c5e14819d77ca2d7b0495d0a0", 16);
    number_set_str(n, "1418185a2d6ddee74accf50dbe5617c6789", 16);
    
    /* Allocate curves */
    
    ec_curve E, E_ed;
    ec_alloc (E, k);
    ec_init(E, k);
    ec_alloc (E_ed, k);
    ec_init(E_ed, k);
    
    /* Create curve */
    
    ec_create (E, "RandP25519_rank2", k, a, b, G, h, n, WEIERSTRASS, PROJECTIVE, STACK_1);
    
    printf("The weierstrass departing curve \nE: \n"); ec_curve_print(E, 16, STACK_1); printf("\n");

    /* Creation of a random point on the elliptic curve */
    
    weierstrass_point_random(res, E, STACK_1);
    printf("random point, res = "); ec_point_print(res, 16, true, k, STACK_1); printf("\n");
    printf("res belongs to E : %d\n\n", weierstrass_belongs(res, E, STACK_1));

    /* Conversion to one of the possible Edwards curve */
    
    weierstrass_to_edwards(E_ed, E, 0, alpha, beta, STACK_1);
                                    /* Here the 0 signifies that neither alpha nor beta are specified and that the function could chose the isomorphism (the first found up to 6 possible) */

    printf("The Edwards curve isomorphic to E calculated with neither alpha nor beta specified \n"); 
    printf("E_ed: \n"); ec_curve_print(E_ed, 16, STACK_1); printf("\n");

    weierstrass_to_edwards(E_ed, E, 2, alpha, beta, STACK_1);
                                   /* Here the 2 signifies that alpha and beta are specified and thus we should get the same curve as previously, 1 could be put just to specify alpha then there will be only two choices for beta */
                                   /* We could also only specify alpha by putting 1 instead of 2 */
    
    printf("The Edwards curve calculated with alpha and beta specified \n"); 
    printf("E_ed: \n"); ec_curve_print(E_ed, 16, STACK_1); printf("\n");

    /* Conversion of a point on an Edwards Elliptic curve*/
    
    weierstrass_point_to_edwards_point (res_ed, res, E, 2, alpha, beta, STACK_1);
                                                        /* Here the 2 signifies that alpha and beta are specified */
    printf("Edwards random point, res_ed = "); ec_point_print(res_ed, 16, true, k, STACK_1); printf("\n");
    printf("res_ed belongs to E_ed : %d\n\n", edwards_belongs(res_ed, E_ed, STACK_1));

    /* Free allocated memory */
    
    field_elt_free(&a, k);
    field_elt_free(&b, k);
    ec_point_free(G, k);
    number_free(&n);
    number_free(&h);
    field_elt_free(&alpha, k);
    field_elt_free(&beta, k);  
    ec_point_free(res, k);
    ec_point_free(res_ed, k);
    number_free(&p);
    field_free(k);
    ec_free(E_ed);
    ec_free(E);
    
    free_mphell();

    return 0;
}
