/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France)

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell_tuto_conversion_weierstrass_to_jacobi_quartic.c
  * \brief Tutorial for Weierstrass elliptic curve conversion to Jacobi Quartic curve
  */

#include <stdio.h>
#include "mphell/mphell.h"

int main()
{
    /* Initialise MPHELL with 256 bits of security strength for the entropy, RANDOM_AES256 as DRBG and DEVURANDOM as entropy source */
    
    init_mphell(256, RANDOM_AES256, DEVURANDOM);
    
    /* Allocate a field of size 6*block_SIZE = 6*64 = 384 on 64 bits architecture */
    
    field k;
    field_alloc(k, FP, bits_to_nblock(384), NULL);
       
    /* Allocate temporary memory */
    
    number p;
    field_elt a, b;
    ec_point G, res, res_jq;
    number n, h;
    field_elt teta;

    number_init(&p, bits_to_nblock(384));

    /* Set the number p from a string in base 16 */
    
    number_set_str(p, "fffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffeffffffff0000000000000000ffffffff", 16);
    
    /* Create the field of characteristic p */
    
    field_create(k, "", STACK_1, 1, p);
    
    ec_point_alloc(res, k);
    ec_point_init(res, k);
    ec_point_alloc(res_jq, k);
    ec_point_init(res_jq, k);

    field_elt_alloc(&a, k);
    field_elt_init(a, k);
    field_elt_alloc(&b, k);
    field_elt_init(b, k);
    ec_point_alloc(G, k);
    ec_point_init(G, k);
    number_init(&n, bits_to_nblock(384));
    number_init(&h, bits_to_nblock(384));
    field_elt_alloc(&teta, k); 
    field_elt_init(teta, k); 
    
    field_elt_set_str(a, "e6d121fc8cfc85448e51524d40b0a3f6146006a7e2001270385be91a918ba0c2066c8ae2615a9bf927b73c9756bce6b2", 16, false, k, STACK_1);
    field_elt_set_str(b, "a36e0555338282573d94a77351044dfad63d30aac6505bcd470ad021e3a4b057d81fc21d6c92ef0190798b386c38f9ba", 16, false, k, STACK_1);
    ec_point_set_aff_str(G, "a3f0d2b9cbda17b9201391877f744cbc69f7efe29f56e2552439582e70238edbdb70193133bae1d9a8b659d2d7692931", "c467049b488e283fdf78669237ca9387b347f63d17ce4fac5e6bbe737b4d8e306dfcb80eee28477ab6a0e06f262e47d2", false, 16, WEIERSTRASS, k, STACK_1);
    number_set_str(h, "15ad3c423d6df1572016199efaeee94ca5242a87a4", 16);
    number_set_str(n, "bcf569c46c00fa8b5f4c7d41b55e14295468dac4097918b25e24499", 16);
    
    /* Allocate curves */
    
    ec_curve E, E_jq;
    ec_alloc (E, k);
    ec_init(E, k);
    ec_alloc (E_jq, k);
    ec_init(E_jq, k);
    
    /* Create curve */
    
    ec_create(E, "RandP384", k, a, b, G, h, n, WEIERSTRASS, PROJECTIVE, STACK_1);
    
    printf("The weierstrass departing curve \nE: \n"); ec_curve_print(E, 16, STACK_1); printf("\n");

    /* Creation of a random point on the weierstras elliptic curve */
    
    weierstrass_point_random(res, E, STACK_1);
    printf("random point, res = "); ec_point_print(res, 16, true, k, STACK_1); printf("\n");
    printf("res belongs to E : %d\n\n", weierstrass_belongs(res, E, STACK_1));

    /* Conversion to one (specific) of the possible Jacobi_Quartic curve */
    
    weierstrass_to_jacobi_quartic(E_jq, E, false, teta, STACK_1);
                                           /* Here false signifies that teta is not specified and that the function have to find the isomorphism for which b_j=1 */

    printf("The Jacobi Quartic curve isomorphic to E calculated with teta not specified \n"); 
    printf("E_jq: \n"); ec_curve_print(E_jq, 16, STACK_1); printf("\n");

    weierstrass_to_jacobi_quartic(E_jq, E, true, teta, STACK_1);
                                           /* Here true signifies that teta is specified  */
    
    printf("The Jacobi Quartic curve calculated with teta specified \n"); 
    printf("E_jq: \n"); ec_curve_print(E_jq, 16, STACK_1); printf("\n");

    /* Conversion of a point on an Jacobi Quartic Elliptic curve*/
    
    weierstrass_point_to_jacobi_quartic_point(res_jq, res, E_jq, STACK_1);
    printf("Jacobi Quartic random point, res_jq = "); ec_point_print(res_jq, 16, true, k, STACK_1); printf("\n");
    printf("res_jq belongs to E_jq : %d\n\n", jacobi_quartic_belongs(res_jq, E_jq, STACK_1));

    ec_free(E_jq); 

    /* Free allocated memory */
    
    field_elt_free(&teta, k);  
    field_elt_free(&a, k);
    field_elt_free(&b, k);
    ec_point_free(G, k);
    number_free(&n);
    number_free(&h);
    ec_point_free(res, k);
    ec_point_free(res_jq, k);
    number_free(&p);
    field_free(k);
    ec_free(E);
    
    free_mphell();

    return 0;
}
