/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France)

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell_tuto_drbg.c
  * \brief Tutorial for DRBG use
  */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <string.h>

#include "mphell/mphell.h"
#include "mphell/mphell-util.h"

int main()
{
    /* AES128 */
    
    printf("CTR DRBG AES 128 test \n");
    
    /* Initialise MPHELL with 128 bits of security strength for the entropy, RANDOM_AES128 as DRBG and DEVURANDOM as entropy source */
    init_mphell(128, RANDOM_AES128, DEVURANDOM);
    
    mphell_status ret;
    int8_t state_handle;
    
    uint8_t returned_bytes[512/8];
    char returned_bytes_hex[512/4];
    
    uint8_t entropy_byte[128];
    uint8_t nonce_byte[64];
    uint8_t entropy_reseed_byte[128];
    uint8_t additionnal_input_reseed_byte[128];
    uint8_t additionnal_input1_byte[128];
    uint8_t additionnal_input2_byte[128];
    uint8_t personalization_string_byte[128];
    
    /* Convert to bytes strings */
    hex_string2bytes(entropy_byte, "03af6c44c7101544ca7888831f3f9f95", 128/8);
    hex_string2bytes(nonce_byte, "58d80fc378569346", 64/8);
    
    hex_string2bytes(entropy_reseed_byte, "410737c3c6112b5113f0ef66b0fa6a07", 128/8);
    hex_string2bytes(additionnal_input_reseed_byte, "ae2ce6d1dba56775db17b8c6d9379f14", 128/8);
    hex_string2bytes(additionnal_input1_byte, "57866375238ea56e97dd6af8c5010618", 128/8);
    hex_string2bytes(additionnal_input2_byte, "34a449f860c0c18fdf854b32de572917", 128/8);
    hex_string2bytes(personalization_string_byte, "7ec2b597c845ae8bbb63609a80c2ab4f", 128/8);
    
    /* Instantiate the DRBG, with reseed interval = 1000 */
    
    drbg_instantiate(ret, &state_handle, 128, DRBG_AES128, DEVURANDOM, 1000, entropy_byte, 128/8, nonce_byte, 64/8, personalization_string_byte, 128/8);
    if(ret->flag != MPHELL_SUCCESS)
    {
        printf("ERROR drbg_instantiate_test_drbg, state_handle = %d\n", state_handle);
        printf("ret->info: %s\n", ret->info);
        return -1;
    }
    
    /* Reseed the DRBG (optional) */
    
    drbg_reseed(ret, state_handle, entropy_reseed_byte, 128/8, additionnal_input_reseed_byte, 128/8);
    if(ret->flag != MPHELL_SUCCESS)
    {
        printf("ERROR drbg_reseed, state_handle = %d\n", state_handle);
        printf("ret->info: %s\n", ret->info);
        return -1;
    }
    
    /* Generate random number */
    
    drbg_generate(ret, returned_bytes, state_handle, 512/8, 128, additionnal_input1_byte, 128/8);
    if(ret->flag != MPHELL_SUCCESS)
    {
        printf("ERROR drbg_generate 1st call, state_handle = %d\n", state_handle);
        printf("ret->info: %s\n", ret->info);
        return -1;
    }
    
    bytes_string2hex(returned_bytes_hex, returned_bytes, 512/8);
    printf("returned_bytes_hex:    %s \n", returned_bytes_hex);
    
    /* Generate random number */
    
    drbg_generate(ret, returned_bytes, state_handle, 512/8, 128, additionnal_input2_byte, 128/8);
    if(ret->flag != MPHELL_SUCCESS)
    {
        printf("ERROR drbg_generate 2nd call, state_handle = %d\n", state_handle);
        printf("ret->info: %s\n", ret->info);
        return -1;
    }
    
    bytes_string2hex(returned_bytes_hex, returned_bytes, 512/8);
    printf("returned_bytes_hex:    %s \n", returned_bytes_hex);
    
    /* Uninstantiate the DRBG */
    drbg_uninstantiate(ret, state_handle);
    if(ret->flag != MPHELL_SUCCESS)
    {
        printf("ERROR drbg_uninstantiate, state_handle = %d\n", state_handle);
        printf("ret->info: %s\n", ret->info);
        return -1;
    }
    
    free_mphell();
}
