/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France)

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell_tuto_eckcdsa.c
  * \brief Tutorial for EC Korean Certificate DSA signature and verification using mphell.
  */

/**
  *  Some documents to get more informations about ECSDSA standard and hash 
  *  standard functions
  *  [KCDSA98] The Korean Certificate-based Digital Signature Algorithm by the 
  *  KCDSA Task Force Team, august 1998, IEEE P1363a, 
  *  (March 2019) https://pdfs.semanticscholar.org/87b2/b3699c14fd165e3f4ee9c5a307863b8cc235.pdf
  *  [FIPS-186] FIPS 186-4: Digital Signature Standard by the National Institute of
  *  Standards and Technology, july 2013, http://dx.doi.org/10.6028/NIST.FIPS.186-4  
  *  [FIPS-202] FIPS 202: SHA 3 Standard: Permutation based hash and extendable
  *  output functions by the National Institute of Standards and Technology, 
  *  august 2015, http://dx.doi.org/10.6028/NIST.FIPS.202
  *  [SP-800] SP 800-107 Recommendation for application using approved hash
  *  algorithms by Quynh Dang of the  National Institute of Standards and 
  *  Technology, august 2012, 
  *  https://nvlpubs.nist.gov/nistpubs/Legacy/SP/nistspecialpublication800-107r1.pdf 
  */

#include <stdio.h>
#include "mphell/mphell.h"
#if MPHELL_USE_MULTITHREADING == 1
#include <omp.h>
#endif

/**
 * @struct  eckcdsa_sig
 * @brief   Define an ECKCDSA signature
 */
struct eckcdsa_sig 
{
    unsigned char r[65]; /*!< Hash of an elliptic curve point, here the size is adapted to 256 bits curve */
    number *s; /*!< Second signature number, belongs to [1, n-1] */
};

/**
 * @typedef  eckcdsa_sig_t
 * @brief    Define the ECKCDSA signature type
 */
typedef struct eckcdsa_sig eckcdsa_sig_t;

/**
 * @typedef eckcdsa_sig
 * @brief   Pointer on an allocated eckcdsa_sig_t structure.
 */
typedef eckcdsa_sig_t eckcdsa_sig[1];

/**
 * @fn     void eckcdsa_sign_alloc(eckcdsa_sig *sig, uint8_t size);
 * @brief  Allocate a signature structure
 * @param  sig Signature
 * @param  size Size (in block) of the signature 
 */
void
eckcdsa_sign_alloc(eckcdsa_sig *sig, uint8_t size)
{

    (*sig)->s = (number*)malloc(sizeof(number));
    number_init((*sig)->s, size);
    number_set_ui(*((*sig)->s), (block)0);  
    int i;    
    for(i=0;i<64;i++)
    {
        ((*sig)->r)[i]='0';
    }
    ((*sig)->r)[64]='\0';  
}

/**
 * @fn     void eckcdsa_sign_free(eckcdsa_sig *sig);
 * @brief  Free a used signature structure
 * @param  sig  Signature
 */
void 
eckcdsa_sign_free(eckcdsa_sig *sig)
{
    if ((*sig)->s)
    {
        number_free((*sig)->s);
        free((*sig)->s);
    }
}

/**
 * @fn     void ec_point_mul_ecdsa(ec_point dst, number_srcptr u1, ec_point_srcptr G, number_srcptr u2, ec_point_srcptr key, ec_curve_srcptr E)
 * @brief  Compute u1 * G + u2 * key
 * @param  dst Destination point
 * @param  u1  Number
 * @param  G   Base point
 * @param  u2  Number
 * @param  key Public key 
 * @param  E   Elliptic curve
 */
void
ec_point_mul_ecdsa(ec_point dst, number_srcptr u1, ec_point_srcptr G,   
                        number_srcptr u2, ec_point_srcptr key, ec_curve_srcptr E)
{
#if MPHELL_USE_MULTITHREADING == 1
    omp_set_num_threads(2);
    ec_point tmp;
    ec_point_get_pool_elt(tmp, E->k, STACK_1);
    #pragma omp parallel
    {
        #pragma omp sections
        {
          #pragma omp section 
          {
              ec_point_mul(tmp, u2, key, E, STACK_1);
          }
          #pragma omp section
          {
              ec_point_mul(dst, u1, G, E, STACK_2);
          }
      }
    }
    ec_point_add(dst, dst, tmp, E, STACK_1);
    ec_point_relax_pool_elt(tmp, E->k, STACK_1);
#else
    ec_point tmp;
    ec_point_get_pool_elt(tmp, E->k, STACK_1);
    ec_point_mul(tmp, u2, key, E, STACK_1);
    ec_point_mul(dst, u1, G, E, STACK_1);
    ec_point_add(dst, dst, tmp, E, STACK_1);
    ec_point_relax_pool_elt(tmp, E->k, STACK_1);
#endif
}

/**
 * @fn     void eckcdsa_sign(unsigned char *hash, eckcdsa_sig *sig, number priv_key, ec_curve *curve);
 * @brief  Sign "hash" with ECKCDSA algorithm using the EC "curve" and the private key "key"
 * @param  hash     H(m) where H is a cryptographic hash function and m the message to sign. Must be written under hexadecimal form. ex: 64 hexadecimal characters for sha256.
 * @param  sig      Destination signature. Pointer on an allocated ecdsa_sig structure.
 * @param  priv_key Private key
 * @param  curve    Elliptic curve
 */
void
eckcdsa_sign(unsigned char *hash, unsigned char *cert, eckcdsa_sig *sig, number priv_key, ec_curve *curve)
{
    number *ss = (*sig)->s;
    
    uint8_t size = (*curve)->k->size;

    uint8_t hashbin[32];
    uint8_t hashbin2[32];
    unsigned char hashval1[65];
    unsigned char hashval2[65];
    char wxy[129];
    char * strx;
    char * stry; 
    uint8_t * message_bin;
    int32_t len_bin;
    uint32_t i;
    
    number k; number_init(&k, 2*size+1);
    number t; number_init(&t, 2*size+1);
        
    ec_point x;
    ec_point_get_pool_elt(x, (*curve)->k, STACK_1);
    
    bool newk = true;  
    while(newk)
    {
        /* Choose k randomly in [1, n-1] */
        /* In fact we need it to be invertible */
        number_random1(k, (*curve)->n, STACK_1);
        number_dec(t, (*curve)->n);    
        while(number_cmp(t, k)==0)
        {
            number_random1(k, (*curve)->n, STACK_1); 
        }
            
        /* Compute W= (wx,wy) = kG  */
        ec_point_mul_unified(x, k, (*curve)->G, *curve, STACK_1);
        ec_point_norm(x, *curve, STACK_1);
        printf("\n Le point W dans la signature:\n");        
        ec_point_print(x, 16, true, (*curve)->k, STACK_1);

        
        /* We compute r = h(wx||wy) */   
        field_elt_str(&strx, x->x, 16, false, (*curve)->k, STACK_1);
        field_elt_str(&stry, x->y, 16, false, (*curve)->k, STACK_1);    
        for(i=0;i<64;i++)
        {
            wxy[i]=strx[i];
        }
        for(i=0;i<64;i++)
        {
            wxy[i+64]=stry[i];
        } 
        wxy[128]= '\0';
        free(strx);
        free(stry);
        
        message_bin =  (uint8_t *)calloc(129*4, sizeof(uint8_t));
        len_bin= hex2bin(wxy, message_bin);
        sha256(hashbin, message_bin, len_bin * 8);
        free(message_bin);
        
        bin2hex(hashbin, 32, hashval1);  
        for(i=0;i<64;i++)
        {
            (*sig)->r[i]=hashval1[i];
        }

        /* We compute z as an hash of the certification data */
        message_bin =  (uint8_t *)calloc(strlen(cert)*4, sizeof(uint8_t));
        len_bin= hex2bin(cert, message_bin);
        sha256(hashbin2, message_bin, len_bin * 8);
        free(message_bin);
        bin2hex(hashbin2, 32, hashval2);

        /* We compute h(z||m) */
        for(i=0;i<64;i++)
        {
            wxy[i]=hashval2[i];
        }
        for(i=0;i<64;i++)
        {
            wxy[i+64]=hash[i];
        } 
        wxy[128]= '\0'; 
        message_bin =  (uint8_t *)calloc(strlen(wxy)*4, sizeof(uint8_t));
        len_bin= hex2bin(wxy, message_bin);
        sha256(hashbin2, message_bin, len_bin * 8);
        bin2hex(hashbin2, 32, hashval2);
        free(message_bin);

        /* We compute e= r ^h(z||m) = h(z||m)^h(W) */
        for(i=0;i<len_bin;i++)
        {
            hashbin2[i] =hashbin2[i] ^ hashbin[i];
        }
        bin2hex(hashbin2, 32, hashval2);
        number_set_str(t, hashval2, 16);
        
        /* We compute e mod n */
        number_mod(t, t, (*curve)->n);
        
        /* We compute s=x(k-e) mod n */
        if (number_cmp(t,k)==1)                   /* We make sure that e<=k */
        {
            number_add(k ,k,(*curve)->n);
        }            
        number_sub(k, k,t);
        number_mod(k, k, (*curve)->n);
        number_mul_mod(*ss, k, priv_key, (*curve)->n, STACK_1);
        
        /* if s == 0 --> new k */
        if(!number_iszero(*ss) ) 
        {
           newk = false; 
        }         
  
    }
    (*sig)->s = ss;
    
    /* Free memory */
    ec_point_relax_pool_elt(x, (*curve)->k, STACK_1);
    number_free(&t);
    number_free(&k);
}


/**
 *  For the hash several functions can be used, the interested reader can refer to [FIPS-180] 
 *    
 */

/**
 * @fn     int8_t eckcdsa_verify(unsigned char *hash, eckcdsa_sig *sig, ec_point *pub_key, ec_curve *curve);
 * @brief  Verify the signature "sig" of hash "hash" using public key "pub_key" and the EC "curve"
 * @param  hash    H(m) where H is a cryptographic hash function and m the message to sign. Must be written under hexadecimal form. ex: 64 hexadecimal characters for sha256.
 * @param  sig     Signature of hash
 * @param  pub_key Public key
 * @param  curve   Elliptic curve
 * @return 1 on success, 0 on failure
 */
int8_t 
eckcdsa_verify(unsigned char *hash, unsigned char *cert, eckcdsa_sig *sig, ec_point *pub_key, ec_curve *curve)
{
    /* Just check that r,s \in {1, ... , q-1} */
    
    if( number_iszero(*((*sig)->s)) || number_cmp(*((*sig)->s),(*curve)->n)>=0)
    {
        return false;
    } 

    char * strx;
    char * stry;
    uint8_t size = (*curve)->k->size;
    uint8_t hashbin[32];
    uint8_t hashbin2[32];
    unsigned char hashval1[65];
    unsigned char hashval2[65];
    unsigned char hzm[129];
    uint8_t * message_bin;
    int32_t len_bin;
    int i;   
    
    number u2; number_init(&u2, size);
    
    
    /* We compute the hash of the certificate data */
    message_bin =  (uint8_t *)calloc(strlen(cert)*4, sizeof(uint8_t));
    len_bin= hex2bin(cert, message_bin);
    sha256(hashbin, message_bin, len_bin * 8);
    free(message_bin);
    bin2hex(hashbin, 32, hashval1);

    /*It is up to the user to define a verification of this data stored in hashval1 */

    /* We compute h(z||m) */
    for(i=0;i<64;i++)
    {
        hzm[i]=hashval1[i];
    }
    for(i=0;i<64;i++)
    {
        hzm[i+64]=hash[i];
    } 
    hzm[128]= '\0'; 
    message_bin =  (uint8_t *)calloc(strlen(hzm)*4, sizeof(uint8_t));
    len_bin= hex2bin(hzm, message_bin);
    sha256(hashbin2, message_bin, len_bin * 8);
    free(message_bin);
    /* Useless bin2hex(hashbin2, 32, hashval2); */

    len_bin= hex2bin((*sig)->r,hashbin);
    /* We compute e= r ^h(z||m) = h(z||m)^h(W) */
    for(i=0;i<len_bin;i++)
    {
        //hashval2[i]=hashval2[i]^(*sig)->r[i];
        hashbin2[i] =hashbin2[i] ^ hashbin[i];
    }
    bin2hex(hashbin2, 32, hashval2);
    number_set_str(u2, hashval2, 16);
    
    /* e mod l */
    number_mod(u2,u2, (*curve)->n); 
    
    ec_point w,x;
    ec_point_get_pool_elt(w, (*curve)->k, STACK_1);
    ec_point_get_pool_elt(x, (*curve)->k, STACK_1);

    /* w' = s*Y + e*G */
    ec_point_mul_ecdsa(w, *((*sig)->s),  *pub_key, u2,(*curve)->G,  *curve);
    ec_point_norm(w, *curve, STACK_1);
    
    /* We compute h(xw'||yw') */
    field_elt_str(&strx, w->x, 16, false, (*curve)->k, STACK_1);
    field_elt_str(&stry, w->y, 16, false, (*curve)->k, STACK_1);    
    for(i=0;i<64;i++)
    {
        hzm[i]=strx[i];
    }
    for(i=0;i<64;i++)
    {
        hzm[i+64]=stry[i];
    } 
    hzm[128]= '\0';
    free(strx);
    free(stry);
        
    message_bin =  (uint8_t *)calloc(strlen(hzm)*4, sizeof(uint8_t));
    len_bin= hex2bin(hzm, message_bin);
    sha256(hashbin, message_bin, len_bin * 8);
    bin2hex(hashbin, 32, hashval1);
    hashval1[64]='\0';  
    free(message_bin);     

    /* Get the result, by comparing h value with r */
    //bool result = number_cmp(h, *((*sig)->r)) == 0;

    bool result=(memcmp((const char *)hashval1, (const char *)(*sig)->r, 12)==0);
    printf("\nWe have the following result \nh(W'): %s\nr: %s \n", (const char *) hashval1, (const char *)(*sig)->r);
    
    /* Free memory */
    ec_point_relax_pool_elt(w, (*curve)->k, STACK_1);
    ec_point_relax_pool_elt(x, (*curve)->k, STACK_1);
    number_free(&u2);

    return result;
}

/**
 * @fn     int8_t eckcdsa_pub_key_validation(ec_point *pub_key, ec_curve *curve);
 * @brief  Verify that the public key is valid (id est check taht pub_key belongs to the curve, that pub_key is not the neutral element and that pub_key is of ordre n) 
 * @param  hash    H(m) where H is a cryptographic hash function and m the message to sign. Must be written under hexadecimal form. ex: 64 hexadecimal characters for sha256.
 * @param  pub_key Public key
 * @param  curve   Elliptic curve
 * @return 1 on success, 0 on failure
 */
int8_t 
eckcdsa_pub_key_validation(ec_point *pub_key, ec_curve *curve)
{
   
    if(!ec_belongs(*pub_key, *curve, STACK_1))
    {
        /* pubkey does not belong  to the curve */
        return false;   
    }   
    if(ec_point_is_neutral(*pub_key, *curve, STACK_1))
    {
        /* pubkey is the neutral element */
        return false;
    }
    /* We finally test the order of the curve */
    ec_point x;
    ec_point_get_pool_elt(x, (*curve)->k, STACK_1);    
    ec_point_mul(x, (*curve)->n, *pub_key, *curve, STACK_1);
    if(!ec_point_is_neutral(x, *curve, STACK_1))
    {
        /* pubkey is the neutral element */
        return false;
    }
    ec_point_relax_pool_elt(x, (*curve)->k, STACK_1);
    
    return true;
}

int main()
{
    /* Initialise MPHELL with 256 bits of security strength for the entropy, RANDOM_AES256 as DRBG and DEVURANDOM as entropy source */
    
    init_mphell(256, RANDOM_AES256, DEVURANDOM);
    
    /* Allocate a field of size 4*block_SIZE = 4*64 = 256 on 64 bits architecture */
    
    field k;
    field_alloc(k, FP, bits_to_nblock(256), NULL);
    
    /* Allocate a number of size 4*block_SIZE = 4*64 = 256 on 64 bits architecture */
    
    number p;
    number_init(&p, bits_to_nblock(256));
    
    /* Set the number p from a string in base 16 */
    
    number_set_str(p, "ffffffff00000001000000000000000000000000ffffffffffffffffffffffff", 16);
    
    /* Create the field of characteristic p */
    
    field_create(k, "", STACK_1, 1, p);
    
    /* Allocate curve */
    
    ec_curve E;
    ec_alloc (E, k);
    ec_init(E, k);
    
    /* Create curve */
    
    field_elt a, b;
    ec_point G;
    number n, h;
    
    field_elt_alloc(&a, k);
    field_elt_init(a, k);
    field_elt_alloc(&b, k);
    field_elt_init(b, k);
    ec_point_alloc(G, k);
    ec_point_init(G, k);
    number_init(&n, bits_to_nblock(256));
    number_init(&h, bits_to_nblock(256));
    
    field_elt_set_str(a, "ffffffff00000001000000000000000000000000fffffffffffffffffffffffc", 16, false, k, STACK_1);
    field_elt_set_str(b, "5ac635d8aa3a93e7b3ebbd55769886bc651d06b0cc53b0f63bce3c3e27d2604b", 16, false, k, STACK_1);
    ec_point_set_aff_str(G, "6b17d1f2e12c4247f8bce6e563a440f277037d812deb33a0f4a13945d898c296", "4fe342e2fe1a7f9b8ee7eb4a7c0f9e162bce33576b315ececbb6406837bf51f5", false, 16, WEIERSTRASS, k, STACK_1);
    number_set_str(h, "1", 16);
    number_set_str(n, "ffffffff00000000ffffffffffffffffbce6faada7179e84f3b9cac2fc632551", 16);
    
    ec_create (E, "Weierstrass_nist_256", k, a, b, G, h, n, WEIERSTRASS, JACOBIAN, STACK_1);
    
    printf("E: \n"); ec_curve_print(E, 16, STACK_1); printf("\n");
    
    field_elt_free(&a, k);
    field_elt_free(&b, k);
    number_free(&h);
    number_free(&n);
    ec_point_free(G, k);

    /* Get private key x from random value between 0 and the size of the group generated by G:E->n*/
    /* Here we follow the notation of [KCDSA98] */
    number priv_key;
    number_init(&priv_key, bits_to_nblock(256));
    number_random1(priv_key, E->n, STACK_1);
       
    /* Test that the value is not equal to q-1 and change it until it is the case*/
    /* In fact we need the private key to be invertible modulo E->n */
    number test;
    number_init(&test, bits_to_nblock(256));
    number_dec(test, E->n);    
    while(number_cmp(test, priv_key)==0)
    {
       number_random1(priv_key, E->n, STACK_1); 
    }
    
    printf("\nThe private key has been created its value is:\n");
    number_print(priv_key, 16);
    
    /* Computation of the public key Y */
    /* Y=x^{-1} G */
    number_invmod(test, priv_key, E->n);
    ec_point pub_key;
    ec_point_alloc(pub_key, E->k);
    ec_point_init(pub_key, E->k);
    ec_point_mul_unified(pub_key, test, E->G, E, STACK_1);
    ec_point_norm(pub_key, E, STACK_1);  
    number_free(&test);    

    /* Print the public key */
    ec_point_norm(pub_key, E, STACK_1);  
    printf("\nThe pubkey has been computed:\n");  
    ec_point_print(pub_key, 16, true, k, STACK_1); 

    /* This test is optional.*/
    printf("\nThe validity of the public key is a:\n");
    int ret = eckcdsa_pub_key_validation(&pub_key, &E);
    if( ret < 1 )
    {
        printf("ERROR\n");
    }
    else
    {
        printf("SUCCESS\n");
    }
    
    /* Digest to sign */
    char * digest = "982c20c2493fc9ae405b74b65a022662c014a38ef3d707217e56e57afac05994";

    /* Certification Data */
    char * cert = "d5c30315f72ed05fe519a1bf75ab5fd0ffec5ac1acb0daf66b6b769598594509";
    
    /* Allocate ecsdsa_sig structure */
    eckcdsa_sig sig;
    eckcdsa_sign_alloc(&sig, bits_to_nblock(256));

    /* Sign the digest */
    eckcdsa_sign((unsigned char *)digest, (unsigned char *)cert, &sig, priv_key, &E);
    
    printf("\nThe ECKCDSA signature is: \n(%s, ",(sig->r));
    number_print(*(sig->s), 16); printf(")\n");
    
    /* Verify the signature */
    printf("\n-> Verify ECKCDSA signature \n");
    ret = eckcdsa_verify((unsigned char *)digest, (unsigned char *)cert, &sig, &pub_key, &E);
    if( ret < 1 )
    {
        printf("ERROR\n");
    }
    else
    {
        printf("SUCCESS\n");
    }
    
    /* Free memory */
    
    ec_point_free(pub_key, k);
    number_free(&priv_key);
    eckcdsa_sign_free(&sig);
    number_free(&p);;
    field_free(k);
    ec_free(E);
    
    free_mphell();

    return 0;
}
