/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France)

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell_tuto_edwards.c
  * \brief Tutorial for Edwards elliptic curve
  */

#include <stdio.h>
#include "mphell/mphell.h"

int main()
{
    /* Initialise MPHELL with 256 bits of security strength for the entropy, RANDOM_AES256 as DRBG and DEVURANDOM as entropy source */
    
    init_mphell(256, RANDOM_AES256, DEVURANDOM);
    
    /* Allocate a field of size 4*block_SIZE = 4*64 = 256 on 64 bits architecture */
    
    field k;
    field_alloc(k, FP, bits_to_nblock(256), NULL);
    
    /* Allocate a number of size 4*block_SIZE = 4*64 = 256 on 64 bits architecture */
    
    number p;
    number_init(&p, bits_to_nblock(256));
    
    /* Set the number p from a string in base 16 */
    
    number_set_str(p, "7fffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffed", 16);
    
    /* Create the field of characteristic p */
    
    field_create(k, "", STACK_1, 1, p);
    
    /* Allocate curve */
    
    ec_curve E;
    ec_alloc (E, k);
    ec_init(E, k);
    
    /* Create curve */
    
    field_elt a, b;
    ec_point G;
    number n, h;
    
    field_elt_alloc(&a, k);
    field_elt_init(a, k);
    field_elt_alloc(&b, k);
    field_elt_init(b, k);
    ec_point_alloc(G, k);
    ec_point_init(G, k);
    number_init(&n, bits_to_nblock(256));
    number_init(&h, bits_to_nblock(256));
    
    field_elt_set_one(a, k);
    field_elt_set_str(b, "2dfc9311d490018c7338bf8688861767ff8ff5b2bebe27548a14b235eca6874a", 16, false, k, STACK_1);
    
    ec_point_set_aff_str(G, "159a6849e44c3c7f061b3d570fc4ed5b5d14c8ba4253df49cc7edf80f533ad9b", "6666666666666666666666666666666666666666666666666666666666666658", false, 16, EDWARDS, k, STACK_1);
    number_set_str(h, "0000000000000000000000000000000000000000000000000000000000000008", 16);
    number_set_str(n, "1000000000000000000000000000000014def9dea2f79cd65812631a5cf5d3ed", 16);
    
    ec_create (E, "Edwards_test", k, a, b, G, h, n, EDWARDS, EXTENDED_EDWARDS, STACK_1);
    
    printf("E: \n"); ec_curve_print(E, 16, STACK_1); printf("\n");
    
    field_elt_free(&a, k);
    field_elt_free(&b, k);
    ec_point_free(G, k);
    number_free(&n);
    number_free(&h);
    
    /* Allocate element of the field k */
    
    ec_point x;
    ec_point y;
    ec_point res;
    ec_point_alloc(x, k);
    ec_point_init(x, k);
    ec_point_alloc(y, k);
    ec_point_init(y, k);
    ec_point_alloc(res, k);
    ec_point_init(res, k);
    
    /* Set field element from string in base 16, which are not under Montgomery form */
    
    ec_point_set_aff_str(x, "67e82d6ecfec167788bc76ffc0165be5b3b5ac785f5bd1d2c427fa5c4ef02548", "5f6b282c46a137a1d612df50487cfa27a4b4df2c77bf6f12e56ec667ee37b181", false, 16, EDWARDS, k, STACK_1);
    ec_point_set_aff_str(y, "1ec6d017b0d11bd1d1b5cfefebfccbc8c8079e0f1030008a2e75207b77a987e6", "20bbd47f960781e9e13ea7cff81a4f5bec6106453cbae87abd3018d7803539be", false, 16, EDWARDS, k, STACK_1);
    
    /* Addition */
    
    ec_point_add(res, x, y, E, STACK_1);
    printf("x+y = "); ec_point_print(res, 16, true, k, STACK_1); printf("\n");
    printf("x+y belongs to E : %d\n\n", ec_belongs(res, E, STACK_1));
    
    /* Substraction */
    
    ec_point_sub(res, x, y, E, STACK_1);
    printf("x-y = "); ec_point_print(res, 16, true, k, STACK_1); printf("\n");
    printf("x-y belongs to E : %d\n\n", ec_belongs(res, E, STACK_1));
    
    /* Doubling */
    
    ec_point_dbl(res, x, E, STACK_1);
    printf("x*2 = "); ec_point_print(res, 16, true, k, STACK_1); printf("\n");
    printf("x*2 belongs to E : %d\n\n", ec_belongs(res, E, STACK_1));
    
    /* Negation */
    
    ec_point_neg(res, x, E);
    printf("-x = "); ec_point_print(res, 16, true, k, STACK_1); printf("\n");
    printf("-x belongs to E : %d\n\n", ec_belongs(res, E, STACK_1));
    
    /* Scalar Multiplication */
    
    number n1;
    number_init(&n1, 4);
    number_set_str(n1, "67e82d6ecfec167788bc76ffc0165be5b3b5ac785f5bd1d2c427fa5c4ef02551", 16);
    ec_point_mul(res, n1, x, E, STACK_1);
    printf("x*67e82d6ecfec167788bc76ffc0165be5b3b5ac785f5bd1d2c427fa5c4ef02551 = "); ec_point_print(res, 16, true, k, STACK_1); printf("\n");
    printf("x*67e82d6ecfec167788bc76ffc0165be5b3b5ac785f5bd1d2c427fa5c4ef02551 belongs to E : %d\n\n", ec_belongs(res, E, STACK_1));
    number_free(&n1);
    
    /* Random point */
    
    ec_point_random(res, E, STACK_1);
    printf("random point, res = "); ec_point_print(res, 16, true, k, STACK_1); printf("\n");
    printf("res belongs to E : %d\n\n", ec_belongs(res, E, STACK_1));

    /* Conversion to a point on a Weierstrass curve  */

    /* Allocation of memory for the destination point */
    
    ec_point convP;
    ec_point_alloc(convP, k);
    ec_point_init(convP, k);
    
    edwards_point_to_weierstrass_point (convP, res, E, STACK_1);
    printf("the image point on a weierstrass curve, convP="); ec_point_print(convP, 16, true, k, STACK_1); printf("\n");
    
    /* Conversion to a Weierstrass curve  */

    /* Allocation of memory for the destination point */
    
    ec_curve convC;
    ec_alloc (convC, k);
    ec_init (convC, k);

    edwards_to_weierstrass(convC, E, STACK_1);
    printf("convC: \n"); ec_curve_print(convC, 16, STACK_1); printf("\n");

    /* Free allocated memory */
    
    ec_point_free(convP, k);
    ec_point_free(x, k);
    ec_point_free(y, k);
    ec_point_free(res, k);
    number_free(&p);
    field_free(k);
    ec_free(E);
    ec_free(convC);
    
    free_mphell();

    return 0;
}
