/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France)

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell_tuto_fp.c
  * \brief Tutorial for primary field (FP) arithmetic
  */

#include <stdio.h>
#include "mphell/mphell.h"

int main()
{
    /* Initialise MPHELL with 128 bits of security strength for the entropy, RANDOM_AES128 as DRBG and DEVURANDOM as entropy source */
    
    init_mphell(128, RANDOM_AES128, DEVURANDOM);
    
    /* Allocate a field of size 3*block_SIZE = 3*64 = 192 on 64 bits architecture */
    
    field k;
    field_alloc(k, FP, bits_to_nblock(192), NULL);
    
    /* Allocate a number of size 3*block_SIZE = 3*64 = 192 on 64 bits architecture */
    
    number p;
    number_init(&p, bits_to_nblock(192));
    
    /* Set the number p from a string in base 16 */
    
    number_set_str(p, "fffffffffffffffffffffffffffffffeffffffffffffffff", 16);
    
    /* Create the field of characteristic p */

    field_create(k, "", STACK_1, 1, p);
    
    /* Allocate element of the field k */
    
    field_elt x;
    field_elt y;
    field_elt res;
    field_elt_alloc(&x, k);
    field_elt_init(x, k);
    field_elt_alloc(&y, k);
    field_elt_init(y, k);
    field_elt_alloc(&res, k);
    field_elt_init(res, k);
    
    /* Set field element from string in base 16, which are not under Montgomery form */
    
    field_elt_set_str(x, "f14b8dbafa22b6ba35626d40d2d00001381bef07d2c3017", 16, false, k, STACK_1);
    field_elt_set_str(y, "85ca4bcae9e156f1eab07c7b6cc500004ebbc0e5e3b237ce", 16, false, k, STACK_1);
    
    printf("x = "); field_elt_print(x, 16, true, k, STACK_1); printf("\n");
    printf("y = "); field_elt_print(y, 16, true, k, STACK_1); printf("\n");
    
    /* Addition */
    
    field_elt_inc(res, x, k);
    printf("x+1 = "); field_elt_print(res, 16, true, k, STACK_1); printf("\n");
    
    field_elt_add(res, x, y, k);
    printf("x+y = "); field_elt_print(res, 16, true, k, STACK_1); printf("\n");
    
    /* Substraction */
    
    field_elt_dec(res, x, k);
    printf("x-1 = "); field_elt_print(res, 16, true, k, STACK_1); printf("\n");
    
    field_elt_sub(res, x, y, k);
    printf("x-y = "); field_elt_print(res, 16, true, k, STACK_1); printf("\n");
    
    /* Multiplication */
    
    field_elt_mul(res, x, y, k, STACK_1);
    printf("x*y = "); field_elt_print(res, 16, true, k, STACK_1); printf("\n");
    
    field_elt_sqr(res, x, k, STACK_1);
    printf("x^2 = "); field_elt_print(res, 16, true, k, STACK_1); printf("\n");
    
    /* Division */
    
    field_elt_div(res, x, y, k, STACK_1);
    printf("x/y = "); field_elt_print(res, 16, true, k, STACK_1); printf("\n");
    
    /* Power */
    
    field_elt_pow_ui(res, x, 3, k, STACK_1);
    printf("x^3 = "); field_elt_print(res, 16, true, k, STACK_1); printf("\n");
    printf("Res is a power of 3 : %d\n", field_elt_ispower_ui(res, 3, k, STACK_1)>0);
    
    number n; number_init(&n, bits_to_nblock(192)); number_set_str(n, "f14b8dbafa22b6ba35626d40d2d00001381bef07d2c3015", 16);
    field_elt_pow_number(res, x, n, k, STACK_1);
    printf("x^n = "); field_elt_print(res, 16, true, k, STACK_1); printf("\n");
    printf("Res is a power of n : %d\n", field_elt_ispower_number(res, n, k, STACK_1)>0);
    number_free(&n);
    
    /* Inverse */
    
    field_elt_inv(res, x, k, STACK_1);
    printf("x^(-1) = "); field_elt_print(res, 16, true, k, STACK_1); printf("\n");
    
    /* Roots */
    
    field_elt_sqrt(res, x, k, STACK_1);
    printf("x is square : %d\n", field_elt_issquare(x, k, STACK_1)>0);
    printf("x^(1/2) = "); field_elt_print(res, 16, true, k, STACK_1); printf("\n");
    
    field_elt_cube_root(res, x, k, STACK_1);
    printf("x is a power of 3 : %d\n", field_elt_ispower_ui(x, 3, k, STACK_1)>0);
    printf("x^(1/3) = "); field_elt_print(res, 16, true, k, STACK_1); printf("\n");
    
    /* Free allocated memory */
    
    field_elt_free(&x, k);
    field_elt_free(&y, k);
    field_elt_free(&res, k);
    number_free(&p);
    field_free(k);
    
    free_mphell();

    return 0;
}
