/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France)

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell_tuto_fp2.c
  * \brief Tutorial for quadratic field extension arithmetic
  */

#include <stdio.h>
#include "mphell/mphell.h"

int main()
{
    /* Initialise MPHELL with 128 bits of security strength for the entropy, RANDOM_AES128 as DRBG and DEVURANDOM as entropy source */
    
    init_mphell(128, RANDOM_AES128, DEVURANDOM);
    
    /* Create base prime field of characteristic p */
    
    field k;
    field_alloc(k, FP, bits_to_nblock(384), NULL);
    number p;
    number_init(&p, bits_to_nblock(384));
    number_set_str(p, "fffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffeffffffff0000000000000000ffffffff", 16);
    field_create(k, "", STACK_1, 1, p);
    
    /* Create finite field GF(p^2) of characteristic p and order p^2 */
    
    field_elt non_residue;
    field_elt_alloc(&non_residue, k);
    field_elt_init(non_residue, k);
    /* Find an irreducible polynomial P for the test: P = X^2 - non_residue */
    field_find_nonsquare(non_residue, k, STACK_1);
    field k2;
    field_alloc(k2, FP2, bits_to_nblock(384), k);
    field_create(k2, "", STACK_1, 2, k, non_residue);
    
    /* Allocate element of the field k2 */
    
    field_elt x;
    field_elt y;
    field_elt res;
    field_elt_alloc(&x, k2);
    field_elt_init(x, k2);
    field_elt_alloc(&y, k2);
    field_elt_init(y, k2);
    field_elt_alloc(&res, k2);
    field_elt_init(res, k2);
    
    /* Set field element from string in base 10, which are not under Montgomery form */
    
    field_elt_set_str(x, "25888030203341481017599806294711784662368639501836779649649722493704107869971378062561359493103614745148646490006495,30818930178847386173826179517114528511268570512817196018110767526684990909388933012679657889738193796300356614506500", 10, false, k2, STACK_1);
    field_elt_set_str(y, "13513975993052998194679233805431829142711099768628667018298570910541613901525492266485906595155323256712960483105824,8583076017547093038452860583029085293811168757648250649837525877560730862107937316367608198520744205561250358605819", 10, false, k2, STACK_1);
    
    /* Addition */
    
    field_elt_inc(res, x, k);
    printf("x+1 = "); field_elt_print(res, 16, true, k, STACK_1); printf("\n");
    
    field_elt_add(res, x, y, k);
    printf("x+y = "); field_elt_print(res, 16, true, k, STACK_1); printf("\n");
    
    /* Substraction */
    
    field_elt_dec(res, x, k);
    printf("x-1 = "); field_elt_print(res, 16, true, k, STACK_1); printf("\n");
    
    field_elt_sub(res, x, y, k);
    printf("x-y = "); field_elt_print(res, 16, true, k, STACK_1); printf("\n");
    
    /* Multiplication */
    
    field_elt_mul(res, x, y, k, STACK_1);
    printf("x*y = "); field_elt_print(res, 16, true, k, STACK_1); printf("\n");
    
    field_elt_sqr(res, x, k, STACK_1);
    printf("x^2 = "); field_elt_print(res, 16, true, k, STACK_1); printf("\n");
    
    /* Division */
    
    field_elt_div(res, x, y, k, STACK_1);
    printf("x/y = "); field_elt_print(res, 16, true, k, STACK_1); printf("\n");
    
    /* Power */
    
    field_elt_pow_ui(res, x, 3, k, STACK_1);
    printf("x^3 = "); field_elt_print(res, 16, true, k, STACK_1); printf("\n");
    printf("Res is a power of 3 : %d\n", field_elt_ispower_ui(res, 3, k, STACK_1)>0);
    
    number n; number_init(&n, bits_to_nblock(192)); number_set_str(n, "f14b8dbafa22b6ba35626d40d2d00001381bef07d2c3015", 16);
    field_elt_pow_number(res, x, n, k, STACK_1);
    printf("x^n = "); field_elt_print(res, 16, true, k, STACK_1); printf("\n");
    printf("Res is a power of n : %d\n", field_elt_ispower_number(res, n, k, STACK_1)>0);
    number_free(&n);
    
    /* Inverse */
    
    field_elt_inv(res, x, k, STACK_1);
    printf("x^(-1) = "); field_elt_print(res, 16, true, k, STACK_1); printf("\n");
    
    /* Roots */
    
    field_elt_sqr(x, x, k, STACK_1);
    printf("x = "); field_elt_print(x, 16, true, k, STACK_1); printf("\n");
    printf("x is square : %d\n", field_elt_issquare(x, k, STACK_1)>0);
    field_elt_sqrt(res, x, k, STACK_1);
    printf("x^(1/2) = "); field_elt_print(res, 16, true, k, STACK_1); printf("\n");
    
    field_elt_pow_ui(x, x, 3, k, STACK_1);
    printf("x = "); field_elt_print(x, 16, true, k, STACK_1); printf("\n");
    printf("x is a power of 3 : %d\n", field_elt_ispower_ui(x, 3, k, STACK_1)>0);
    field_elt_cube_root(res, x, k, STACK_1);
    printf("x^(1/3) = "); field_elt_print(res, 16, true, k, STACK_1); printf("\n");
    
    /* Free allocated memory */
    
    field_elt_free(&x, k2);
    field_elt_free(&y, k2);
    field_elt_free(&res, k2);
    number_free(&p);
    field_elt_free(&non_residue, k);
    field_free(k);
    field_free(k2);
    
    free_mphell();

    return 0;
}
