/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France)

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell_tuto_fp3.c
  * \brief Tutorial for cubic field extension arithmetic
  */

#include <stdio.h>
#include "mphell/mphell.h"

int main()
{
    /* Initialise MPHELL with 128 bits of security strength for the entropy, RANDOM_AES128 as DRBG and DEVURANDOM as entropy source */
    
    init_mphell(128, RANDOM_AES128, DEVURANDOM);
    
    /* Create base prime field of characteristic p */
    
    field k;
    field_alloc(k, FP, bits_to_nblock(384), NULL);
    number p;
    number_init(&p, bits_to_nblock(384));
    number_set_str(p, "8CB91E82A3386D280F5D6F7E50E641DF152F7109ED5456B412B1DA197FB71123ACD3A729901D1A71874700133107EC53", 16);
    field_create(k, "", STACK_1, 1, p);
    
    /* Create finite field GF(p^3) of characteristic p and order p^3 */
    
    field_elt non_residue;
    field_elt_alloc(&non_residue, k);
    field_elt_init(non_residue, k);
    /* Find an irreducible polynomial P for the test: P = X^3 - non_residue */
    printf("Find a non cube element in k : %d\n", field_find_nonpower_ui(non_residue, 3, k, STACK_1));
    field k3;
    field_alloc(k3, FP3, bits_to_nblock(384), k);
    field_create(k3, "", STACK_1, 2, k, non_residue);
    
    /* Allocate element of the field k2 */
    
    field_elt x;
    field_elt y;
    field_elt res;
    field_elt_alloc(&x, k3);
    field_elt_init(x, k3);
    field_elt_alloc(&y, k3);
    field_elt_init(y, k3);
    field_elt_alloc(&res, k3);
    field_elt_init(res, k3);
    
    /* Set field element from string in base 10, which are not under Montgomery form */
    
    field_elt_set_str(x, "17533160620960747720658538821194093754122675622398475560762133052267072585559185687559064016240090086860997857609409,6500634679458907167017356094867727707826954836406037804481994709882642877127237360869164454175728327118666667333650,11742847707788052444522195706157812780258053127409735266675287415078850342578399252850159043780127493098080513256739", 10, false, k3, STACK_1);
    field_elt_set_str(y, "4126110149158568452410698021138511225673440764619173039319485451554017348466776134677497966604444001579710560363922,15158636090660409006051880747464877271969161550611610795599623793938447056898724461367397528668805761322041750639681,9916423062331263728547041136174792199538063259607913333406331088742239591447562569386402939064406595342627904716592", 10, false, k3, STACK_1);
    
    /* Addition */
    
    field_elt_inc(res, x, k);
    printf("x+1 = "); field_elt_print(res, 16, true, k, STACK_1); printf("\n");
    
    field_elt_add(res, x, y, k);
    printf("x+y = "); field_elt_print(res, 16, true, k, STACK_1); printf("\n");
    
    /* Substraction */
    
    field_elt_dec(res, x, k);
    printf("x-1 = "); field_elt_print(res, 16, true, k, STACK_1); printf("\n");
    
    field_elt_sub(res, x, y, k);
    printf("x-y = "); field_elt_print(res, 16, true, k, STACK_1); printf("\n");
    
    /* Multiplication */
    
    field_elt_mul(res, x, y, k, STACK_1);
    printf("x*y = "); field_elt_print(res, 16, true, k, STACK_1); printf("\n");
    
    field_elt_sqr(res, x, k, STACK_1);
    printf("x^2 = "); field_elt_print(res, 16, true, k, STACK_1); printf("\n");
    printf("res is square : %d\n", field_elt_issquare(res, k, STACK_1)>0);
    
    /* Division */
    
    field_elt_div(res, x, y, k, STACK_1);
    printf("x/y = "); field_elt_print(res, 16, true, k, STACK_1); printf("\n");
    
    /* Power */
    
    field_elt_pow_ui(res, x, 3, k, STACK_1);
    printf("x^3 = "); field_elt_print(res, 16, true, k, STACK_1); printf("\n");
    printf("Res is a power of 3 : %d\n", field_elt_ispower_ui(res, 3, k, STACK_1)>0);
    
    number n; number_init(&n, bits_to_nblock(192)); number_set_str(n, "f14b8dbafa22b6ba35626d40d2d00001381bef07d2c3015", 16);
    field_elt_pow_number(res, x, n, k, STACK_1);
    printf("x^n = "); field_elt_print(res, 16, true, k, STACK_1); printf("\n");
    printf("Res is a power of n : %d\n", field_elt_ispower_number(res, n, k, STACK_1)>0);
    number_free(&n);
    
    /* Inverse */
    
    field_elt_inv(res, x, k, STACK_1);
    printf("x^(-1) = "); field_elt_print(res, 16, true, k, STACK_1); printf("\n");
    
    /* Roots */
    
    field_elt_sqr(x, x, k, STACK_1);
    printf("x = "); field_elt_print(x, 16, true, k, STACK_1); printf("\n");
    printf("x is square : %d\n", field_elt_issquare(x, k, STACK_1)>0);
    field_elt_sqrt(res, x, k, STACK_1);
    printf("x^(1/2) = "); field_elt_print(res, 16, true, k, STACK_1); printf("\n");
    
    field_elt_pow_ui(x, x, 3, k, STACK_1);
    printf("x = "); field_elt_print(x, 16, true, k, STACK_1); printf("\n");
    printf("x is a power of 3 : %d\n", field_elt_ispower_ui(x, 3, k, STACK_1)>0);
    field_elt_cube_root(res, x, k, STACK_1);
    printf("x^(1/3) = "); field_elt_print(res, 16, true, k, STACK_1); printf("\n");
    
    /* Free allocated memory */
    
    field_elt_free(&x, k3);
    field_elt_free(&y, k3);
    field_elt_free(&res, k3);
    number_free(&p);
    field_elt_free(&non_residue, k);
    field_free(k);
    field_free(k3);
    
    free_mphell();

    return 0;
}
