/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France)

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell_tuto_jacobi_quartic.c
  * \brief Tutorial for Jacobi Quartic elliptic curve
  */

#include <stdio.h>
#include "mphell/mphell.h"

int main()
{
    /* Initialise MPHELL with 256 bits of security strength for the entropy, RANDOM_AES256 as DRBG and DEVURANDOM as entropy source */
    
    init_mphell(256, RANDOM_AES256, DEVURANDOM);
    
    /* Allocate a field of size 4*block_SIZE = 4*64 = 256 on 64 bits architecture */
    
    field k;
    field_alloc(k, FP, bits_to_nblock(256), NULL);
    
    /* Allocate a number of size 4*block_SIZE = 4*64 = 256 on 64 bits architecture */
    
    number p;
    number_init(&p, bits_to_nblock(256));
    
    /* Set the number p from a string in base 16 */
    
    number_set_str(p, "be6dc5f3bca8726edc8d23dfc89f5922e12826336befbd97ef9ad755aed3ed6f", 16);
    
    /* Create the field of characteristic p */
    
    field_create(k, "", STACK_1, 1, p);
    
    /* Allocate curve */
    
    ec_curve E;
    ec_alloc (E, k);
    ec_init(E, k);
    
    /* Create curve */
    
    field_elt a, b;
    ec_point G;
    number h,n;
    
    field_elt_alloc(&a, k);
    field_elt_init(a, k);
    field_elt_alloc(&b, k);
    field_elt_init(b, k);
    ec_point_alloc(G, k);
    ec_point_init(G, k);
    number_init(&n, bits_to_nblock(256));
    number_init(&h, bits_to_nblock(256));
    
    field_elt_set_str(a, "49f59d6eae84c3d20838901b12eb680503e0c84c9438b2343e7cd5f750b2ccdf", 16, false, k, STACK_1);
    field_elt_set_one(b, k); /* Useless for Jacobi Quartic EC */
    
    ec_point_set_str(G, "27e09c4f00f1e6f7763e63e8cd0939c910b5cff8e9deecafe41b2933ab8adf27", "b18a89ba34ad9fee919d49cbf816a22be34778e54ac6498ba29a17db22732673", "78289546d596f34a9fc8e2fd46ebe745c354d1853183a42deac29ea956b114d0", "4c075c65a97f10ff49bc6c8a223957fc99eca98df4a998165d96d31b138a5a97", false, 16, k, STACK_1);
    number_set_str(n, "2f9b717cef2a1c9bb72348f7f227d648e20a240c39f62f5f1cf3643f55d42cb1", 16);
    number_set_str(h, "0000000000000000000000000000000000000000000000000000000000000004", 16);
    
    ec_create (E, "Jacobi_test", k, a, b, G, h, n, JACOBI_QUARTIC, PROJECTIVE, STACK_1);
    
    printf("E: \n"); ec_curve_print(E, 16, STACK_1); printf("\n");
    
    field_elt_free(&a, k);
    field_elt_free(&b, k);
    ec_point_free(G, k);
    number_free(&n);
    number_free(&h);
    
    /* Allocate element of the field k */
    
    ec_point x;
    ec_point y;
    ec_point res;
    ec_point_alloc(x, k);
    ec_point_init(x, k);
    ec_point_alloc(y, k);
    ec_point_init(y, k);
    ec_point_alloc(res, k);
    ec_point_init(res, k);
    
    /* Set field element from string in base 16, which are not under Montgomery form */
    
    ec_point_set_aff_str (x, "38255193717143657873171169431252446125914975555408103986804661333889508186792", "56879352098683955683913475804414516257891196368009685803867282765611829589550", false, 10, JACOBI_QUARTIC, k, STACK_1);
    ec_point_set_aff_str (y, "6481597564264008301482198835779793377095770727839417835648754067898733644793", "72836348899108920838208922130674404413346456085043969930205677190313902908769", false, 10, JACOBI_QUARTIC, k, STACK_1);
    
    /* Addition */
    
    ec_point_add(res, x, y, E, STACK_1);
    printf("x+y = "); ec_point_print(res, 16, true, k, STACK_1); printf("\n");
    printf("x+y belongs to E : %d\n\n", ec_belongs(res, E, STACK_1));
    
    /* Substraction */
    
    ec_point_sub(res, x, y, E, STACK_1);
    printf("x-y = "); ec_point_print(res, 16, true, k, STACK_1); printf("\n");
    printf("x-y belongs to E : %d\n\n", ec_belongs(res, E, STACK_1));
    
    /* Doubling */
    
    ec_point_dbl(res, x, E, STACK_1);
    printf("x*2 = "); ec_point_print(res, 16, true, k, STACK_1); printf("\n");
    printf("x*2 belongs to E : %d\n\n", ec_belongs(res, E, STACK_1));
    
    /* Negation */
    
    ec_point_neg(res, x, E);
    printf("-x = "); ec_point_print(res, 16, true, k, STACK_1); printf("\n");
    printf("-x belongs to E : %d\n\n", ec_belongs(res, E, STACK_1));
    
    /* Scalar Multiplication */
    
    number n1;
    number_init(&n1, 4);
    number_set_str(n1, "6481597564264008301482198835779793377095770727839417835648754067898733644745", 10);
    ec_point_mul(res, n1, x, E, STACK_1);
    printf("x*6481597564264008301482198835779793377095770727839417835648754067898733644745 = "); ec_point_print(res, 16, true, k, STACK_1); printf("\n");
    printf("x*6481597564264008301482198835779793377095770727839417835648754067898733644745 belongs to E : %d\n\n", ec_belongs(res, E, STACK_1));
    number_free(&n1);
    
    /* Random point */
    
    ec_point_random(res, E, STACK_1);
    printf("random point, res = "); ec_point_print(res, 16, true, k, STACK_1); printf("\n");
    printf("res belongs to E : %d\n\n", ec_belongs(res, E, STACK_1));
    
    /* Free allocated memory */
    
    ec_point_free(x, k);
    ec_point_free(y, k);
    ec_point_free(res, k);
    number_free(&p);
    field_free(k);
    ec_free(E);
    
    free_mphell();

    return 0;
}
