/* 
  MPHELL-4.0 
  Author(s): The MPHELL team

 (C) Copyright 2015-2018 - Institut Fourier / Univ. Grenoble Alpes (France)

 This file is part of the MPHELL Library.
 MPHELL is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, version 3 of the License.
 
 MPHELL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with MPHELL.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * \file  mphell_tuto_weierstrass.c
  * \brief Tutorial for Weierstrass elliptic curve
  */

#include <stdio.h>
#include "mphell/mphell.h"

int main()
{
    /* Initialise MPHELL with 256 bits of security strength for the entropy, RANDOM_AES256 as DRBG and DEVURANDOM as entropy source */
    
    init_mphell(256, RANDOM_AES256, DEVURANDOM);
    
    /* Allocate a field of size 3*block_SIZE = 3*64 = 192 on 64 bits architecture */
    
    field k;
    field_alloc(k, FP, bits_to_nblock(192), NULL);
    
    /* Allocate a number of size 3*block_SIZE = 3*64 = 192 on 64 bits architecture */
    
    number p;
    number_init(&p, bits_to_nblock(192));
    
    /* Set the number p from a string in base 16 */
    
    number_set_str(p, "fffffffffffffffffffffffffffffffeffffffffffffffff", 16);
    
    /* Create the field of characteristic p */
    
    field_create(k, "", STACK_1, 1, p);
    
    /* Allocate curve */
    
    ec_curve E;
    ec_alloc(E, k);
    ec_init(E, k);
    
    /* Create curve */
    
    field_elt a, b;
    ec_point G;
    number h, n;
    
    field_elt_alloc(&a, k);
    field_elt_init(a, k);
    field_elt_alloc(&b, k);
    field_elt_init(b, k);
    ec_point_alloc(G, k);
    ec_point_init(G, k);
    number_init(&h, bits_to_nblock(192));
    number_init(&n, bits_to_nblock(192));
    
    field_elt_set_str(a, "fffffffffffffffffffffffffffffffefffffffffffffffc", 16, false, k, STACK_1);
    field_elt_set_str(b, "64210519e59c80e70fa7e9ab72243049feb8deecc146b9b1", 16, false, k, STACK_1);
    ec_point_set_aff_str(G, "188da80eb03090f67cbf20eb43a18800f4ff0afd82ff1012", "07192b95ffc8da78631011ed6b24cdd573f977a11e794811", false, 16, WEIERSTRASS, k, STACK_1);
    number_set_str(h, "000000000000000000000000000000000000000000000001", 16);
    number_set_str(n, "ffffffffffffffffffffffff99def836146bc9b1b4d22831", 16);
    
    ec_create(E, "Weierstrass_test", k, a, b, G, h, n, WEIERSTRASS, PROJECTIVE, STACK_1);
    
    printf("E: \n"); ec_curve_print(E, 16, STACK_1); printf("\n");
    
    field_elt_free(&a, k);
    field_elt_free(&b, k);
    ec_point_free(G, k);
    number_free(&n);
    number_free(&h);
    
    /* Allocate element of the field k */
    
    ec_point x;
    ec_point y;
    ec_point res;
    ec_point_alloc(x, k);
    ec_point_init(x, k);
    ec_point_alloc(y, k);
    ec_point_init(y, k);
    ec_point_alloc(res, k);
    ec_point_init(res, k);
    
    /* Set field element from string in base 16, which are not under Montgomery form */
    
    ec_point_set_aff_str(x, "3a6e0e0135079698f34344bb2261a5fa0730419b639b45ca", "52c86dc001c77804f6a1e9ab8e5ac4627727919a68993692", false, 16, WEIERSTRASS, k, STACK_1);
    ec_point_set_aff_str(y, "8cf9058484386d660331a5ef852002360c00d5c940010c33", "2ec3235f6b5eeec13bc3c206b47e60cf715966c67c6e292a", false, 16, WEIERSTRASS, k, STACK_1);
    
    /* Addition */
    
    ec_point_add(res, x, y, E, STACK_1);
    printf("x+y = "); ec_point_print(res, 16, true, k, STACK_1); printf("\n");
    printf("x+y belongs to E : %d\n\n", ec_belongs(res, E, STACK_1));
    
    /* Substraction */
    
    ec_point_sub(res, x, y, E, STACK_1);
    printf("x-y = "); ec_point_print(res, 16, true, k, STACK_1); printf("\n");
    printf("x-y belongs to E : %d\n\n", ec_belongs(res, E, STACK_1));
    
    /* Doubling */
    
    ec_point_dbl(res, x, E, STACK_1);
    printf("x*2 = "); ec_point_print(res, 16, true, k, STACK_1); printf("\n");
    printf("x*2 belongs to E : %d\n\n", ec_belongs(res, E, STACK_1));
    
    /* Negation */
    
    ec_point_neg(res, x, E);
    printf("-x = "); ec_point_print(res, 16, true, k, STACK_1); printf("\n");
    printf("-x belongs to E : %d\n\n", ec_belongs(res, E, STACK_1));
    
    /* Scalar Multiplication */
    
    number m;
    number_init(&m, 3);
    number_set_str(m, "3a6e0e0135079698f34344bb2261a5fa0730419b639b45ca", 16);
    ec_point_mul(res, m, x, E, STACK_1);
    printf("x*3a6e0e0135079698f34344bb2261a5fa0730419b639b45ca = "); ec_point_print(res, 16, true, k, STACK_1); printf("\n");
    printf("x*3a6e0e0135079698f34344bb2261a5fa0730419b639b45ca belongs to E : %d\n\n", ec_belongs(res, E, STACK_1));
    number_free(&m);
    
    
    /* Random point */
    
    ec_point_random(res, E, STACK_1);
    printf("random point, res = "); ec_point_print(res, 16, true, k, STACK_1); printf("\n");
    printf("res belongs to E : %d\n\n", ec_belongs(res, E, STACK_1));
    
    ec_point_free(res, k);
    number_free(&p);
    ec_point_free(x, k);
    ec_point_free(y, k);
    field_free(k);
    ec_free(E);
    free_mphell();

    return 0;
}
